/*************************************************************************\
**  source       * ssscan.c
**  directory    * ss
**  description  * Portable sscan utilities 
**               * 
**               * Copyright (C) 2006 Solid Information Technology Ltd
\*************************************************************************/
/*
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
*/

#ifdef DOCUMENTATION
**************************************************************************

Implementation:
--------------

WINDOWS SDK DLL C-libraries do not offer sscanf(). That is why we have to
implement services that can do the same thing.

There are usually available functions (or macros) like
atoi(), atol(), atof() etc, but we can not use them because they return
zero (0) also when the operation is not succesfull. We usually want to
test if stream contains something valid, and we surely can not do it with
atoi() etc.

So, we do the whole thing by scanning the input stream and checking
it character by character.

Then, the actual return value is generated by atof() and atol().


Limitations:
-----------


Error handling:
--------------

None

Objects used:
------------

None

Preconditions:
-------------


Multithread considerations:
--------------------------

This code is re-entrant

Example:
-------

See tscan.c

**************************************************************************
#endif /* DOCUMENTATION */


#include "ssc.h"
#include "ssstdio.h"
#include "ssstdlib.h"
#include "ssdebug.h"
#include "ssctype.h"
#include "ssfloat.h"
#include "sstraph.h"
#include "ssscan.h"

/*##**********************************************************************\
 * 
 *		SsStrScanDouble
 * 
 * Converts an ascii string to double. 
 * 
 * A valid double is of form 
 * 
 *  [whitespace][{+|-}][digits][.digits][{d|D|e|E]}[sign]digits]
 * 
 * Parameters : 
 * 
 *	src - in, use
 *		source string
 * 
 *	p_d - out
 *		pointer to double
 *
 *      p_mismatch - out
 *		pointer to the first position in src that
 *          does not belong to the double returned
 *
 * Return value : 
 * 
 *      TRUE, if succesfull. p_d and p_mismatch are updated.
 *      FALSE, if failed.
 * 
 * Limitations  : 
 * 
 * Globals used : 
 */
bool SsStrScanDouble(src, p_d, p_mismatch)
	char* src;
	double* p_d;
        char** p_mismatch;
{
        bool valid = FALSE;
        char* p_ch;
        char* org_src = src;

        p_ch = src++;


        /* Allow white spaces in the beginning */
        while (ss_isspace((uchar)*p_ch)) {
            p_ch = src++;
        }

        /* Next, allow sign */
        if ((*p_ch == '-') || (*p_ch == '+')) {
            p_ch = src++;
        } 

        /* Then there may be digits */
        while (ss_isdigit((uchar)*p_ch)) {
            valid = TRUE;
            p_ch = src++;
        }

        /* Allow decimal point */
        if (*p_ch == '.') {
            p_ch = src++;
        }

        /* Then there may be digits */
        while (ss_isdigit((uchar)*p_ch)) {
            valid = TRUE;
            p_ch = src++;
        }

        /* There has to be some digits for now */
        if (!valid) {
            return(FALSE);
        }

        /* we handle here the possible {e|E|d|D}{exponent} */

        if ((*p_ch == 'e') || (*p_ch == 'E') ||
            (*p_ch == 'd') || (*p_ch == 'D') ) {

            int exp;
            bool bo;
            char* ptr;

            p_ch = src++;

            bo = SsStrScanInt(p_ch, &exp, &ptr);

            if (bo) {
                p_ch = ptr;
            } else {
                p_ch--;
            }
        }

        if (valid) {
            static const double dbl_zero = 0.0;
            ss_trapcode_t trapcode = SS_TRAP_NONE;

            SS_TRAP_HANDLERSECTION
                case SS_TRAP_FPE:
                case SS_TRAP_FPE_INVALID:
                case SS_TRAP_FPE_INEXACT:
                case SS_TRAP_FPE_STACKFAULT:
                case SS_TRAP_FPE_STACKOVERFLOW:
                case SS_TRAP_FPE_STACKUNDERFLOW:
                case SS_TRAP_FPE_BOUND:
                case SS_TRAP_FPE_DENORMAL:
                case SS_TRAP_FPE_INTOVFLOW:
                case SS_TRAP_FPE_OVERFLOW:
                case SS_TRAP_FPE_INTDIV0:
                case SS_TRAP_FPE_ZERODIVIDE:
                case SS_TRAP_FPE_UNDERFLOW:
                case SS_TRAP_FPE_EXPLICITGEN:
                    trapcode = SS_TRAP_GETCODE();
                    valid = FALSE;
                    ss_dprintf_1(("SsStrScanDouble FP trap %d\n", trapcode));
                    break;
                default:
                    SS_TRAP_RERAISE();
                    ss_error;
            SS_TRAP_RUNSECTION
                *p_mismatch = p_ch;
#if defined(SS_CLIX) || defined(SS_QNX)
                valid = (sscanf(org_src, "%lg", p_d) == 1);
                if (valid) {
                    if (*p_d == 0.0) {
                        *p_d = dbl_zero;
                    }
                    valid = SS_DOUBLE_IS_PORTABLE(*p_d);
                }
#else
                *p_d = atof(org_src);
                if (*p_d == 0.0) {
                    *p_d = dbl_zero;
                }
                valid = SS_DOUBLE_IS_PORTABLE(*p_d);
#endif
            SS_TRAP_END
        }
        return(valid);
}

/*##**********************************************************************\
 * 
 *		SsStrScanLong
 * 
 * Converts an ascii string to long integer. 
 * Leading whitespaces are skipped.  Handles hexnumbers with 0x prefix too.
 * 
 *  [whitespace][{+|-}][digits]
 *  [whitespace]0x[hexdigits]
 * 
 * Parameters : 
 * 
 *	s - in, use
 *		source string
 *
 *	p_l - out
 *		pointer to long integer
 *
 *      p_mismatch - out
 *		pointer to the first position in src that
 *          does not belong to the long integer returned
 *
 * Return value : 
 * 
 *      TRUE, if successful
 *      FALSE, if failed.
 * 
 * Limitations  : 
 * 
 * Globals used : 
 */
bool SsStrScanLong(char* s, long* p_l, char** p_mismatch)
{
        bool valid;
        bool negative;
        bool hex;
        ulong base;
        ulong v;
        ulong t;

        ss_dassert(s != NULL);
        ss_dassert(p_l != NULL);
        ss_dassert(p_mismatch != NULL);
        negative = FALSE;
        valid    = FALSE;
        hex      = FALSE;
        base = 10;

        while (ss_isspace((uchar)*s)) {
            s++;
        }
        switch (*s) {
            case '-':
                negative = TRUE;
                /* FALLTHROUGH */
            case '+':
                s++;
                break;
            case '0':
                if (s[1] == 'x' || s[1] == 'X') {
                    hex  = TRUE;
                    base = 16;
                    s += 2;
                }
                break;
            default:
                break;
        }
        for (v = 0, valid = FALSE;
             ;
             v = v * base + t, s++, valid = TRUE)
        {
            switch (*s) {
                case '9':
                    t = 9;
                    continue;
                case '8':
                    t = 8;
                    continue;
                case '7':
                    t = 7;
                    continue;
                case '6':
                    t = 6;
                    continue;
                case '5':
                    t = 5;
                    continue;
                case '4':
                    t = 4;
                    continue;
                case '3':
                    t = 3;
                    continue;
                case '2':
                    t = 2;
                    continue;
                case '1':
                    t = 1;
                    continue;
                case '0':
                    t = 0;
                    continue;
                    /* Hex handling */
                case 'f':
                case 'F':
                    t = 15;
                    if (hex) {
                        continue;
                    }
                    break;                        
                case 'e':
                case 'E':
                    t = 14;
                    if (hex) {
                        continue;
                    }
                    break;                        
                case 'd':
                case 'D':
                    t = 13;
                    if (hex) {
                        continue;
                    }
                    break;                        
                case 'c':
                case 'C':
                    t = 12;
                    if (hex) {
                        continue;
                    }
                    break;                        
                case 'b':
                case 'B':
                    t = 11;
                    if (hex) {
                        continue;
                    }
                    break;                        
                case 'a':
                case 'A':
                    t = 10;
                    if (hex) {
                        continue;
                    }
                    break;                        
                default:
                    break;
            }
            break; /* quit loop */
        }
        if (negative) {
            *p_l = -(long)v;
			if(*p_l > 0) {
				valid = FALSE;
			}
        } else {
            *p_l = (long)v;
			if(*p_l < 0) {
				valid = FALSE;
			}
        }
        *p_mismatch = s;
        return (valid);
}


/*##**********************************************************************\
 * 
 *		SsStrScanInt
 * 
 * Converts an ascii string to integer. 
 *  [whitespace][{+|-}][digits]
 *  [whitespace]0x[hexdigits]
 * 
 * Parameters : 
 * 
 *	src - in, use
 *		source string
 *
 *	p_l - out
 *		pointer to integer
 *
 *      p_mismatch - out
 *		pointer to the first position in src that
 *          does not belong to the integer returned
 *
 * Return value : 
 * 
 *      TRUE, if successful. p_i and p_mismatch are updated.
 *      FALSE, if failed.
 * 
 * Limitations  : 
 * 
 * Globals used : 
 */
bool SsStrScanInt(src, p_i, p_mismatch)
	char* src;
	int* p_i;
        char** p_mismatch;
{
        long l;

        if (SsStrScanLong(src, &l, p_mismatch)) {
            *p_i = (int)l;
            return(TRUE);
        } else {
            return(FALSE);
        }
}

/*##**********************************************************************\
 *
 *      SsStrScanYesNo
 *
 * Convert a yes or no to bool.
 *
 * Accepts (Y[ES]|N[O]), case-insensitive.
 *
 * Any string beginning with 'y' will be TRUE, but the difference is in
 * where p_mismatch points to.  "yesxxx" and "yepxxx" difference is that
 * in former p_mismatch point to "xxx" and in latter "epxxx".
 *
 * If there shouldn't be extra in the string then check that p_mismatch points
 * to '\0' value after the call.
 *
 * Parameters:
 *      src - in
 *          source string
 *
 *      p_b - out
 *          output, true -> yes, false -> no.
 *
 *      p_mismatch - out
 *          points to point where value ended.
 *
 * Return value:
 *      TRUE, if success.
 *      FALSE, if not found.
 *
 * Limitations:
 *
 * Globals used:
 */
bool SsStrScanYesNo(char* src, bool* p_b, char** p_mismatch)
{
        bool value = FALSE;
        bool valid = TRUE;
        
        if (ss_tolower(*src) == 'y') {
            src++;
            if (*src != '\0') { /* not just 'y' */
                if (SsStrnicmp("es", src, 2) == 0) { /* "yes" */
                    src += 2;
                }
            }
            value = TRUE;
            
        } else if (ss_tolower(*src) == 'n') {
            src++;
            if (*src != '\0') { /* not just 'n' */
                if (ss_tolower(*src) == 'o') { /* "no" */
                    src++;
                }
            }
            value = FALSE;
            
        } else {
            valid = FALSE;
        }
        
        if (p_mismatch != NULL) {
            *p_mismatch = src;
        }
        *p_b = value;
        
        return valid;
}

/*  EOF  */
