/* Copyright (C) 2006-2007 MySQL AB & Solid Information Technology Ltd


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA

*/

/*
  Please read ha_exmple.cc before reading this file.
  Please keep in mind that the example storage engine implements all methods
  that are required to be implemented. handler.h has a full list of methods
  that you can implement.
*/

#define ENGINENAME_SOLIDDB        "solidDB"
#define ENGINENAME_SOLIDDB_ALIAS  "solid"

#ifdef USE_PRAGMA_INTERFACE
#pragma interface			/* gcc class implementation */
#endif

/*
  EXAMPLE_SHARE is a structure that will be shared amoung all open handlers
  The example implements the minimum of what you will probably need.
*/

typedef struct st_soliddb_share {
  THR_LOCK lock;
  pthread_mutex_t mutex;
  char *table_name;
  uint table_name_length,use_count;
} SOLIDDB_SHARE;

/*
  Class definition for the storage engine
*/

class ha_soliddb: public handler
{
  THR_LOCK_DATA lock;                     /* MySQL lock */
  SOLIDDB_SHARE *share;                   /* Shared lock info */
  bool rnd_mustinit;                      /* rnd_init called */
  bool for_update;                        /* write lock on read */
  bool extra_keyread;                     /* HA_EXTRA_KEYREAD */
  bool extra_retrieve_primary_key;        /* HA_EXTRA_RETRIEVE_PRIMARY_KEY */
  bool extra_retrieve_all_cols;           /* HA_EXTRA_RETRIEVE_ALL_COLS */
  bool mainmemory;

  ulong int_table_flags;

  struct st_solid_connection* solid_conn; /* Reference to Solid connection. */
  struct rsrelhandlestruct* solid_rsrelh; /* Solid table handle. */
  struct solid_relcur_st* solid_relcur;   /* Solid cursor handle. */

  st_solid_connection* get_solid_connection(
#if MYSQL_VERSION_ID >= 50100
      handlerton* hton,
#endif
      THD*        thd,
      const char* table_name)
  {
      return get_solid_connection(table_name);
  }

  st_solid_connection* get_solid_connection( const char * table_name = NULL);

  int solid_fetch(byte *buf, bool nextp, THD* thd);
  int solid_index_read_idx(
          byte* buf, 
          uint index, 
          const byte* key,
          uint key_len,
          enum ha_rkey_function find_flag,
          bool firstp, 
          THD* thd);
  struct solid_relcur_st* solid_relcur_create(
          struct st_solid_connection* con,
          struct solid_relcur_st* old_relcur,
          struct rsrelhandlestruct* rsrelh,
          TABLE* table,
          THD* thd,
          int index,
          KEY* key_info);

public:

#if MYSQL_VERSION_ID >= 50100
  ha_soliddb(handlerton *hton, TABLE_SHARE *table_arg);
#else
  ha_soliddb(TABLE *table_arg);
#endif
  
  ~ha_soliddb();

  /* Return solidDB specific error message */
  bool get_error_message(int error, String *buf);

  /* The name that will be used for display purposes */
  const char *table_type() const { return(ENGINENAME_SOLIDDB); }
  /*
    The name of the index type that will be used for display
    don't implement this method unless you really have indexes
   */
  const char *index_type(uint inx) { return "BTREE"; }
  const char **bas_ext() const;
  /*
    This is a list of flags that says what the storage engine
    implements. The current table flags are documented in
    handler.h
  */
#if MYSQL_VERSION_ID >= 50100
  ulonglong table_flags() const { return int_table_flags; }
#else
  ulong table_flags() const { return int_table_flags; }
#endif
  
  /*
    This is a bitmap of flags that says how the storage engine
    implements indexes. The current index flags are documented in
    handler.h. If you do not implement indexes, just return zero
    here.

    part is the key part to check. First key part is 0
    If all_parts it's set, MySQL want to know the flags for the combined
    index up to and including 'part'.
  */
  ulong index_flags(uint inx, uint part, bool all_parts) const
  {
      /* TODO: support for all, check handler.h for documentation.
                return (HA_READ_NEXT |
	                  HA_READ_PREV |
	                  HA_READ_ORDER |
	                  HA_READ_RANGE |
                          HA_KEYREAD_ONLY);
      */
      if (mainmemory) {
        /* gotoend does not yet work */
        return (HA_READ_NEXT |
                HA_READ_ORDER |
                HA_READ_RANGE |
                HA_KEYREAD_ONLY);
      } else {
          return (HA_READ_NEXT |
                  HA_READ_PREV |
                  HA_READ_ORDER |
                  HA_READ_RANGE |
                  HA_KEYREAD_ONLY);
      }
  }
  /*
    unireg.cc will call the following to make sure that the storage engine can
    handle the data it is about to send.

    Return *real* limits of your storage engine here. MySQL will do
    min(your_limits, MySQL_limits) automatically

    There is no need to implement ..._key_... methods if you don't suport
    indexes.
  */
  // TODO: we can get correct values for these.
  // for now we return something usefull
  uint max_supported_record_length() const { return HA_MAX_REC_LENGTH; }
  uint max_supported_keys()          const { return MAX_KEY; }
  uint max_supported_key_length() const { return HA_MAX_REC_LENGTH; }
  uint max_supported_key_part_length() const { return HA_MAX_REC_LENGTH; }
  const key_map *keys_to_use_for_scanning() { return &key_map_full; }

  bool has_transactions()  { return 1;}

  /*
    Called in test_quick_select to determine if indexes should be used.
  */
  double scan_time();
  /*
    The next method will never be called if you do not implement indexes.
  */
  double read_time(uint index, uint ranges, ha_rows rows);

  /*
    Everything below are methods that we implment in ha_example.cc.

    Most of these methods are not obligatory, skip them and
    MySQL will treat them as not implemented
  */
  int open(const char *name, int mode, uint test_if_locked);    // required
  int close(void);                                              // required

  int write_row(byte * buf);
  int update_row(const byte * old_data, byte * new_data);
  int delete_row(const byte * buf);
  int index_read(byte * buf, const byte * key,
                 uint key_len, enum ha_rkey_function find_flag);
  int index_read_last(byte * buf, const byte * key,
                 uint key_len);
  int index_read_idx(byte * buf, uint idx, const byte * key,
                     uint key_len, enum ha_rkey_function find_flag);
  int index_next(byte * buf);
  int index_next_same(byte * buf, const byte *key, uint keylen);
  int index_prev(byte * buf);
  int index_first(byte * buf);
  int index_last(byte * buf);
  /*
    unlike index_init(), rnd_init() can be called two times
    without rnd_end() in between (it only makes sense if scan=1).
    then the second call should prepare for the new table scan
    (e.g if rnd_init allocates the cursor, second call should
    position it to the start of the table, no need to deallocate
    and allocate it again
  */
  int rnd_init(bool scan);                                      //required
  int rnd_end();
  int rnd_next(byte *buf);                                      //required
  int rnd_pos(byte * buf, byte *pos);                           //required
  void position(const byte *record);                            //required

#if MYSQL_VERSION_ID >= 50030
  int info(uint);
#else
  void info(uint);                                              //required
#endif
  
  int start_stmt(THD *thd, thr_lock_type lock_type);

  int extra(enum ha_extra_function operation);
  int external_lock(THD *thd, int lock_type);                   //required
  int delete_all_rows(void);
  ha_rows records_in_range(uint inx, key_range *min_key,
                           key_range *max_key);
  int delete_table(const char *from);
  int rename_table(const char * from, const char * to);
  int create(const char *name, TABLE *form,
             HA_CREATE_INFO *create_info);                      //required

  THR_LOCK_DATA **store_lock(THD *thd, THR_LOCK_DATA **to,
                             enum thr_lock_type lock_type);     //required

  uint8 table_cache_type();

#if MYSQL_VERSION_ID < 50100  
  bool update_auto_increment();
#endif
  
  int check(THD* thd, HA_CHECK_OPT* check_opt);

  char* get_foreign_key_create_info();

  int get_foreign_key_list(THD *thd, List<FOREIGN_KEY_INFO> *f_key_list);

  /* Used in REPLACE, return 1 if a table is referenced by a foreign key */
  uint referenced_by_foreign_key();

  void free_foreign_key_create_info(char* str);

#if MYSQL_VERSION_ID >= 50100

  int index_init(uint keynr, bool sorted);

  void get_auto_increment(ulonglong offset, ulonglong increment,
                                  ulonglong nb_desired_values,
                                  ulonglong *first_value,
                                  ulonglong *nb_reserved_values);

  bool check_if_incompatible_data(HA_CREATE_INFO *info, uint table_changes);

  int add_index(TABLE *table_arg, KEY *key_info, uint num_of_keys);

  int prepare_drop_index(TABLE *table_arg, uint *key_num, uint num_of_keys);

  int final_drop_index(TABLE *table_arg);

  int reset();
  
#endif
};

/* Configuration parameters for SolidDB storage engine */

#if MYSQL_VERSION_ID >= 50100
extern SHOW_VAR soliddb_status_variables[];
#else
extern struct show_var_st soliddb_status_variables[];
#endif

extern char* soliddb_logdir;
extern char* soliddb_backupdir;
extern char* soliddb_filespec;
extern longlong soliddb_cache_size;
extern long soliddb_durability_level;
extern my_bool soliddb_checkpoint_deletelog;
extern long soliddb_lock_wait_timeout;
extern long soliddb_db_block_size;
extern long soliddb_log_block_size;
extern long soliddb_backup_block_size;
extern longlong soliddb_checkpoint_interval;
extern long soliddb_io_threads;
extern longlong soliddb_lockhash_size;
extern long soliddb_checkpoint_time;
extern my_bool soliddb_log_enabled;
extern long soliddb_maxcursors;
extern long soliddb_threads;
extern my_bool soliddb_pessimistic;
extern long soliddb_write_threads;
extern long soliddb_extend_increment;
extern long soliddb_readahead;
extern char* soliddb_admin_command;

bool soliddb_add_filespec(THD *, set_var *var);
int soliddb_check_filespec(THD *, set_var *var);
void soliddb_update_durability_level(THD *, enum_var_type type);
void soliddb_update_checkpoint_time(THD *, enum_var_type type);
void soliddb_update_lock_wait_timeout(THD *, enum_var_type type);
void soliddb_update_checkpoint_interval(THD *, enum_var_type type);
bool soliddb_update_backupdir(THD *, set_var *var);
int soliddb_check_backupdir(THD *, set_var *var);
int soliddb_check_admin_command(THD*, set_var *var);
bool soliddb_update_admin_command(THD*, set_var *var);

int solid_get_error_message(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD*,
        TABLE **,
        int,
        String *);

void solid_drop_database(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        char* path);

bool solid_flush_logs(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton
#endif
    );

void soliddb_export_status(void);

#if MYSQL_VERSION_ID < 50100
int solid_end(void);
bool soliddb_show_status(THD* thd);
bool soliddb_show_mutex_status(THD *thd);
#endif
