/* Copyright (C) 2006-2007 MySQL AB & Solid Information Technology Ltd


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; only under version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA

*/

/*
  ha_soliddb is a storage engine interface for solidDB.
  You can enable it in your build by doing the following during your
  build process:
  ./configure --with-soliddb-storage-engine

  Once this is done mysql will let you create tables with:
  CREATE TABLE A (...) ENGINE=SOLIDDB;

  Please read the object definition in ha_soliddb.h before reading the rest
  if this file.

  To get an idea of what occurs here is an example select that would do a
  scan of an entire table:
  ha_soliddb::store_lock
  ha_soliddb::external_lock
  ha_soliddb::info
  ha_soliddb::rnd_init
  ha_soliddb::extra
  ENUM HA_EXTRA_CACHE   Cash record in HA_rrnd()
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::rnd_next
  ha_soliddb::extra
  ENUM HA_EXTRA_NO_CACHE   End cacheing of records (def)
  ha_soliddb::external_lock
  ha_soliddb::extra
  ENUM HA_EXTRA_RESET   Reset database to after open

  In the above example has 9 row called before rnd_next signalled that it was
  at the end of its data. In the above example the table was already opened
  (or you would have seen a call to ha_soliddb::open(). Calls to
  ha_soliddb::extra() are hints as to what will be occuring to the request.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#if defined(WITH_SOLIDDB_STORAGE_ENGINE) || defined(HAVE_SOLID_DB)

#ifdef USE_PRAGMA_IMPLEMENTATION
#pragma implementation        // gcc: Class implementation
#endif

#define INSIDE_HA_SOLIDDB_CC

#define _WIN32_WINNT 0x0403 /* enable TryEnterCriticalSection etc. */

#include <mysql_priv.h>
#include <mysys_err.h>
#include <my_sys.h>
#include <my_base.h>

#include <ssmyenv.h>

/* We cannot use bool defined in Solid headers because it
* conflicts with the bool defined in Linux.
*/
#define bool_defined
#define solid_bool  int
#ifdef SS_LINUX
#define bool        int
#endif

#ifdef __cplusplus
extern "C" {                    /* Assume C declarations for C++   */
#endif  /* __cplusplus */

#include "ssmyenv.h"
#include "sswindow.h"
#include "ssstdio.h"
#include "ssstring.h"

#include "ssc.h"
#include "ssdebug.h"
#include "ssmem.h"
#include "sssprint.h"
#include "ssfnsplt.h"
#include "sstime.h"
#include "ssthread.h"
#include "ssmsglog.h"

#include "uti0vtpl.h"

#include "ui0msg.h"

#include "su0list.h"
#include "su0rbtr.h"
#include "su0err.h"
#include "su0error.h"
#include "su0vers.h"
#include "su0time.h"
#include "su0inifi.h"
#include "su0cfgst.h"
#include "su0param.h"
#include "su0pars.h"
#include "su0usrid.h"

#include "rs0types.h"
#include "rs0atype.h"
#include "rs0aval.h"
#include "rs0ttype.h"
#include "rs0sysi.h"
#include "rs0cons.h"
#include "rs0key.h"
#include "rs0pla.h"
#include "rs0relh.h"
#include "rs0aval.h"
#include "rs0vbuf.h"

#include "dbe0type.h"
#include "dbe0curs.h"
#include "dbe0rel.h"
#include "dbe0db.h"
#include "dbe7cfg.h"
#include "dbe6bkey.h"

#include "est1est.h"
#include "est1pla.h"

#include "tab0tran.h"
#include "tab0conn.h"
#include "tab0admi.h"
#include "tab0relh.h"
#include "tab0relc.h"
#include "tab0sche.h"
#include "tab1dd.h"
#include "tab0seq.h"
#include "tab0srv.h"
#include "tab1refact.h"
#include "su0vfil.h"
#include "ctype.h"
#include "my_dir.h"
#include "my_getopt.h"
#ifdef __WIN__
#include <direct.h>
#endif

#ifdef __cplusplus
}                                    /* End of extern "C" { */
#endif  /* __cplusplus */

#define USE_DDL_TABLE_LOCKING
#define DDL_LOCK_TIMEOUT 0 /* in miliseconds */

#ifdef SS_LINUX
#undef bool
#endif

#define ss_printf       printf

#if defined(SS_NT) && defined(SS_MYSQL_PERFCOUNT)
#define ss_win_perf(c)              c
#define ss_win_perf_start           if (mysql_enable_perfcount) QueryPerformanceCounter((LARGE_INTEGER*)&startcount)
#define ss_win_perf_stop(pc, cc)    if (mysql_enable_perfcount) {\
    QueryPerformanceCounter((LARGE_INTEGER*)&endcount);\
    (pc) += endcount - startcount;\
    (cc)++;\
}
#else
#define ss_win_perf(c)
#define ss_win_perf_start
#define ss_win_perf_stop(pc, cc)
#endif

#ifdef SS_DEBUG
#define SDB_DBUG_ENTER(fname) DBUG_ENTER(fname); ss_pprintf_1((fname "\n"));
#define SDB_DBUG_RETURN(rc) do { ss_pprintf_1(( "%s:return %d\n", &_db_func_, rc)); DBUG_RETURN(rc); } while(0)
#else
#define SDB_DBUG_ENTER(fname)
#define SDB_DBUG_RETURN(rc) return(rc)
#endif

#include "ha_soliddb.h"

void solid_set_relname( char* schema, char* relname, char* mysqlname, bool lower_case_names);
void solid_split_myrelname( const char* mysqlname, char** schema, char** relname );
inline rs_entname_t* solid_split_myrelname( const char* mysqlname, rs_entname_t *solrel )
{
    solid_split_myrelname(mysqlname, &solrel->en_schema, &solrel->en_name);
    return solrel;
}

void solid_relname( rs_entname_t *solrel, const char* mysqlname);
void solid_relname( rs_entname_t *solrel, const char* mysqldb, const char* mysqlrel);

bool is_soliddb_table( const char* dbname, const char* tablename );
bool is_soliddb_table( File &frm_file );

bool is_DDL_command( THD* thd );
bool operation_requires_xtable_lock( THD* thd );

/* MySQL related utility functions. */
bool copy_file( File &src, const char* dstdir, const char* fname, const char* ext );

#define SQL_WHITESPACES " \t\n\r"
#define BACKUP_LIST_FILE_NAME "solid_restore.cnf"
#define BACKUP_CONFIG_FILE_NAME "my.cnf"

#define MAX_REF_LENGTH      8000                        /* TODO: neeed to check this. */
#define TABLE_SCAN_INDEX    -1
#define MAX_FOREIGN_LEN     64000

#define CHKVAL_CONN         986543
#define CHKVAL_RELCUR       329874

#define CHK_CONN(c)         ss_dassert((c) != NULL && (c)->sc_chk == CHKVAL_CONN)
#define CHK_RELCUR(r)       ss_dassert((r) != NULL && (r)->sr_chk == CHKVAL_RELCUR)

typedef enum {
        SOLID_NEW_CURSOR_CREATE,
        SOLID_NEW_CURSOR_REUSE,
        SOLID_NEW_CURSOR_USEOLD
} solid_new_cursor_t;

typedef struct st_solid_connection {
        int           sc_chk;
        tb_connect_t* sc_tbcon;        // connection
        rs_sysi_t*    sc_cd;           // client data
        tb_trans_t*   sc_trans;        // transaction for the connection
        bool          sc_isvalid;      // is this connection valid
        int           sc_nlink;        // number of refences to the connection
        su_err_t*     sc_errh;
        su_list_t*    sc_relcurlist;   // list of relcur objects allocated from
                                    // this connection
        su_list_t*    sc_relcurdonelist;  // list of relcur objects that should
                                       // be released
        uint          sc_errkey_num;   // Key number on duplicate key error
        TABLE*        sc_table;        // Table where duplicate key error was found
        SsSemT*       sc_mutex;        // Own mutex for a connection
        rs_key_t*     sc_errkey;       // Key on constraint error
        rs_tval_t*    sc_tval;         // Value to be inserted or updated

        int           con_n_tables; // #-of tables in use

} SOLID_CONN;

typedef struct solid_relcur_st {
        int                   sr_chk;
        SOLID_CONN*           sr_con;
        dbe_cursor_t*         sr_relcur;
        rs_ttype_t*           sr_ttype;
        rs_tval_t*            sr_fetchtval;
        dbe_bkey_t*           sr_bkeybuf;
        int                   sr_index;
        bool                  sr_full_scan;
        rs_key_t*             sr_key;       /* Used key, or NULL for ::rnd_ routines. */
        rs_sysi_t*            sr_cd;
        tb_trans_t*           sr_trans;
        su_list_t*            sr_relcurlist;
        su_list_node_t*       sr_relcurlistnode;
        rs_relh_t*            sr_rsrelh;
        bool                  sr_for_update;   // Use 'for update' in select
        dbe_cursor_type_t     sr_cursor_type;  // Cursor type
        int*                  sr_selectlist;   // Attribute number list of selected columns
        int*                  sr_plaselectlist;   // Attribute number list of selected columns
        su_list_t*            sr_constraints;
        rs_tval_t*            sr_constval;
        rs_tval_t*            sr_postval;
        bool*                 sr_usedfields;   // Boolean array of used fields
        rs_pla_t*             sr_pla;
        rs_key_t*             sr_plakey;
    	query_id_t            sr_query_id;      /* the latest query id with this cursor. */
        bool                  sr_open_cursor;
        bool                  sr_extra_keyread;                     /* HA_EXTRA_KEYREAD */
        bool                  sr_extra_retrieve_primary_key;        /* HA_EXTRA_RETRIEVE_PRIMARY_KEY */
        bool                  sr_extra_retrieve_all_cols;           /* HA_EXTRA_RETRIEVE_ALL_COLS */
        bool                  sr_unique_row;
        bool                  sr_mainmem;
        int                   sr_prevnextp;
        rs_vbuf_t*            sr_vbuf;
} solid_relcur_t;

typedef struct {
        char*   srm_mysqlname;
        char*   srm_solidname;
} solid_sysrel_map_t;

/* Table to map special MySQL table names to Solid system tables. You meed to
 * create these tables in MySQL with a set of columns that are compatible
 * with solidDB system tables.
 */
static solid_sysrel_map_t solid_sysrel_map[] =
{
        { (char *)"SOLIDDB_SYS_TABLES",  (char *)RS_RELNAME_TABLES },
        { (char *)"SOLIDDB_SYS_COLUMNS", (char *)RS_RELNAME_COLUMNS },
        { (char *)"SOLIDDB_SYS_COLUMNS_AUX", (char *)RS_RELNAME_COLUMNS_AUX },
        { (char *)"SOLIDDB_SYS_KEYS",    (char *)RS_RELNAME_KEYS },
        { (char *)"SOLIDDB_SYS_SCHEMAS", (char *)RS_RELNAME_SCHEMAS },
        { (char *)"SOLIDDB_SYS_KEYPARTS", (char *)RS_RELNAME_KEYPARTS },
        { (char *)"SOLIDDB_SYS_FORKEYS", (char *)RS_RELNAME_FORKEYS },
        { (char *)"SOLIDDB_SYS_FORKEYPARTS", (char *)RS_RELNAME_FORKEYPARTS },
        { (char *)"SOLIDDB_SYS_SEQUENCES", (char *)RS_RELNAME_SEQUENCES },
        { (char *)"SOLIDDB_SYS_BLOBS", (char *)RS_RELNAME_BLOBS },
        { (char *)"SOLIDDB_SYS_CARDINAL", (char *)RS_RELNAME_CARDINAL },
        { (char *)"SOLIDDB_SYS_INFO", (char *)RS_RELNAME_INFO },
        { (char *)"SOLIDDB_SYS_TABLEMODES", (char *)RS_RELNAME_TABLEMODES },
        { NULL, NULL }
};

/* System tables */
static const char* soliddb_sysrel_names[] =
{
    RS_RELNAME_TABLES,
    RS_RELNAME_COLUMNS,
    RS_RELNAME_COLUMNS_AUX,
    RS_RELNAME_USERS,
    RS_RELNAME_UROLE,
    RS_RELNAME_RELAUTH,
    RS_RELNAME_ATTAUTH,
    RS_RELNAME_VIEWS,
    RS_RELNAME_KEYPARTS,
    RS_RELNAME_KEYS,
    RS_RELNAME_CARDINAL,
    RS_RELNAME_INFO,
    RS_RELNAME_SYNONYM,
    RS_RELNAME_TYPES,
    RS_RELNAME_SQL_LANG,
    RS_RELNAME_PROCEDURES,
    RS_RELNAME_TRIGGERS,
    RS_RELNAME_EVENTS,
    RS_RELNAME_TABLEMODES,
    RS_RELNAME_FORKEYS,
    RS_RELNAME_FORKEYPARTS,
    RS_RELNAME_SEQUENCES,
    RS_RELNAME_CATALOGS,
    RS_RELNAME_SCHEMAS,
    RS_RELNAME_PROCEDURE_COL_INFO,
    RS_RELNAME_SYSPROPERTIES,
    RS_RELNAME_HOTSTANDBY,
    RS_RELNAME_BLOBS,
    NULL
};

/* solidDB for MySQL status variables */

typedef struct export_vars_st {
        longlong soliddb_memory_statistics;
        longlong soliddb_transaction_commits;
        longlong soliddb_transaction_rollbacks;
        longlong soliddb_transaction_aborts;
        longlong soliddb_transaction_total;
        longlong soliddb_transaction_readonly;
        longlong soliddb_transaction_trxbuf;
        longlong soliddb_transaction_active;
        longlong soliddb_transaction_validate;
        long soliddb_cache_hitrate;
        longlong soliddb_cache_find;
        longlong soliddb_cache_read;
        longlong soliddb_cache_write;
        bool soliddb_indexmerge_active;
        longlong soliddb_index_writes;
        longlong soliddb_index_writesincemerge;
        longlong soliddb_log_writes;
        longlong soliddb_log_writessincecp;
        longlong soliddb_active;
        longlong soliddb_activeavg;
        longlong soliddb_filesize;
        longlong soliddb_logsize;
        bool soliddb_cp_active;
        bool soliddb_backup_active;
} export_vars_t;

export_vars_t export_vars; 

#if MYSQL_VERSION_ID >= 50100
SHOW_VAR soliddb_status_variables[] = {
#else
struct show_var_st soliddb_status_variables[] = {
#endif
    {"memory_statistics", (char *) &export_vars.soliddb_memory_statistics, SHOW_LONGLONG},
    {"transaction_commits", (char *) &export_vars.soliddb_transaction_commits, SHOW_LONGLONG},
    {"transaction_rollbacks", (char *) &export_vars.soliddb_transaction_rollbacks, SHOW_LONGLONG},
    {"transaction_aborts", (char *) &export_vars.soliddb_transaction_aborts, SHOW_LONGLONG},
    {"transaction_total", (char *) &export_vars.soliddb_transaction_total, SHOW_LONGLONG},
    {"transaction_readonly", (char *) &export_vars.soliddb_transaction_readonly, SHOW_LONGLONG},
    {"transaction_trxbuf", (char *) &export_vars.soliddb_transaction_trxbuf, SHOW_LONGLONG},
    {"transaction_activecnt", (char *) &export_vars.soliddb_transaction_active, SHOW_LONGLONG},
    {"transaction_validatecnt", (char *) &export_vars.soliddb_transaction_validate, SHOW_LONGLONG},
    {"cache_hit_rate", (char *) &export_vars.soliddb_cache_hitrate, SHOW_LONG},
    {"cache_finds", (char *) &export_vars.soliddb_cache_find, SHOW_LONGLONG},
    {"cache_reads", (char *) &export_vars.soliddb_cache_read, SHOW_LONGLONG},
    {"cache_writes", (char *) &export_vars.soliddb_cache_write, SHOW_LONGLONG},
    {"indexmerge_active", (char *) &export_vars.soliddb_indexmerge_active, SHOW_BOOL},
    {"index_writes", (char *) &export_vars.soliddb_index_writes, SHOW_LONGLONG},
    {"index_writes_after_merge", (char *) &export_vars.soliddb_index_writesincemerge, SHOW_LONGLONG},
    {"log_writes", (char *) &export_vars.soliddb_log_writes, SHOW_LONGLONG},
    {"log_writes_after_cp", (char *)&export_vars.soliddb_log_writessincecp, SHOW_LONGLONG},
    {"active_searches", (char *)&export_vars.soliddb_active, SHOW_LONGLONG},
    {"average_searches", (char *)&export_vars.soliddb_activeavg, SHOW_LONGLONG},
    {"database_size", (char *)&export_vars.soliddb_filesize, SHOW_LONGLONG}, 
    {"checkpoint_active", (char *)&export_vars.soliddb_cp_active, SHOW_BOOL},
    {"backup_active", (char *)&export_vars.soliddb_backup_active, SHOW_BOOL},
    {NullS, NullS, SHOW_LONG}
};

/* Prototypes for admin functions */

static su_ret_t do_backup( rs_sysi_t* cd, char* parameters, su_err_t** p_errh );
static su_ret_t do_performance_monitor(rs_sysi_t* cd, char* parameters, su_err_t** p_errh);
static su_ret_t do_checkpoint(rs_sysi_t* cd, char* parameters, su_err_t** p_errh);
static su_ret_t do_command(rs_sysi_t* cd, const char* cmd, const char* parameters, su_err_t** p_errh);

/* Structure to hold admin command strings and their functions */
typedef struct solid_admin_commands_st {
        const char* command_str;
        su_ret_t(*command_func)(rs_sysi_t*, char *, su_err_t**);
} solid_admin_commands_t;

/* Init admin commands table */
static solid_admin_commands_t solid_admin_commands [] =
{
    {"backup", do_backup},
    {"pmon", do_performance_monitor},
    {"checkpoint", do_checkpoint},
    {NULL, NULL }
};

static bool solid_handler_init(void);
static int solid_close_connection(THD* thd);

static int solid_commit(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton,
#endif
        THD* thd,
        bool all);

static SOLID_CONN* get_solid_ha_data_connection(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton,
#endif
        THD* thd);

static int solid_rollback(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD* thd,
        bool all);

#if MYSQL_VERSION_ID >= 50100
static uint solid_alter_table_flags(uint flags);

bool soliddb_show_status(handlerton *hton, THD* thd, stat_print_fn*, enum ha_stat_type);
#endif /* MYSQL_VERSION_ID >= 50100 I.E. 5.1 */

static const char solid_hton_name[]= "solidDB";

static const char solid_hton_comment[]=
"Fully transactional disk-based engine with multiversion optimistic/pessimistic concurrency control";

#if MYSQL_VERSION_ID >= 50100
/*
  This needs to exist until the query cache callback is removed
  or learns to pass hton.
*/
static handlerton *legacy_soliddb_hton;
#else
handlerton solid_hton= {
    solid_hton_name,
    SHOW_OPTION_YES,
    solid_hton_comment,
    DB_TYPE_SOLID_DB,
    solid_handler_init,    /* We do need to write one! */
    0,       /* slot */
    0,       /* savepoint size. */

    solid_close_connection, /* close_connection */

    NULL,    /* savepoint */
    NULL,    /* rollback to savepoint */
    NULL,    /* release savepoint */
    solid_commit,    /* commit */
    solid_rollback,    /* rollback */
    NULL,    /* prepare */
    NULL,    /* recover */
    NULL,    /* commit_by_xid */
    NULL,    /* rollback_by_xid */
    NULL,    /* create_cursor_read_view */
    NULL,    /* set_cursor_read_view */
    NULL,    /* close_cursor_read_view */
    solid_get_error_message, /* get_error_message */
    HTON_NO_FLAGS
};
#endif /* MYSQL_VERSION_ID */

static int solid_started = 0;
static bool solid_printmsg = TRUE;
static SsSemT* solid_mutex = NULL;
static SsMsgLogT* solid_msglog;
static su_inifile_t* solid_inifile = NULL;

/* Variables for example share methods */
static HASH soliddb_open_tables; // Hash used to track open tables
pthread_mutex_t soliddb_mutex;   // This is the mutex we use to init the hash
static int soliddb_init= 0;      // Variable for checking the init state of hash

/* Configuration parameters for MySQL/soliDB storage engine. Note that these
 should include default values! */

#define SS_SEMNUM_SOLIDDB_BACKUP_DIR (SS_SEMNUM_SS_LIB)

char* soliddb_logdir = NULL;
char  soliddb_backupdir_buf[FN_REFLEN] = "";
char *soliddb_backupdir = soliddb_backupdir_buf;
static SsFlatMutexT soliddb_backupdir_mutex;
char* soliddb_filespec = NULL;
char* soliddb_admin_command = NULL;
longlong soliddb_cache_size = DBE_DEFAULT_INDEXCACHESIZE;
long soliddb_durability_level = DBE_DURABILITY_STRICT;
my_bool soliddb_checkpoint_deletelog = FALSE; /* By default solidDB should delete log files
                                                only after backup. */
long soliddb_lock_wait_timeout = DBE_DEFAULT_PESSIMISTIC_LOCK_TO;
long soliddb_optimistic_lock_wait_timeout = DBE_DEFAULT_OPTIMISTIC_LOCK_TO;
long soliddb_log_block_size = DBE_DEFAULT_LOGBLOCKSIZE;
long soliddb_db_block_size = DBE_DEFAULT_INDEXBLOCKSIZE;
long soliddb_backup_block_size = DBE_DEFAULT_BACKUP_BLOCKSIZE;
longlong soliddb_checkpoint_interval = DBE_DEFAULT_CPINTERVAL;
long soliddb_io_threads = DBE_DEFAULT_NUMIOTHREADS;
long soliddb_write_threads = DBE_DEFAULT_NUMWRITERIOTHREADS;
longlong soliddb_lockhash_size = DBE_DEFAULT_LOCKHASHSIZE;
long soliddb_checkpoint_time = DBE_DEFAULT_CPMINTIME;
my_bool soliddb_log_enabled = DBE_DEFAULT_LOGENABLED;
my_bool soliddb_pessimistic = DBE_DEFAULT_PESSIMISTIC;
long soliddb_extend_increment = DBE_DEFAULT_INDEXEXTENDINCR;
long soliddb_readahead = DBE_DEFAULT_READAHEADSIZE;

static bool internal_solid_end_trans_and_stmt(
    rs_sysi_t* cd,
    tb_trans_t* trans,
    bool commitp,
    su_err_t** p_errh);
static int ha_solid_mysql_error(
                                THD *thd,
                                su_err_t* errh);
static int ha_solid_mysql_error(THD *thd,su_err_t* errh);
static int handle_error(su_err_t* errh);

static int ha_solid_mysql_error(THD *thd, su_err_t* errh);

typedef struct ddl_monitoring
{
    unsigned int number_of_backups;
    unsigned int number_of_ddls;
    SsFlatMutexT ddl_monitoring_mutex;
    SsMesT* ddl_finish_event;
} ddl_monitoring_t;

#define SS_SEMNUM_SOLIDDB_DDL_MONITOR (SS_SEMNUM_SS_LIB)

static inline bool is_soliddb_system_relation (const char *sol_schema, const char *sol_relname)
{
    bool systab = false;
    if ( !strcmp(sol_schema, SNC_SCHEMA) ) {
        for(unsigned i = 0; soliddb_sysrel_names[i]; ++i) {
            if (!strcmp(soliddb_sysrel_names[i], sol_relname)) {
                systab = true;
                break;
            }
        }
    }
    return systab;
}

static inline bool is_soliddb_system_relation (const char *full_myrelname)
{
    rs_entname_t en;
    solid_relname(&en, full_myrelname);
    bool systab = is_soliddb_system_relation(en.en_schema,en.en_name);
    rs_entname_done_buf(&en);
    return systab;
}

static inline bool is_soliddb_system_relation (const rs_entname_t &solrel)
{
    return is_soliddb_system_relation(solrel.en_schema, solrel.en_name);
}


static inline void init_ddl_monitoring( ddl_monitoring_t *ddl_monitor )
{
        ddl_monitor->number_of_backups = ddl_monitor->number_of_ddls = 0;
        SsFlatMutexInit( &ddl_monitor->ddl_monitoring_mutex, SS_SEMNUM_SOLIDDB_DDL_MONITOR );
        ddl_monitor->ddl_finish_event = SsMesCreateLocal();
        SsMesSend( ddl_monitor->ddl_finish_event );
}

static inline void destroy_ddl_monitoring( ddl_monitoring_t *ddl_monitor )
{
        SsFlatMutexDone( ddl_monitor->ddl_monitoring_mutex );
        SsMesFree( ddl_monitor->ddl_finish_event );
}

static ddl_monitoring_t solid_ddl_monitoring;

static inline void block_solid_DDL()
{
        SsFlatMutexLock( solid_ddl_monitoring.ddl_monitoring_mutex );
        if (solid_ddl_monitoring.number_of_backups < UINT_MAX) {
            solid_ddl_monitoring.number_of_backups++;
        }
        bool is_ddl_going = solid_ddl_monitoring.number_of_ddls > 0;
        SsFlatMutexUnlock( solid_ddl_monitoring.ddl_monitoring_mutex );
        if ( is_ddl_going ) {
            SsMesWait( solid_ddl_monitoring.ddl_finish_event );
        }
}

static inline void unblock_solid_DDL()
{
        SsFlatMutexLock( solid_ddl_monitoring.ddl_monitoring_mutex );
        if (solid_ddl_monitoring.number_of_backups) {
            solid_ddl_monitoring.number_of_backups--;
        }
        SsFlatMutexUnlock( solid_ddl_monitoring.ddl_monitoring_mutex );
}

static inline bool is_solid_DDL_blocked() /* is not actually going to be used */
{
        SsFlatMutexLock( solid_ddl_monitoring.ddl_monitoring_mutex );
        bool is_solid_DDL_disabled = solid_ddl_monitoring.number_of_backups > 0;
        SsFlatMutexUnlock( solid_ddl_monitoring.ddl_monitoring_mutex );
        return is_solid_DDL_disabled;
}

static inline bool start_solid_DDL( int &rc )
{
        SsFlatMutexLock( solid_ddl_monitoring.ddl_monitoring_mutex );
        bool is_solid_DDL_disabled = solid_ddl_monitoring.number_of_backups > 0;
        if ( !is_solid_DDL_disabled ) {
            if (solid_ddl_monitoring.number_of_ddls < UINT_MAX) {
                solid_ddl_monitoring.number_of_ddls++;
            }
            SsMesReset( solid_ddl_monitoring.ddl_finish_event );
        }
        SsFlatMutexUnlock( solid_ddl_monitoring.ddl_monitoring_mutex );

        if( is_solid_DDL_disabled ) {
            rc = HA_ERR_WRONG_COMMAND;
            sql_print_error( "DDL commands are temporarily disabled, "
                             "possibly because on-line backup is in progress. "
                             "Try again later.\n" );
            /* ui_msg_message(BACKUP_MSG_FAILED_S, su_err_geterrstr(*p_errh)); */
            ss_pprintf(("DDL commands are temporarily disabled, "
                        "possibly because on-line backup is in progress. "
                        "Try again later.\n"));
        }

        return !is_solid_DDL_disabled;
}

static inline void finish_solid_DDL()
{
        SsFlatMutexLock( solid_ddl_monitoring.ddl_monitoring_mutex );
        unsigned int prev_number_of_ddls = solid_ddl_monitoring.number_of_ddls;
        if (solid_ddl_monitoring.number_of_ddls) {
            solid_ddl_monitoring.number_of_ddls--;
        }
        if ( prev_number_of_ddls && !solid_ddl_monitoring.number_of_ddls ) {
            /* this DDL was the last one, now there is no DDL active */
            SsMesSend( solid_ddl_monitoring.ddl_finish_event );
        }
        SsFlatMutexUnlock( solid_ddl_monitoring.ddl_monitoring_mutex );
}

static su_err_t* lock_table(SOLID_CONN* con, rs_relh_t* tableh,
                            bool exclusive = /*exclusive*/TRUE, 
                            long timeout = /* default timeout*/ -1L );
static su_err_t* lock_table(SOLID_CONN* con, rs_entname_t* en,
                            bool exclusive = /*exclusive*/TRUE, 
                            long timeout = /* default timeout*/ -1L );

static su_err_t* lock_table(SOLID_CONN* con, rs_relh_t* tableh,
                            bool exclusive /*exclusive = TRUE*/, 
                            long timeout /* default timeout = -1L*/  )
{
    su_err_t* errh = NULL;
    solid_bool finished = FALSE;
    bool succp = FALSE;

    SDB_DBUG_ENTER("solid::lock_table");
    ss_dassert(con != NULL);
    ss_dassert(tableh != NULL);
    
    do {

        succp = tb_trans_lockrelh(con->sc_cd,
                                  con->sc_trans,
                                  tableh, 
                                  exclusive,
                                  timeout,
                                  &finished,
                                  &errh);
        
    } while (succp && !finished && rs_sysi_lockwait(con->sc_cd));
    
    SDB_DBUG_RETURN(errh);
}

static su_err_t* lock_table(SOLID_CONN* con, rs_entname_t* en,
                            bool exclusive /*exclusive = TRUE */, 
                            long timeout /* default timeout = -1L */  )
{
        su_err_t* errh = NULL;
        rs_relh_t* tableh = NULL;

        SDB_DBUG_ENTER("solid::lock_table");
        ss_dassert(con != NULL);
        ss_dassert(en != NULL);

        tableh = tb_dd_getrelh(con->sc_cd, con->sc_trans, en, NULL, &errh);
        if (tableh && !errh) {
            errh = lock_table(con, tableh, exclusive, timeout);
            rs_relh_done(con->sc_cd, tableh);
        }
        SDB_DBUG_RETURN(errh);
}

/* not used so far.
static int check_and_start_DDL(SOLID_CONN* con) 
{
    SDB_DBUG_ENTER("solid::lock_table");
    ss_dassert(con != NULL);

    THD* thd = current_thd;
    su_err_t* errh = NULL;
    int rc = 0;
    rs_entname_t en;
    solid_relname(&en, ((TABLE_LIST*)thd->lex->select_lex.table_list.first)->db,
        ((TABLE_LIST*)thd->lex->select_lex.table_list.first)->table_name);

    if (is_soliddb_system_relation(en.en_schema, en.en_name)) {
        internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, FALSE, NULL);
        sql_print_error("Trying to alter SolidDB system table %s\n", en.en_name);
        finish_solid_DDL();
        SDB_DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
    }

    // lock the table exclusively if the operation requires this.
    if (operation_requires_xtable_lock(thd) && (errh = lock_table(con, &en))) {
        internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, FALSE, NULL);
        sql_print_error("%s\n", su_err_geterrstr(errh));
        int rc = ha_solid_mysql_error(thd, errh);
        su_err_done(errh);
        finish_solid_DDL();
        SDB_DBUG_RETURN(rc);
    }
    rs_entname_done_buf(&en);
    SDB_DBUG_RETURN(0);
}
*/

/*#***********************************************************************\
 *
 *              solid_create_handler
 *
 * Create solidDB handler
 *
 * Parameters :
 *
 *      TABLE_SHARE* table, in out, use
 *      MEM_ROOT*    mem_root, in out, use
 *
 * Return value : Initialized handler interface
 *
 * Limitations: MySQL 5.1.x
 *
 * Globals used :
 */

#if MYSQL_VERSION_ID >= 50100
static handler *solid_create_handler(handlerton *hton, TABLE_SHARE *table, MEM_ROOT *mem_root)
{
  return new (mem_root) ha_soliddb(hton, table);
}
#endif

/*#***********************************************************************\
 *
 *              ha_map_solid_mysql_fkaction
 *
 * Maps a solid foreign key referential action code to MySQL foreign key
 * referential action. Prints referential action string to buffer if a
 * buffer is given.
 *
 * Parameters :
 *
 *     sqlrefact_t solid_ref_action, in, solid foreign key referential action
 *     char*       buffer, in, buffer where to print action string or NULL
 *
 * Return value : uint, MySQL foreign key referential action code
 *
 * Globals used : -
 */
static inline uint ha_map_solid_mysql_fkaction(
        sqlrefact_t solid_ref_action,
        char *buffer)
{
        uint mysql_ref_action;

        switch (solid_ref_action) {
            case SQL_REFACT_RESTRICT:
                mysql_ref_action = foreign_key::FK_OPTION_RESTRICT;

                if (buffer) {
                    strcat(buffer, "RESTRICT");
                }
                break;
            case SQL_REFACT_CASCADE:
                mysql_ref_action = foreign_key::FK_OPTION_CASCADE;

                if (buffer) {
                    strcat(buffer, "CASCADE");
                }
                break;
            case SQL_REFACT_SETNULL:
                mysql_ref_action = foreign_key::FK_OPTION_SET_NULL;

                if (buffer) {
                    strcat(buffer, "SET NULL");
                }
                break;
            case SQL_REFACT_NOACTION:
                mysql_ref_action = foreign_key::FK_OPTION_NO_ACTION;

                if (buffer) {
                    strcat(buffer, "NO ACTION");
                }
                break;
            case SQL_REFACT_SETDEFAULT:
                mysql_ref_action = foreign_key::FK_OPTION_DEFAULT;

                if (buffer) {
                    strcat(buffer, "SET DEFAULT");
                }
                break;
            default:
                mysql_ref_action = foreign_key::FK_OPTION_UNDEF;
                break;
        }

        return (mysql_ref_action);
}

static rs_relh_t* solid_getrelhbyid(
        rs_sysi_t* cd,
        tb_trans_t* trans,
        long relid,
        su_err_t** p_errh)
{
        rs_relh_t* relh;
        dbe_ret_t rc;

        /* Check if relid is not valid. */
        rc = dbe_trx_readrelh(tb_trans_dbtrx(cd, trans), relid);
        if (rc != SU_SUCCESS) {
            if (p_errh) {
                rs_error_create(p_errh, E_DDOP);
            }
            return(NULL);
        }

        /* Read the relh. */
        relh = tb_dd_getrelhbyid(cd, trans, relid, NULL, p_errh);

        return(relh);
}

/*#***********************************************************************\
 *
 *              solid_get_error_message
 *
 * Get error key, table and message
 *
 * Parameters :
 *
 *      handlerton*, hton, in, use, MySQL handlerton structure on 5.1
 *      THD*     thd, in, use, MySQL thread
 *      TABLE**  table, out, MySQL table where error occurred
 *      int      error_num, in, not used, error value
 *      String*  error_string, inout, not used, error message
 *
 * Return value : byle*  table name
 *
 * Globals used :
 */
int solid_get_error_message(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD* thd,
        TABLE **table,
        int error_num,
        String* error_string)
{
        SOLID_CONN* con = NULL;
        int key_num = 0;
        rs_relh_t* relh;
        char* relname;
        char* schema;
        char* relname_ptr;
        char* schema_ptr;
        char* orig_name;
        rs_entname_t en;
        int sysrelid;
        uint len;

        ss_dprintf_1(("solid_get_error_message\n"));

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif
        ss_dassert(con != NULL);

        SsSemEnter(con->sc_mutex);

        key_num = con->sc_errkey_num;
        *table = con->sc_table;

#if MYSQL_VERSION_ID >= 50100
        ss_dassert(con->sc_table != NULL);
        ss_dassert(con->sc_table->s != NULL);
        orig_name = (char *)con->sc_table->s->path.str;
#else
        ss_dassert(con->sc_table != NULL);
        ss_dassert(con->sc_table->s != NULL);
        orig_name = (char *)con->sc_table->s->path;
#endif

        len = strlen(orig_name) + 1;

        relname = (char *)SsMemAlloc(len);
        schema = (char *)SsMemAlloc(len);

        solid_set_relname(schema, relname, orig_name, lower_case_table_names);

        rs_entname_initbuf(&en,
                           NULL,       // catalog,
                           schema,
                           relname);

        relh = tb_dd_getrelh(con->sc_cd,
                             con->sc_trans,
                             &en,
                             NULL,       // p_priv
                             NULL);      // rs_err_t** p_errh
        ss_dassert(relh != NULL);

        SsMemFree(relname);
        SsMemFree(schema);

        if (con->sc_errkey) {
            char *foreign_str = NULL;

            if (error_num == HA_ERR_ROW_IS_REFERENCED ||
                error_num == HA_ERR_NO_REFERENCED_ROW) {

                if (rs_key_type(con->sc_cd, con->sc_errkey) == RS_KEY_FORKEYCHK) {
                    rs_ano_t j;
                    rs_ano_t k;
                    rs_ano_t n_parts;
                    rs_relh_t *refrelh;
                    rs_ano_t k_part;
                    rs_err_t* errh = NULL;
                    bool first_part = TRUE;
                    su_pa_t* forrefkeys = NULL;
                    rs_key_t* forkey = NULL;

                    refrelh = tb_dd_getrelhbyid(
                                    con->sc_cd,
                                    con->sc_trans,
                                    rs_key_refrelid(con->sc_cd, con->sc_errkey),
                                    NULL,
                                    NULL);

                    if (refrelh != NULL) {

                        foreign_str = (char *)SsMemAlloc(MAX_FOREIGN_LEN+1);

                        strcpy(foreign_str, "CONSTRAINT `");
                        strcat(foreign_str, rs_key_name(con->sc_cd, con->sc_errkey));
                        strcat(foreign_str, "` FOREIGN KEY (");
                        n_parts = rs_key_nparts(con->sc_cd, con->sc_errkey);

                        for(k_part = 0; k_part < n_parts; k_part++) {
                            rs_ano_t     kp_ano;

                            kp_ano = rs_keyp_ano(con->sc_cd, con->sc_errkey, k_part);

                            if (rs_keyp_parttype(con->sc_cd, con->sc_errkey, k_part) == RSAT_USER_DEFINED) {

                                if (first_part) {
                                    first_part = FALSE;
                                } else {
                                    strcat(foreign_str, ", ");
                                }

                                strcat(foreign_str, "`");
                                strcat(foreign_str, rs_ttype_aname(con->sc_cd, rs_relh_ttype(con->sc_cd, relh), kp_ano));
                                strcat(foreign_str, "`");
                            }
                        }

                        strcat(foreign_str, ") REFERENCES ");

                        strcat(foreign_str, "`");
                        strcat(foreign_str, rs_relh_name(con->sc_cd, refrelh));
                        strcat(foreign_str, "`");
                        strcat(foreign_str, " (");
                        first_part = TRUE;

                        tb_dd_resolverefkeys(con->sc_cd, refrelh);
                        forrefkeys = rs_relh_refkeys(con->sc_cd, refrelh);
                        k = 0;

                        su_pa_do(forrefkeys, k) {
                            forkey = (rs_key_t*)su_pa_getdata(forrefkeys, k);
                            n_parts = rs_key_nparts(con->sc_cd, forkey);

                            if (rs_key_type(con->sc_cd, forkey) != RS_KEY_FORKEYCHK &&
                                rs_key_refrelid(con->sc_cd, forkey) == rs_relh_relid(con->sc_cd, relh)) {

                                for(j = 0; j < n_parts; j++) {
                                    rs_ano_t key_ano;

                                    key_ano = rs_keyp_ano(con->sc_cd, forkey, j);

                                    if(rs_keyp_parttype(con->sc_cd, forkey, j) == RSAT_USER_DEFINED) {
                                        if (first_part) {
                                            first_part = FALSE;
                                        } else {
                                            strcat(foreign_str, ", ");
                                        }

                                        strcat(foreign_str, "`");
                                        strcat(foreign_str, rs_ttype_aname(con->sc_cd,
                                                                   rs_relh_ttype(con->sc_cd, refrelh), key_ano));
                                        strcat(foreign_str, "`");
                                    }
                                }
                            }
                        }

                        strcat(foreign_str, ") ON DELETE ");
                        ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_delete_action(con->sc_cd, con->sc_errkey), foreign_str);
                        strcat(foreign_str, " ON UPDATE ");
                        ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_update_action(con->sc_cd, con->sc_errkey), foreign_str);

                        rs_relh_done(con->sc_cd, refrelh);
                    }
                }
            }
            if (foreign_str == NULL) {
                foreign_str = (char *)SsMemStrdup((char *)rs_key_name(con->sc_cd, con->sc_errkey));
            }

            error_string->append(foreign_str);
            SsMemFree(foreign_str);
            con->sc_errkey = NULL;
        }

        con->sc_errkey_num = 0;
        con->sc_table = NULL;

        SsSemExit(con->sc_mutex);

        rs_relh_done(con->sc_cd, relh);

        return (key_num);
}

/*#***********************************************************************\
 *
 *              solid_get_key
 *
 * Create our own hash to get a key
 *
 * Parameters :
 *
 *      SOLIDDB_SHARE* share, in out, use
 *
 * Return value : byle*  table name
 *
 * Globals used :
 */
static byte* soliddb_get_key(SOLIDDB_SHARE *share,uint *length,
                             my_bool not_used __attribute__((unused)))
{
        *length=share->table_name_length;
        return (byte*) share->table_name;
}

/*#***********************************************************************\
 *
 *              check_solid_sysrel
 *
 *
 *
 * Parameters :
 *
 *      char* relname, in, use
 *
 * Return value : Index to system relation or -1
 *
 * Globals used :
 */
static int check_solid_sysrel(char* relname)
{
        int i;

        for (i = 0; solid_sysrel_map[i].srm_mysqlname != NULL; i++) {
            if (strcmp(solid_sysrel_map[i].srm_mysqlname, relname) == 0) {
                return(i);
            }
        }
        return(-1);
}

/*#***********************************************************************\
 *
 *              get_share
 *
 * Example of simple lock controls. The "share" it creates is structure we will
 * pass to each example handler. Do you have to have one of these? Well, you have
 * pieces that are used for locking, and they are needed to function.
 *
 * Parameters :
 *
 *      const char* table_name, in, use
 *      TABLE*      table
 *
 * Return value : SOLIDDB_SHARE *
 *
 * Globals used :
 */
static SOLIDDB_SHARE *get_share(const char *table_name, TABLE *table)
{
        SOLIDDB_SHARE *share;
        uint length;
        char *tmp_name;

        /*
           So why does this exist? There is no way currently to init a storage engine.
           Innodb and BDB both have modifications to the server to allow them to
           do this. Since you will not want to do this, this is probably the next
           best method.
        */
        if (!soliddb_init)
        {
            /* Hijack a mutex for init'ing the storage engine */
            pthread_mutex_lock(&LOCK_mysql_create_db);
            if (!soliddb_init)
            {
                soliddb_init++;
                VOID(pthread_mutex_init(&soliddb_mutex,MY_MUTEX_INIT_FAST));
                (void) hash_init(&soliddb_open_tables,system_charset_info,32,0,0,
                                (hash_get_key) soliddb_get_key,0,0);
            }
            pthread_mutex_unlock(&LOCK_mysql_create_db);
        }
        pthread_mutex_lock(&soliddb_mutex);
        length=(uint) strlen(table_name);

        if (!(share= ((SOLIDDB_SHARE*)hash_search(&soliddb_open_tables,
                                                (byte*) table_name,
                                                length))))
        {
            if (!(share=(SOLIDDB_SHARE *)
                    my_multi_malloc(MYF(MY_WME | MY_ZEROFILL),
                                    &share, sizeof(*share),
                                    &tmp_name, length+1,
                                    NullS)))
            {
                pthread_mutex_unlock(&soliddb_mutex);
                return NULL;
            }

            share->use_count=0;
            share->table_name_length=length;
            share->table_name=tmp_name;
            strmov(share->table_name,table_name);
            if (my_hash_insert(&soliddb_open_tables, (byte*) share))
                goto error;
            thr_lock_init(&share->lock);
            pthread_mutex_init(&share->mutex,MY_MUTEX_INIT_FAST);
        }
        share->use_count++;
        pthread_mutex_unlock(&soliddb_mutex);

        return share;

error:
        pthread_mutex_destroy(&share->mutex);
        pthread_mutex_unlock(&soliddb_mutex);
        my_free((gptr) share, MYF(0));

        return NULL;
}

/*#***********************************************************************\
 *
 *              free_share
 *
 * Free lock controls. We call this whenever we close a table. If the table had
 * the last reference to the share then we free memory associated with it.
 *
 * Parameters :
 *
 *     SOLIDDB_SHARE* share
 *
 * Return value : 0
 *
 * Globals used :
 */
static int free_share(SOLIDDB_SHARE *share)
{
        pthread_mutex_lock(&soliddb_mutex);

        if (!--share->use_count)
        {
            hash_delete(&soliddb_open_tables, (byte*) share);
            thr_lock_delete(&share->lock);
            pthread_mutex_destroy(&share->mutex);
            my_free((gptr) share, MYF(0));
        }

        pthread_mutex_unlock(&soliddb_mutex);

        return 0;
}

/*#***********************************************************************\
 *
 *              solid_message_fp
 *
 *
 *
 * Parameters :
 *
 *     ui_msgtype_t type
 *     char*        msg
 *     int          newline
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_message_fp(ui_msgtype_t type, su_msgret_t msgcode, char* msg, int newline)
{
        if (newline && msg[strlen(msg)-1] == '\n') {
            newline = FALSE;
        }

        if (solid_printmsg) {
            if (newline) {
                printf("%s\n", msg);
            } else {
                printf("%s", msg);
            }
        }

        if (solid_msglog != NULL) {
            char *format;

            format = (char *)SsMemAlloc(SS_MSGLOG_BUFSIZE-100+1);

            if (newline) {
                SsSprintf(format, (char *)"%%.%ds\n", SS_MSGLOG_BUFSIZE-100);
            } else {
                SsSprintf(format, (char *)"%%.%ds", SS_MSGLOG_BUFSIZE-100);
            }
            SsMsgLogPrintfWithTime(solid_msglog, format, msg);
            SsMsgLogFlush(solid_msglog);

            SsMemFree(format);
        }
}

/*#***********************************************************************\
 *
 *              assert_message_fp
 *
 *
 *
 * Parameters :
 *
 *     char*        msg
 *
 * Return value : -
 *
 * Globals used :
 */
static void assert_message_fp(const char* msg)
{
        solid_printmsg = TRUE;
        solid_message_fp(UI_MSG_ERROR, SRV_MSG_ASSERTION, (char *)msg, TRUE);
}

/*#***********************************************************************\
 *
 *              solid_set_logdir
 *
 * Set up directory for log files.
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use
 *     const char*   soliddb_logdir, in, use, directory name
 *
 * Return value : TRUE if successfull
 *                FALSE otherwice
 *
 * Globals used :
 */
static bool solid_set_logdir(
        su_inifile_t* inifile,
        const char *  soliddb_logdir)
{
        ss_pprintf_1(("LogDir=%s\n", soliddb_logdir));

        su_inifile_putstring(inifile, (char *)SU_DBE_LOGSECTION,
                            (char *)SU_DBE_LOGDIR, soliddb_logdir);

        return (TRUE);
}

/*#***********************************************************************\
 *
 *              solid_set_backupdir
 *
 * Set up directory for backup files.
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use, solid configuration
 *     const char*   soliddb_backupdir, in, use, backup directory
 *
 * Return value : TRUE
 *
 * Globals used :
 */
static bool solid_set_backupdir(
        su_inifile_t* inifile,
        const char *  soliddb_backupdir)
{
        ss_pprintf_1(("BackupDir=%s\n", soliddb_backupdir));

        su_inifile_putstring(inifile, (char *)SU_DBE_GENERALSECTION,
                            (char *)SU_DBE_BACKUPDIR, soliddb_backupdir);

        return (TRUE);
}

/*#***********************************************************************\
 *
 *              soliddb_check_backupdir
 *
 * Check directory for backup files.
 *
 * Parameters :
 *
 *     THD*      thd, in, use
 *     set_var*  var, in, use, variable value from MySQL
 *
 * Return value : 0
 *
 * Globals used :
 */
int soliddb_check_backupdir(
        THD* thd,
        set_var* var)
{
        ss_pprintf_1(("BackupDir= [%s] [%s]\n", var->value->name, var->value->str_value.c_ptr_quick() ));

        return var ? 0 : -1;
}

/*#***********************************************************************\
 *
 *              solid_update_backupdir
 *
 * Update directory for backup files.
 *
 * Parameters :
 *
 *     THD*      thd, in, use
 *     set_var*  var, in, use
 *
 * Return value : TRUE if successfull
 *                FALSE otherwice
 *
 * Globals used :
 */
bool soliddb_update_backupdir(
        THD* thd,
        set_var* var)
{
        char* new_value;

        ss_pprintf_1(("BackupDir=%s\n", var->value->name));

        new_value = var->value->str_value.c_ptr_quick();
        if (var->value->null_value || !new_value) {
            new_value = (char*)"";
        }

        new_value = SsStrTrimLeft(new_value);
        SsStrTrimRight(new_value);

        SsFlatMutexLock(soliddb_backupdir_mutex);
        strcpy(soliddb_backupdir_buf, new_value);
        SsFlatMutexUnlock(soliddb_backupdir_mutex);

        return 0;
}

/*#***********************************************************************\
 *
 *              solid_parseandset_filespec
 *
 * Set up paths for database files given on a MySQL configuration file
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use, solid configuration
 *     const char *  soliddb_filespec, in, use, database file specification
 *
 * Return value : TRUE in case of success
 *                FALSE on error
 *
 * Globals used :
 */
static bool solid_parseandset_filespec(
        su_inifile_t* inifile,
        const char*   soliddb_filespec)
{
        char* FileSpec = NULL;
        char* FileSize = NULL;
        char* FileName = NULL;
        char* InputString = NULL;
        char* saveptr = NULL;
        char* saveptr2 = NULL;
        char* NewFileSpec = NULL;
        char* OriginalPtr = NULL;
        int   n_filespec = 1;
        int   len;

        ss_dassert(soliddb_filespec != NULL);
        ss_dassert(inifile != NULL);

        len = strlen(soliddb_filespec);
        InputString = (char *)SsMemAlloc(len+2);
        OriginalPtr = InputString;
        memcpy(InputString, soliddb_filespec,len+1);

        FileSpec = strtok_r(InputString, "," , &saveptr);

        if (FileSpec == NULL) {
            return (FALSE);
        }

        while (FileSpec) {
            FileName = strtok_r(FileSpec, " ", &saveptr2);
            FileSize = strtok_r(NULL, ",", &saveptr2);

            if (FileName && FileSize) {
                char* KeyName;

                NewFileSpec = (char *)SsMemAlloc(strlen(FileSpec) + strlen(FileSize) + 2);
                sprintf(NewFileSpec, "%s %s",FileSpec, FileSize);
                ss_pprintf_1(((char *)"FileSpec_%u=%s\n", n_filespec, NewFileSpec));
                KeyName = (char *)SsMemAlloc(strlen(SU_DBE_FILESPEC) + 32);
                sprintf(KeyName, SU_DBE_FILESPEC, n_filespec);
                ss_pprintf_1(((char *)"KeyName=%s\n", KeyName));
                su_inifile_putstring(inifile, (char *)SU_DBE_INDEXSECTION, KeyName, NewFileSpec);
                n_filespec++;

                SsMemFree(KeyName);
                SsMemFree(NewFileSpec);

                FileSpec = strtok_r(NULL, ",", &saveptr);
            } else {
                SsMemFree(OriginalPtr);
                return (FALSE);
            }
        }

        SsMemFree(OriginalPtr);

        return (TRUE);
}

/*#***********************************************************************\
 *
 *              solid_set_longlong
 *
 * Set up longlong parameter given in a MySQL configuration file.
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use
 *     longlong      value, in, use, configuration parameter value
 *     const char*   section, in, use, section on inifile
 *     const char*   param, in, use, parameter in inifile
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_set_longlong(
        su_inifile_t* inifile,
        longlong      value,
        const char*   section,
        const char*   param)
{
        char* string;

        string = (char *)SsMemAlloc(64);
        sprintf(string, "%lld", value);
        ss_pprintf(("%s=%s\n", param, string));

        su_inifile_putstring(inifile, section, param, string);

        SsMemFree(string);
}

/*#***********************************************************************\
 *
 *              solid_set_bool
 *
 * Set up bool parameter given in a MySQL configuration file
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use
 *     bool          value, in, use, configuration parameter value
 *     const char*   section, in, use, section on inifile
 *     const char*   param, in, use, parameter in inifile
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_set_bool(
        su_inifile_t* inifile,
        bool          value,
        const char*   section,
        const char*   param)
{
        char* string;

        string = (char *)SsMemAlloc(4);

        if (value) {
            sprintf(string, "yes");
        } else {
            sprintf(string, "no");
        }

        ss_pprintf(("%s=%s\n", param, string));
        su_inifile_putstring(inifile, section, param, string);

        SsMemFree(string);
}

/*#***********************************************************************\
 *
 *              solid_set_long
 *
 * Set up long parameter given in a MySQL configuration file
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use
 *     long          value, in, use, configuration parameter value
 *     const char*   section, in, use, section on inifile
 *     const char*   param, in, use, parameter in inifile
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_set_long(
        su_inifile_t* inifile,
        long          value,
        const char*   section,
        const char*   param)
{
        char* string;

        string = (char *)SsMemAlloc(64);

        sprintf(string, "%ld", value);
        ss_pprintf(("%s=%s\n", param, string));

        su_inifile_putstring(inifile, section, param, string);

        SsMemFree(string);
}

/*#***********************************************************************\
 *
 *              solid_set_block_size
 *
 * Set up log block size parameter given in a MySQL configuration file
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use
 *     long          block_size, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_set_block_size(
        su_inifile_t* inifile,
        long          block_size)
{
        char* blocksize_string;

        blocksize_string = (char *)SsMemAlloc(64);
        sprintf(blocksize_string, "%ld", block_size);
        ss_pprintf(("BlockSize=%s\n", blocksize_string));

        su_inifile_putstring(inifile, SU_DBE_LOGSECTION, SU_DBE_BLOCKSIZE,
                             blocksize_string);

        SsMemFree(blocksize_string);
}

/*#***********************************************************************\
 *
 *              solid_inifile_saveas
 *
 * Save solidDB configuration to a file
 *
 * Parameters :
 *
 *     su_inifile_t* inifile, in out, use, solid configuration,
 *     char*         fname, in, use, file name
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_inifile_saveas(
        su_inifile_t* inifile,
        char*         fname)
{
        su_vfile_t *vfp = NULL;
        SS_FILE* fp = NULL;

        vfp = su_vfp_init_txt(fname, (char *)"w");
        fp = su_vfp_access(vfp);
        su_inifile_savefp(inifile, fp);
        su_vfp_done(vfp);
}

/*#***********************************************************************\
 *
 *              solid_startedif
 *
 * Start solidDB server if it has not been started.
 *
 * Parameters : -
 *
 * Return value : TRUE if solidDB started
 *                FALSE if solidDB can't be started
 *
 * Globals used :
 */
static bool solid_startedif(void)
{
        bool success = TRUE;

        if (!solid_started) {
            su_inifile_t* inifile;
            bool succp = TRUE;

            tb_globalinit();

            SS_PUSHNAME("solid_startedif");

            SsSetAssertMessageHeader((char *)"solidDB Error");
            SsSetAssertMessageFunction(assert_message_fp);

            ui_msg_setdba((char *)"mysql", (char *)"mysql");
            ui_msg_setdefcatalog((char *)"dba");

            solid_printmsg = TRUE;
            ui_msg_setmessagefp(solid_message_fp);
            solid_msglog = SsMsgLogInit((char *)SU_SOLID_MSGFILENAME, 1024*1024);

            ui_msg_message(SRV_MSG_VERSION_S, SsGetVersionstring(FALSE));
            ui_msg_message(0, "This is a solidDB for MySQL build.");
            ui_msg_message(SRV_MSG_COPYRIGHT_S, ss_copyright);

            /* Create a inifile and read the contents of the inifile */
            inifile = su_inifile_init(SU_SOLINI_FILENAME, NULL);
            ss_pprintf_1(("solid_startedif:tb_init\n"));

            /* Now set MySQL/solidDB startup options to inifile. This means
               that MySQL startup options overdrive Solid configuration
               options. */

            /* Set up CacheSize */
            solid_set_longlong(inifile, soliddb_cache_size,
                               SU_DBE_INDEXSECTION, SU_DBE_CACHESIZE);

            /* Set up CheckPointDeleteLog */
            solid_set_bool(inifile, soliddb_checkpoint_deletelog,
                           SU_DBE_GENERALSECTION, SU_DBE_CPDELETELOG);

            /* Set up LogEnabled */
            solid_set_bool(inifile, soliddb_log_enabled,
                           SU_DBE_LOGSECTION, SU_DBE_LOGENABLED);

            /* Set up Pessimistic */
            solid_set_bool(inifile, soliddb_pessimistic,
                           SU_DBE_GENERALSECTION, SU_DBE_PESSIMISTIC);

            /* Set up LockWaitTimeout */
            solid_set_long(inifile, soliddb_lock_wait_timeout,
                           SU_DBE_GENERALSECTION, SU_DBE_PESSIMISTIC_LOCK_TO);

            solid_set_long(inifile, soliddb_optimistic_lock_wait_timeout,
                           SU_DBE_GENERALSECTION, SU_DBE_OPTIMISTIC_LOCK_TO);

            /* Set up DurabilityLevel */
            solid_set_long(inifile, soliddb_durability_level,
                           SU_DBE_LOGSECTION, SU_DBE_DURABILITYLEVEL);

            /* Set up BlockSize */
            if (soliddb_log_block_size == 0) {
                soliddb_log_block_size = DBE_DEFAULT_LOGBLOCKSIZE;
            }

            solid_set_long(inifile, soliddb_log_block_size,
                           SU_DBE_LOGSECTION, SU_DBE_BLOCKSIZE);

            if (soliddb_db_block_size == 0) {
                soliddb_db_block_size = DBE_DEFAULT_INDEXBLOCKSIZE;
            }

            solid_set_long(inifile, soliddb_db_block_size,
                           SU_DBE_INDEXSECTION, SU_DBE_BLOCKSIZE);

            if (soliddb_backup_block_size == 0) {
                soliddb_backup_block_size = DBE_DEFAULT_BACKUP_BLOCKSIZE;
            }

            solid_set_long(inifile, soliddb_backup_block_size,
                           SU_DBE_GENERALSECTION, SU_DBE_BACKUP_BLOCKSIZE);

            /* Set up CheckPointInterval */
            solid_set_longlong(inifile, soliddb_checkpoint_interval,
                               SU_DBE_GENERALSECTION, SU_DBE_CPINTERVAL);

            /* Set up I/O threads */
            solid_set_long(inifile, soliddb_io_threads,
                           SU_DBE_GENERALSECTION, SU_DBE_NUMIOTHREADS);

            /* Set up LockHashSize */
            solid_set_long(inifile, soliddb_lockhash_size,
                               SU_DBE_GENERALSECTION, SU_DBE_LOCKHASHSIZE);

            /* Set up MinCheckpointTime */
            solid_set_long(inifile, soliddb_checkpoint_time,
                           SU_DBE_GENERALSECTION, SU_DBE_CPMINTIME);

            /* Set up WriteThreads */
            solid_set_long(inifile, soliddb_write_threads,
                           SU_DBE_GENERALSECTION, SU_DBE_NUMWRITERIOTHREADS);

            /* Set up ExtendIncrement */
            solid_set_long(inifile, soliddb_extend_increment,
                           SU_DBE_LOGSECTION, SU_DBE_EXTENDINCR);

            /* Set up ReadAhead */
            solid_set_long(inifile, soliddb_readahead,
                           SU_DBE_LOGSECTION, SU_DBE_READAHEADSIZE);

            /* Set up LogDir. */
            if (soliddb_logdir != NULL) {
                success = solid_set_logdir(inifile, soliddb_logdir);
            }

            /* Set up Backup directory. */
            if (soliddb_backupdir != soliddb_backupdir_buf) {
                if (soliddb_backupdir != NULL && soliddb_backupdir[0]) {
                    strncpy(soliddb_backupdir_buf,soliddb_backupdir,sizeof(soliddb_backupdir_buf)-1);
                }
                soliddb_backupdir = soliddb_backupdir_buf;
                success = solid_set_backupdir(inifile, soliddb_backupdir);
            }

            /* Parse and set up FileSpec */

            if (success && soliddb_filespec != NULL) {
                success = solid_parseandset_filespec(inifile, soliddb_filespec);
            }

            if (success) {
                SsFlatMutexInit( &soliddb_backupdir_mutex, SS_SEMNUM_SOLIDDB_BACKUP_DIR );
                init_ddl_monitoring( &solid_ddl_monitoring );
                solid_started = tb_srv_init(inifile, NULL);
                solid_inifile = inifile;
            }

#ifdef SS_DEBUG
            solid_inifile_saveas(inifile, (char *)"mysql_soliddb_inifile.out");
#endif

            if (solid_started) {
                /* Create a global mutex for relation cursor list access.
                 * TODO: Semnum is not exactly correct but works ok here.
                 */
                solid_mutex = SsSemCreateLocal(SS_SEMNUM_SSE_LOCSRV);
                succp = tb_srv_start(tb_getlocal());
                ss_dassert(succp);
                ui_msg_message(SRV_MSG_DB_STARTED);
                solid_printmsg = FALSE;
                SS_POPNAME;
            } else {
                SsMsgLogT* msglog;
                SS_POPNAME;
                solid_printmsg = TRUE;
                msglog = solid_msglog;
                solid_msglog = NULL;
                SsMsgLogDone(msglog);
                tb_globaldone();
                su_inifile_done(inifile);
            }
        }

        return(solid_started);
}

/*#***********************************************************************\
 *
 *              solid_end
 *
 * Shutdown solidDB server
 *
 * Parameters : unused
 *
 * Return value : 0
 *
 * Globals used :
 */
#if MYSQL_VERSION_ID >= 50100
static int solid_end(
        handlerton *hton,
        ha_panic_function type)
#else
       int solid_end(void)
#endif
{
        DBUG_ENTER("solid_end");
        ss_pprintf_1(("solid_end\n"));

        if (solid_started) {
            SsMsgLogT* msglog;
            ss_pprintf_1(("Stop solid server...\n"));
            solid_printmsg = TRUE;
            tb_srv_stop();
            ui_msg_message(SRV_MSG_SHUTDOWN_S, SS_SERVER_NAME);
            SsSemFree(solid_mutex); solid_mutex = NULL;
            SsFlatMutexDone( soliddb_backupdir_mutex );
            destroy_ddl_monitoring( &solid_ddl_monitoring );
            msglog = solid_msglog;
            solid_msglog = NULL;
            SsMsgLogDone(msglog);

            /* it seems all deinitialization should not appear
             * after following lines as the solid's heap is destroyed
             * in these lines. To this point all solid's dynamically
             * allocated objects must be freed.
             */
            tb_srv_done();
            ss_debug(SsMemChkPrintList());
            tb_globaldone();

            solid_started = 0;
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              solid_handler_init
 *
 * Initialize solidDB server
 *
 * Parameters : -
 *
 * Return value : 0 if success
 *                1 on error
 *
 * Globals used :
 */
static bool solid_handler_init(void)
{
        ss_pprintf_1(("solid_handler_init\n"));

#if MYSQL_VERSION_ID < 50100
        if (have_soliddb != SHOW_OPTION_YES) {
            return(1); /* solidDB disabled */
        }
#endif

        solid_startedif();

        if (solid_started) {
            return(0);
        }

        return(1);
}

/*#***********************************************************************\
 *
 *              solid_link_connection
 *
 * Link to connection
 *
 * Parameters :
 *
 *     SOLID_CONN* con, in out, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_link_connection(SOLID_CONN* con)
{
        CHK_CONN(con);
        ss_dassert(con->sc_nlink >= 1);
        SsSemEnter(solid_mutex);
        con->sc_nlink++;
        SsSemExit(solid_mutex);
}

static void free_constr_list(rs_sysi_t* cd, su_list_t* list)
{
        su_list_node_t* listnode;
        rs_cons_t* cons;

        ss_dprintf_3(("free_constr_list\n"));

        su_list_do(list, listnode) {
            cons = (rs_cons_t*)su_listnode_getdata(listnode);
            rs_cons_done(cd, cons);
        }
        su_list_done(list);
}

static void free_end_of_constr_list(rs_sysi_t* cd, su_list_t* list, su_list_node_t* last_node)
{
        su_list_node_t* listnode;
        rs_cons_t* cons;

        ss_dprintf_3(("free_end_of_constr_list\n"));
        ss_dassert(last_node != NULL);

        for (;;) {
            listnode = su_list_last(list);
            ss_assert(listnode != NULL);
            cons = (rs_cons_t*)su_listnode_getdata(listnode);
            rs_cons_done(cd, cons);
            su_list_remove(list, listnode);
            if (listnode == last_node) {
                break;
            }
        }
}

/*#***********************************************************************\
 *
 *              solid_relcur_free_cursor
 *
 * Free solidDB relation cursor
 *
 * Parameters :
 *
 *     rs_sysi_t*      cd, in out, use
 *     solid_relcur_t* solid_relcur, in out, use
 *     SOLID_CONN*     con, in out, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_relcur_free_cursor(
        rs_sysi_t* cd,
        solid_relcur_t* solid_relcur,
        SOLID_CONN* con)
{
        ss_dassert(con != NULL);
        CHK_CONN(con);

        SsSemEnter(con->sc_mutex);
        CHK_CONN(con);

        if (solid_relcur->sr_relcurlist != NULL) {
            su_list_remove(solid_relcur->sr_relcurlist, solid_relcur->sr_relcurlistnode);
            solid_relcur->sr_relcurlist = NULL;
            solid_relcur->sr_relcurlistnode = NULL;
        } else {
            ss_dassert(solid_relcur->sr_relcurlistnode == NULL);
        }
        if (solid_relcur->sr_fetchtval != NULL) {
            ss_debug(rs_tval_resetexternalflatva(cd, solid_relcur->sr_ttype, solid_relcur->sr_fetchtval));
            if (!solid_relcur->sr_mainmem) {
                rs_tval_free(cd, solid_relcur->sr_ttype, solid_relcur->sr_fetchtval);
            }
            solid_relcur->sr_fetchtval = NULL;
        }
        if (solid_relcur->sr_bkeybuf != NULL) {
            dbe_bkey_done_ex(cd, solid_relcur->sr_bkeybuf);
            solid_relcur->sr_bkeybuf = NULL;
        }
        if (solid_relcur->sr_relcur != NULL) {
            ss_pprintf_2(("solid_relcur_free_cursor:close cursor\n"));
            if (solid_relcur->sr_constraints != NULL) {
                free_constr_list(con->sc_cd, solid_relcur->sr_constraints);
                solid_relcur->sr_constraints = NULL;
            }
            if (solid_relcur->sr_constval != NULL) {
                rs_tval_free(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_constval);
                solid_relcur->sr_constval = NULL;
            }
            if (solid_relcur->sr_postval != NULL) {
                rs_tval_free(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_postval);
                solid_relcur->sr_postval = NULL;
            }
            if (solid_relcur->sr_vbuf != NULL) {
                rs_vbuf_done(con->sc_cd, solid_relcur->sr_vbuf);
                solid_relcur->sr_vbuf = NULL;
            }
            SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_CLOSE);
            dbe_cursor_done(solid_relcur->sr_relcur, NULL);
            solid_relcur->sr_relcur = NULL;
        }
        if (solid_relcur->sr_pla != NULL) {
            rs_pla_done(cd, solid_relcur->sr_pla);
            solid_relcur->sr_pla = NULL;
            solid_relcur->sr_plakey = NULL;
        }
        SsSemExit(con->sc_mutex);
}

/*#***********************************************************************\
 *
 *              solid_relcur_free
 *
 * Free solidDB relation
 *
 * Parameters :
 *
 *     rs_sysi_t*      cd, in out, use
 *     solid_relcur_t* solid_relcur, in out, use
 *     SOLID_CONN*     con, in out, use
 *
 * Return value: -
 *
 * Globals used :
 */
static void solid_relcur_free(
        rs_sysi_t* cd,
        solid_relcur_t* solid_relcur,
        SOLID_CONN* con,
        bool add_to_donelist)
{
        ss_pprintf_1(("solid_relcur_free:bool add_to_donelist=%d\n", add_to_donelist));

        if (solid_relcur != NULL) {
            ss_pprintf_2(("solid_relcur_free:free cursor %ld\n", (long)solid_relcur));
            CHK_RELCUR(solid_relcur);
            if (add_to_donelist) {
                SsSemEnter(con->sc_mutex);
                if (con->sc_relcurdonelist == NULL) {
                    con->sc_relcurdonelist = su_list_init(NULL);
                    su_list_startrecycle(con->sc_relcurdonelist);
                }
                su_list_insertfirst(con->sc_relcurdonelist, solid_relcur);
                SsSemExit(con->sc_mutex);
            } else {
                solid_relcur_free_cursor(cd, solid_relcur, con);
                solid_relcur->sr_chk = 0;
                rs_relh_done(cd, solid_relcur->sr_rsrelh);
                SsMemFree(solid_relcur->sr_selectlist);
                SsMemFree(solid_relcur->sr_usedfields);
                ss_dassert(solid_relcur->sr_pla == NULL);
                SsMemFree(solid_relcur);
            }
        }
}

/*#***********************************************************************\
 *
 *		solid_relcur_clearrelcurdonelist
 *
 * Clears all relation cursors that are added to donelist.
 *
 * Parameters :
 *
 *		con -
 *
 *
 * Return value :
 *
 * Comments :
 *
 * Globals used :
 *
 * See also :
 */
static void solid_relcur_clearrelcurdonelist(SOLID_CONN* con, int maxlen)
{
        solid_relcur_t* solid_relcur;

        ss_pprintf_3(("solid_relcur_clearrelcurdonelist\n"));

        while (su_list_length(con->sc_relcurdonelist) > (uint)maxlen) {
            SsSemEnter(con->sc_mutex);
            solid_relcur = (solid_relcur_t*)su_list_removelast(con->sc_relcurdonelist);
            SsSemExit(con->sc_mutex);
            if (solid_relcur != NULL) {
                solid_relcur_free(con->sc_cd, solid_relcur, con, FALSE);
            } else {
                break;
            }
        }
}

/*#***********************************************************************\
 *
 *              solid_free_connection_nomutex
 *
 * Free solid connection
 *
 * Parameters :
 *
 *     SOLID_CONN* con, in out, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_free_connection_nomutex(SOLID_CONN* con)
{
        solid_relcur_t* solid_relcur;

        ss_pprintf_1(("solid_free_connection_nomutex:con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
        CHK_CONN(con);

        if (con->sc_relcurlist != NULL) {
            while ((solid_relcur = (solid_relcur_t*)su_list_removefirst(con->sc_relcurlist)) != NULL) {
                ss_pprintf_2(("solid_free_connection_nomutex:close cursor, con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
                CHK_RELCUR(solid_relcur);
                ss_dassert(solid_relcur->sr_relcurlist != NULL);
                solid_relcur->sr_relcurlist = NULL;
                solid_relcur->sr_relcurlistnode = NULL;
                SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_CLOSE);
                if (solid_relcur->sr_fetchtval != NULL) {
                    ss_debug(rs_tval_resetexternalflatva(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_fetchtval));
                    if (!solid_relcur->sr_mainmem) {
                        rs_tval_free(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_fetchtval);
                    }
                    solid_relcur->sr_fetchtval = NULL;
                }
                if (solid_relcur->sr_bkeybuf != NULL) {
                    dbe_bkey_done_ex(con->sc_cd, solid_relcur->sr_bkeybuf);
                    solid_relcur->sr_bkeybuf = NULL;
                }
                if (solid_relcur->sr_relcur != NULL) {
                    if (solid_relcur->sr_constraints != NULL) {
                        free_constr_list(con->sc_cd, solid_relcur->sr_constraints);
                        solid_relcur->sr_constraints = NULL;
                    }
                    if (solid_relcur->sr_constval != NULL) {
                        rs_tval_free(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_constval);
                        solid_relcur->sr_constval = NULL;
                    }
                    if (solid_relcur->sr_postval != NULL) {
                        rs_tval_free(con->sc_cd, solid_relcur->sr_ttype, solid_relcur->sr_postval);
                        solid_relcur->sr_postval = NULL;
                    }
                    if (solid_relcur->sr_vbuf != NULL) {
                        rs_vbuf_done(con->sc_cd, solid_relcur->sr_vbuf);
                        solid_relcur->sr_vbuf = NULL;
                    }
                    dbe_cursor_done(solid_relcur->sr_relcur, NULL);
                    solid_relcur->sr_relcur = NULL;
                    rs_pla_done(con->sc_cd, solid_relcur->sr_pla);
                    solid_relcur->sr_pla = NULL;
                    solid_relcur->sr_plakey = NULL;
                }
            }
            su_list_done(con->sc_relcurlist);
            con->sc_relcurlist = NULL;
        }
        if (con->sc_tbcon != NULL) {
            tb_srv_donecd(con->sc_cd);
            tb_disconnect(con->sc_tbcon);
            con->sc_tbcon = NULL;
        }
        if (con->sc_errh != NULL) {
            su_err_done(con->sc_errh);
            con->sc_errh = NULL;
        }
}

/*#***********************************************************************\
 *
 *              solid_unlink_connection
 *
 * Unlink from connection
 *
 * Parameters :
 *
 *     SOLID_CONN* con, in out, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void solid_unlink_connection(SOLID_CONN* con)
{
        ss_pprintf_1(("solid_unlink_connection:con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
        CHK_CONN(con);

        SsSemEnter(solid_mutex);
        ss_dassert(con->sc_nlink > 0);
        con->sc_nlink--;
        if (con->sc_nlink == 0) {
            ss_pprintf_1(("solid_unlink_connection:physically free connection, con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
            CHK_CONN(con);
            if (con->sc_relcurdonelist != NULL) {
                solid_relcur_clearrelcurdonelist(con, 0);
                su_list_done(con->sc_relcurdonelist);
                con->sc_relcurdonelist = NULL;
            }
            SsSemEnter(con->sc_mutex);
            solid_free_connection_nomutex(con);
            ss_dassert(con->con_n_tables == 0);
            con->sc_chk = 0;

            con->sc_errkey = NULL;
            con->sc_errkey_num = 0;

            SsSemExit(con->sc_mutex);
            SsSemFree(con->sc_mutex);
            SsMemFree(con);
        }
        SsSemExit(solid_mutex);
}

/*#***********************************************************************\
 *
 *              solid_clear_connection
 *
 * Clear connection
 *
 * Parameters :
 *
 *     SOLID_CONN*     con, in, use
 *     rs_relh_t*      rsrelh, in, use
 *     solid_relcur_t* solid_relcur, in, use
 *
 * Return value : 0
 *
 * Globals used :
 */
static void solid_clear_connection(
        SOLID_CONN* con,
        rs_relh_t* rsrelh,
        solid_relcur_t* solid_relcur)
{
        ss_pprintf_1(("solid_clear_connection\n"));
        CHK_CONN(con);
        solid_relcur_free(con->sc_cd, solid_relcur, con, TRUE);
        if (rsrelh != NULL) {
            rs_relh_done(con->sc_cd, rsrelh);
        }
        solid_unlink_connection(con);
}

/*#***********************************************************************\
 *
 *              get_solid_ha_data_connection
 *
 * Gets a connection from thd if exists or generates a new connection.
 *
 * Parameters :
 *
 *     THD* thd, in, use
 *
 * Return value :
 *
 *     SOLID_CONN* con if connection established or NULL
 *
 * Globals used :
*/
static SOLID_CONN* get_solid_ha_data_connection(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton,
#endif
        THD* thd)
{
        SOLID_CONN* con;
        uint solid_ha_data_slot =
            #if MYSQL_VERSION_ID >= 50100
                    hton->slot;
            #else
                    solid_hton.slot;
            #endif

        ss_pprintf_1(("get_solid_ha_data_connection\n"));

        if (thd == NULL) {
            ss_pprintf_2(("get_solid_ha_data_connection:thd == NULL\n"));
            return(NULL);
        }

        con = (SOLID_CONN*)thd->ha_data[solid_ha_data_slot];

        if (con == NULL) {
            /* No connection yet in ha_data, generate a new connection.
             */

            SsThrRegister();

            ss_pprintf_1(("get_solid_ha_data_connection:tb_connect_local\n"));
            con = (SOLID_CONN*)SsMemAlloc(sizeof(struct st_solid_connection));

            con->con_n_tables = 0;
            con->sc_isvalid = TRUE;
            con->sc_nlink = 1;
            con->sc_errh = NULL;
            con->sc_chk = CHKVAL_CONN;
            con->sc_errkey_num = 0;
            con->sc_table = NULL;
            con->sc_errkey = NULL;

            con->sc_tbcon = tb_connect_local(-1, (char *)"mysql", (char *)"mysql");
            ss_dassert(con->sc_tbcon != NULL);

            con->sc_cd = tb_getclientdata(con->sc_tbcon);
            ss_dassert(con->sc_cd != NULL);
            ss_dassert(rs_sysi_getconnecttype(con->sc_cd) == RS_SYSI_CONNECT_USER);

            tb_srv_initcd(con->sc_cd);

            rs_sysi_setflag(con->sc_cd, RS_SYSI_FLAG_MYSQL);
            rs_sysi_setallowduplicatedelete(con->sc_cd, TRUE);

            con->sc_trans = tb_getsqltrans(con->sc_tbcon);
            ss_dassert(con->sc_trans != NULL);

            con->sc_relcurlist = su_list_init(NULL);
            su_list_startrecycle(con->sc_relcurlist);

            con->sc_relcurdonelist = NULL;

            con->sc_mutex = SsSemCreateLocal(SS_SEMNUM_SSE_CONNECT);

            thd->ha_data[solid_ha_data_slot] = con;

            SS_PMON_ADD(SS_PMON_MYSQL_CONNECT);
        }
        CHK_CONN(con);
        return(con);
}

/*#***********************************************************************\
 *
 *              ::get_solid_connection
 *
 * Get connection to solidDB
 *
 * Parameters :
 *
 *     THD*        thd, in, use
 *     const char* table_name, in, use
 *
 * Return value : 0
 *
 * Globals used :
 */
SOLID_CONN* ha_soliddb::get_solid_connection(const char* table_name)
{
        SOLID_CONN* con;
        SOLID_CONN* old_con = NULL;
        struct rsrelhandlestruct* old_rsrelh = NULL;
        solid_relcur_t* old_relcur = NULL;

        THD* thd = current_thd;

        ss_pprintf_1(("ha_solid::get_solid_connection\n"));

        if (thd == NULL) {
            ss_pprintf_1(("ha_solid::get_solid_connection:thd == NULL\n"));
            ss_dassert(solid_conn == NULL);
            return(NULL);
        }

        con = (SOLID_CONN*)thd->ha_data[this->ht->slot];

        if (con != solid_conn && solid_conn != NULL) {
            ss_pprintf_1(("ha_solid::get_solid_connection:con != solid_conn\n"));
            ss_dassert(solid_conn != NULL);
            CHK_CONN(solid_conn);
            solid_clear_connection(solid_conn, solid_rsrelh, solid_relcur);
            solid_conn = NULL;
            solid_rsrelh = NULL;
            solid_relcur = NULL;
        }

        con = solid_conn;

        if (con != NULL && !con->sc_isvalid) {
            /* We have a connection in handler that is no
             * longer valid.
             */
            ss_pprintf_1(("ha_solid::get_solid_connection:invalid connection in handler\n"));
            CHK_CONN(solid_conn);
            old_con = solid_conn;
            old_rsrelh = solid_rsrelh;
            old_relcur = solid_relcur;

            solid_conn = NULL;
            solid_rsrelh = NULL;
            solid_relcur = NULL;

            con = NULL;
        }

        if (con == NULL) {
            /* We do not have a valid connection in handler.
             */
            ss_pprintf_1(("ha_solid::get_solid_connection:no valid connection in handler\n"));

            con = get_solid_ha_data_connection(
                #if MYSQL_VERSION_ID >= 50100
                                this->ht,
                #endif
                thd);

            /* Link connection to this handler. */
            solid_link_connection(con);
            solid_conn = con;
        }
        if (old_con != NULL) {
            /* Clear old connection.
             */
            ss_pprintf_1(("ha_solid::get_solid_connection:clear old connection\n"));
            solid_clear_connection(old_con, old_rsrelh, old_relcur);
        }

        if (solid_rsrelh && rs_relh_isaborted(NULL, solid_rsrelh)) {
            rs_relh_done(con->sc_cd, solid_rsrelh);
            solid_rsrelh = NULL;
        }

        if (solid_rsrelh == NULL && table_name != NULL) {

            /* Get a new table handle.
             */
            char* relname;
            char* schema;
            char* relname_ptr;
            char* schema_ptr;
            rs_entname_t en;
            int sysrelid;
            int len = strlen(table_name) + 1;

            relname = (char *)SsMemAlloc(len + 1);
            schema = (char *)SsMemAlloc(len + 1);

            ss_pprintf_1(("ha_solid::get_solid_connection:get new table handle\n"));

            solid_set_relname(schema, relname, (char *)table_name, lower_case_table_names);
            ss_pprintf_1(("ha_solid::get_solid_connection:get new table handle:%s->%s.%s\n", (char *)table_name, schema, relname));

            sysrelid = check_solid_sysrel(relname);

            if (sysrelid != -1) {
                relname_ptr = solid_sysrel_map[sysrelid].srm_solidname;
                schema_ptr = (char *)RS_AVAL_SYSNAME;
            } else {
                relname_ptr = relname;
                schema_ptr = schema;
            }

            rs_entname_initbuf(
                &en,
                NULL,       // catalog,
                schema_ptr,
                relname_ptr);

            /* TODO:
               - how quoted identifiers (case sensitive) are handled?
             */

            solid_rsrelh = tb_dd_getrelh(
                            con->sc_cd,
                            con->sc_trans,
                            &en,
                            NULL,       // p_priv
                            NULL);      // rs_err_t** p_errh ? TODO: add errh processing ?

            SsMemFree(relname);
            SsMemFree(schema);

            if (solid_rsrelh != NULL) {
                mainmemory = (rs_relh_reltype(con->sc_cd, solid_rsrelh) == RS_RELTYPE_MAINMEMORY);
            }
        }

        CHK_CONN(con);

        return(con);
}

/*#***********************************************************************\
 *
 *              solid_close_connection
 *
 * Frees a possible connection object associated with the current THD
 *
 * Parameters :
 *
 *      THD* thd, in, use
 *
 * Return value :
 *
 *      0 or error number
 *
 * Globals used :
 */
static int solid_close_connection(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD*	thd)
{
        SOLID_CONN* con;

#if MYSQL_VERSION_ID >= 50100
        con = (SOLID_CONN*)thd->ha_data[hton->slot];
#else
        con = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        if (con != NULL) {
            CHK_CONN(con);

#if MYSQL_VERSION_ID >= 50100
            thd->ha_data[hton->slot] = NULL;
#else
            thd->ha_data[solid_hton.slot] = NULL;
#endif

            ss_pprintf_1(("solid_close_connection:tb_disconnect, con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
            SsSemEnter(con->sc_mutex);
            CHK_CONN(con);
            con->sc_isvalid = FALSE;
            SS_PMON_ADD(SS_PMON_MYSQL_ROLLBACK);
            tb_trans_rollback_onestep(con->sc_cd, con->sc_trans, FALSE, NULL);
            solid_free_connection_nomutex(con);
            SsSemExit(con->sc_mutex);
            solid_unlink_connection(con);

            SsThrUnregister();
        }
        return(0);
}

/*#***********************************************************************\
 *
 *              ha_solid_mysql_error
 *
 * Map solidDB error code to MySQL error code
 *
 * Parameters :
 *
 *      THD* thd, in, use, MySQL thread structure
 *      su_err_t* errh, in, use, solidDB error structure
 *
 * Return value :
 *
 *      MySQL error code
 *
 * Globals used :
 */
static int ha_solid_mysql_error(
        THD *thd,
        su_err_t* errh)
{
        int solrc;
        int myrc;

        ss_pprintf_1(("ha_solid_mysql_error:%s\n", errh != NULL ? su_err_geterrstr(errh) : "NULL"));

        if (errh == NULL) {
            solrc = -1;
        } else {
            solrc = su_err_geterrcode(errh);
        }

        switch (solrc) {
            case E_FORKEYREFEXIST_S:
                myrc = HA_ERR_ROW_IS_REFERENCED;
                break;
            case E_KEYNAMEEXIST_S:
                myrc = HA_ERR_WRONG_COMMAND;
                break;
            case DBE_ERR_UNIQUE_S:
            case DBE_ERR_PRIMUNIQUE_S:
                myrc = HA_ERR_FOUND_DUPP_KEY;
                break;
            case DBE_ERR_NOTFOUND:
            case DBE_ERR_NOACTSEARCH:
            case E_UPDNOCUR:
            case E_DELNOCUR:
                myrc = HA_ERR_NO_ACTIVE_RECORD;
                break;
            case DBE_ERR_LOCKED:
            case DBE_ERR_DDOPACT:
            case DBE_ERR_LOSTUPDATE:
            case DBE_ERR_NOTSERIALIZABLE:
            case DBE_ERR_TRXTIMEOUT:    // Transaction timed out
                myrc = HA_ERR_LOCK_WAIT_TIMEOUT;
                break;
            case DBE_ERR_DEADLOCK:

                /* Roll back the whole transaction, tell also to
                   MySQL that we rolled the transaction */

#if MYSQL_VERSION_ID >= 50100
                solid_rollback(legacy_soliddb_hton, thd, TRUE);
#else
                solid_rollback(thd, TRUE);
#endif

                if (thd) {
                    ha_rollback(thd);
                }

                myrc = HA_ERR_LOCK_DEADLOCK;
                break;
            case DBE_ERR_PARENTNOTEXIST_S:
                myrc = HA_ERR_NO_REFERENCED_ROW;
                break;
            case DBE_ERR_CHILDEXIST_S:
                myrc = HA_ERR_ROW_IS_REFERENCED;
                break;
            case SU_ERR_FILE_WRITE_DISK_FULL:
            case SU_ERR_FILE_WRITE_CFG_EXCEEDED:
                myrc = HA_ERR_RECORD_FILE_FULL;
                break;
            case DBE_ERR_TOOLONGKEY:
            case E_TOOLONGCONSTR:
                myrc = HA_ERR_TO_BIG_ROW;
                break;
            case DBE_ERR_TRXREADONLY:
            case DBE_ERR_DBREADONLY:
                myrc = HA_ERR_READ_ONLY_TRANSACTION;
                break;
            case E_RELNOTEXIST_S:
                myrc = HA_ERR_NO_SUCH_TABLE;
                break;
            case E_RELEXIST_S:
            case E_DDOP:
                myrc = HA_ERR_TABLE_EXIST;
                break;
            default:
                myrc = -1;
                break;
        }

        ss_pprintf_1(("ha_solid_mysql_error:%d->%d\n", solrc, myrc));

        return(myrc);
}
/*#***********************************************************************\
 *
 *              handle_error
 *
 * Handler errors in solidDB for MySQL
 *
 * Parameters :
 *
 *    su_err_t* errh, in, use, solidDB error handler
 *
 * Return value : MySQL error code
 *
 * Globals used : current_thd
 */
static inline int handle_error(
        su_err_t* errh)
{
    THD *thd = current_thd;
    int rc = 0;

    if (errh) {
        rc = ha_solid_mysql_error(thd, errh);
    }

    if (errh) {
        sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
        su_err_done(errh); 
    }
    
    return rc;
}

/*#***********************************************************************\
 *
 *              ha_solid_map_mysql_solid_isolation
 *
 * Maps a MySQL trx isolation level code to the solidDB (tab-level)
 * isolation level code.
 *
 * NOTE: needed MySQL isolation is in thd->variables.tx_isolation or
 * in thd->session_tx_isolation
 *
 * Parameters :
 *
 *      enum_tx_isolation iso, in, use
 *
 * Return value :
 *
 *      solidDB isolation code
 *
 * Globals used :
 */
static inline int ha_solid_map_mysql_solid_isolation(
        enum_tx_isolation iso)
{
        switch(iso) {
            case ISO_REPEATABLE_READ:
                return(TB_TRANSOPT_ISOLATION_REPEATABLEREAD);
            case ISO_READ_COMMITTED:
                return(TB_TRANSOPT_ISOLATION_READCOMMITTED);
            case ISO_SERIALIZABLE:
                return(TB_TRANSOPT_ISOLATION_SERIALIZABLE);
            case ISO_READ_UNCOMMITTED:
                return(TB_TRANSOPT_ISOLATION_READCOMMITTED);
            default:
                ss_error;
                return(0);
        }
}

/*#***********************************************************************\
 *
 *              solid_ignore_duplicate
 *
 * Check whether we should ignore duplicate key error
 *
 * Parameters :
 *
 *      THD*  thd, in, use, MySQL thread
 *
 * Return value :
 *
 *      TRUE, if we should update duplicate key
 *      FALSE, otherwise
 *
 * Globals used :
 */
static inline bool solid_ignore_duplicate(
        THD *thd)
{
        /* REPLACE ... [SELECT] || LOAD REPLACE */
        if (thd->lex->sql_command == SQLCOM_REPLACE ||
            thd->lex->sql_command == SQLCOM_REPLACE_SELECT ||
            (thd->lex->sql_command == SQLCOM_LOAD &&
             thd->lex->duplicates == DUP_REPLACE)) {
            return(TRUE);
        }

        /* INSERT ... [SELECT] ... ON DUPLICATE KEY ... */
        if ((thd->lex->sql_command == SQLCOM_INSERT ||
             thd->lex->sql_command == SQLCOM_INSERT_SELECT) &&
            thd->lex->duplicates  == DUP_UPDATE) {
            return(TRUE);
        }

        /* {INSERT|UPDATE} IGNORE [SELECT] */
        if ((thd->lex->sql_command == SQLCOM_INSERT ||
             thd->lex->sql_command == SQLCOM_INSERT_SELECT ||
             thd->lex->sql_command == SQLCOM_UPDATE ||
             thd->lex->sql_command == SQLCOM_UPDATE_MULTI) &&
            thd->lex->ignore == TRUE){
            return(TRUE);
        }

        return(FALSE);
}

/*#***********************************************************************\
 *
 *              solid_ignore_duplicate_update
 *
 * Check whether we should ignore duplicate key error on update
 *
 * Parameters :
 *
 *      THD*  thd, in, use, MySQL thread
 *
 * Return value :
 *
 *      TRUE, if we should update duplicate key
 *      FALSE, otherwise
 *
 * Globals used :
 */
static inline bool solid_ignore_duplicate_update(
        THD *thd)
{
        /* INSERT ... [SELECT] ... ON DUPLICATE KEY ... */
        if ((thd->lex->sql_command == SQLCOM_INSERT ||
             thd->lex->sql_command == SQLCOM_INSERT_SELECT) &&
            thd->lex->duplicates  == DUP_UPDATE) {
            return(TRUE);
        }

        /* {INSERT|UPDATE} IGNORE [SELECT] */
        if ((thd->lex->sql_command == SQLCOM_INSERT ||
             thd->lex->sql_command == SQLCOM_INSERT_SELECT ||
             thd->lex->sql_command == SQLCOM_UPDATE ||
             thd->lex->sql_command == SQLCOM_UPDATE_MULTI) &&
            thd->lex->ignore == TRUE){
            return(TRUE);
        }

        return(FALSE);
}

/* The flag trx->active_trans is set to 1 in
 *
 * 1. ::external_lock(),
 * 2. ::start_stmt(),
 * 3. innobase_query_caching_of_table_permitted(),
 * 4. innobase_savepoint(),
 * 5. ::init_table_handle_for_HANDLER(),
 * 6. innobase_start_trx_and_assign_read_view(),
 * 7. ::transactional_table_lock()

 * and it is only set to 0 in a commit or a rollback. If it is 0 we know
 * there cannot be resources to be freed and we could return immediately.
 * For the time being, we play safe and do the cleanup though there should
 * be nothing to clean up.
 */

/*#***********************************************************************\
 *
 *              solid_commit
 *
 *  Commits a transaction in an solid database or marks an SQL statement
 *  ended.
 *
 * Parameters :
 *
 *      THD*  thd, in, use
 *      bool  all, in, use, TRUE - commit transaction
 *                          FALSE - commit statement
 *
 * Return value : 0
 *
 * Globals used :
 */
static int solid_commit(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD*    thd,
        bool    all)
{
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   solid_trans;
        bool          succp=TRUE;
        int           rc = 0;
        su_err_t*     errh = NULL;
        dbe_ret_t     commit_rc = DBE_RC_SUCC;
        solid_bool    finished = FALSE;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)

        ss_win_perf_start;

        DBUG_ENTER("solid_commit");
        DBUG_PRINT("trans", ("ending transaction"));

        ss_pprintf_1(("solid_commit:all=%d thd->options=%d, AUTOCOMMIT = %s\n", all, thd->options,
                      thd->options&OPTION_NOT_AUTOCOMMIT ? "OFF" : "ON" ));

        con = get_solid_ha_data_connection(
#if MYSQL_VERSION_ID >= 50100
                hton,
#endif
                thd);

        CHK_CONN(con);

        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        solid_trans = con->sc_trans;

        if (all || !(thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN))) {

            /* We were instructed to commit the whole transaction, or
             * this is an SQL statement end and AUTOCOMMIT is ON.
             */
            ss_pprintf_2(("solid_commit:commit transaction, con=%ld, trans=%ld\n", (long)con, (long)solid_trans));
            SS_PMON_ADD(SS_PMON_MYSQL_COMMIT);

            tb_trans_setusertrxcleanup(cd, solid_trans, TRUE);

            do {
                succp = tb_trans_commit(cd, solid_trans, &finished, &errh);
            } while (rs_sysi_lockwait(cd) || !finished);

        } else {
            ss_pprintf_2(("solid_commit:commit statement, con=%ld, trans=%ld\n", (long)con, (long)solid_trans));

            do {
                succp = tb_trans_stmt_commit(cd, solid_trans, &finished, &errh);
            } while (!finished);
        }

        if (!succp) {

            dbe_trx_t* dbe_trx;
            THD_TRANS *trans= all ? &thd->transaction.all : &thd->transaction.stmt;

            /* If more than one handler participate to this transaction
               issue a warning that whole transaction might not been rolled back. */

            if (trans->nht > 1) {
                push_warning_printf(thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                 ER_WARNING_NOT_COMPLETE_ROLLBACK,
                 ER(ER_WARNING_NOT_COMPLETE_ROLLBACK));
            }

            rc = ha_solid_mysql_error(thd, errh);
            if (errh) {
                su_err_done(errh);
                errh = NULL;
            }

            SsSemEnter(con->sc_mutex);
            dbe_trx = tb_trans_dbtrx(cd, solid_trans);
            if (dbe_trx) {
                con->sc_errkey = dbe_trx_geterrkey(dbe_trx);
            }
            SsSemExit(con->sc_mutex);
        }

        if (solid_trans && tb_trans_isusertrxcleanup(cd, solid_trans)) {
            tb_trans_setusertrxcleanup(cd, solid_trans, FALSE);

            if(tb_trans_dbtrx(cd, solid_trans)) {
                tb_trans_commit_cleanup(cd, solid_trans, tb_trans_getdberet(cd, solid_trans), &errh);

                if (errh) {
                    su_err_done(errh);
                }
            }
        }

        ss_win_perf_stop(commit_perfcount, commit_callcount);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              internal_solid_end_trans_and_stmt
 *
 *
 *
 * Parameters :
 *
 *      rs_sysi_t*   cd,
 *      tb_trans_t*  trans,
 *      bool         commitp,
 *      su_err_t**   p_errh
 *
 * Return value :
 *
 *      TRUE
 *      FALSE
 *
 * Globals used :
 */
static bool internal_solid_end_trans_and_stmt(
        rs_sysi_t* cd,
        tb_trans_t* trans,
        bool commitp,
        su_err_t** p_errh)
{
        bool succp = true;
        solid_bool finished;

        ss_pprintf_1(("ha_solid::internal_solid_end_trans_and_stmt:%s, trans=%ld\n",
                       commitp ? "commit" : "rollback", (long)trans));

        if (commitp) {
            succp = tb_trans_stmt_commit_onestep(cd, trans, p_errh);
            if (!succp) {
                ss_pprintf_2(("ha_solid::internal_solid_end_trans_and_stmt:stmt commit failed, %s\n",
                               su_err_geterrstr(*p_errh)));
                commitp = FALSE;
            }
        } else {
            tb_trans_stmt_rollback_onestep(cd, trans, NULL);
        }

        if (commitp) {
            SS_PMON_ADD(SS_PMON_MYSQL_COMMIT);
            do {
                succp = tb_trans_commit(cd, trans, &finished, p_errh);
            } while (rs_sysi_lockwait(cd) || !finished);
            if (!succp) {
                ss_pprintf_2(("ha_solid::internal_solid_end_trans_and_stmt:stmt commit failed, %s\n",
                               su_err_geterrstr(*p_errh)));
                ss_error;   /* TODO: Check this, maybe we should not assert here */
            }
        } else {
            SS_PMON_ADD(SS_PMON_MYSQL_ROLLBACK);
            tb_trans_rollback_onestep(cd, trans, FALSE, NULL);
        }
        return(succp);
}

/*#***********************************************************************\
 *
 *              solid_init
 *
 * Initialize solidDB handlerton
 *
 * Parameters : -
 *
 * Return value : 0
 *
 * Globals used :
 */
#if MYSQL_VERSION_ID >= 50100
int solid_init(void *p)
{
        DBUG_ENTER("solid_init");

        handlerton *soliddb_hton = (handlerton *)p;
        legacy_soliddb_hton = soliddb_hton;

        soliddb_hton->state = have_soliddb;
        soliddb_hton->db_type = DB_TYPE_SOLID_DB;
        soliddb_hton->savepoint_offset = 0;
        soliddb_hton->close_connection = solid_close_connection;
        soliddb_hton->savepoint_set = NULL;     /* TODO: Savepoints */
        soliddb_hton->savepoint_rollback = NULL;
        soliddb_hton->savepoint_release = NULL;
        soliddb_hton->commit = solid_commit;
        soliddb_hton->rollback = solid_rollback;
        soliddb_hton->prepare = NULL; /* TODO: X/Open XA */
        soliddb_hton->recover = NULL;
        soliddb_hton->commit_by_xid = NULL;
        soliddb_hton->rollback_by_xid = NULL;
        soliddb_hton->create_cursor_read_view = NULL; /* TODO: Cursors */
        soliddb_hton->set_cursor_read_view = NULL;
        soliddb_hton->close_cursor_read_view = NULL;
        soliddb_hton->create = solid_create_handler;
        soliddb_hton->drop_database = solid_drop_database;
        soliddb_hton->panic = solid_end;
        soliddb_hton->start_consistent_snapshot = NULL; /* TODO: Consistent snapshots */
        soliddb_hton->flush_logs = solid_flush_logs;
        soliddb_hton->show_status = soliddb_show_status;
        soliddb_hton->flags = HTON_NO_FLAGS;
        soliddb_hton->release_temporary_latches = NULL; /* TODO: needed ? */
        soliddb_hton->alter_table_flags = solid_alter_table_flags;
        soliddb_hton->get_error_message = solid_get_error_message;

        if (have_soliddb != SHOW_OPTION_YES) {
            DBUG_RETURN(1); /* Nothing else to do */
        }

        solid_handler_init();

        DBUG_RETURN(0);

}
#endif

/*#***********************************************************************\
 *
 *              solid_rollback
 *
 *  Rollback a transaction in an solid database or marks an SQL statement
 *  ended
 *
 * Parameters :
 *
 *      THD*  thd, in, use
 *      bool  all, in, use, TRUE - rollback transaction
 *                          FALSE - rollback statement
 *
 * Return value : 0
 *
 * Globals used :
 */
static int solid_rollback(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD*    thd,
        bool    all)
{
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        su_err_t*     errh = NULL;
        bool          succp=TRUE;
        int           rc;
        solid_bool    finished = FALSE;

        DBUG_ENTER("solid_rollback");
        DBUG_PRINT("trans", ("ending transaction"));

        ss_pprintf_1(("solid_rollback\n"));

        con = get_solid_ha_data_connection(
#if MYSQL_VERSION_ID >= 50100
                hton,
#endif
                thd);

        CHK_CONN(con);

        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;

        if (all || (!(thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)))) {
            /* We were instructed to commit the whole transaction, or
               this is an SQL statement end and autocommit is on */
            ss_pprintf_2(("solid_rollback:rollback transaction, con=%ld, trans=%ld\n", (long)con, (long)trans));
            SS_PMON_ADD(SS_PMON_MYSQL_ROLLBACK);
            do {
                succp = tb_trans_rollback(cd, trans, NULL, &finished, TRUE, &errh);
            } while (!finished);
        } else {
            ss_pprintf_2(("solid_rollback:rollback statement, con=%ld, trans=%ld\n", (long)con, (long)trans));
            do {
                succp = tb_trans_stmt_rollback(cd, trans, &finished, &errh);
            } while (!finished);

        }

        if (!succp) {
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
        } else {
            rc = 0;
        }

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              solid_trans_beginif
 *
 * Start a new transaction or statement if it has not been already started
 *
 * Parameters :
 *
 *      THD*  thd, in, use
 *      bool  start_stmt, in, use, TRUE - start a statement
 *                                 FALSE - start a transaction
 *
 * Return value : -
 *
 * Globals used :
 */

static void solid_trans_beginif(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        THD* thd,
        bool start_stmt)
{
        SOLID_CONN* con;
        bool new_transaction_started;

        ss_pprintf_1(("solid_trans_beginif:thd->options %d, start_stmt = %d.\n", thd->options, (int)start_stmt));

        if (thd->options & (OPTION_NOT_AUTOCOMMIT)) {
            ss_pprintf_1(("solid_trans_beginif:OPTION_NOT_AUTOCOMMIT\n"));
        }

#if MYSQL_VERSION_ID >= 50100
        con = (SOLID_CONN*)thd->ha_data[hton->slot];
#else
        con = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        if (con == NULL) {
            ss_error;
            ss_pprintf_1(("solid_trans_beginif:con == NULL\n"));
            return;
        }

        CHK_CONN(con);

        if (start_stmt) {
            new_transaction_started = tb_trans_begintransandstmt(con->sc_cd, con->sc_trans);
        } else {
            new_transaction_started = tb_trans_beginif(con->sc_cd, con->sc_trans);
        }

        if (new_transaction_started && thd->transaction.on) {
            if (thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)) {

#if MYSQL_VERSION_ID >= 50100
                trans_register_ha(thd, TRUE, hton);
#else
                trans_register_ha(thd, TRUE, &solid_hton);
#endif
            }

#if MYSQL_VERSION_ID >= 50100
            trans_register_ha(thd, FALSE, hton);
#else
            trans_register_ha(thd, FALSE, &solid_hton);
#endif

            ss_pprintf_1(("solid_trans_beginif:new transaction registered, con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
        } else {
            ss_pprintf_1(("solid_trans_beginif:old transaction, NOT registered, con=%ld, trans=%ld\n", (long)con, (long)con->sc_trans));
        }
}

static void ha_solid_mysql_atype_setmysqldatatype(rs_sysi_t* cd, rs_atype_t* atype, Field* field)
{
        rs_mysqldatatype_t mysqldatatype = RS_MYSQLTYPE_NONE;

        switch (field->type()) {
            case MYSQL_TYPE_VAR_STRING: /* old <= 4.1 VARCHAR */
                mysqldatatype = RS_MYSQLTYPE_VAR_STRING;
                break;
            case MYSQL_TYPE_VARCHAR:    /* new >= 5.0.3 true VARCHAR */
                mysqldatatype = RS_MYSQLTYPE_VARCHAR;
                break;
            case MYSQL_TYPE_STRING:
                mysqldatatype = RS_MYSQLTYPE_STRING;
                break;
            case FIELD_TYPE_NEWDECIMAL:
                mysqldatatype = RS_MYSQLTYPE_NEWDECIMAL;
                break;
            case FIELD_TYPE_LONG:
                mysqldatatype = RS_MYSQLTYPE_LONG;
                break;
            case FIELD_TYPE_LONGLONG:
                mysqldatatype = RS_MYSQLTYPE_LONGLONG;
                break;
            case FIELD_TYPE_DATE:
                mysqldatatype = RS_MYSQLTYPE_DATE;
                break;
            case FIELD_TYPE_YEAR:
                mysqldatatype = RS_MYSQLTYPE_YEAR;
                break;
            case FIELD_TYPE_NEWDATE:
                mysqldatatype = RS_MYSQLTYPE_NEWDATE;
                break;
            case FIELD_TYPE_TIME:
                mysqldatatype = RS_MYSQLTYPE_TIME;
                break;
            case FIELD_TYPE_TIMESTAMP:
                mysqldatatype = RS_MYSQLTYPE_TIMESTAMP;
                break;
            case FIELD_TYPE_DATETIME:
                mysqldatatype = RS_MYSQLTYPE_DATETIME;
                break;
            case FIELD_TYPE_TINY:
                mysqldatatype = RS_MYSQLTYPE_TINY;
                break;
            case FIELD_TYPE_SHORT:
                mysqldatatype = RS_MYSQLTYPE_SHORT;
                break;
            case FIELD_TYPE_INT24:
                mysqldatatype = RS_MYSQLTYPE_INT24;
                break;
            case FIELD_TYPE_FLOAT:
                mysqldatatype = RS_MYSQLTYPE_FLOAT;
                break;
            case FIELD_TYPE_DOUBLE:
                mysqldatatype = RS_MYSQLTYPE_DOUBLE;
                break;
            case FIELD_TYPE_DECIMAL:
                mysqldatatype = RS_MYSQLTYPE_DECIMAL;
                break;
            case FIELD_TYPE_GEOMETRY:
                mysqldatatype = RS_MYSQLTYPE_GEOMETRY;
                break;
            case FIELD_TYPE_TINY_BLOB:
                mysqldatatype = RS_MYSQLTYPE_TINY_BLOB;
                break;
            case FIELD_TYPE_MEDIUM_BLOB:
                mysqldatatype = RS_MYSQLTYPE_MEDIUM_BLOB;
                break;
            case FIELD_TYPE_BLOB:
                mysqldatatype = RS_MYSQLTYPE_BLOB;
                break;
            case FIELD_TYPE_LONG_BLOB:
                mysqldatatype = RS_MYSQLTYPE_LONG_BLOB;
                break;
            case MYSQL_TYPE_BIT:
                mysqldatatype = RS_MYSQLTYPE_BIT;
                break;
            default:
                ss_error;
        }

        rs_atype_setmysqldatatype(cd, atype, mysqldatatype);
}

/*#***********************************************************************\
 *
 *              ha_solid_mysql_atype
 *
 * Map MySQL field type to solidDB attribute type
 *
 * Parameters :
 *
 *      rs_sysi_t* cd, in, use
 *      Field*     field, in, use
 *
 * Return value :
 *
 *      rs_atype_t*
 *
 * Globals used :
 */
static rs_atype_t* ha_solid_mysql_atype(rs_sysi_t* cd, Field* field)
{
        rs_atype_t* atype;
        rs_sqldatatype_t sqldatatype = RSSQLDT_INTEGER;
        ulong length;
        long scale;

        length = field->pack_length();
        scale = field->decimals();

        switch (field->type()) {
            case MYSQL_TYPE_VAR_STRING: /* old <= 4.1 VARCHAR */
            case MYSQL_TYPE_VARCHAR:    /* new >= 5.0.3 true VARCHAR */
                if (field->type() == MYSQL_TYPE_VARCHAR) {
                    length -= ((Field_varstring*)field)->length_bytes;
                }
                scale = -1;
                if (field->binary()) {
                    sqldatatype = RSSQLDT_LONGVARBINARY;
                } else {
                    sqldatatype = RSSQLDT_LONGVARCHAR;
                }
                break;
            case MYSQL_TYPE_STRING:
                sqldatatype = RSSQLDT_LONGVARCHAR;
                break;
            case FIELD_TYPE_NEWDECIMAL:
                // sqldatatype = RSSQLDT_VARCHAR;
                sqldatatype = RSSQLDT_LONGVARBINARY;
                break;
            case FIELD_TYPE_LONG:
            case FIELD_TYPE_LONGLONG:
                sqldatatype = RSSQLDT_BIGINT;
                break;
            case FIELD_TYPE_DATE:
            case FIELD_TYPE_NEWDATE:
                sqldatatype = RSSQLDT_DATE;
                break;
            case FIELD_TYPE_TIME:
                sqldatatype = RSSQLDT_TIME;
                break;
            case FIELD_TYPE_TIMESTAMP:
            case FIELD_TYPE_DATETIME:
                sqldatatype = RSSQLDT_TIMESTAMP;
                break;
            case FIELD_TYPE_YEAR:
            case FIELD_TYPE_TINY:
            case FIELD_TYPE_SHORT:
            case FIELD_TYPE_INT24:
                sqldatatype = RSSQLDT_INTEGER;
                break;
            case FIELD_TYPE_FLOAT:
            case FIELD_TYPE_DOUBLE:
                sqldatatype = RSSQLDT_DOUBLE;
                break;
            case FIELD_TYPE_DECIMAL:
                // sqldatatype = RSSQLDT_DECIMAL;
                sqldatatype = RSSQLDT_LONGVARBINARY;
                break;
            case FIELD_TYPE_GEOMETRY:
            case FIELD_TYPE_TINY_BLOB:
            case FIELD_TYPE_MEDIUM_BLOB:
            case FIELD_TYPE_BLOB:
            case FIELD_TYPE_LONG_BLOB:
                length = field->field_length;
                scale = -1;
                if (field->binary()) {
                    sqldatatype = RSSQLDT_LONGVARBINARY;
                } else {
                    sqldatatype = RSSQLDT_LONGVARCHAR;
                }
                break;

            case MYSQL_TYPE_BIT:
                length = field->field_length;
                scale = 0;
                sqldatatype = RSSQLDT_LONGVARBINARY;
                break;

            default:
                ss_error;
        }

        atype = rs_atype_initbysqldt(
                    cd,
                    sqldatatype,
                    length,
                    scale);

        rs_atype_setautoinc(cd, atype, (field->flags & AUTO_INCREMENT_FLAG), 0);
        ha_solid_mysql_atype_setmysqldatatype(cd, atype, field);

        ss_pprintf_3(("%s(%d,%d)\n", rs_atype_name(cd, atype), length, scale));

        return(atype);
}

/*#***********************************************************************\
 *
 *              is_mysql_primary_key
 *
 * Return TRUE if given key is a primary key
 *
 * Parameters :
 *
 *      st_key* key, in, use, MySQL key definition
 *
 * Return value : TRUE if key is primary key, FALSE otherwise
 *
 *
 * Globals used :
 */
static inline bool is_mysql_primary_key(st_key* key)
{
        return(strcmp(key->name, "PRIMARY") == 0);
}

/*#***********************************************************************\
 *
 *              ha_mysql_solid_create_unique_init
 *
 * Create a unique key for solidDB
 *
 * Parameters :
 *
 *      st_key*  key, in, use
 *      uint     nkey, in, use
 *
 * Return value :
 *
 *      tb_sqlunique_t*
 *
 * Globals used :
 */
static tb_sqlunique_t* ha_mysql_solid_create_unique_init(st_key* key, uint nkey)
{
        tb_sqlunique_t* skey = NULL;
        uint kp;
        size_t skey_size;

        if (nkey < 1) {
            ss_pprintf_1(("ha_mysql_solid_create_unique:NO PRIMARY KEY:nkey\n", nkey));
            return(NULL);
        }

        if (key == NULL) {
            ss_pprintf_1(("ha_mysql_solid_create_unique:NO PRIMARY KEY:st_key* == NULL\n"));
            return(NULL);
        }
        if (key->key_parts < 1) {
            ss_pprintf_1(("ha_mysql_solid_create_unique:NO PRIMARY KEY:key->key_parts %d\n", key->key_parts));
            return(NULL);
        }

        skey_size = sizeof(tb_sqlunique_t);
        ss_pprintf_1(("ha_mysql_solid_create_unique:PRIMARY KEY:key->key_parts %d, skey_size %d\n", key->key_parts, skey_size));

        ss_dassert(strcmp(key->name, "PRIMARY") == 0);

        skey = (tb_sqlunique_t*)SsMemAlloc(skey_size);
        skey->name = NULL;
        skey->len  = key->key_parts;
        skey->fields = (uint*)SsMemAlloc(skey->len*sizeof(uint));
        for (kp = 0;kp<key->key_parts;kp++) {
            ss_pprintf_1(("ha_mysql_solid_create_unique:ano %d\n", key->key_part[kp].fieldnr-1));
            skey->fields[kp] = key->key_part[kp].fieldnr-1;
        }

        return(skey);
}

/*#***********************************************************************\
 *
 *              keyname_from_mysql_to_solid
 *
 * Map MySQL keyname to solidDB keyname
 *
 * Parameters :
 *
 *      rs_sysi_t*  cd, in, use
 *      rs_relh_t*  relh, in, use, solidDB relation
 *      char*       mysql_keyname, in, use
 *      char*       solid_keyname_buffer,in out, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void keyname_from_mysql_to_solid(
        rs_sysi_t* cd,
        rs_relh_t* relh,
        char* mysql_keyname,
        char* solid_keyname_buffer,
        bool  lower_case_names)
{
        long relid;

        relid = rs_relh_relid(cd, relh);

        SsSprintf(solid_keyname_buffer, "%s_%ld", mysql_keyname, relid);

        if (lower_case_names) {
            SsStrupr(solid_keyname_buffer);
        }
}

/*#***********************************************************************\
 *
 *              ha_mysql_solid_create_unique_done
 *
 * Free unique index definition structure
 *
 * Parameters :
 *
 *     tb_sqlunique_t* skey, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
static void ha_mysql_solid_create_unique_done(tb_sqlunique_t* skey)
{
        if (skey != NULL) {
            ss_pprintf_1(("ha_mysql_solid_create_unique_done:\n"));
            ss_dassert(skey->len > 0);
            ss_dassert(skey->fields != NULL);
            SsMemFree(skey->fields);
            SsMemFree(skey);
        }
}

/*#***********************************************************************\
 *
 *              ha_mysql_solid_foreign_done
 *
 * Free foreign key definition structure
 *
 * Parameters :
 *
 *     tb_sqlforkey_t* fkey, in, use, foreign key definitions
 *     uint         n_forkeys, in, use, # of foreign key definitions
 *
 * Return value : -
 *
 * Globals used :
 */
static void ha_mysql_solid_foreign_done(
        tb_sqlforkey_t* fkey,
        uint         n_forkeys)
{
        uint n_forkey;

        ss_dassert(fkey != NULL);
        ss_dassert(n_forkeys > 0);

        if (fkey != NULL) {
            ss_pprintf_1(("ha_mysql_solid_foreign_done:\n"));

            for (n_forkey=0; n_forkey < n_forkeys; n_forkey++) {
                ss_dassert(fkey[n_forkey].len > 0);
                ss_dassert(fkey[n_forkey].fields != NULL);

                SsMemFree(fkey[n_forkey].fields);
                if (fkey[n_forkey].name != NULL) {
                    SsMemFree(fkey[n_forkey].name);
                }
                if (fkey[n_forkey].reftable != NULL) {
                    SsMemFree(fkey[n_forkey].reftable);
                }
                if (fkey[n_forkey].refschema != NULL) {
                    SsMemFree(fkey[n_forkey].refschema);
                }
                if (fkey[n_forkey].refcatalog != NULL) {
                    SsMemFree(fkey[n_forkey].refcatalog);
                }

                if (fkey[n_forkey].reffields != NULL) {
                    uint n_cols;
                    uint n_col;

                    n_cols = fkey[n_forkey].len;
                    for (n_col=0; n_col < n_cols; n_col++) {
                        ss_dassert(fkey[n_forkey].reffields[n_col]);
                        SsMemFree(fkey[n_forkey].reffields[n_col]);
                    }
                    SsMemFree(fkey[n_forkey].reffields);
                }
            }

            SsMemFree(fkey);
        }
}

/*#***********************************************************************\
 *
 *              ha_solid_createindex
 *
 * Create a index for a solidDB table
 *
 * Parameters :
 *
 *     rs_relh_t*    relh, in, use, solidDB relation
 *     st_key*       key,in, use, MySQL key definition
 *     tb_connect_t* tbcon, in, use, solidDB table connection
 *     rs_sysi_t*    cd, in, use
 *     tb_trans_t*   trans, in, use, solidDB transaction
 *     char*         authid, in, use, index name
 *     char*         catalog, in, use
 *     char*         extrainfo, in, NULL,
 *     su_err_t*     p_errh, out, error status or NULL
 *
 * Return value : TRUE if success or FALSE in case of error
 *
 * Globals used :
 */
static bool ha_solid_createindex(
        rs_relh_t* relh,
        st_key* key,
        char* tabname,
        tb_connect_t* tbcon,
        rs_sysi_t* cd,
        tb_trans_t* trans,
        char*  authid,
        char*  catalog,
        char*  extrainfo,
        su_err_t**  p_errh,
        bool lower_case_names)
{
        uint   kp;
        bool   succp=TRUE;
        char*  indexname;
        char*  tauthid;
        char*  tcatalog;
        char*  textrainfo;
        bool   unique;
        uint   attr_c;
        char** attrs;
        solid_bool*  desc;
        void*  cont;

        ss_dassert(key != NULL);

        ss_pprintf_1(("ha_solid_createindex:%s, table %s (flags)\n", key->name, tabname, key->flags));

        indexname = (char *)SsMemAlloc(strlen(key->name)+64);
        keyname_from_mysql_to_solid(cd, relh, key->name, indexname, lower_case_names);

        tauthid = authid;
        tcatalog = catalog;
        textrainfo = extrainfo;
        unique = (key->flags & HA_NOSAME);
        cont = NULL;

        ss_dassert(key->key_parts > 0);
        attr_c = key->key_parts;
        attrs = (char**)SsMemAlloc(attr_c*sizeof(char*));
        desc = (solid_bool*)SsMemAlloc(attr_c*sizeof(solid_bool));

        for (kp=0;kp<attr_c;kp++) {
            ss_pprintf_1(("ha_solid_createindex:col %s\n" ,key->key_part[kp].field->field_name));
            attrs[kp] = (char*)SsMemStrdup((char*)key->key_part[kp].field->field_name);

            if (lower_case_names) {
                SsStrupr(attrs[kp]);
            }

            if (key->key_part[kp].key_part_flag & HA_REVERSE_SORT) {
                desc[kp] = TRUE;
            } else {
                desc[kp] = FALSE;
            }
        }

        succp = tb_createindex(
                     cd,
                     trans,
                     indexname,
                     authid,
                     catalog,
                     extrainfo,
                     tabname,
                     tauthid,
                     tcatalog,
                     textrainfo,
                     unique,
                     attr_c,
                     attrs,
                     desc,
                     &cont,
                     p_errh);

        ss_dassert(cont == NULL);

        for (kp=0;kp<attr_c;kp++) {
            SsMemFree(attrs[kp]);
        }

        SsMemFree(attrs);
        SsMemFree(desc);
        SsMemFree(indexname);

        return(succp);
}

/*#***********************************************************************\
 *
 *              ::ha_soliddb
 *
 * Constructor
 *
 * TODO: Check that we really can support these handler flags !
 *
 * Parameters : see below
 *
 * Return value : -
 *
 * Globals used :
 */
ha_soliddb::ha_soliddb(
        #if MYSQL_VERSION_ID >= 50100
             handlerton *hton, TABLE_SHARE *table_arg
        #else
             TABLE *table_arg
        #endif
     )
    : handler(
        #if MYSQL_VERSION_ID >= 50100
             hton
        #else
             &solid_hton
        #endif
        , table_arg ),
     int_table_flags(
        HA_TABLE_SCAN_ON_INDEX |
        HA_NULL_IN_KEY |
        HA_CAN_INDEX_BLOBS |
        HA_CAN_GEOMETRY | /*TODO: GEOMETRY SUPPORT*/
        HA_AUTO_PART_KEY |
        HA_CAN_SQL_HANDLER |
        HA_PRIMARY_KEY_IN_READ_INDEX |
        HA_REC_NOT_IN_SEQ |
        #if MYSQL_VERSION_ID >= 50100
            HA_PARTIAL_COLUMN_READ |
            HA_PRIMARY_KEY_REQUIRED_FOR_POSITION
        #else
            HA_NOT_EXACT_COUNT
        #endif
        ),
        solid_conn(NULL),
        solid_rsrelh(NULL),
        solid_relcur(NULL),
        rnd_mustinit(FALSE),
        for_update(FALSE),
        mainmemory(FALSE)
{
        DBUG_ENTER("ha_solid::ha_solid");
        ss_pprintf_1(("ha_solid::ha_solid #%p.\n", this));
        THD* thd = current_thd;
        if( thd && thd->lex ) {
          ss_pprintf_2((" SQL: type %d, text [%.*s]\n", thd->lex->sql_command, thd->query_length, thd->query ));
        }
        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              ::~ha_soliddb
 *
 * Destructor
 *
 * Parameters : -
 *
 * Return value : -
 *
 * Globals used :
 */
ha_soliddb::~ha_soliddb()
{
        ss_pprintf_1(("ha_solid::~ha_solid #%p.\n", this));

        if (solid_conn != NULL) {

            solid_clear_connection(
                solid_conn,
                solid_rsrelh,
                solid_relcur);

            solid_conn = NULL;
            solid_relcur = NULL;
            solid_rsrelh = NULL;
        }
        ss_dassert(solid_conn == NULL);
        ss_dassert(solid_relcur == NULL);
        ss_dassert(solid_rsrelh == NULL);
}

/*#***********************************************************************\
 *
 *              ::get_error_message
 *
 * This function is called from handler::print_error() to get solidDB
 * specific error message (see handler::print_error() for details)
 *
 * Parameters :
 *
 *     error  in, use   - error code (returned by ha_solid_mysql_error())
 *     buf    out, use  - corresponding error message
 *
 * Return value : 1 - this is a temporary error (error will be reported
 *                    to client as a temporary one)
 *                0 - otherwise
 *                    (see handler::print_error() for details)
 *
 * Globals used :
 */
bool ha_soliddb::get_error_message(int error, String* buf)
{
        switch (error) {
        case HA_ERR_TO_BIG_ROW:
              buf->append(STRING_WITH_LEN("Too big row"));
            break;
        default:
            break;
        }

	      return FALSE;
}

/*#***********************************************************************\
 *
 *              ::start_stmt
 *
 * MySQL calls this function at the start of each SQL statement inside LOCK
 * TABLES. Inside LOCK TABLES the ::external_lock method does not work to
 * mark SQL statement borders. Note also a special case: if a temporary table
 * is created inside LOCK TABLES, MySQL has not called external_lock() at all
 * on that table.
 * MySQL-5.0 also calls this before each statement in an execution of a stored
 * procedure. To make the execution more deterministic for binlogging, MySQL-5.0
 * locks all tables involved in a stored procedure with full explicit table
 * locks (thd->in_lock_tables is true in ::store_lock()) before executing the
 * procedure.
 *
 * Registers that solid takes part in an SQL statement, so that MySQL knows to
 * roll back the statement if the statement results in an error. This MUST be
 * called for every SQL statement that may be rolled back by MySQL. Calling this
 * several times to register the same statement is allowed, too.
 *
 * Registers an solid transaction in MySQL, so that the MySQL XA code knows
 * to call the solid prepare and commit, or rollback for the transaction. This
 * MUST be called for every transaction for which the user may call commit or
 * rollback. Calling this several times to register the same transaction is
 * allowed, too.
 *
 * This function also registers the current SQL statement.
 *
 * Parameters :
 *
 *     THD*          thd, in, use
 *     thr_lock_type lock_type, in, use
 *
 * Return value : 0
 *
 * Globals used :
 */
int ha_soliddb::start_stmt(
        THD*            thd,
        thr_lock_type   lock_type)
{
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        su_err_t*     errh = NULL;

        DBUG_ENTER("ha_solid::start_stmt");

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif
        CHK_CONN(con);

        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;

#if MYSQL_VERSION_ID >= 50100
        thd->set_current_stmt_binlog_row_based_if_mixed();
#endif

        ss_pprintf_1(("ha_solid::start_stmt, con=%ld, trans=%ld, lock_type=%d, query=%s\n",
                      (long)con, (long)trans, lock_type, thd->query));

        extra_keyread = FALSE;
        extra_retrieve_primary_key = FALSE;
        extra_retrieve_all_cols = TRUE; /* TODO: Should be FALSE */

        /* Register the statement */

        if (thd->transaction.on) {
            trans_register_ha(thd, FALSE, (handlerton *)ht);

            ss_pprintf_1(("Statement registered\n"));

            solid_commit(
#if MYSQL_VERSION_ID >= 50100
                    ht,
#endif
                    thd, FALSE);
        }

        tb_trans_begintransandstmt(cd, trans);

        if (thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN) &&
            thd->transaction.on) {
            /* No AUTOCOMMIT mode, register for a transaction */

            trans_register_ha(thd, FALSE, (handlerton *)ht);
            ss_pprintf_1(("Transaction registered\n"));
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::bas_ext
 *
 * If frm_error() is called then we will use this to to find out what file extentions
 * exist for the storage engine. This is also used by the default rename_table and
 * delete_table method in handler.cc.
 *
 * Parameters : -
 *
 * Return value :
 *
 *     char**
 *
 * Globals used :
 */
/*
*/
static const char *ha_soliddb_exts[] = {
        NullS
};

const char **ha_soliddb::bas_ext() const
{
    ss_pprintf_1(("ha_solid::bas_ext\n"));
    return ha_soliddb_exts;
}

/*#***********************************************************************\
 *
 *              ::table_caching_type
 *
 * Return table caching type
 *
 * #define HA_CACHE_TBL_NONTRANSACT 0
 * #define HA_CACHE_TBL_NOCACHE     1
 * #define HA_CACHE_TBL_ASKTRANSACT 2
 * #define HA_CACHE_TBL_TRANSACT    4
 *
 * Parameters : -
 *
 * Return value : uint8
 *
 * Globals used :
 */
uint8 ha_soliddb::table_cache_type()
{
        ss_pprintf_1(("ha_solid::table_cache_type:HA_CACHE_TBL_TRANSACT\n"));
        return HA_CACHE_TBL_TRANSACT; /* Use transactional query cache */
}

/*#***********************************************************************\
 *
 *              get_ref_length
 *
 * Return length of the clustered key
 *
 * Parameters :
 *
 *     rs_sysi_t* cd, in, use
 *     rs_relh_t* rsrelh, in, use, solidDB relation
 *
 * Return value : int
 *
 * Globals used :
 */
static int get_ref_length(rs_sysi_t* cd, rs_relh_t* rsrelh)
{
        rs_key_t* key;
        rs_ano_t nrefparts;
        rs_ano_t i;
        rs_ttype_t* ttype;
        rs_ano_t ano;
        rs_atype_t* atype;
        long atype_maxlen;
        int ref_length;

        DBUG_ENTER("ha_soliddb::get_ref_length");
        ss_pprintf_1(("ha_solid::get_ref_length\n"));

        key = rs_relh_clusterkey(cd, rsrelh);
        nrefparts = rs_key_nrefparts(cd, key);
        ttype = rs_relh_ttype(cd, rsrelh);

        ref_length = VA_LENGTHMAXLEN;

        for (i = 0; i < nrefparts; i++) {
            ano = rs_keyp_ano(cd, key, i);
            if (ano == RS_ANO_NULL) {
                /* Not a real column. */
                atype = rs_keyp_constatype(cd, key, i);
                ss_dassert(atype != NULL);
            } else {
                atype = rs_ttype_atype(cd, ttype, ano);
            }
            atype_maxlen = rs_atype_maxstoragelength(cd, atype);

            ref_length = ref_length + atype_maxlen;
        }
        if (rs_keyp_parttype(cd, key, i) == RSAT_TUPLE_VERSION) {
            ano = rs_keyp_ano(cd, key, i);
            atype = rs_ttype_atype(cd, ttype, ano);
            atype_maxlen = rs_atype_maxstoragelength(cd, atype);

            ref_length = ref_length + atype_maxlen;
        }

        DBUG_RETURN(ref_length);
}

/*#***********************************************************************\
 *
 *              ::open
 *
 * Used for opening tables. The name will be the name of the file.
 * A table is opened when it needs to be opened. For instance
 * when a request comes in for a select on the table (tables are not
 * open and closed for each request, they are cached).
 *
 * Called from handler.cc by handler::ha_open(). The server opens all tables by
 * calling ha_open() which then calls the handler specific open().
 *
 * Parameters :
 *
 *     const char* name, in, use
 *     int         mode, in, not used
 *     uint        test_if_locked, in, not used
 *
 * Return value : 0
 *
 * Globals used :
 */
int ha_soliddb::open(const char *name, int mode, uint test_if_locked)
{
        SOLID_CONN* con;
        THD* thd;

        DBUG_ENTER("ha_soliddb::open");
        ss_pprintf_1(("ha_solid::open\n"));

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

        if (solid_rsrelh != NULL) {
            ref_length = get_ref_length(
                            con->sc_cd,
                            solid_rsrelh);

            if (ref_length > MAX_REF_LENGTH) { /* TODO: Need a better check. */
                ref_length = MAX_REF_LENGTH;
            }
        } else {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        if (!(share = get_share(name, table))) {
            DBUG_RETURN(1);
        }
        thr_lock_data_init(&share->lock,&lock,NULL);

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::close
 *
 * Closes a table. We call the free_share() function to free any resources
 * that we have allocated in the "shared" structure.
 *
 * Called from sql_base.cc, sql_select.cc, and table.cc.
 * In sql_select.cc it is only used to close up temporary tables or during
 * the process where a temporary table is converted over to being a
 * myisam table.
 *
 * For sql_base.cc look at close_data_tables().
 *
 *
 * Parameters : -
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::close(void)
{
        DBUG_ENTER("ha_soliddb::close");
        ss_pprintf_1(("ha_solid::close a table:NOT IMPLEMENTED:nothing to do\n"));
        DBUG_RETURN(free_share(share));
}

/*#***********************************************************************\
 *
 *              ha_print_aval
 *
 * Print attribute value
 *
 * Parameters :
 *
 *     rs_sysi_t*   cd, in, use
 *     rs_atype_t*  atype, in, use
 *     rs_aval_t*   aval, in, use
 *     char*        infostr, in
 *
 * Return value : -
 *
 * Globals used :
 */
static void ha_print_aval(
        rs_sysi_t* cd,
        rs_atype_t* atype,
        rs_aval_t* aval,
        char* infostr)
{
        char* tmp;
        tmp = rs_aval_print(cd, atype, aval);
        ss_pprintf_2(("%s:aval='%.255s'\n", infostr, tmp));
        SsMemFree(tmp);
}

/*#***********************************************************************\
 *
 *              set_mysql_field_to_aval_or_dynva
 *
 * Set mysql field value to solidDB attribute value or dynamic value
 *
 * Parameters :
 *
 *     THD*        thd, in, use, MySQL thread
 *     rs_sysi_t*  cd, in, use, solid system info
 *     rs_atype_t* atype, in out, use, attribute type
 *     rs_aval_t*  aval, in out, use, attribute value
 *     dynva_t*    dva, in out use,
 *     Field*      field, in, use, MySQL attribute
 *     const byte* ptr, in, MySQL value
 *     bool        iskeyptr, in, use,
 *
 * Return value : -
 *
 * Globals used :
 */
static bool set_mysql_field_to_aval_or_dynva(
        THD* thd,
        TABLE* table,
        rs_sysi_t* cd,
        rs_atype_t* atype,
        rs_aval_t* aval,
        va_t* dva,
        Field* field,
        const byte* ptr,
        bool iskeyptr,
        su_err_t** p_errh)
{
        RS_AVALRET_T avret;
        int32 i4 = 0;
        uint32 ui4 = 0;
        int mysqltype;
        Field_blob* field_blob = NULL;
        bool succp = TRUE;
#if MYSQL_VERSION_ID >= 50100
        my_bitmap_map *old_map = dbug_tmp_use_all_columns(table, table->read_set);
#endif

        mysqltype = field->type();

        switch (rs_atype_datatype(cd, atype)) {
            case RSDT_CHAR:
                {
                    bool ismysqlblob;
                    uint length;

                    ismysqlblob =
                    ismysqlblob = (mysqltype == FIELD_TYPE_TINY_BLOB   ||
                                  mysqltype == FIELD_TYPE_MEDIUM_BLOB ||
                                  mysqltype == FIELD_TYPE_BLOB        ||
                                  mysqltype == FIELD_TYPE_LONG_BLOB);

                    if (ismysqlblob) {
                        char* buf;
                        uint length;

                        ss_dassert(mysqltype == FIELD_TYPE_TINY_BLOB   ||
                                  mysqltype == FIELD_TYPE_MEDIUM_BLOB ||
                                  mysqltype == FIELD_TYPE_BLOB        ||
                                  mysqltype == FIELD_TYPE_LONG_BLOB);

                        if (iskeyptr) {
                            short j;
                            shortget(j, ptr);
                            length = j;
                            buf = (char*)ptr + 2;
                        } else {
                            field_blob = (Field_blob*)field;
                            ss_pprintf_1(("BLOB:type %d, real_type %d, field_length %d, field->binary() == %d\n",
                                      mysqltype, field->real_type(), field->field_length, field->binary()));

                            length = field_blob->get_length();
                            field_blob->get_ptr(&buf);
                        }
                        ss_pprintf_1(("FIELD_BLOB:CHAR:iskeyptr %d, field_length %d value=(%.20s)\n", iskeyptr, length, buf));

                        if (dva != NULL) {
                            va_setdataandnull(dva, buf, length);
                        } else {
                            avret = rs_aval_set8bitcdata_ext(cd, atype, aval, buf, length, p_errh);
                            succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                        }

                    } else {

                        ss_dassert(mysqltype == FIELD_TYPE_STRING ||
                                  mysqltype == MYSQL_TYPE_VAR_STRING ||
                                  mysqltype == MYSQL_TYPE_VARCHAR);

                        /* Logic here is copied from another ha_ file (which one??)
                         * The slightly different behaviour is done because lenght bytes
                         * are different if the value is given as a row or as a constraint.
                         * Also I could not find a method that would give the out from
                         * a key constraint data buffer.
                         */
                        switch (field->real_type()) {
                        case MYSQL_TYPE_SET:
                        case MYSQL_TYPE_ENUM:
                            length = field->pack_length();
                            break;
                        case MYSQL_TYPE_VARCHAR:
                            /* varstring */
                            if (iskeyptr) {
                                length = uint2korr(ptr);
                                ptr = ptr + 2;
                            } else {
                                uint length_bytes = field->pack_length() - field->key_length();
                                length = length_bytes == 1 ? (uint) (uchar) *ptr : uint2korr(ptr);
                                ptr = ptr + length_bytes;
                            }
                            break;
                        default:
                            /* string */
                            length = field->field_length;
                            break;
                        }

                        if (dva != NULL) {
                            va_setdataandnull(dva, (void*)ptr, length);
                        } else {
                            avret = rs_aval_set8bitcdata_ext(cd, atype, aval, (char*)ptr, length, p_errh);
                            succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                        }
                    }
                }
                break;
            case RSDT_INTEGER: {

                switch(mysqltype) {
                   case FIELD_TYPE_YEAR:
                       i4 = field->val_int();
                       break;
                   case FIELD_TYPE_LONG:
                        /* Logic copied to here from field.cpp because I could not
                         * find a method to read key constraint value data. So we
                         * just as well always use this direct read method.
                         */
#ifdef WORDS_BIGENDIAN
                        if (table->s->db_low_byte_first) {
                            i4 = sint4korr(ptr);
                        } else
#endif
                        {
                            if (((Field_num*)field)->unsigned_flag) {
                                ulongget(i4, ptr);
                            } else {
                                longget(i4, ptr);
                            }
                        }
                        break;
                    case FIELD_TYPE_TINY:
                        /* Copied from field.cpp
                         */
                        i4 = ((Field_num*)field)->unsigned_flag
                                ? (int32) ((uchar*) ptr)[0] :
                                (int32) ((signed char*) ptr)[0];
                        break;
                    case FIELD_TYPE_SHORT:
                        {
                            /* Copied from field.cpp
                             */
                            short j;
#ifdef WORDS_BIGENDIAN
                            if (table->s->db_low_byte_first) {
                                j = sint2korr(ptr);
                            } else
#endif
                            {
                                shortget(j, ptr);
                            }
                            i4 = ((Field_num*)field)->unsigned_flag
                                    ? (int32) (unsigned short) j
                                    : (int32) j;
                        }
                        break;
                    case FIELD_TYPE_INT24:
                        /* Copied from field.cpp
                         */
                        i4 = ((Field_num*)field)->unsigned_flag
                                ? (int32) uint3korr(ptr)
                                : sint3korr(ptr);
                        break;
                    default:
                       ss_error;
                }

                if (dva != NULL) {
                    va_setlong(dva, i4);
                } else {
                    avret = rs_aval_setlong_ext(cd, atype, aval, i4, p_errh);
                    succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                }

                break;
            }
            case RSDT_FLOAT:
            case RSDT_DOUBLE:
                {
                    /* Copied from field.cpp
                     */
                    if (mysqltype == FIELD_TYPE_FLOAT) {
                        float j;

#ifdef WORDS_BIGENDIAN
                        if (table->s->db_low_byte_first) {
                            float4get(j, ptr);
                        } else
#endif
                        {
                            memcpy_fixed((byte*)&j, ptr, sizeof(j));
                        }
                        if (dva != NULL) {
                            va_setdouble(dva, (double)j);
                        } else {
                            avret = rs_aval_setdouble_ext(cd, atype, aval, (double)j, p_errh);
                            succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                        }
                    } else if (mysqltype == FIELD_TYPE_DOUBLE) {

                        double j;

#ifdef WORDS_BIGENDIAN
                        if (table->s->db_low_byte_first) {
                            float8get(j,ptr);
                        } else
#endif
                        {
                            doubleget(j,ptr);
                        }
                        if (dva != NULL) {
                            va_setdouble(dva, j);
                        } else {
                            avret = rs_aval_setdouble_ext(cd, atype, aval, j, p_errh);
                            succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                        }
                    } else {
                        ss_error;
                    }
                }
                break;
            case RSDT_DATE: {
                TIME ltime;
                dt_date_t solid_date;
                dt_datesqltype_t datesqltype = DT_DATE_SQLDATE;
                solid_bool succ = TRUE;

                /* See field.cc for more documentation about these types */

                memset(&ltime, 0, sizeof(TIME));


                switch(field->real_type()) {

                    case FIELD_TYPE_NEWDATE:
                        ((Field_newdate*)field)->get_date(&ltime, 0);
                        datesqltype = DT_DATE_SQLDATE;
                        break;
                    case FIELD_TYPE_DATE:
                        ((Field_date*)field)->get_date(&ltime, 0);
                        datesqltype = DT_DATE_SQLDATE;
                        break;
                    case FIELD_TYPE_TIME:
                        ((Field_time*)field)->get_time(&ltime);
                        datesqltype = DT_DATE_SQLTIME;
                        break;
                    case FIELD_TYPE_TIMESTAMP:
                        ((Field_timestamp*)field)->get_date(&ltime, 0);
                        datesqltype = DT_DATE_SQLTIMESTAMP;
                        break;
                    case FIELD_TYPE_DATETIME:
                        ((Field_datetime*)field)->get_date(&ltime, 0);
                        datesqltype = DT_DATE_SQLTIMESTAMP;
                        break;
                   default:
                        ss_error;
                        break;
                }

                succ = dt_date_setdata(&solid_date,
                                    ltime.year,  /* year  */
                                    ltime.month, /* month */
                                    ltime.day,   /* day   */
                                    ltime.hour,  /* hour  */
                                    ltime.minute,/* minute*/
                                    ltime.second,/* second*/
                                    0);

                ss_dassert(succp != FALSE);

                ss_pprintf_2(("Stored date: %d-%d-%d %d:%d:%d\n",
                              dt_date_year(&solid_date), dt_date_month(&solid_date),
                              dt_date_mday(&solid_date), dt_date_hour(&solid_date),
                              dt_date_min(&solid_date), dt_date_sec(&solid_date)));

                if (dva != NULL) {
                    dt_date_datetova(&solid_date, dva);
                } else {

                    avret = rs_aval_setdate_ext(cd, atype, aval, &solid_date, datesqltype, p_errh);
                    succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                }

                break;
            }
            case RSDT_DFLOAT:
                ss_error;
                break;
            case RSDT_BINARY:
    	        switch (mysqltype) {
                    case MYSQL_TYPE_VAR_STRING: /* old <= 4.1 VARCHAR */
                    case MYSQL_TYPE_VARCHAR:    /* new >= 5.0.3 true VARCHAR */
                        {
                            uint length;
                            ss_dassert(field->binary());
                            if (field->real_type() != MYSQL_TYPE_VARCHAR) {
                                /* string */
                                length = field->field_length;
                            } else {
                                /* varstring */
                                if (iskeyptr) {
                                    length = uint2korr(ptr);
                                    ptr = ptr + 2;
                                } else {
                                    uint length_bytes = field->pack_length() - field->key_length();
                                    length = length_bytes == 1 ? (uint) (uchar) *ptr : uint2korr(ptr);
                                    ptr = ptr + length_bytes;
                                }
                            }
                            if (dva != NULL) {
                                va_setdataandnull(dva, (void*)ptr, length);
                            } else {
                                avret = rs_aval_setbdata_ext(cd, atype, aval, (char*)ptr, length, p_errh);
                                succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                            }
                        }
                        break;
                    case FIELD_TYPE_NEWDECIMAL:
                    case FIELD_TYPE_DECIMAL:
                    case MYSQL_TYPE_BIT:
                        if (dva != NULL) {
                            va_setdataandnull(dva, (void*)ptr, field->pack_length());
                        } else {
                            avret = rs_aval_setbdata_ext(
                                        cd,
                                        atype,
                                        aval,
                                        (char*)ptr,
                                        field->pack_length(),
                                        p_errh);
                            succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                        }
                        break;
                    case FIELD_TYPE_GEOMETRY:
                    case FIELD_TYPE_TINY_BLOB:
                    case FIELD_TYPE_MEDIUM_BLOB:
                    case FIELD_TYPE_BLOB:
                    case FIELD_TYPE_LONG_BLOB:
                        {
                            char* buf;
                            uint length;
                            if (iskeyptr) {
                                short j;
                                shortget(j, ptr);
                                length = j;
                                buf = (char*)ptr + 2;
                            } else {
                                field_blob = (Field_blob*)field;
                                ss_pprintf_1(("BLOB:type %d, real_type %d, field_length %d, field->binary() == %d\n",
                                          mysqltype, field->real_type(), field->field_length, field->binary()));

                                length = field_blob->get_length();
                                field_blob->get_ptr(&buf);
                            }
                            ss_pprintf_1(("FIELD_BLOB:BINARY:iskeyptr %d, field_length %d value=(%.20s)\n", iskeyptr, length, buf));
                            if (dva != NULL) {
                                va_setdataandnull(dva, buf, length);
                            } else {
                                avret = rs_aval_setbdata_ext(cd, atype, aval, buf, length, p_errh);
                                succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                            }
                        }
                        break;
                    default:
                        ss_error;
                }
                break;
            case RSDT_UNICODE:
                ss_error;
                break;
            case RSDT_BIGINT:
                {
                    ulonglong j;
                    ss_int8_t i8;

                    switch(field->real_type()) {
                        case FIELD_TYPE_LONG: {
#ifdef WORDS_BIGENDIAN
                            if (table->s->db_low_byte_first) {
                                ulong k;
                                k = sint4korr(ptr);
                                j = (ulonglong)k;
                            } else
#endif
                            {
                                if (((Field_num*)field)->unsigned_flag) {
                                    ulong k;

                                    ulongget(k, ptr);
                                    j = (ulonglong)k;
                                } else {
                                    long k;

                                    longget(k, ptr);
                                    j = (ulonglong)k;
                                }
                            }

                            break;
                        }
                        default: {
#ifdef WORDS_BIGENDIAN
                            if (table->s->db_low_byte_first) {
                                j = sint8korr(ptr);
                            } else
#endif
                            {
                                longlongget(j, ptr);
                            }
                            break;
                        }
                    }

                    SsInt8SetNativeUint8(&i8, j);

                    if (dva != NULL) {
                        va_setint8(dva, i8);
                    } else {
                        avret = rs_aval_setint8_ext(cd, atype, aval, i8, p_errh);
                        succp = (avret == RSAVR_SUCCESS || avret == RSAVR_TRUNCATION);
                    }
                }
                break;
            default:
                ss_error;
        }
        if (!succp && p_errh != NULL) {
            sql_print_error("%s (%s)", su_err_geterrstr(*p_errh), thd->query);
        }

#ifdef SS_DEBUG
        if (aval) {
            char* tmp;
            tmp = rs_aval_print(cd, atype, aval);
            ss_pprintf_1(("Field %s stored value '%s'\n", field->field_name, tmp));
            SsMemFree(tmp);
        }
#endif

#if MYSQL_VERSION_ID >= 50100
        dbug_tmp_restore_column_map(table->read_set, old_map);
#endif

        return(succp);
}

/*#***********************************************************************\
 *
 *              do_checkpoint
 *
 * This function flushes a logs and does a checkpoint for the MySQL/solidDB
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     char* parameters,  in, use, parameters
 *     su_err_t**  p_errh, in out, NULL
 *
 * Return value : solidDB return code
 *
 * Globals used: current_thd
 */
static su_ret_t do_checkpoint(
        rs_sysi_t* cd,
        char* parameters __attribute__((unused)),
        su_err_t** p_errh)
{
        su_ret_t rc = SU_SUCCESS;
        tb_connect_t* tc = NULL;
        dbe_db_t* soliddb = NULL;

        DBUG_ENTER("solid:do_checkpoint");
        ss_pprintf_1(("do_checkpoint\n"));

        ss_dassert(cd != NULL);

        soliddb = (dbe_db_t*)rs_sysi_db(cd);

        dbe_db_logflushtodisk(soliddb);

        tc = tb_sysconnect_init((tb_database_t*)rs_sysi_tabdb(cd));

        /* adding a checkpoint record to the end of log */

        rc = tb_createcheckpoint(tc,/*splitlog = */ TRUE);

        if (rc != DBE_RC_SUCC) {
            su_err_init(p_errh, rc);
        }

        tb_sysconnect_done(tc);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              solid_clear_dir
 *
 * This function removes files from the directory
 *
 * Parameters :
 *
 *     const char* path, in, use, directory name
 *
 * Return value : 0  for SUCCESS
 *                -1 in case of error
 *
 * Globals used: -
 */
static int solid_clear_dir(
        const char* path )
{
    SDB_DBUG_ENTER("solid_clear_dir");

    MY_DIR *dirp = my_dir( path, MYF(MY_DONT_SORT|MY_WANT_STAT) );

    if( !dirp ) {
        goto ok;
    }

    for ( uint i = 0; i < dirp->number_off_files; ++i ) {
        FILEINFO *file = dirp->dir_entry + i;
        char filePath[FN_REFLEN];

        /* skipping this directory and parent directory references. */
        if ( MY_S_ISDIR(file->mystat->st_mode) &&
            (!strcmp( file->name, "." ) || !strcmp( file->name, ".." )) ) {
            continue;
        }

        strxmov( filePath, path, "/", file->name, NullS );

        if ( MY_S_ISDIR(file->mystat->st_mode) ?
                      solid_clear_dir( filePath ) || rmdir( filePath ) : /* for dirs call itself recursively */
                      my_delete_with_symlink( filePath, MYF(MY_WME) ) /* otherwise just remove it */
            ) {
            goto err;
        }
    }

ok:
    my_dirend(dirp);
    SDB_DBUG_RETURN(0);

err:
    my_dirend( dirp );
    SDB_DBUG_RETURN(-1);
}

static void handle_backslashes(const char* src, char* dst)
{
    while (*src)
    {
        if (*src == '\\')
            *dst++ = '\\';
        *dst++ = *src++;
    }
    *dst = 0;
}

static bool fwrite_string(FILE *file, const char *str)
{
    uint cnt = strlen(str);
    return my_fwrite(file, (byte*) str, cnt, MYF(0)) == cnt;
}

static int backup_config(const char *backup_dir)
{
    char cnf_file_name[FN_REFLEN];
    FILE *file;
    char buf[255];
    char valbuf[255];
    int errcode = -1;

    strxmov( cnf_file_name, backup_dir, "/", BACKUP_CONFIG_FILE_NAME, NULL );

    if ( !(file = my_fopen(cnf_file_name, O_CREAT|O_WRONLY, MYF(0))) ) {
        return -1;
    }

    if (!fwrite_string(file, "[mysqld]\n")) {
        goto FINISH;
    }

    for (struct my_option *opt = my_long_options; opt->name; opt++) {
        gptr *pvalue = opt->value; // = (opt->var_type & GET_ASK_ADDR ? (*getopt_get_addr)("", 0, opt) : opt->value);

        if (!pvalue) {
            continue;
        }

        switch (opt->var_type & GET_TYPE_MASK)
        {
        case GET_NO_ARG:
        case GET_BOOL:
            {
                my_bool value = *(my_bool*) pvalue;
                if (value == (my_bool) opt->def_value)
                    continue;
                if (value)
                    sprintf(buf, "%s\n", opt->name);
                else
                    sprintf(buf, "skip-%s\n", opt->name);
                break;
            }
        case GET_INT:
            {
                int value = *(int*) pvalue;
                if (value == (int) opt->def_value)
                    continue;
                sprintf(buf, "%s=%d\n", opt->name, value);
                break;
            }
        case GET_UINT:
            {
                uint value = *(uint*) pvalue;
                if (value == (uint) opt->def_value)
                    continue;
                sprintf(buf, "%s=%u\n", opt->name, value);
                break;
            }
        case GET_LONG:
            {
                long value = *(long*) pvalue;
                if (value == (long) opt->def_value)
                    continue;
                sprintf(buf, "%s=%ld\n", opt->name, value);
                break;
            }
        case GET_ULONG:
            {
                ulong value = *(ulong*) pvalue;
                if (value == (ulong) opt->def_value)
                    continue;
                sprintf(buf, "%s=%lu\n", opt->name, value);
                break;
            }
        case GET_LL:
            {
                longlong value = *(longlong*) pvalue;
                if (value == (longlong) opt->def_value)
                    continue;
                longlong2str(value, valbuf, -10);
                sprintf(buf, "%s=%s\n", opt->name, valbuf);
                break;
            }
        case GET_ULL:
            {
                ulonglong value = *(ulonglong*) pvalue;
                if (value == (ulonglong) opt->def_value)
                    continue;
                longlong2str(value, valbuf, 10);
                sprintf(buf, "%s=%s\n", opt->name, valbuf);
                break;
            }
        case GET_STR:
        case GET_STR_ALLOC:
            {
                char* value = *(char**) pvalue;
                if (!value)
                    continue;
                if (!SsStricmp(opt->name, "datadir"))
                    value = mysql_real_data_home;
                if (opt->def_value && *(char**) opt->def_value && !SsStricmp(value, *(char**) opt->def_value))
                    continue;
                if ((!SsStricmp(opt->name, "myisam-recover") || !SsStricmp(opt->name, "sql-mode")) &&
                    !SsStricmp(value, "off"))
                    continue;
                handle_backslashes(value, valbuf);
                sprintf(buf, "%s=%s\n", opt->name, valbuf);
                break;
            }
        default:
            continue;
        }

        if (!fwrite_string(file, buf)) {
            goto FINISH;
        }
    }

    errcode = 0;

FINISH:
    my_fclose(file, MYF(0));
    return errcode;
}

/*#***********************************************************************\
 *
 *              copy_mysql_files_related_to_solid
 *
 * Performs backup of Solid database.
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     char* parameters,  in, use
 *     su_err_t**  p_errh, in out, NULL
 *
 * Return value : solidDB return code
 */
static int copy_mysql_files_related_to_solid( const char* backup_dir )
{
        SDB_DBUG_ENTER("copy_mysql_files_related_to_solid");

        FILE *backup_cnf; /* file where the list of all backed up tables is written.*/
        {
            char backup_list_path[FN_REFLEN];
            strxmov( backup_list_path, backup_dir, "/", BACKUP_LIST_FILE_NAME, NULL );
            backup_cnf = my_fopen( backup_list_path, O_WRONLY|O_CREAT|O_TRUNC, MYF(0) );
        }

        MY_DIR *databases = my_dir( mysql_data_home, MYF(MY_DONT_SORT|MY_WANT_STAT) );
        if ( !databases ) {
            return 1;
        }

        /* fill list of all databases. */
        for ( uint d = 0; d < databases->number_off_files; ++d ) {

            /* skipping not directories, this directory and parent directory references. */
            if ( !MY_S_ISDIR(databases->dir_entry[d].mystat->st_mode) ||
                 MY_S_ISDIR(databases->dir_entry[d].mystat->st_mode) &&
                (!strcmp( databases->dir_entry[d].name, "." ) ||
                 !strcmp( databases->dir_entry[d].name, ".." )) ) {
                continue;
            }

            const char * const db = databases->dir_entry[d].name;

            ss_pprintf_1(("\t : database is %s.\n", db));

            char db_path[FN_REFLEN];
            strxmov( db_path, mysql_data_home, "/", db, NULL );

            MY_DIR *dirp = my_dir( db_path, MYF(MY_DONT_SORT|MY_WANT_STAT) );
            if ( !dirp ) {
                continue;
            }

            //List_iterator_fast<char> frmi(frm_files);
            //while((frm = frmi++)) {
            for ( uint i = 0; i < dirp->number_off_files; ++i ) {

                /* skipping this directory and parent directory references. */
                if ( MY_S_ISDIR(dirp->dir_entry[i].mystat->st_mode) &&
                    (!strcmp( dirp->dir_entry[i].name, "." ) ||
                     !strcmp( dirp->dir_entry[i].name, ".." )) ) {
                    continue;
                }

                char *afile = dirp->dir_entry[i].name;

                ss_pprintf_1(("\t : file is %s.\n", afile));

                char path[FN_REFLEN];
                strxmov( path, db_path, "/", afile, NULL );

                /* TODO check open mode: it is better to restrict other access. */
                File frm_file = my_open( path, O_RDONLY | O_SHARE, MYF(0) );

                if ( frm_file >= 0 ) {
                    ss_pprintf_1(("\t : file %s has been opened.\n", afile));

                    if ( true || is_soliddb_table( frm_file ) ) {

                        ss_pprintf_1(("\t : file %s is a SolidDB table and will be copied.\n", afile));

                        /* copying FRM file to the backup directory. */
                        char dstdir[FN_REFLEN];

                        strxmov( dstdir, backup_dir, "/", db, NULL );
                        ss_pprintf_2(("copy_mysql_files_related_to_solid: SolidDB FRM: %s\n", path));

                        if ( !copy_file( frm_file, dstdir, afile, "" ) ) {
                            my_close( frm_file, MYF(MY_WME) );
                            my_fclose( backup_cnf, MYF(MY_WME) );
                            SDB_DBUG_RETURN(1);
                        }

                        fprintf( backup_cnf, "%s.%s\n", db, afile );
                    }
                    my_close( frm_file, MYF(MY_WME) );
                }
            }
        }

        my_fclose( backup_cnf, MYF(MY_WME) );
        SDB_DBUG_RETURN(0);
}


/*#***********************************************************************\
 *
 *              do_backup
 *
 * Performs backup of Solid database.
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     char* parameters,  in, use
 *     su_err_t**  p_errh, in out, NULL
 *
 * Return value : solidDB return code
 */
static su_ret_t do_backup(
        rs_sysi_t* cd,
        char* parameters,
        su_err_t** p_errh )
{
        su_ret_t rc = SU_SUCCESS;
        THD* thd = current_thd;
        dbe_db_t* soliddb = NULL;
        char *db = NULL;
        char* backup_dir = parameters;
        tb_connect_t* tc;
        char backup_dir_buf[FN_REFLEN];

        SDB_DBUG_ENTER("solid:do_backup");

        block_solid_DDL();

        soliddb = (dbe_db_t*)rs_sysi_db(cd);
        ss_assert( soliddb );

        if (!backup_dir || !*backup_dir) {
            SsFlatMutexLock(soliddb_backupdir_mutex);
            strncpy(backup_dir_buf, soliddb_backupdir, sizeof(backup_dir_buf) - 1);
            backup_dir = backup_dir_buf;
            SsFlatMutexUnlock(soliddb_backupdir_mutex);
        }

        if (!*backup_dir) {
            rc = DBE_ERR_FAILED;
            goto epilogue;
        }

        if( (rc = dbe_db_backupcheck( soliddb, backup_dir, p_errh ) ) != DBE_RC_SUCC ) {
            ui_msg_message(BACKUP_MSG_FAILED_S, su_err_geterrstr(*p_errh));
            rc = DBE_ERR_FAILED;
            goto epilogue;
        }

        if( solid_clear_dir( backup_dir ) ) {
            goto epilogue;
        }

        if( backup_config( backup_dir ) ) {
            goto epilogue;
        }

        /* adding a checkpoint record to the end of log */
        tc = tb_sysconnect_init( (tb_database_t*)rs_sysi_tabdb(cd) );

        if ( (rc = tb_createcheckpoint( tc, /*splitlog = */ TRUE )) != DBE_RC_SUCC ) {
            su_err_init( p_errh, rc );
        }
        tb_sysconnect_done(tc);

        if ( rc != DBE_RC_SUCC ) {
            goto epilogue;
        }

        if ( (rc = dbe_db_backupstart(soliddb, backup_dir, FALSE, p_errh)) != DBE_RC_SUCC) {
            goto epilogue;
        }

        ui_msg_message(BACKUP_MSG_STARTED_SS, "directory", backup_dir );

        do {
            rc = dbe_db_backupadvance(soliddb, p_errh);
        } while(rc == DBE_RC_CONT);

        if (rc == DBE_RC_END) {
            rc = SU_SUCCESS;
        } else {
            ui_msg_message(BACKUP_MSG_FAILED_S, su_err_geterrstr(*p_errh));
            rc = DBE_ERR_FAILED;
        }

        dbe_db_backupstop(soliddb);

        if( rc != SU_SUCCESS ) {
            goto epilogue;
        }

        if( copy_mysql_files_related_to_solid( backup_dir ) ) {
            rc = DBE_ERR_FAILED;
            goto epilogue;
        }

epilogue:
        unblock_solid_DDL();
        if ( rc == SU_SUCCESS || rc == DBE_RC_SUCC ) {
            ui_msg_message(BACKUP_MSG_COMPLETED);
        } else {
            if (*p_errh == NULL) {
                su_err_init(p_errh, rc);
            }

            ui_msg_message(BACKUP_MSG_START_FAILED_WITH_MSG_S, su_err_geterrstr(*p_errh));
        }
        SDB_DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              do_performance_monitor
 *
 * Performs performance monitor of the database.
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     char* parameters,  in, use
 *     su_err_t**  p_errh, in out, NULL
 *
 * Return value : solidDB return code
 */
static su_ret_t do_performance_monitor(
        rs_sysi_t* cd,
        char* parameters,
        su_err_t** p_errh)
{
        uint interval;
        bool append = FALSE;
        bool raw = FALSE;
        char filename[256];
        char comment[256];
        bool pmon_clear = FALSE;
        bool pmon_start = FALSE;
        bool pmon_stop = FALSE;
        bool succp = FALSE;
        su_ret_t rc = SU_SUCCESS;

        if (parameters == NULL) {
            return (DBE_ERR_FAILED);
        }

        su_pars_match_t m;
        su_pars_match_init( &m, parameters );

        if (su_pars_match_const(&m, (char *)"diff")) {
            if (su_pars_match_const(&m, (char *)"clear")) {
                pmon_clear = TRUE;
                succp = TRUE;
            } else if (su_pars_match_const(&m, (char *)"stop")) {
                pmon_stop = TRUE;
                succp = TRUE;
            } else if (su_pars_match_const(&m, (char *)"start")) {


                if (su_pars_get_filename(&m, filename, sizeof(filename))) {
                } else {
                    strcpy(filename, (char *)"pmondiff.out");
                }

                if (!su_pars_get_uint(&m, &interval)) {
                    interval = 1000;
                }

                if (su_pars_match_const(&m, (char *)"append")) {
                    append = TRUE;
                }
                raw = FALSE;

                if (su_pars_match_const(&m, (char *)"raw")) {
                    raw = TRUE;
                }

                if (!su_pars_get_stringliteral(&m, comment, sizeof(comment))) {
                    comment[0] = 0;
                }

                pmon_start = TRUE;
                succp = TRUE;
            }
        }

        if (su_pars_match_const(&m, (char *)"") && succp) {

            if (pmon_clear) {
                ss_pprintf_3(("do_performance_monitor:pmon clear\n"));
                tb_srv_pmondiff_clear(cd);

            } else if (pmon_stop) {
                ss_pprintf_3(("do_performance_monitor:pmon stop\n"));

                if (!tb_srv_pmondiff_stop(cd, p_errh)) {
                    return(DBE_ERR_FAILED);
                }

            } else if (pmon_start) {
                ss_pprintf_3(("do_performance_monitor:pmon start\n"));

                if (!tb_srv_pmondiff_start(cd, interval, filename, append, raw, comment, p_errh)) {
                    return(DBE_ERR_FAILED);
                }
            } else {
                return(DBE_ERR_FAILED);
            }

            return SU_SUCCESS;
        } else {
            return(DBE_ERR_FAILED);
        }
}
/*#***********************************************************************\
 *
 *              do_command
 *
 * Finds an implementation function for administrative command and executes it.
 * From the command parameters are extracted. Parameters are believed to follow
 * the command. Leading whitespaces are removed from the beginning of the parameters.
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use, system info data type
 *     const char* cmd, in, use, admin command
 *     const char* param, in, use, parameters for admin command or NULL
 *     su_err_t**  p_errh, in out, NULL, error structure
 *
 * Return value : solidDB return code
 *
 * Globals used :
 */
static su_ret_t do_command(
        rs_sysi_t* cd,
        const char* cmd,
        const char* parameters,
        su_err_t**  p_errh)
{
        uint n_command = 0;
        const char* command = NULL;
        unsigned long len = 0;

        SDB_DBUG_ENTER("solid:do_command");

        len = strlen(cmd);

        for( n_command = 0; solid_admin_commands[n_command].command_str != NULL; ++n_command ) {

            const char * const this_command = solid_admin_commands[n_command].command_str;

            command = SsStristr(cmd, this_command);

            if ( command && command == cmd ) {

                if ( !parameters && len > strlen(this_command) ) {
                    parameters = cmd + strlen(this_command);
                    /* shoot whitespaces */
                    for( ; *parameters && strchr(SQL_WHITESPACES, *parameters); ++parameters );

                    if ( !*parameters ) {
                        parameters = NULL;
                    }
                }

                /* now really do the command */
                su_ret_t rc = solid_admin_commands[n_command].command_func(cd, (char *)parameters, p_errh);

                SDB_DBUG_RETURN(rc);
            }
        }

        SDB_DBUG_RETURN(DBE_ERR_FAILED);
}

/*#***********************************************************************\
 *
 *              check_admin_command
 *
 * Check wheather user has issued a admin command e.g. backup
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use, system info data type
 *     tb_trans_t* trans, in, use, transaction
 *     rs_ttype_t* ttype, in, use, tuple type
 *     rs_tval_t*  tval, in, use, tuple value
 *     su_err_t**  p_errh, in out, NULL, error structure
 *
 * Return value : solidDB return code
 *
 * Globals used :
 */
static su_ret_t check_admin_command(
        rs_sysi_t* cd,
        tb_trans_t* trans,
        rs_ttype_t* ttype,
        rs_tval_t* tval,
        su_err_t** p_errh)
{
        char* cmd;
        su_ret_t rc = SU_SUCCESS;
        rs_aval_t* cmd_val;
        rs_aval_t* params_val;
        char *params;
        size_t nattrs = rs_ttype_sql_nattrs(cd, ttype);
        rs_atype_t* cmd_type = rs_ttype_sql_atype(cd, ttype, 0);
        rs_atype_t* params_type = rs_ttype_sql_atype( cd, ttype, 1 );
        rs_datatype_t pdatatype;

        ss_pprintf_3(("check_admin_command\n"));

        if ( (pdatatype = rs_atype_datatype(cd, cmd_type)) != RSDT_CHAR ) {
            ss_pprintf_3(("check_admin_command: datatype of 1st field is wrong: %d\n", pdatatype));
            goto admin_command_error;
        }

        if ( (pdatatype = rs_atype_datatype( cd, params_type )) != RSDT_CHAR) {
            ss_pprintf_3(("check_admin_command: datatype of 2nd field is wrong: %d\n", pdatatype));
            goto admin_command_error;
        }

        cmd_val = rs_tval_sql_aval(cd, ttype, tval, 0);

        if ( rs_aval_isnull(cd, cmd_type, cmd_val) ) {
            ss_pprintf_3(("check_admin_command: COMMAND should not be null\n"));
            goto admin_command_error;
        }

        cmd = rs_aval_getasciiz(cd, cmd_type, cmd_val);

        params_val = rs_tval_sql_aval( cd, ttype, tval, 1 );
        params = NULL;

        if ( !rs_aval_isnull( cd, params_type, params_val ) ) {
            params = rs_aval_getasciiz( cd, params_type, params_val );
        }

        //ss_pprintf_100(("check_admin_command: backup parameters are: [%s]\n", params));

        ss_pprintf_3(("ADMIN COMMAND = %s\n", cmd));

        rc = do_command(cd, cmd, params, p_errh);

        if (rc == SU_SUCCESS || p_errh != NULL) {
            return (rc);
        }

admin_command_error:
        ss_pprintf_3(("check_admin_command:wrong command\n"));
        su_err_init(p_errh, DBE_ERR_FAILED);

        return DBE_ERR_FAILED;
}

/*#***********************************************************************\
 *
 *              soliddb_check_admin_command
 *
 * Check admin command given by the user
 *
 * Parameters :
 *
 *     THD*      thd, in, use, thread data
 *     set_var*  var, in, use, variable value from MySQL
 *
 * Return value : 0 or -1 (error)
 *
 * Globals used :
 */
int soliddb_check_admin_command(
        THD*thd,
        set_var *var)
{
        su_ret_t rc = SU_SUCCESS;
        su_err_t* errh=NULL;
        SOLID_CONN* con=NULL;

        DBUG_ENTER("soliddb_check_admin_command");

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_pprintf_1(("Admin command = %s\n", var->value->name));

        soliddb_admin_command = var->value->name;

        rc = soliddb_admin_command ?
            do_command(con->sc_cd, var->value->name, NULL, &errh) : DBE_ERR_FAILED;

        if (rc != SU_SUCCESS) {
            if (errh) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
                su_err_done(errh);
                errh = NULL;
            }

            DBUG_RETURN(-1);
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              soliddb_update_admin_command
 *
 * Update admin command string given by the user
 *
 * Parameters :
 *
 *     THD*      thd, in, use, thread data
 *     set_var*  var, in, use, variable value from MySQL
 *
 * Return value : 0
 *
 * Globals used :
 */
bool soliddb_update_admin_command(
        THD*thd,
        set_var *var)
{
        DBUG_ENTER("soliddb_update_admin_command");

        soliddb_admin_command = var->value->name;

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::write_row
 *
 * write_row() inserts a row. No extra() hint is given currently if a bulk load
 * is happeneding. buf() is a byte array of data. You can use the field
 * information to extract the data from the native byte array type.
 * Example of this would be:
 * for (Field **field=table->field ; *field ; field++)
 * {
 *   ...
 * }
 *
 * See ha_tina.cc for an example of extracting all of the data as strings.
 * ha_berekly.cc has an example of how to store it intact by "packing" it
 * for ha_berkeley's own native storage type.
 *
 * See the note for update_row() on auto_increments and timestamps. This
 * case also applied to write_row().
 *
 * Called from item_sum.cc, item_sum.cc, sql_acl.cc, sql_insert.cc,
 * sql_insert.cc, sql_select.cc, sql_table.cc, sql_udf.cc, and sql_update.cc.
 *
 *
 * Parameters :
 *
 *     byte*   buf, in, use, values
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* thd
 */
int ha_soliddb::write_row(byte * buf)
{
        int rc;
        bool succp = TRUE;
        Field** fieldptr;
        Field* field;
        rs_ttype_t* ttype;
        rs_tval_t* tval;
        rs_atype_t* atype;
        rs_aval_t* aval;
    	int	ano;
        int nattrs;
        su_ret_t suret;
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        THD* thd;
        void* cont = NULL;
        su_err_t* errh = NULL;
        solid_bool finished = FALSE;

        DBUG_ENTER("ha_solid::write_row");
        ss_pprintf_1(("ha_solid::write_row\n"));

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);
        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;
        con->sc_table = table;

        if (solid_rsrelh == NULL) {
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            DBUG_RETURN(rc);
        }

        if (rs_relh_issysrel(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        if (rs_relh_isaborted(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

      	statistic_increment(thd->status_var.ha_write_count, &LOCK_status);

        ttype = rs_relh_ttype(cd, solid_rsrelh);
        tval = rs_tval_create(cd, ttype);
        nattrs = rs_ttype_sql_nattrs(cd, ttype);

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        ss_pprintf_2(("ha_solid::write_row, table %s\n",
                                         rs_relh_name(cd, solid_rsrelh)));

        if (table->next_number_field && buf == table->record[0]) {
            /* This is the case where the table has an
		    auto-increment column */

            /* TODO: Here we should use transactional lock mechanism on the
               auto-inc counter of the table to ensure that replication and
               roll-forward of the binlog exactly imitates also the given
               auto-inc values at least for statement based replication.*/

            update_auto_increment();
        }

        /* If we have a timestamp column, update it to the current time */
        if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_INSERT) {
            table->timestamp_field->set_time();
        }

        for (fieldptr = table->field, ano = 0; *fieldptr != NULL; fieldptr++, ano++) {
        	int	phys_ano;

            field = *fieldptr;
            phys_ano = rs_ttype_quicksqlanotophys(cd, ttype, ano);
            atype = rs_ttype_atype(cd, ttype, phys_ano);
            aval = rs_tval_aval(cd, ttype, tval, phys_ano);

            if (field->null_ptr != NULL && (field->null_bit & *field->null_ptr)) {
                ss_pprintf_1(("ha_solid::write_row:NULL\n"));
                rs_aval_setnull(cd, atype, aval);
            } else {
                succp = set_mysql_field_to_aval_or_dynva(
                            thd,
                            table,
                            cd,
                            atype,
                            aval,
                            NULL,
                            field,
#if MYSQL_VERSION_ID >= 50100
                            (const byte*)(buf + field->offset(field->table->record[0])),
#else
                            (const byte*)(buf + field->offset()),
#endif
                            FALSE,
                            &errh);
                if (!succp) {
                    break;
                }
            }
            ss_poutput_2(ha_print_aval(cd, atype, aval, (char *)"ha_solid::write_row"));
        }
        ss_poutput_2(rs_tval_print(cd, ttype, tval));

        if (succp) {
            if (strcmp(rs_relh_name(cd, solid_rsrelh), "SOLIDDB_ADMIN_COMMANDS") == 0) {
                suret = check_admin_command(cd, trans, ttype, tval, &errh);
            } else {
                do {
                    suret = dbe_rel_insert(
                                tb_trans_dbtrx(cd, trans),
                                solid_rsrelh,
                                tval,
                                &errh);
                } while (rs_sysi_lockwait(cd) || suret == DBE_RC_WAITLOCK);
            }
        }
        
        if (succp && suret == DBE_RC_SUCC) {
            succp = TRUE;
        } else {
            succp = FALSE;
            ss_dassert(errh != NULL);
        }

        /* If we should ignore duplicate keys we need to do a
           statement commit to get those errors. Similarly, if this
           is a alter table or create index we need to produce all
           errors now. */

        if (succp &&
            (solid_ignore_duplicate(thd) ||
             thd->lex->sql_command == SQLCOM_ALTER_TABLE ||
             thd->lex->sql_command == SQLCOM_CREATE_INDEX)) {
            do {
                succp = tb_trans_stmt_commit(cd, trans, &finished, &errh);
            } while (!finished);
        }

        if (!succp &&
            (thd->lex->sql_command == SQLCOM_REPLACE || thd->lex->sql_command == SQLCOM_REPLACE_SELECT)) {
            /* For REPLACE we can store created tval to connection for update */
            con->sc_tval = tval; 
        } else {
            rs_tval_free(cd, ttype, tval);
            con->sc_tval = NULL;
        }


        if (!succp) {
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);

        } else {
            rc = 0;
        }

        ss_pprintf_1(("ha_solid::write_row:return %d\n", rc));

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ha_solid_cascade
 *
 * Do a foreign key cascade operations for a update or delete
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     tb_trans_t* trans, in, use
 *     rs_relh_t*  solid_rsrelh, in, use
 *     solid_bool* flags, in, use or NULL in delete
 *     rs_tval_t*  old_tval, in, use
 *     rs_tval_t*  new_tval, in, use or NULL in delete
 *     rs_ttype_t* ttype, in, use
 *     su_err_t**  p_errh, in out, NULL
 *
 * Return value : solidDB error code or DBE_RC_SUCC
 *
 * Globals used :
 */
static su_ret_t ha_solid_cascade(
    rs_sysi_t*  cd,
    tb_trans_t* trans,
    rs_relh_t*  relh,
    solid_bool* flags,
    rs_tval_t*  old_tval,
    rs_tval_t*  new_tval,
    rs_ttype_t* ttype,
    su_err_t**  errh)
{
        uint i = 0;
        su_pa_t *keys = rs_relh_refkeys(cd, relh);
        su_ret_t rc = DBE_RC_SUCC;

        su_pa_do(keys, i) {
            rs_key_t *key = (rs_key_t *)su_pa_getdata(keys, i);

            if (key != NULL && rs_key_type(cd, key) == RS_KEY_PRIMKEYCHK) {
                uint rc = 0;
                tb_trans_keyaction_state_t *casc_state = NULL;

                do {
                    rc = tb_ref_keyaction(cd, trans,
                                          key, ttype, flags,
                                          old_tval, new_tval,
                                          &casc_state, errh);
                } while (rc == TB_CHANGE_CONT);

                if (rc == TB_CHANGE_ERROR) {
                    rc = DBE_ERR_CHILDEXIST_S;
                    goto cascade_error;
                }
            }
        }

        rc = DBE_RC_SUCC;

cascade_error:

        return (rc);
}

/*#***********************************************************************\
 *
 *              ::update_row
 *
 * Yes, update_row() does what you expect, it updates a row. old_data will have
 * the previous row record in it, while new_data will have the newest data in
 * it.
 *
 * Keep in mind that the server can do updates based on ordering if an ORDER BY
 * clause was used. Consecutive ordering is not guarenteed.
 * Currently new_data will not have an updated auto_increament record, or
 * and updated timestamp field. You can do these for example by doing these:
 * if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
 *   table->timestamp_field->set_time();
 * if (table->next_number_field && record == table->record[0])
 *   update_auto_increment();
 *
 * Called from sql_select.cc, sql_acl.cc, sql_update.cc, and sql_insert.cc.
 *
 * Parameters :
 *
 *     const byte*  old_data, in, use, old data
 *     byte*        new_data, in, use, new data
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* thd
 */
int ha_soliddb::update_row(
        const byte * old_data,
        byte * new_data)
{
        int rc;
        su_ret_t suret;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        rs_ttype_t* ttype;
        rs_tval_t* tval;
        rs_atype_t* atype;
        rs_aval_t* aval;
        Field** fieldptr;
        Field* field;
    	int	ano;
        solid_bool* selflags;
        int nattrs;
        int physnattrs;
        THD* thd;
        SOLID_CONN*   con;
        bool succp = TRUE;
        su_err_t* errh = NULL;
        solid_bool finished = FALSE;

        DBUG_ENTER("ha_soliddb::update_row");
        ss_pprintf_1(("ha_solid::update_row\n"));

        thd = current_thd;

        ss_dassert(solid_relcur != NULL);
        ss_dassert(solid_relcur->sr_relcur != NULL);
        CHK_RELCUR(solid_relcur);

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);
        cd = solid_relcur->sr_cd;
        ss_dassert(cd != NULL);
        ss_dassert(solid_rsrelh != NULL);

        if (rs_relh_issysrel(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }
        if (rs_relh_isaborted(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        statistic_increment(thd->status_var.ha_update_count, &LOCK_status);

        trans = solid_relcur->sr_trans;
        ttype = solid_relcur->sr_ttype;

        ss_dassert(ttype != NULL);

        if (thd->lex->sql_command != SQLCOM_REPLACE && thd->lex->sql_command != SQLCOM_REPLACE_SELECT) {
            tval = rs_tval_create(cd, ttype);
        } else {
            tval = con->sc_tval;
            ss_dassert(tval != NULL);
        }
        
        physnattrs = rs_ttype_nattrs(cd, ttype);
        selflags = (solid_bool*)SsMemCalloc(physnattrs, sizeof(selflags[0]));
        nattrs = rs_ttype_sql_nattrs(cd, ttype);
        con->sc_table = table;

        ss_pprintf_2(("ha_solid::update_row, table %s\n", rs_relh_name(cd, solid_rsrelh)));

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
        my_bitmap_map *old_map = dbug_tmp_use_all_columns(table, table->read_set);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        if (thd->lex->sql_command != SQLCOM_REPLACE && thd->lex->sql_command != SQLCOM_REPLACE_SELECT) {
            /* If we have a timestamp column, update it to the current time */
            if (table->timestamp_field_type & TIMESTAMP_AUTO_SET_ON_UPDATE)
                table->timestamp_field->set_time();

            for (fieldptr = table->field, ano = 0; *fieldptr != NULL; fieldptr++, ano++) {
                field = *fieldptr;

                if (table->found_next_number_field &&
                    field == table->found_next_number_field &&
                    field->val_int() != 0
#if MYSQL_VERSION_ID >= 50100
                    && bitmap_is_set(table->write_set, field->field_index)) {
#else
                    ) {
#endif

                    table->next_number_field = field;
                    update_auto_increment();
                }

                if (solid_relcur->sr_usedfields[ano]) {
                    int	phys_ano;
                    phys_ano = rs_ttype_quicksqlanotophys(cd, ttype, ano);
                    atype = rs_ttype_atype(cd, ttype, phys_ano);
                    aval = rs_tval_aval(cd, ttype, tval, phys_ano);
                    selflags[phys_ano] = TRUE;
                    ss_dprintf_2(("ha_solid::update_row, ano=%d, aname=%s\n", phys_ano, rs_ttype_aname(cd, ttype, phys_ano)));

                    if (field->null_ptr != NULL && (field->null_bit & *field->null_ptr)) {
                        ss_pprintf_1(("ha_solid::update_row:NULL\n"));
                        rs_aval_setnull(cd, atype, aval);
                    } else {
                        succp = set_mysql_field_to_aval_or_dynva(
                                thd,
                                table,
                                cd,
                                atype,
                                aval,
                                NULL,
                                field,
#if MYSQL_VERSION_ID >= 50100
                                (const byte*)(new_data + field->offset(field->table->record[0])),
#else
                                (const byte*)(new_data + field->offset()),
#endif
                                FALSE,
                                &errh);
                        if (!succp) {
                            break;
                        }
                    }
                    ss_poutput_2(ha_print_aval(cd, atype, aval, (char *)"ha_solid::update_row"));
                }
            }
        } else {
            for (fieldptr = table->field, ano = 0; *fieldptr != NULL; fieldptr++, ano++) {
                field = *fieldptr;

                if (solid_relcur->sr_usedfields[ano]) {
                    int	phys_ano;
                    phys_ano = rs_ttype_quicksqlanotophys(cd, ttype, ano);
                    selflags[phys_ano] = TRUE;
                }
            }
        }
        
#if MYSQL_VERSION_ID >= 50100
        dbug_tmp_restore_column_map(table->read_set, old_map);
#endif

        ss_poutput_2(rs_tval_print(cd, ttype, tval));

        if (succp) {
            do {
                suret = dbe_cursor_update(
                            solid_relcur->sr_relcur,
                            solid_relcur->sr_fetchtval,
                            tb_trans_dbtrx(cd, trans),
                            solid_rsrelh,
                            selflags,
                            tval,
                            NULL,
                            &errh);
            } while (rs_sysi_lockwait(cd) || suret == DBE_RC_WAITLOCK);
        }

#ifdef SOLIDDB_FOREIGN_CASCADE
        if (succp && suret == DBE_RC_SUCC) {
            solid_bool referenced = FALSE;

            referenced = tb_dd_isforkeyref(cd, trans, solid_rsrelh, NULL);

            if (referenced) {
                suret = ha_solid_cascade(cd, trans, solid_rsrelh,
                                         selflags, solid_relcur->sr_fetchtval,
                                         tval, ttype, &errh);
            }
        }
#endif

        SsMemFree(selflags);
        rs_tval_free(cd, ttype, tval);
        con->sc_tval = NULL;

        if (succp && suret == DBE_RC_SUCC) {
            succp = TRUE;
        } else {
            succp = FALSE;
        }

        /* If we should ignore duplicate keys we need to do a
           statement commit to get those errors and pass them to
           MySQL. Note that in replace we should not ignore duplicate
           key errors at this stage. */
        if (succp &&
            solid_ignore_duplicate_update(thd)) {
            do {
                succp = tb_trans_stmt_commit(cd, trans, &finished, &errh);
            } while (!finished);

            /* Ignore errors */
            if (!succp) {
                succp = TRUE;
                su_err_done(errh);
            }
        }

        if (succp) {
            rc = 0;
        } else {
            ss_dassert(errh != NULL);
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
        }
        ss_pprintf_2(("ha_solid::update_row:rc=%d\n", rc));

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::delete_row
 *
 * This will delete a row. buf will contain a copy of the row to be deleted.
 * The server will call this right after the current row has been called (from
 * either a previous rnd_nexT() or index call).
 * If you keep a pointer to the last row or can access a primary key it will
 * make doing the deletion quite a bit easier.
 * Keep in mind that the server does no guarentee consecutive deletions. ORDER BY
 * clauses can be used.
 *
 * Called in sql_acl.cc and sql_udf.cc to manage internal table information.
 * Called in sql_delete.cc, sql_insert.cc, and sql_select.cc. In sql_select it is
 * used for removing duplicates while in insert it is used for REPLACE calls.
 *
 *
 * Parameters :
 *
 *     const byte* buf, in, use, deleted row
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* thd
 */
int ha_soliddb::delete_row(const byte * buf)
{
        int rc;
        su_ret_t suret;
        rs_sysi_t* cd;
        tb_trans_t* trans;
        su_err_t* errh = NULL;
        THD* thd;
        SOLID_CONN* con;

        DBUG_ENTER("ha_soliddb::delete_row");
        ss_pprintf_1(("ha_solid::delete_row\n"));
        ss_dassert(solid_relcur != NULL);
        ss_dassert(solid_relcur->sr_relcur != NULL);
        CHK_RELCUR(solid_relcur);

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif
        ss_dassert(con != NULL);

        cd = solid_relcur->sr_cd;
        trans = solid_relcur->sr_trans;

        if (rs_relh_issysrel(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }
        if (rs_relh_isaborted(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        statistic_increment(thd->status_var.ha_delete_count, &LOCK_status);

        con->sc_table = table;

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        ss_pprintf_2(("ha_solid::delete_row, table %s\n", rs_relh_name(cd, solid_rsrelh)));

        do {
            suret = dbe_cursor_delete(
                        solid_relcur->sr_relcur,
                        solid_relcur->sr_fetchtval,
                        tb_trans_dbtrx(cd, trans),
                        solid_rsrelh,
                        &errh);
        } while (rs_sysi_lockwait(cd) || suret == DBE_RC_WAITLOCK);

#ifdef SOLIDDB_FOREIGN_CASCADE
        if (suret == DBE_RC_SUCC) {
            solid_bool referenced = FALSE;

            referenced = tb_dd_isforkeyref(cd, trans, solid_rsrelh, NULL);

            if (referenced) {
                suret = ha_solid_cascade(cd, trans, solid_rsrelh,
                                         NULL, solid_relcur->sr_fetchtval,
                                         NULL, solid_relcur->sr_ttype, &errh);
            }
        }
#endif

        if (suret == DBE_RC_SUCC || su_err_geterrcode(errh) == DBE_ERR_NOTFOUND) {
            rc = 0;
        } else {
            rc = ha_solid_mysql_error(thd, errh);
        }
        if (errh != NULL) {
            su_err_done(errh);
        }

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              solid_resolve_key
 *
 * Resolve solidDB key from MySQL key
 *
 * Parameters :
 *
 *     rs_sysi_t*  cd, in, use
 *     rs_relh_t*  rsrelh, in, use, solidDB table relation
 *     KEY*        key_info, in, use, MySQL key definition
 *     int         keyindex, in, use, MySQL key number
 *
 * Return value :
 *
 *     rs_key_t*, solidDB key definition or NULL
 *
 * Globals used :
 */
static rs_key_t* solid_resolve_key(
        rs_sysi_t* cd,
        rs_relh_t* relh,
        KEY* key_info,
        int keyindex,
        bool lower_case_names)
{
        rs_key_t* key;
        rs_entname_t en;

        if (key_info == NULL) {
            ss_dprintf_3(("solid_resolve_key:key_info == NULL, use clusterkey\n"));
            key = rs_relh_clusterkey(cd, relh);
            ss_dassert(key != NULL);
        } else {
            char* keyname;

            keyname = (char *)SsMemAlloc(strlen(key_info->name)+32);
            keyname_from_mysql_to_solid(cd, relh, key_info->name, keyname, lower_case_names);
            rs_entname_initbuf(&en, NULL, rs_relh_schema(cd, relh), keyname);
            key = rs_relh_keybyname(cd, relh, &en);
            if (key == NULL && keyindex == 0) {
                key = rs_relh_clusterkey(cd, relh);
            }
            ss_dassert(key != NULL);
            ss_dprintf_3(("solid_resolve_key:key %s\n", keyname));
            SsMemFree(keyname);
        }

        return(key);
}

/*#***********************************************************************\
 *
 *              relcur_checkreuse
 *
 * Check if we can reuse existing relcur.
 *
 * Parameters :
 *
 *     SOLID_CONN*       con, in, use, solid connection
 *     solid_relcur_t*   old_relcur, in, use, existing relation cursor
 *     rs_relh_t*        rsrelh, in, use, relation
 *     int               index, in, use, index
 *     dbe_cursor_type_t cursor_type, in, use, new cursor type
 *
 * Return value :
 *
 *     solid_new_cursor_t, new cursor type
 *
 * Globals used :
 */
static solid_new_cursor_t relcur_checkreuse(
        SOLID_CONN* con,
        solid_relcur_t* old_relcur,
        rs_relh_t* rsrelh,
        int index,
        dbe_cursor_type_t cursor_type,
        bool extra_keyread)
{
        solid_new_cursor_t new_cursor_type;

        if (old_relcur == NULL
            || old_relcur->sr_rsrelh != rsrelh
            || cursor_type != old_relcur->sr_cursor_type
            || old_relcur->sr_extra_keyread != extra_keyread
            || (extra_keyread && old_relcur->sr_index != index))
        {
            ss_pprintf_2(("relcur_checkreuse:SOLID_NEW_CURSOR_CREATE\n"));
            new_cursor_type = SOLID_NEW_CURSOR_CREATE;
        } else if (old_relcur->sr_index != index) {
            ss_pprintf_2(("relcur_checkreuse:SOLID_NEW_CURSOR_REUSE\n"));
            ss_dassert(old_relcur->sr_con == con);
            new_cursor_type = SOLID_NEW_CURSOR_REUSE;
        } else {
            ss_pprintf_2(("relcur_checkreuse:SOLID_NEW_CURSOR_USEOLD\n"));
            ss_dassert(old_relcur->sr_con == con);
            new_cursor_type = SOLID_NEW_CURSOR_USEOLD;
        }
        return(new_cursor_type);
}

/*#***********************************************************************\
 *
 *              ha_solid_cursor_type
 *
 * Set up solidDB cursor type
 *
 * Parameters :
 *
 *     THD*             thd, in, use
 *     bool             for_update, in, use
 *
 * Return values :
 *
 *     DBE_CURSOR_UPDATE
 *     DBE_CURSOR_DELETE
 *     DBE_CURSOR_FORUPDATE
 *     DBE_CURSOR_SELECT
 *
 * Globals used :
 */
static inline dbe_cursor_type_t ha_solid_cursor_type(
        THD* thd,
        bool for_update)
{
        if (thd->lex->sql_command == SQLCOM_UPDATE ||
            thd->lex->sql_command == SQLCOM_UPDATE_MULTI) {
            return (DBE_CURSOR_UPDATE);
        } else if (thd->lex->sql_command == SQLCOM_DELETE ||
                   thd->lex->sql_command == SQLCOM_DELETE_MULTI) {
            return (DBE_CURSOR_DELETE);
        } else {
            if (for_update) {
                return (DBE_CURSOR_FORUPDATE);
            } else {
                return (DBE_CURSOR_SELECT);
            }
        }
}

/*#***********************************************************************\
 *
 *              ::solid_relcur_create
 *
 * Create solidDB relation cursor
 *
 * Parameters :
 *
 *     SOLID_CONN*      con, in, use, solidDB connection
 *     solid_relcur_t*  old_relcur, in, use, old relation cursor or NULL
 *     rs_relh_t*       rsrelh, in, use, table relation
 *     TABLE*           table, in, use, MySQL table
 *     THD*             thd, in, use
 *     int              index, in, use, index number
 *     KEY*             key_info, in, use, MySQL key definition
 *
 * Return value :
 *
 *     solid_relcur_t*  relation cursor
 *
 * Globals used :
 */
solid_relcur_t* ha_soliddb::solid_relcur_create(
        SOLID_CONN* con,
        solid_relcur_t* old_relcur,
        rs_relh_t* rsrelh,
        TABLE* table,
        THD* thd,
        int index,
        KEY* key_info)
{
        solid_relcur_t* new_relcur = NULL;
        int nattrs;
        int ano;
        int physano;
        int i;
        rs_sysi_t* cd;
        tb_trans_t* trans;
        solid_new_cursor_t new_cursor_type;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)

        ss_win_perf_start;
        cd = con->sc_cd;
        ss_pprintf_1(("ha_solid::solid_relcur_create:table %.30s, query %.250s\n", rs_relh_name(cd, rsrelh), thd->query));
        SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_CACHE_FIND);

        trans = con->sc_trans;

        new_cursor_type = relcur_checkreuse(con, old_relcur, rsrelh, index,
                                            ha_solid_cursor_type(thd, for_update),
                                            extra_keyread);

        if (old_relcur != NULL && new_cursor_type == SOLID_NEW_CURSOR_CREATE) {
            ss_pprintf_2(("ha_solid::solid_relcur_create, free old cursor\n"));
            solid_relcur_free(cd, old_relcur, con, FALSE);
            old_relcur = NULL;
        }

        /* Could not use old cursor, try to find one from con->sc_relcurdonelist
         */
        if (old_relcur == NULL && con->sc_relcurdonelist != NULL) {
            su_list_node_t* n;
            solid_relcur_t* best_old_relcur = NULL;
            su_list_node_t* best_n = NULL;
            solid_new_cursor_t best_new_cursor_type = SOLID_NEW_CURSOR_CREATE;
            void* p;

            ss_pprintf_2(("ha_solid::solid_relcur_create, could not use old cursor, try to find one from con->sc_relcurdonelist\n"));
            SsSemEnter(con->sc_mutex);
            su_list_do_get(con->sc_relcurdonelist, n, p) {
                old_relcur = (solid_relcur_t*)p;
                new_cursor_type = relcur_checkreuse(con, old_relcur, rsrelh, index,
                                                    ha_solid_cursor_type(thd, for_update),
                                                    extra_keyread);

                if (new_cursor_type == SOLID_NEW_CURSOR_USEOLD) {
                    best_old_relcur = old_relcur;
                    best_n = n;
                    best_new_cursor_type = new_cursor_type;
                    break;
                }
                if (new_cursor_type == SOLID_NEW_CURSOR_REUSE
                &&  best_old_relcur == NULL)
                {
                    best_old_relcur = old_relcur;
                    best_n = n;
                    best_new_cursor_type = new_cursor_type;
                }
            }
            old_relcur = best_old_relcur;
            n = best_n;
            new_cursor_type = best_new_cursor_type;
            if (new_cursor_type != SOLID_NEW_CURSOR_CREATE) {
                /* Found a cursor that can be used.
                 */
                ss_pprintf_2(("ha_solid::solid_relcur_create, found a cursor that can be used\n"));
                su_list_remove(con->sc_relcurdonelist, n);
            } else {
                ss_pprintf_2(("ha_solid::solid_relcur_create, no cursor found from relcurdonelist\n"));
                ss_dassert(old_relcur == NULL);
            }
            SsSemExit(con->sc_mutex);

        }

        if (old_relcur != NULL) {
            CHK_RELCUR(old_relcur);
            if (new_cursor_type == SOLID_NEW_CURSOR_CREATE) {
                ss_pprintf_2(("ha_solid::solid_relcur_create, free old cursor\n"));
                solid_relcur_free(cd, old_relcur, con, FALSE);
                new_relcur = NULL;
            } else {
                if (new_cursor_type == SOLID_NEW_CURSOR_REUSE) {
                    ss_pprintf_2(("ha_solid::solid_relcur_create, free old dbe cursor\n"));
                    solid_relcur_free_cursor(cd, old_relcur, con);
                } else {
                    ss_dassert(new_cursor_type == SOLID_NEW_CURSOR_USEOLD);
                }
                new_relcur = old_relcur;
            }
        }

        if (new_cursor_type == SOLID_NEW_CURSOR_CREATE) {
            Field** fieldptr;
            Field* field;
            rs_key_t* key;
            rs_key_t* primkey;

            ss_pprintf_2(("ha_solid::solid_relcur_create, create a new cursor, for_update=%d\n", for_update));
            SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_CREATE);

            new_relcur = SSMEM_NEW(solid_relcur_t);

            new_relcur->sr_chk = CHKVAL_RELCUR;
            new_relcur->sr_con = con;
            new_relcur->sr_mainmem = (rs_relh_reltype(cd, rsrelh) == RS_RELTYPE_MAINMEMORY);
            new_relcur->sr_ttype = rs_relh_ttype(cd, rsrelh);
            if (new_relcur->sr_mainmem) {
                new_relcur->sr_fetchtval = NULL;
            } else {
                new_relcur->sr_fetchtval = rs_tval_create(cd, new_relcur->sr_ttype);
            }
            new_relcur->sr_bkeybuf = NULL;
            new_relcur->sr_full_scan = TRUE;
            new_relcur->sr_key = solid_resolve_key(cd, rsrelh, key_info, index, lower_case_table_names);
            new_relcur->sr_index = index;
            new_relcur->sr_cd = cd;
            new_relcur->sr_trans = trans;
            new_relcur->sr_rsrelh = rsrelh;
            new_relcur->sr_for_update = for_update;
            new_relcur->sr_prevnextp = 2;
            new_relcur->sr_unique_row = FALSE;

            new_relcur->sr_cursor_type = ha_solid_cursor_type(thd, for_update);

            new_relcur->sr_vbuf = NULL;

            new_relcur->sr_extra_keyread = extra_keyread;                     /* HA_EXTRA_KEYREAD */
            new_relcur->sr_extra_retrieve_primary_key = extra_retrieve_primary_key;        /* HA_EXTRA_RETRIEVE_PRIMARY_KEY */
            new_relcur->sr_extra_retrieve_all_cols = extra_retrieve_all_cols;           /* HA_EXTRA_RETRIEVE_ALL_COLS */
            new_relcur->sr_query_id = thd->query_id;
            new_relcur->sr_pla = NULL;
            new_relcur->sr_plakey = NULL;
            rs_relh_link(cd, rsrelh);
            SsSemEnter(con->sc_mutex);
            CHK_CONN(con);
            new_relcur->sr_relcurlist = con->sc_relcurlist;
            new_relcur->sr_relcurlistnode = su_list_insertlast(con->sc_relcurlist, (void*)new_relcur);
            SsSemExit(con->sc_mutex);

            nattrs = rs_ttype_nattrs(cd, new_relcur->sr_ttype);
            new_relcur->sr_selectlist = (int*)SsMemAlloc((nattrs + 1) * sizeof(new_relcur->sr_selectlist[0]));
            new_relcur->sr_plaselectlist = NULL;
            new_relcur->sr_constraints = NULL;
            new_relcur->sr_constval = NULL;
            new_relcur->sr_postval = NULL;
            new_relcur->sr_usedfields = (bool*)SsMemCalloc(nattrs, sizeof(new_relcur->sr_usedfields[0]));
            key = new_relcur->sr_key;
            primkey = rs_relh_clusterkey(cd, rsrelh);
            ss_dassert(primkey != NULL);
            for (fieldptr = table->field, ano = 0, i = 0; *fieldptr != NULL; fieldptr++, ano++) {
                bool add_to_selectlist;

                physano = rs_ttype_sqlanotophys(cd, new_relcur->sr_ttype, ano);
                field = *fieldptr;
                add_to_selectlist = FALSE;
                if (extra_keyread) {
                    if (rs_key_searchkpno(cd, key, physano) == RS_ANO_NULL) {
                        ss_pprintf_2(("ha_solid::solid_relcur_create, skip column, extra_keyread, not in key, ano=%d\n", ano));
                        continue;
                    }
                    add_to_selectlist = TRUE;
                }
                if (extra_retrieve_all_cols) {
                    add_to_selectlist = TRUE;
                }
                if (!add_to_selectlist && extra_retrieve_primary_key) {
                    if (rs_key_searchkpno(cd, primkey, physano) != RS_ANO_NULL) {
                        add_to_selectlist = TRUE;
                    }
                }

#if MYSQL_VERSION_ID >= 50100
                if (bitmap_is_set(table->read_set, ano) ||
                    bitmap_is_set(table->write_set, ano)) {
                    add_to_selectlist = TRUE;
                }
#endif

                if (add_to_selectlist) {
                    ss_pprintf_2(("ha_solid::solid_relcur_create, add to select list, ano=%d, physano=%d\n", ano, physano));
                    new_relcur->sr_selectlist[i++] = physano;
                    new_relcur->sr_usedfields[ano] = TRUE;
                } else {
                    ss_pprintf_2(("ha_solid::solid_relcur_create, skip column, ano=%d\n", ano));
                }
            }
            new_relcur->sr_selectlist[i++] = -1;

            new_relcur->sr_relcur = NULL;
        } else if (new_cursor_type == SOLID_NEW_CURSOR_REUSE) {
            ss_dassert(new_relcur->sr_relcur == NULL);
            new_relcur->sr_key = solid_resolve_key(cd, rsrelh, key_info, index, lower_case_table_names);
            new_relcur->sr_cd = cd;
            new_relcur->sr_trans = trans;
            new_relcur->sr_for_update = for_update;
            new_relcur->sr_cursor_type = ha_solid_cursor_type(thd, for_update);
            new_relcur->sr_extra_keyread = extra_keyread;                     /* HA_EXTRA_KEYREAD */
            new_relcur->sr_extra_retrieve_primary_key = extra_retrieve_primary_key;        /* HA_EXTRA_RETRIEVE_PRIMARY_KEY */
            new_relcur->sr_extra_retrieve_all_cols = extra_retrieve_all_cols;           /* HA_EXTRA_RETRIEVE_ALL_COLS */
            new_relcur->sr_query_id = thd->query_id;
            SsSemEnter(con->sc_mutex);
            CHK_CONN(con);
            new_relcur->sr_relcurlist = con->sc_relcurlist;
            new_relcur->sr_relcurlistnode = su_list_insertlast(con->sc_relcurlist, (void*)new_relcur);
            SsSemExit(con->sc_mutex);
        } else {
            ss_dassert(new_cursor_type == SOLID_NEW_CURSOR_USEOLD);
            SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_CACHE_HIT);
        }

        ss_dassert(new_relcur->sr_relcurlist != NULL);
        ss_dassert(new_relcur->sr_relcurlistnode != NULL);

        ss_win_perf_stop(relcur_create_perfcount, relcur_create_callcount);
        CHK_RELCUR(new_relcur);
        ss_pprintf_2(("ha_solid::solid_relcur_create, return cursor %ld\n", (long)new_relcur));

        return(new_relcur);
}

/*#***********************************************************************\
 *
 *              mysql_record_is_null
 *
 * Check if a field in a record is NULL.
 *
 * Parameters:
 *
 *     TABLE*           table, in, use, table
 *     Field*           field, in, use, field
 *     char*            record, in, use, a row in MySQL format
 *
 * Return value: 1 if record is NULL, 0 otherwise
 *
 * Globals used :
 */
static inline int mysql_record_is_null(
        TABLE* table,
        Field* field,
        char * record)
{
        uint null_offset;

        if (!field->null_ptr) {
            return (0);
        } else {
            null_offset = (uint) ((char *)field->null_ptr
                                  - (char *) table->record[0]);

            if (record[null_offset] & field->null_bit) {
                return (1);
            } else {
                return (0);
            }
        }
}

/*#***********************************************************************\
 *
 *              solid_relcur_setconstr
 *
 * Set constraints for relation cursor
 *
 * Parameters :
 *
 *     rs_sysi_t*       cd, in, use
 *     tb_trans_t*      trans, in, use, transaction
 *     rs_relh_t*       rsrelh, in, use, table relation
 *     solid_relcur_t*  old_relcur, in, use, old relation cursor or NULL
 *     int              keyindex, in, use, index number
 *     KEY*             key_info, in, use, MySQL key definition
 *     const byte*      keyconstr_ptr, in, use, key constraint
 *     uint             keyconstr_len, in, use, key constraint length
 *     enum ha_rkey_function find_flag, in, use
 *
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
static int solid_relcur_setconstr(
        THD *thd,
        rs_sysi_t* cd,
        tb_trans_t* trans,
        rs_relh_t* rsrelh,
        solid_relcur_t* solid_relcur,
        uint keyindex,
        KEY* key_info,
        TABLE* table,
        const byte* keyconstr_ptr,
        uint keyconstr_len,
        enum ha_rkey_function find_flag,
        bool lower_case_names)
{
        KEY_PART_INFO* key_part;
        int kpno;
        const byte* ptr;
        uint key_len;
        enum ha_rkey_function current_find_flag;
        rs_key_t* key;
        vtpl_t* range_start;
        vtpl_t* range_end;
        va_t* va;
        va_t* dva = NULL;
        va_t* range_start_last_va;
        va_t* range_end_last_va;
        su_list_t* tuple_reference;
        rs_ttype_t* ttype;
        su_err_t* errh = NULL;
        bool newindex;
        bool newplan;
        int rc;
        int cons_len;
        uint nequal = 0;
        uint relop = RS_RELOP_EQUAL;
        bool mainmem;
        rs_cons_t* cons;
        su_list_node_t* cons_n=NULL;
        bool dereference;
        bool replan = FALSE;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)

        ss_win_perf_start;
        DBUG_ENTER("ha_solid::solid_relcur_setconstr");

        ss_pprintf_1(("ha_solid::solid_relcur_setconstr:key %s, keyindex %d, cursor %ld, find_flag %d\n",
            key_info == NULL ? "" : key_info->name, keyindex, (long)solid_relcur, (int)find_flag));


        CHK_RELCUR(solid_relcur);
        SS_PUSHNAME("solid_relcur_setconstr");

        mainmem = solid_relcur->sr_mainmem;

        ttype = rs_relh_ttype(cd, rsrelh);

        if (solid_relcur->sr_key != NULL && (uint)solid_relcur->sr_index == keyindex) {
            key = solid_relcur->sr_key;
        } else {
            /* Resolve key. */
            key = solid_resolve_key(cd, rsrelh, key_info, keyindex, lower_case_names);
        }

        if (solid_relcur->sr_plakey == NULL || solid_relcur->sr_plakey != key) {
            newindex = TRUE;
        } else {
            newindex = FALSE;
        }

        solid_relcur->sr_index = keyindex;
        solid_relcur->sr_key = key;
        solid_relcur->sr_plakey = key;
        solid_relcur->sr_full_scan = FALSE;

        if (mainmem) {
            range_start = (vtpl_t*)NULL;
            range_end = (vtpl_t*)NULL;
            dva = (va_t*)NULL;
        } else {
            /* Calculate max constraint lenght. Assume all fields can be long
             * v-attributes and assume that we end null byte at the end of
             * each value.
             */
            cons_len = keyconstr_len
                       + (rs_key_nparts(cd, key) * sizeof(va_t))
                       + (rs_key_nparts(cd, key) + 1) * VA_LENGTHMAXLEN;
            range_start = (vtpl_t*)SsMemAlloc(cons_len + 1);
            range_end = (vtpl_t*)SsMemAlloc(cons_len + 1);
            dva = (va_t*)SsMemAlloc(cons_len);

            /* Set upper and lower limit for the key. */
            vtpl_setvtpl(range_start, VTPL_EMPTY);
            vtpl_setvtpl(range_end, VTPL_EMPTY);
        }
        nequal++;

        kpno = 0;
        ss_dassert(rs_keyp_isconstvalue(cd, key, kpno));

        if (mainmem) {
            va = NULL;
            range_start_last_va = NULL;
            range_end_last_va = NULL;
        } else {
            va = rs_keyp_constvalue(cd, key, kpno);

            range_start_last_va = va;
            range_end_last_va = va;
        }

        if (mainmem) {
            if (solid_relcur->sr_constraints == NULL) {
                solid_relcur->sr_constraints = su_list_init(NULL);
                solid_relcur->sr_constval = rs_tval_create(cd, ttype);
            }
            cons_n = su_list_first(solid_relcur->sr_constraints);
        }

        if (keyconstr_ptr != NULL) {

            ptr = keyconstr_ptr;
            key_len = keyconstr_len;
            current_find_flag = find_flag;

            /* Get MySQL constraints.
             *
             * Some of the code here is copied from ha_federated.cpp.
             * TODO: Need to recheck this.
             */
            for (key_part = key_info->key_part, kpno++; ; key_part++, kpno++) {
                rs_ano_t ano;
                rs_atype_t* atype;
                rs_aval_t* aval;
                Field* field;
                uint store_length;
                bool isnull;
                bool null_value;
                int offset;

                if (rs_keyp_parttype(cd, key, kpno) != RSAT_USER_DEFINED) {
                    ss_pprintf_2(("ha_solid::solid_relcur_setconstr:kpno=%d, not RSAT_USER_DEFINED, break\n", kpno));
                    break;
                }

                field = key_part->field;
                store_length = key_part->store_length;

                ano = rs_keyp_ano(cd, key, kpno);
                ss_dassert(ano != RS_ANO_NULL);

                ss_pprintf_2(("ha_solid::solid_relcur_setconstr:kpno=%d, ano=%d\n", kpno, ano));

                atype = rs_ttype_atype(cd, ttype, ano);

                offset = 0;

                isnull = mysql_record_is_null(table, field, (char *)ptr);

                if (key_part->null_bit && (*ptr)) {
                    /* Constraint is IS NULL or IS NOT NULL */

                    null_value = TRUE;

                    switch (current_find_flag) {
                        case HA_READ_KEY_EXACT:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_EXACT:RS_RELOP_ISNULL\n"));
                            if (!mainmem) {
                                if (range_start_last_va != NULL) {
                                    vtpl_appva(range_start, range_start_last_va);
                                }
                                range_start_last_va = VA_NULL;
                                if (range_end_last_va != NULL) {
                                    vtpl_appva(range_end, range_end_last_va);
                                }
                                range_end_last_va = VA_NULL;
                            }
                            break;
                        case HA_READ_KEY_OR_NEXT:
                        case HA_READ_PREFIX:
                        case HA_READ_PREFIX_LAST:
                        case HA_READ_PREFIX_LAST_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_OR_NEXT...:RS_RELOP_GE NULL\n"));
                            if (!mainmem) {
                                if (range_start_last_va != NULL) {
                                    vtpl_appva(range_start, range_start_last_va);
                                }
                                range_start_last_va = VA_NULL;
                            }
                            relop = RS_RELOP_GE;
                            break;
                        case HA_READ_BEFORE_KEY:
                        case HA_READ_KEY_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_BEFORE_KEY...:RS_RELOP_LE NOTNULL\n"));
                            if (!mainmem) {
                                if (range_end_last_va != NULL) {
                                    vtpl_appva(range_end, range_end_last_va);
                                }
                                range_end_last_va = VA_NULL;
                            }
                            relop = RS_RELOP_LE;
                            break;
                        case HA_READ_AFTER_KEY:
                            if (isnull) {
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_AFTER_KEY:RS_RELOP_ISNULL\n"));
                                relop = RS_RELOP_GE;
                            } else {
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_AFTER_KEY:RS_RELOP_NE NOTNULL\n"));
                                relop = RS_RELOP_GT;
                            }
                            if (!mainmem) {
                                if (range_start_last_va != NULL) {
                                    vtpl_appva(range_start, range_start_last_va);
                                }
                                range_start_last_va = VA_NULL;
                            }
                            break;
                        default:
                            ss_pprintf_1(("ha_solid::solid_relcur_setconstr:**UNKNOWN** find_flag=%d\n",
                                          current_find_flag));

                            if (!mainmem) {
                                SsMemFree(dva);
                            }
                            if (range_start) {
                                dynvtpl_free(&range_start);
                            }

                            if (range_end) {
                                dynvtpl_free(&range_end);
                            }
                            
                            ss_error;        /* We should not come here */
                            SS_POPNAME;
                            DBUG_RETURN(1);
                    }

                } else {

                    null_value = FALSE;

                    if (key_part->null_bit) {
                        offset = 1;
                    }

                    /* Constraint has a value */
                    switch (current_find_flag) {
                        case HA_READ_KEY_EXACT:
                            if (store_length >= key_len
                                || key_part->type == HA_KEYTYPE_BIT
                                || field->result_type() != STRING_RESULT)
                            {
                                int mysqltype = field->type();
                                
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_EXACT => RS_RELOP_EQUAL\n"));
                                if (field->result_type() == STRING_RESULT &&
                                    (mysqltype == MYSQL_TYPE_VAR_STRING ||
                                     mysqltype == FIELD_TYPE_STRING ||
                                     mysqltype == MYSQL_TYPE_VARCHAR)) {
                                    ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_EXACT(STRING) => RS_RELOP_GE\n"));
                                    relop = RS_RELOP_GE;
                                }
                            } else {
                                /* TODO: LIKE */
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:LIKE => RS_RELOP_GE\n"));
                                relop = RS_RELOP_GE;
                            }
                            break;
                        case HA_READ_AFTER_KEY:
                            if (store_length >= key_len) {
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_AFTER_KEY => RS_RELOP_GT\n"));
                                relop = RS_RELOP_GT;
                                break;
                            }
                            /* FALLTHROUGH */
                        case HA_READ_KEY_OR_NEXT:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_OR_NEXT => RS_RELOP_GE\n"));
                            relop = RS_RELOP_GE;
                            break;
                        case HA_READ_BEFORE_KEY:
                            if (store_length >= key_len) {
                                ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_BEFORE_KEY => RS_RELOP_LT\n"));
                                relop = RS_RELOP_LT;
                                break;
                            }
                            /* FALLTHROUGH */
                        case HA_READ_KEY_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_KEY_OR_PREV => RS_RELOP_LE\n"));
                            relop = RS_RELOP_LE;
                            break;

                        case HA_READ_PREFIX:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_PREVIX => RS_RELOP_GE\n"));
                            relop = RS_RELOP_GE;
                            break;

                        case HA_READ_PREFIX_LAST:
                        case HA_READ_PREFIX_LAST_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_relcur_setconstr:HA_READ_PREFIX_LAST => RS_RELOP_LE\n"));
                            relop = RS_RELOP_LE;
                            break;

                        default:
                            ss_pprintf_1(("ha_solid::solid_relcur_setconstr:**UNKNOWN** find_flag=%d\n",
                                          current_find_flag));

                            SsMemFree(dva);
                            if (range_start) {
                                dynvtpl_free(&range_start);
                            }

                            if (range_end) {
                                dynvtpl_free(&range_end);
                            }
                            
                            ss_error;        /* We should not come here */
                            SS_POPNAME;
                            DBUG_RETURN(1);
                    }

                    switch (relop) {
                        case RS_RELOP_EQUAL:
                            if (range_start_last_va != NULL) {
                                vtpl_appva(range_start, range_start_last_va);
                                range_start_last_va = NULL;
                            }
                            if (range_end_last_va != NULL) {
                                vtpl_appva(range_end, range_end_last_va);
                                range_end_last_va = NULL;
                            }
                            break;
                        case RS_RELOP_GE:
                        case RS_RELOP_GT:
                            if (range_start_last_va != NULL) {
                                vtpl_appva(range_start, range_start_last_va);
                                range_start_last_va = NULL;
                            }
                            break;
                        case RS_RELOP_LT:
                        case RS_RELOP_LE:
                            if (range_end_last_va != NULL) {
                                vtpl_appva(range_end, range_end_last_va);
                                range_end_last_va = NULL;
                            }
                           break;
                        default:
                            SsMemFree(dva);

                            if (range_start) {
                                dynvtpl_free(&range_start);
                            }

                            if (range_end) {
                                dynvtpl_free(&range_end);
                            }

                            ss_derror;
                            ss_pprintf_1(("ha_solid::solid_relcur_setconstr:**UNKNOWN** relop=%d\n", relop));
                            SS_POPNAME;
                            DBUG_RETURN(1);
                    }

                }

                if (mainmem) {
                    bool new_cons;

                    if (cons_n == NULL) {
                        ss_pprintf_2(("ha_solid::solid_relcur_setconstr:no old cons\n"));
                        new_cons = TRUE;
                    } else {
                        cons = (rs_cons_t*)su_listnode_getdata(cons_n);
                        aval = rs_tval_aval(cd, ttype, solid_relcur->sr_constval, ano);
                        ss_dassert(rs_cons_ano(cd, cons) == ano);
                        ss_dassert(rs_cons_atype(cd, cons) == atype);
                        ss_dassert(rs_cons_aval(cd, cons) == aval);
                        if (rs_cons_relop(cd, cons) != relop) {
                            ss_pprintf_2(("ha_solid::solid_relcur_setconstr:relop in old cons is different\n"));
                            new_cons = TRUE;
                            free_end_of_constr_list(cd, solid_relcur->sr_constraints, cons_n);
                            cons_n = NULL;
                        } else {
                            ss_pprintf_2(("ha_solid::solid_relcur_setconstr:use old cons\n"));
                            new_cons = FALSE;
                            cons_n = su_list_next(solid_relcur->sr_constraints, cons_n);
                        }
                    }
                    if (new_cons) {
                        rs_sqlcons_t sqlcons;

                        replan = TRUE;
#if 0
                        aval = rs_tval_aval(cd, ttype, solid_relcur->sr_constval, ano);
                        if (aval == NULL) {
                            aval = rs_aval_create(cd, atype);
                            rs_tval_insertaval(cd, ttype, solid_relcur->sr_constval, ano, aval);
                        }
#endif

                        /* Get a reference to aval inside tval */
                        aval = rs_tval_aval(cd, ttype, solid_relcur->sr_constval, ano);

                        sqlcons.sc_relop = relop;
                        sqlcons.sc_attrn = ano;
                        sqlcons.sc_atype = atype;
                        sqlcons.sc_aval = aval;
                        sqlcons.sc_escatype = NULL;
                        sqlcons.sc_escaval = NULL;
                        sqlcons.sc_alias = TRUE;
                        sqlcons.sc_tabcons = FALSE;

                        cons = rs_cons_init(
                                    cd,
                                    relop,
                                    ano,
                                    atype,
                                    aval,
                                    0,          /* su_bflag_t flags */
                                    &sqlcons,   /* rs_sqlcons_t* sqlcons */
                                    0,          /* int escchar */
                                    atype,      /* rs_atype_t* org_atype */
                                    NULL);      /* rs_err_t** p_errh */
                        ss_assert(cons != NULL);
                        su_list_insertlast(solid_relcur->sr_constraints, cons);
                    }
                    if (null_value) {
                        rs_aval_setnull(cd, atype, aval);
                    } else {
                        set_mysql_field_to_aval_or_dynva(thd, table, cd, atype, aval, NULL, field, ptr + offset, TRUE, NULL);
                    }

                } else if (!null_value) {

                    ss_dassert(!mainmem);

                    set_mysql_field_to_aval_or_dynva(thd, table, cd, atype, NULL, dva, field, ptr + offset, TRUE, NULL);

                    switch (relop) {
                        case RS_RELOP_EQUAL:
                            range_start_last_va = dva;
                            range_end_last_va = dva;
                            nequal++;
                            break;
                        case RS_RELOP_GE:
                        case RS_RELOP_GT:
                            range_start_last_va = dva;
                            break;
                        case RS_RELOP_LT:
                        case RS_RELOP_LE:
                            range_end_last_va = dva;
                            break;
                        default:
                            SsMemFree(dva);
                            
                            if (range_start) {
                                dynvtpl_free(&range_start);
                            }

                            if (range_end) {
                                dynvtpl_free(&range_end);
                            }

                            ss_derror;
                            ss_pprintf_1(("ha_solid::solid_relcur_setconstr:**UNKNOWN** relop=%d\n", relop));
                            SS_POPNAME;
                            DBUG_RETURN(1);
                    }
                }

                if (store_length >= key_len) {
                    break;
                }
                key_len -= store_length;
                ptr += store_length;
            }
        }

        if (mainmem) {
            if (cons_n != NULL) {
                /* Old list has more constraints than the new one, free extra
                 * constrains. 
                 */
                free_end_of_constr_list(cd, solid_relcur->sr_constraints, cons_n);
                replan = TRUE;
            }
            ss_dassert(range_start == NULL);
            ss_dassert(range_end == NULL);
        } else {
            if (range_start_last_va != NULL) {
                if (relop ==  RS_RELOP_GT) {
                    vtpl_appvawithincrement(range_start, range_start_last_va);
                } else {
                    vtpl_appva(range_start, range_start_last_va);
                }
            }
            if (range_end_last_va != NULL) {
                if (relop ==  RS_RELOP_LT) {
                    vtpl_appva(range_end, range_end_last_va);
                } else {
                    vtpl_appvawithincrement(range_end, range_end_last_va);
                }
            }
            SsMemFree(dva);

            ss_output_1((vtpl_dprintvtpl(1, range_start)));
            ss_output_1((vtpl_dprintvtpl(1, range_end)));
        }

        if (rs_key_isprimary(cd, key)) {
            dereference = FALSE;
        } else {
            dereference = !solid_relcur->sr_extra_keyread;
        }

        if (solid_relcur->sr_pla == NULL) {
            solid_bool dereference2;
            ss_dassert(newindex);
            solid_relcur->sr_pla = rs_pla_alloc(cd);
            solid_relcur->sr_plaselectlist = tb_pla_form_select_list(
                                                cd,
                                                rs_relh_clusterkey(cd, rsrelh),
                                                key,
                                                solid_relcur->sr_selectlist,
                                                dereference,
                                                &dereference2);
            ss_dassert(dereference == dereference2);
            newplan = TRUE;
        } else {
            newplan = FALSE;
        }

        if (newindex) {
            rs_pla_clear_tuple_reference_buf(cd, solid_relcur->sr_pla);

            tuple_reference = tb_pla_form_tuple_reference(
                                cd,
                                rs_relh_clusterkey(cd, rsrelh),
                                rs_pla_get_tuple_reference_buf(cd, solid_relcur->sr_pla),
                                key);
        } else {
            tuple_reference = rs_pla_get_tuple_reference_buf(cd, solid_relcur->sr_pla);
        }
        ss_dassert(su_list_length(tuple_reference) > 0);

        rc = 0;

        if (newplan) {
            ss_pprintf_2(("ha_solid::solid_relcur_setconstr:rs_pla_initbuf\n"));

            rs_pla_initbuf(
                cd,
                solid_relcur->sr_pla,
                rsrelh,
                key,
                TRUE,
                range_start,
                TRUE,
                range_end,
                TRUE,
                NULL,
                NULL,
                solid_relcur->sr_constraints,
                tuple_reference,
                solid_relcur->sr_plaselectlist,
                dereference,
                0,
                0,
                0,
                FALSE);

        } else {
            ss_pprintf_2(("ha_solid::solid_relcur_setconstr:rs_pla_reset\n"));

            ss_debug(rs_pla_check_reset(
                cd,
                solid_relcur->sr_pla,
                rsrelh,
                key,
                TRUE,
                range_start,
                TRUE,
                range_end,
                TRUE,
                NULL,
                NULL,
                solid_relcur->sr_constraints,
                tuple_reference,
                solid_relcur->sr_plaselectlist,
                dereference,
                0,
                0,
                0));

            rs_pla_reset(
                cd,
                solid_relcur->sr_pla,
                TRUE,
                range_start,
                range_end);

            if (mainmem) {
                rs_pla_setreplan(cd, solid_relcur->sr_pla, replan);
            }
        }

        solid_relcur->sr_open_cursor = TRUE;
        if (!solid_relcur->sr_mainmem) {
            solid_relcur->sr_unique_row = rs_key_isunique(cd, key) &&
                                          nequal == (rs_key_lastordering(cd, key) + 1) &&
                                          solid_relcur->sr_cursor_type == DBE_CURSOR_SELECT &&
                                          thd->lex->sql_command == SQLCOM_SELECT;
        }
        ss_pprintf_2(("ha_solid::solid_relcur_setconstr:sr_unique_row=%d, nequal=%d, key lastordering=%d, key unique=%d\n",
            solid_relcur->sr_unique_row, nequal, rs_key_lastordering(cd, key),
            rs_key_isunique(cd, key)));

        ss_win_perf_stop(relcur_setconstr_perfcount, relcur_setconstr_callcount);
        SS_POPNAME;

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ha_solid_time_mysql
 *
 * Converts solidDB time structure to time_t
 *
 *
 * Parameters :
 *
 *  dt_date_t* solid_time, in, use, time in solid structure
 *
 * Return value : time_t structure
 *
 * Globals used :
 */
static inline time_t ha_solid_time_mysql(
        dt_date_t* solid_time)
{
        SsTmT ss_time;
        time_t mysql_ttime;

        memset(&ss_time, 0, sizeof(SsTmT));

        ss_time.tm_sec = dt_date_sec(solid_time);
        ss_time.tm_min = dt_date_min(solid_time);
        ss_time.tm_hour = dt_date_hour(solid_time);
        ss_time.tm_mday = dt_date_mday(solid_time);
        ss_time.tm_mon = dt_date_month(solid_time);
        ss_time.tm_year = dt_date_year(solid_time) - 1900;
        ss_time.tm_yday = 0;
        ss_time.tm_isdst = -1;

        mysql_ttime = (time_t)SsMktime(&ss_time);

        return (mysql_ttime);
}

/*#***********************************************************************\
 *
 *              ::solid_fetch
 *
 *   Fetch the next row in a result set. This is a separate routine mainly
 *   to have serarate statictics_increment call for different external
 *   fetch calls.
 *
 *
 * Parameters :
 *
 *     byte*  rowbuf, in out, row buffer
 *     bool   nextp, in, use,
 *     THD*   thd, in, use, MySQL thread
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::solid_fetch(
        byte *rowbuf,
        bool nextp,
        THD* thd)
{
        rs_ttype_t* ttype;
        rs_atype_t* atype;
        rs_aval_t* aval;
    	int	ano;
        int steps;
        Field** fieldptr;
        Field* field;
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        rs_relh_t* rsrelh;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)
        void* cont = NULL;
        dbe_ret_t rc = DBE_RC_SUCC;
        su_err_t* errh = NULL;
        bool isfetch = FALSE;
        int sret = 0;
        size_t smallblobsizemax = INT_MAX; //TODO: what is max blob length in mysql and how long can constraint be for blobs

        ss_win_perf_start;
        DBUG_ENTER("ha_soliddb::solid_fetch");
        ss_dassert(solid_relcur != NULL);
        CHK_RELCUR(solid_relcur);

#if MYSQL_VERSION_ID >= 50100
        solid_conn = (SOLID_CONN*)thd->ha_data[ht->slot];
#else
        solid_conn = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        con = solid_conn;
        ss_dassert(con != NULL);
        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;
        rsrelh = solid_rsrelh;

        ss_pprintf_1(("ha_solid::solid_fetch, table=%s, nextp=%d\n", rs_relh_name(cd, rsrelh), nextp));

        if (solid_relcur->sr_open_cursor) {
            bool unique_search = FALSE;

            ss_win_perf(__int64 startcount;)
            ss_win_perf(__int64 endcount;)

            ss_win_perf_start;

            solid_relcur->sr_open_cursor = FALSE;

            if (solid_relcur->sr_unique_row && nextp) {

                if (solid_relcur->sr_fetchtval == NULL) {
                    ss_dassert(!solid_relcur->sr_mainmem);
                    solid_relcur->sr_fetchtval = rs_tval_create(cd,
                                                                solid_relcur->sr_ttype);
                }

                ss_dassert(solid_relcur->sr_ttype != NULL);

                rc = dbe_cursor_getunique(
                        tb_trans_dbtrx(cd, trans),
                        solid_relcur->sr_ttype,
                        solid_relcur->sr_selectlist,
                        solid_relcur->sr_pla,
                        solid_relcur->sr_fetchtval,
                        &solid_relcur->sr_bkeybuf,
                        &errh);
                unique_search = (rc != DBE_RC_NOTFOUND);
                isfetch = unique_search;
                ss_dassert(rc != DBE_RC_WAITLOCK);
            } else {
                unique_search = FALSE;
            }

            if (!unique_search) {
                if (solid_relcur->sr_relcur == NULL) {
                    ss_pprintf_2(("ha_solid::solid_fetch:dbe_cursor_init\n"));
                    SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_RESET_FULL);

                    solid_relcur->sr_relcur = dbe_cursor_init(
                                                tb_trans_dbtrx(cd, trans),
                                                solid_relcur->sr_ttype,
                                                solid_relcur->sr_selectlist,
                                                solid_relcur->sr_pla,
                                                solid_relcur->sr_cursor_type,
                                                NULL,
                                                &errh);

                    ss_dassert(solid_relcur->sr_relcur != NULL);

                    if (solid_relcur->sr_mainmem) {
                        ss_dassert(solid_relcur->sr_vbuf == NULL);
                        if (solid_relcur->sr_cursor_type == DBE_CURSOR_SELECT) {
                            solid_relcur->sr_vbuf = rs_vbuf_init(cd, solid_relcur->sr_ttype, 10);
                        } else {
                            solid_relcur->sr_vbuf = rs_vbuf_init(cd, solid_relcur->sr_ttype, 2);
                        }
                    }

                } else {
                    ss_pprintf_2(("ha_solid::solid_fetch:dbe_cursor_reset\n"));

                    if (nextp && rowbuf != NULL && !solid_relcur->sr_mainmem) {
                        SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_RESET_FETCH);
                        rc = dbe_cursor_reset_fetch(
                                solid_relcur->sr_relcur,
                                tb_trans_dbtrx(cd, trans),
                                solid_relcur->sr_ttype,
                                solid_relcur->sr_selectlist,
                                solid_relcur->sr_pla,
                                &solid_relcur->sr_fetchtval,
                                &errh);
                        isfetch = (rc == DBE_RC_FOUND || rc == DBE_RC_END);
                        ss_dassert(rc != DBE_RC_WAITLOCK);

                    } else {
                        SS_PMON_ADD(SS_PMON_MYSQL_CURSOR_RESET_SIMPLE);
                        if (solid_relcur->sr_vbuf != NULL) {
                            rs_vbuf_reset(cd, solid_relcur->sr_vbuf);
                        }
                        dbe_cursor_reset(
                            solid_relcur->sr_relcur,
                            tb_trans_dbtrx(cd, trans),
                            solid_relcur->sr_ttype,
                            solid_relcur->sr_selectlist,
                            solid_relcur->sr_pla);
                        isfetch = FALSE;
                    }
                }
            }
            ss_win_perf_stop(relcur_open_perfcount, relcur_open_callcount);

            if (rs_relh_isaborted(cd, rsrelh)) {
                int rc;
                ss_pprintf_2(("ha_solid::solid_fetch, rs_relh_isaborted\n"));
                ss_dassert(errh == NULL);
                rs_error_create(&errh, E_DDOP);
                SsLogErrorMessage(su_err_geterrstr(errh));
                rc = ha_solid_mysql_error(thd, errh);
                su_err_done(errh);
                DBUG_RETURN(rc);
            }

            if (rowbuf == NULL) {
                ss_pprintf_2(("ha_solid::solid_fetch, rowbuf == NULL, return without fetching\n"));
                DBUG_RETURN(0);
            }
        }

        if (!isfetch) {
            ss_win_perf(__int64 startcount);
            ss_win_perf(__int64 endcount);

            if (solid_relcur->sr_relcur == NULL) {
                DBUG_RETURN(HA_ERR_END_OF_FILE);
            }

            ss_dassert(solid_relcur->sr_relcur != NULL);

            ss_win_perf_start;
            SS_PMON_ADD(SS_PMON_MYSQL_FETCH_NEXT);
            steps = 0;
            if (solid_relcur->sr_mainmem) {
                bool waitlock;
                waitlock = FALSE;
                do {
                    if (waitlock) {
                        rc = dbe_cursor_relock(
                                solid_relcur->sr_relcur,
                                tb_trans_dbtrx(cd, trans),
                                &errh);
                        if (rc != SU_SUCCESS) {
                            break;
                        }
                        waitlock = FALSE;
                    }
                    /* Reset the vbuf if we've changed directions. */
                    if (solid_relcur->sr_prevnextp != nextp) {
                        rs_vbuf_rewind(cd, solid_relcur->sr_vbuf);
                        solid_relcur->sr_prevnextp = nextp;
                    }
                    if (!rs_vbuf_hasdata(cd, solid_relcur->sr_vbuf)) {
                        rc = dbe_cursor_nextorprev_n(
                                solid_relcur->sr_relcur,
                                nextp,
                                tb_trans_dbtrx(cd, trans),
                                solid_relcur->sr_vbuf,
                                &errh);
                        ss_rc_assert(steps < 100000, steps);
                        steps++;
                    } else {
                        rc = DBE_RC_FOUND;
                    }
                    if (rc == DBE_RC_FOUND) {
                        solid_relcur->sr_fetchtval = rs_vbuf_readtval(cd, solid_relcur->sr_vbuf);
                        if (solid_relcur->sr_fetchtval != NULL) {
                            rc = DBE_RC_FOUND;
                        } else {
                            rc = DBE_RC_END;
                        }
                    }
                    if (rc == DBE_RC_WAITLOCK) {
                        waitlock = TRUE;
                    }
                } while (rs_sysi_lockwait(cd) || rc == DBE_RC_WAITLOCK);
                ss_dassert(rc != DBE_RC_NOTFOUND);
            } else {
                do {
                    do {
                        rc = dbe_cursor_nextorprev(
                                solid_relcur->sr_relcur,
                                nextp,
                                tb_trans_dbtrx(cd, trans),
                                &solid_relcur->sr_fetchtval,
                                &errh);
                        ss_rc_assert(steps < 100000, steps);
                        steps++;
                    } while (rs_sysi_lockwait(cd) || rc == DBE_RC_WAITLOCK);
                } while (rc == DBE_RC_NOTFOUND);
            }

            ss_win_perf_stop(relcur_next_perfcount, relcur_next_callcount);
        }

        if (rc != DBE_RC_FOUND && rc != DBE_RC_END) {
            int rc;
            ss_dassert(errh != NULL);
            SsLogErrorMessage(su_err_geterrstr(errh));
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
            DBUG_RETURN(rc);
        }

        if (rc == DBE_RC_END) {
            ss_pprintf_1(("ha_solid::solid_fetch:end of set\n"));
            table->status = STATUS_NOT_FOUND;
            ss_win_perf_stop(fetch_perfcount, fetch_callcount);

            DBUG_RETURN(HA_ERR_END_OF_FILE);

        } else if (thd->lex->sql_command != SQLCOM_REPLACE &&
                   thd->lex->sql_command != SQLCOM_REPLACE_SELECT) {

#if MYSQL_VERSION_ID >= 50100
            my_bitmap_map *old_map = dbug_tmp_use_all_columns(table, table->write_set);
#endif

            ss_pprintf_1(("ha_solid::solid_fetch:found row\n"));
            ss_dassert(rc == DBE_RC_FOUND);

            ttype = solid_relcur->sr_ttype;
            memset(rowbuf, 0, table->s->null_bytes);
            table->status = 0;

            ss_poutput_2(rs_tval_print(cd, ttype, solid_relcur->sr_fetchtval));

            for (fieldptr = table->field, ano = 0; *fieldptr != NULL; fieldptr++, ano++) {
            	int	phys_ano;
                field = *fieldptr;

                if (!solid_relcur->sr_usedfields[ano]) {
                    ss_pprintf_1(("ha_solid::solid_fetch:skip field\n"));
                    continue;
                }
                phys_ano = rs_ttype_quicksqlanotophys(cd, ttype, ano);
                atype = rs_ttype_atype(cd, ttype, phys_ano);
                aval = rs_tval_aval(cd, ttype, solid_relcur->sr_fetchtval, ano);

                if (rs_aval_isnull(cd, atype, aval)) {
                    ss_pprintf_1(("ha_solid::solid_fetch:NULL\n"));
                    field->set_null();
                } else {
                    char* str;
                    ulong length;
                    char* tmp_ptr;
                    ulong aval_length;

                    field->set_notnull();
                    tmp_ptr = field->ptr;

#if MYSQL_VERSION_ID >= 50100
                    field->ptr = (char*)rowbuf + field->offset(field->table->record[0]);
#else
                    field->ptr = (char*)rowbuf + field->offset();
#endif

                    switch (rs_atype_datatype(cd, atype)) {
                        case RSDT_CHAR:
                            if (rs_aval_isblob(cd, atype, aval)) {
                                tb_blobg2_loadblobtoaval_limit(cd, atype, aval, smallblobsizemax);
                            }

                            str = (char*)rs_aval_getdata(cd, atype, aval, &aval_length);

                            switch (field->real_type()) {
                                case MYSQL_TYPE_SET:
                                case MYSQL_TYPE_ENUM:
                                    length = field->pack_length();
                                    memcpy(field->ptr, str, length);
                                    break;
                                default:

                                    sret = field->store(
                                                str,
                                                aval_length,
                                                field->charset());
                                    break;
                            }
                            break;
                        case RSDT_INTEGER:

                            if (((Field_num*)field)->unsigned_flag) {
                                sret = field->store((longlong)(unsigned long)
                                                    rs_aval_getlong(cd, atype, aval), TRUE);
                            } else {
                                sret = field->store((longlong)rs_aval_getlong(cd, atype, aval), FALSE);

                            }

                            break;
                        case RSDT_FLOAT:
                        case RSDT_DOUBLE:

                            sret = field->store(rs_aval_getdouble(cd, atype, aval));
                            break;
                        case RSDT_DATE: {
                            dt_date_t *solid_date;
                            timestamp_type mysql_type;

                            solid_date = rs_aval_getdate(cd, atype, aval);
                            ss_dassert(solid_date != NULL);

                            ss_pprintf_2(("Fetched date: %d-%d-%d %d:%d:%d\n",
                                          dt_date_year(solid_date), dt_date_month(solid_date),
                                          dt_date_mday(solid_date), dt_date_hour(solid_date),
                                          dt_date_min(solid_date), dt_date_sec(solid_date)));

                            /* See field.cc for documentation */
                            switch(field->real_type()) {
                                case FIELD_TYPE_DATE: {
                                    uint32 tmp = 0;

                                    tmp = (uint32) dt_date_year(solid_date) *10000L +
                                        (uint32) (dt_date_month(solid_date)+dt_date_mday(solid_date));

                                    sret = field->store(tmp, (field->flags & UNSIGNED_FLAG));
                                    break;
                                }
                                case FIELD_TYPE_TIME:
                                case FIELD_TYPE_NEWDATE:
                                case FIELD_TYPE_DATETIME: {
                                    TIME ltime;

                                    ltime.year   = dt_date_year(solid_date);
                                    ltime.month  = dt_date_month(solid_date);
                                    ltime.day    = dt_date_mday(solid_date);
                                    ltime.hour   = dt_date_hour(solid_date);
                                    ltime.minute = dt_date_min(solid_date);
                                    ltime.second = dt_date_sec(solid_date);
                                    ltime.second_part = 0;
                                    ltime.neg    = 0;

                                    sret = field->store_time(&ltime, MYSQL_TIMESTAMP_DATE);
                                    break;
                                }
                                case FIELD_TYPE_TIMESTAMP: {
                                    longlong ts = 0;

                                    ts = dt_date_year(solid_date) * LL(10000000000) +
                                        dt_date_month(solid_date) * LL(100000000) +
                                        dt_date_mday(solid_date) * 1000000L +
                                        dt_date_hour(solid_date) * 10000L +
                                        dt_date_min(solid_date) * 100 +
                                        dt_date_sec(solid_date);

                                    sret = field->store(ts, (field->flags & UNSIGNED_FLAG));

                                    break;
                                }
                                default:
                                    ss_error;
                                    break;
                            }

                            break;

                        }
                        case RSDT_DFLOAT:
                            ss_error;
                            break;
                        case RSDT_BINARY:
                            {
                                void* data;
                                ulong length;
                                int mysqltype = field->type();

                                if (rs_aval_isblob(cd, atype, aval)) {
                                    tb_blobg2_loadblobtoaval_limit(cd, atype, aval, smallblobsizemax);
                                }

                                data = rs_aval_getdata(cd, atype, aval, &length);

                                switch (field->type()) {
                                    case MYSQL_TYPE_GEOMETRY:
                                    case MYSQL_TYPE_VAR_STRING: /* old <= 4.1 VARCHAR */
                                    case MYSQL_TYPE_VARCHAR:    /* new >= 5.0.3 true VARCHAR */
                                    case FIELD_TYPE_TINY_BLOB:
                                    case FIELD_TYPE_MEDIUM_BLOB:
                                    case FIELD_TYPE_BLOB:
                                    case FIELD_TYPE_LONG_BLOB:
                                        ss_pprintf_1(("ha_solid::solid_fetch:mysqltype %d, length %ld\n", mysqltype, (long)length));

                                        sret = field->store(
                                                    (const char *)data,
                                                    (uint)length,
                                                    field->charset());
                                        break;
                                    default:
                                        ss_dassert(field->pack_length() == length);
                                        memcpy(field->ptr, data, length);
                                        break;
                                }
                            }
                            break;
                        case RSDT_UNICODE:
                            if (rs_aval_isblob(cd, atype, aval)) {
                                tb_blobg2_loadblobtoaval_limit(cd, atype, aval, smallblobsizemax);
                            }

                            str = (char*)rs_aval_getdata(cd, atype, aval, &aval_length);

                            switch (field->real_type()) {
                                case MYSQL_TYPE_SET:
                                case MYSQL_TYPE_ENUM:
                                    ss_error;
                                    break;
                                default:
                                    sret = field->store(
                                                str,
                                                aval_length,
                                                field->charset());
                                    break;
                            }
                            break;
                        case RSDT_BIGINT:
                            {
                                ss_int8_t i8;

                                i8 = rs_aval_getint8(cd, atype, aval);

                                switch(field->real_type()) {
                                    case FIELD_TYPE_LONG:
                                    {
                                        longlong ll;

                                        if (field->flags & UNSIGNED_FLAG) {
                                            ll = (ulong) SsInt8GetNativeUint8(i8);
                                        } else {
                                            ll = (long) SsInt8GetNativeUint8(i8);
                                        }

                                        sret = field->store(ll, (field->flags & UNSIGNED_FLAG));
                                        break;
                                    }
                                    default:
                                    {
                                        if (field->flags & UNSIGNED_FLAG) {
                                            ulonglong ll;

                                            ll = SsInt8GetNativeUint8(i8);
                                            sret = field->store(ll, TRUE);
                                        } else {
                                            longlong ll;

                                            ll = (longlong) SsInt8GetNativeUint8(i8);
                                            sret = field->store(ll, FALSE);
                                        }
                                        break;
                                    }
                                }
                            }
                            break;
                        default:
                            ss_error;
                    }
                    field->ptr = tmp_ptr;
                    ss_poutput_2(ha_print_aval(cd, atype, aval, (char *)"ha_solid::solid_fetch:"));
                }
            }

            ss_win_perf_stop(fetch_perfcount, fetch_callcount);

#if MYSQL_VERSION_ID >= 50100
            dbug_tmp_restore_column_map(table->write_set, old_map);
#endif

            DBUG_RETURN(0);
        } else {
            DBUG_RETURN(0);
        }
}

/*#***********************************************************************\
 *
 *              ::index_init
 *
 *   Initialize a handle to use an index.
 *
 *
 * Parameters :
 *
 *     uint   keynr, in, use, key number
 *     bool   sorted, not used
 *
 * Return value : 0
 *
 * Globals used :
 */

#if MYSQL_VERSION_ID >= 50100
int ha_soliddb::index_init(
        uint keynr,
        bool sorted __attribute__(unused))
{
        int	error	= 0;
        DBUG_ENTER("index_init");

        active_index = keynr;

        /* TODO: Set up solidDB index ready here. */

        DBUG_RETURN(0);
}
#endif /* MYSQL_VERSION_ID >= 50100 */

/*#***********************************************************************\
 *
 *              ::solid_index_read_idx
 *
 * Index read
 *
 * Parameters :
 *
 *     byte*                  buf, in out, row buffer
 *     uint                   index, in, index number
 *     const byte*            key, in, use
 *     uint                   key_len, in, use, key length
 *     enum ha_rkey:function  find_flag, in, use
 *     bool                   firstp, in, use
 *     THD*                   thd, in, use
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::solid_index_read_idx(
        byte* buf,
        uint index,
        const byte* key,
        uint key_len,
        enum ha_rkey_function find_flag,
        bool firstp,
        THD* thd)
{
        int           rc;
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)

        ss_win_perf_start;
        DBUG_ENTER("ha_soliddb::solid_index_read_idx");
        ss_pprintf_1(("ha_solid::solid_index_read\n"));

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);
        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            DBUG_RETURN(rc);
        }

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, FALSE);
#else
        solid_trans_beginif(thd, FALSE);
#endif

        solid_relcur = solid_relcur_create(
                            con,
                            solid_relcur,
                            solid_rsrelh,
                            table,
                            thd,
                            index,
                            &table->key_info[index]);

        rc = solid_relcur_setconstr(
                thd,
                cd,
                trans,
                solid_rsrelh,
                solid_relcur,
                index,
                &table->key_info[index],
                table,
                key,
                key_len,
                find_flag,
                lower_case_table_names);

        if (rc == 0) {
            if (!firstp) {
                dbe_ret_t dberc;
                su_err_t* errh = NULL;

                /* Create dbe cursor with NULL buf in fetch. */
                rc = solid_fetch(NULL, firstp, thd);

                if (rc == 0) {
                    ss_pprintf_2(("ha_solid::solid_index_read, dbe_cursor_gotoend\n"));
                    do {
                        dberc = dbe_cursor_gotoend(
                                    solid_relcur->sr_relcur,
                                    tb_trans_dbtrx(cd, trans),
                                    &errh);
                    } while (rs_sysi_lockwait(cd) ||dberc == DBE_RC_WAITLOCK);
                    su_rc_assert(dberc == DBE_RC_SUCC, dberc);
                }
            }
            if (rc == 0) {
                rc = solid_fetch(buf, firstp, thd);
            }
        }

        ss_win_perf_stop(index_read_idx_perfcount, index_read_idx_callcount);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_read
 *
 * Positions an index cursor to the index specified in the handle. Fetches the
 * row if available. If the key value is null, begin at the first key of the
 * index.
 *
 * Parameters :
 *
 *     byte*                 buf, in out, row buffer
 *     const byte*           key, in, use,
 *     uint key_len          key_len, in, use
 *     enum ha_rkey_function find_flag, in
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::index_read(
        byte * buf,
        const byte * key,
        uint key_len,
        enum ha_rkey_function find_flag)
{
        int rc;
        THD* thd = current_thd;

        DBUG_ENTER("ha_soliddb::index_read");
        ss_pprintf_1(("ha_solid::index_read\n"));

      	statistic_increment(thd->status_var.ha_read_key_count, &LOCK_status);

        if (find_flag == HA_READ_KEY_OR_PREV ||
            find_flag == HA_READ_PREFIX_LAST_OR_PREV ||
            find_flag == HA_READ_BEFORE_KEY) {
            rc = solid_index_read_idx(buf, active_index, key, key_len, find_flag, FALSE, thd);
        } else {
            rc = solid_index_read_idx(buf, active_index, key, key_len, find_flag, TRUE, thd);
        }

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_read_last
 *
 * Position to the last row in the index with the current key value or prefix.
 *
 *
 * Parameters :
 *
 *     byte*                 buf, in out, row buffer
 *     const byte*           key, in, use,
 *     uint key_len          key_len, in, use
 *  *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::index_read_last(
        byte* buf,
        const byte* key,
        uint key_len)
{
        int rc;
        THD *thd = current_thd;

        DBUG_ENTER("ha_soliddb::index_read_last");
        ss_pprintf_1(("ha_solid::index_read_last\n"));

      	statistic_increment(thd->status_var.ha_read_key_count, &LOCK_status);

        rc = solid_index_read_idx(buf, active_index, key, key_len, HA_READ_PREFIX_LAST_OR_PREV, FALSE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_read_idx
 *
 * Positions an index cursor to the index specified in key. Fetches the
 * row if any.  This is only used to read whole keys.
 *
 * Parameters :
 *
 *     byte*                  buf, in out, row buffer
 *     uint                   index, in, index number
 *     const byte*            key, in, use
 *     uint                   key_len, in, use, key length
 *     enum ha_rkey_function  find_flag, in, use
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_read_idx(
        byte* buf,
        uint index,
        const byte* key,
        uint key_len,
        enum ha_rkey_function find_flag)
{
        int rc = 0;
        THD* thd;

        DBUG_ENTER("ha_soliddb::index_read_idx");
        ss_pprintf_1(("ha_solid::index_read_idx\n"));
        SS_PMON_ADD(SS_PMON_MYSQL_INDEX_READ);

        thd = current_thd;

      	statistic_increment(thd->status_var.ha_read_key_count, &LOCK_status);

        rc = solid_index_read_idx(buf, index, key, key_len, find_flag, TRUE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_next
 *
 * Used to read forward through the index.
 *
 * Parameters :
 *
 *     byte*  buf, in out, row buffer
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_next(byte * buf)
{
        int rc;
        THD* thd;

        DBUG_ENTER("ha_soliddb::index_next");
        ss_pprintf_1(("ha_solid::index_next\n"));

        thd = current_thd;

      	statistic_increment(thd->status_var.ha_read_next_count, &LOCK_status);

        rc = solid_fetch(buf, TRUE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_next_same
 *
 * Read the next row matching to the key value given.
 *
 * Parameters :
 *
 *     byte*         buf, in out, row buffer
 *     const byte*   key, in, use, key value
 *     uint          keylen, in, use, key value length
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_next_same(
        byte*        buf,
        const byte*  key,
        uint         keylen)
{
        int rc;
        THD* thd;

        DBUG_ENTER("ha_soliddb::index_next_same");
        ss_pprintf_1(("ha_solid::index_next_same\n"));

        thd = current_thd;

      	statistic_increment(thd->status_var.ha_read_next_count, &LOCK_status);

        rc = solid_fetch(buf, TRUE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_prev
 *
 * Used to read backwards through the index.
 *
 * Parameters :
 *
 *     byte*  buf, in out, row buffer
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_prev(byte * buf)
{
        int rc;
        THD* thd;

        DBUG_ENTER("ha_soliddb::index_prev");
        ss_pprintf_1(("ha_solid::index_prev\n"));

        thd = current_thd;

        statistic_increment(thd->status_var.ha_read_prev_count,  &LOCK_status);

        rc = solid_fetch(buf, FALSE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_first
 *
 * index_first() asks for the first key in the index.
 *
 * Parameters :
 *
 *     byte*   buf, in out, row buffer
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_first(byte * buf)
{
      int rc;
      THD* thd;

      DBUG_ENTER("ha_soliddb::index_first");
      ss_pprintf_1(("ha_solid::index_first\n"));

      thd = current_thd;

      statistic_increment(thd->status_var.ha_read_first_count, &LOCK_status);

      rc = solid_index_read_idx(buf, active_index, NULL, 0, HA_READ_AFTER_KEY, TRUE, thd);

      DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::index_last
 *
 * index_last() asks for the last key in the index.
 *
 * Parameters :
 *
 *     byte*   buf, in out, row buffer
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::index_last(byte * buf)
{
        int rc;
        THD* thd;

        DBUG_ENTER("ha_soliddb::index_last");
        ss_pprintf_1(("ha_solid::index_last\n"));

        thd = current_thd;

        statistic_increment(thd->status_var.ha_read_last_count, &LOCK_status);

        rc = solid_index_read_idx(buf, active_index, NULL, 0, HA_READ_KEY_OR_PREV, FALSE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::rnd_init
 *
 * rnd_init() is called when the system wants the storage engine to do a table
 * scan.
 *
 * See the example in the introduction at the top of this file to see when
 * rnd_init() is called.
 *
 * Parameters :
 *
 *     bool  scan, in, not used
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::rnd_init(bool scan)
{
        DBUG_ENTER("ha_soliddb::rnd_init");
        ss_pprintf_1(("ha_solid::rnd_init\n"));
        SS_PMON_ADD(SS_PMON_MYSQL_RND_INIT);

        rnd_mustinit = TRUE;

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::rnd_end
 *
 * rnd_init() is called when the system wants the storage engine to do stop
 * a table scan.
 *
 * Parameters : -
 *
 * Return value : 0
 *
 * Globals used :
 */
int ha_soliddb::rnd_end()
{
        DBUG_ENTER("ha_soliddb::rnd_end");
        ss_pprintf_1(("ha_solid::rnd_end\n"));
        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::rnd_next
 *
 * This is called for each row of the table scan. When you run out of records
 * you should return HA_ERR_END_OF_FILE. Fill buff up with the row information.
 * The Field structure for the table is the key to getting data into buf
 * in a manner that will allow the server to understand it.
 *
 * Parameters :
 *
 *     byte*   buf, in out, row buffer
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::rnd_next(byte *buf)
{
        int rc;
        THD* thd;
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;

        DBUG_ENTER("ha_soliddb::rnd_next");
        ss_pprintf_1(("ha_solid::rnd_next\n"));

        thd = current_thd;

      	statistic_increment(thd->status_var.ha_read_rnd_next_count, &LOCK_status);

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        if (solid_rsrelh == NULL) {
            rc = HA_ERR_NO_SUCH_TABLE;
            DBUG_RETURN(rc);
        }

        ss_dassert(solid_rsrelh != NULL);
        ss_dassert(con != NULL);
        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;

        if (rnd_mustinit) {

#if MYSQL_VERSION_ID >= 50100
            solid_trans_beginif(ht, thd, FALSE);
#else
            solid_trans_beginif(thd, FALSE);
#endif

            solid_relcur = solid_relcur_create(
                                con,
                                solid_relcur,
                                solid_rsrelh,
                                table,
                                thd,
                                (unsigned int)TABLE_SCAN_INDEX,
                                NULL);

            solid_relcur->sr_full_scan = TRUE;
            solid_relcur->sr_key = NULL;
            solid_relcur->sr_index = TABLE_SCAN_INDEX;

            rc = solid_relcur_setconstr(
                    thd,
                    cd,
                    trans,
                    solid_rsrelh,
                    solid_relcur,
                    (unsigned int)TABLE_SCAN_INDEX,
                    NULL,
                    NULL,
                    NULL,
                    0,
                    HA_READ_AFTER_KEY,
                    lower_case_table_names);

            rnd_mustinit = FALSE;
        }

        rc = solid_fetch(buf, TRUE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              print_ref
 *
 * TODO: add description of this function
 *
 * Parameters :
 *
 *     SOLID_CONN*      con, in, use, solidDB connection
 *     solid_relcur_t*  solid_relcur, in, use, relation cursor
 *     byte*            pos, in ,use
 *
 * Return value : -
 *
 * Globals used :
 */
static void print_ref(SOLID_CONN* con, solid_relcur_t* solid_relcur, byte *pos)
{
        vtpl_t* vtpl;
        rs_relh_t* rsrelh;
        rs_key_t* key;
        rs_ano_t nrefparts;
        rs_ano_t i;
        rs_sysi_t* cd;
        rs_ttype_t* ttype;
        rs_tval_t* tval;

        cd = con->sc_cd;
        vtpl = (vtpl_t*)pos;
        rsrelh = solid_relcur->sr_rsrelh;
        key = rs_relh_clusterkey(cd, rsrelh);
        nrefparts = rs_key_nrefparts(cd, key);
        ttype = solid_relcur->sr_ttype;
        tval = rs_tval_create(cd, ttype);

        for (i = 0; i < nrefparts; i++) {
            rs_ano_t ano;

            ano = rs_keyp_ano(cd, key, i);
            if (ano != RS_ANO_NULL) {
                va_t* va;

                va = vtpl_getva_at(vtpl, i);
                ss_dassert(va != NULL);
                rs_tval_setva(cd, ttype, tval, ano, va);
            }
        }

        rs_tval_print(cd, ttype, tval);

        rs_tval_free(cd, ttype, tval);
}

/*#***********************************************************************\
 *
 *              ::position
 *
 * position() is called after each call to rnd_next() if the data needs
 * to be ordered. You can do something like the following to store
 * the position:
 * my_store_ptr(ref, ref_length, current_position);
 *
 * The server uses ref to store data. ref_length in the above case is
 * the size needed to store current_position. ref is just a byte array
 * that the server will maintain. If you are using offsets to mark rows, then
 * current_position should be the offset. If it is a primary key like in
 * BDB, then it needs to be a primary key.
 *
 * Parameters :
 *
 *     const byte*  record, in, use
 *
 * Return value : -
 *
 * Globals used : THD* current_thd
 */
void ha_soliddb::position(const byte *record)
{
        SOLID_CONN* con;
        rs_sysi_t* cd;
        void* tref;
        vtpl_t* vtpl;
        size_t data_len;
        THD* thd;

        DBUG_ENTER("ha_soliddb::position");
        ss_pprintf_1(("ha_solid::position\n"));
        CHK_RELCUR(solid_relcur);
        ss_dassert(solid_relcur->sr_relcur != NULL);

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);
        cd = con->sc_cd;

        tref = dbe_cursor_gettref(solid_relcur->sr_relcur, solid_relcur->sr_fetchtval);
        if (solid_relcur->sr_mainmem) {
            void* refdata;
            refdata = dbe_tref_getrefrvaldata((dbe_tref_t*)tref, &data_len);

            /* Copy tuple reference to buffer. */
            memcpy(ref, refdata, data_len);
        } else {
            vtpl = dbe_tref_getvtpl((dbe_tref_t*)tref);
            data_len = vtpl_grosslen(vtpl);
            ss_dassert(data_len <= ref_length);

            /* Copy tuple reference v-tuple to buffer. */
            memcpy(ref, vtpl, data_len);
        }

        /* Clear rest of buffer because MySQL may call memcmp
         * for ref_length length.
         */
        memset((char*)ref + data_len, '\0', ref_length - data_len);

        ss_poutput_1(if (!solid_relcur->sr_mainmem) print_ref(con, solid_relcur, (byte*)ref));

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              ::rnd_pos
 *
 * This is like rnd_next, but you are given a position to use
 * to determine the row. The position will be of the type that you stored in
 * ref. You can use ha_get_ptr(pos,ref_length) to retrieve whatever key
 * or position you saved when position() was called.
 *
 * Parameters :
 *
 *     byte*  buf, in out, row buffer
 *     byte*  pos, in, use, position
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::rnd_pos(byte * buf, byte *pos)
{
        int rc;
        rs_relh_t* rsrelh;
        rs_ano_t nrefparts;
        rs_ano_t i;
        rs_sysi_t* cd;
        tb_trans_t* trans;
        rs_ttype_t* ttype;
        rs_tval_t* tval;
        SOLID_CONN* con;
        THD* thd;
        su_ret_t suret;
        rs_key_t* key;

        DBUG_ENTER("ha_soliddb::rnd_pos");
        ss_pprintf_1(("ha_solid::rnd_pos\n"));

        thd = current_thd;

        statistic_increment(thd->status_var.ha_read_rnd_count, &LOCK_status);

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);
        cd = con->sc_cd;
        trans = con->sc_trans;

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            DBUG_RETURN(rc);
        }

        if (rnd_mustinit) {

#if MYSQL_VERSION_ID >= 50100
            solid_trans_beginif(ht, thd, FALSE);
#else
            solid_trans_beginif(thd, FALSE);
#endif

            solid_relcur = solid_relcur_create(
                                con,
                                solid_relcur,
                                solid_rsrelh,
                                table,
                                thd,
                                TABLE_SCAN_INDEX,
                                NULL);

            solid_relcur->sr_full_scan = TRUE;
            solid_relcur->sr_key = NULL;
            solid_relcur->sr_index = TABLE_SCAN_INDEX;

            rc = solid_relcur_setconstr(
                    thd,
                    cd,
                    trans,
                    solid_rsrelh,
                    solid_relcur,
                    (unsigned int)TABLE_SCAN_INDEX,
                    NULL,
                    NULL,
                    NULL,
                    0,
                    HA_READ_AFTER_KEY,
                    lower_case_table_names);

            rnd_mustinit = FALSE;

            if (rc == 0) {
                /* Just reset the cursor, buf == NULL. */
                rc = solid_fetch(NULL, TRUE, thd);
            }
        }

        CHK_RELCUR(solid_relcur);
        ss_dassert(solid_relcur->sr_relcur != NULL);

        ss_poutput_1(if (!solid_relcur->sr_mainmem) print_ref(con, solid_relcur, pos));

        rsrelh = solid_rsrelh;
        key = rs_relh_clusterkey(cd, rsrelh);
        ttype = solid_relcur->sr_ttype;
        if (solid_relcur->sr_postval == NULL) {
            solid_relcur->sr_postval = rs_tval_create(cd, ttype);
        }
        tval = solid_relcur->sr_postval;

        if (solid_relcur->sr_mainmem) {
            dbe_tref_projectrvaltotval(
                cd,
                (ss_byte_t*)pos,
                ref_length,
                ttype,
                tval,
                key);
        } else {
            vtpl_t* vtpl;

            vtpl = (vtpl_t*)pos;
            nrefparts = rs_key_nrefparts(cd, key);

            for (i = 0; i < nrefparts; i++) {
                rs_ano_t ano;

                ano = rs_keyp_ano(cd, key, i);
                if (ano != RS_ANO_NULL) {
                    va_t* va;

                    va = vtpl_getva_at(vtpl, i);
                    ss_dassert(va != NULL);
                    rs_tval_setvaref_flat(cd, ttype, tval, ano, va);
                }
            }
        }

        do {
            suret = dbe_cursor_setposition(
                        solid_relcur->sr_relcur,
                        tb_trans_dbtrx(cd, trans),
                        tval,
                        NULL);
        } while (rs_sysi_lockwait(cd) || suret == DBE_RC_WAITLOCK);
        ss_rc_assert(suret == DBE_RC_SUCC, suret);

        rc = solid_fetch(buf, TRUE, thd);

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::info
 *
 * ::info() is used to return information to the optimizer.
 * Currently this table handler doesn't implement most of the fields
 * really needed. SHOW also makes use of this data
 * Another note, you will probably want to have the following in your
 * code:
 * if (records < 2)
 *   records = 2;
 * The reason is that the server will optimize for cases of only a single
 * record. If in a table scan you don't know the number of records
 * it will probably be better to set records to two so you can return
 * as many records as you need.
 * Along with records a few more variables you may wish to set are:
 *   records
 *   deleted
 *   data_file_length
 *   index_file_length
 *   delete_length
 *   check_time
 *
 * Parameters :
 *
 *     uint  flag, in, use, status variable
 *
 * Return value : -
 *
 * Globals used : THD* current_thd
 */
#if MYSQL_VERSION_ID >= 50030
  int
#else
  void
#endif
ha_soliddb::info(uint flag)
{
        ss_int8_t     i8;
        SOLID_CONN*   con;
        rs_sysi_t*    cd;
        double        ntuples;
        double        nbytes;
        double        nindexbytes;
        bool          succp=TRUE;
        rs_relh_t*    rsrelh;
        uint          i;
        uint          j;
        su_pa_t*      keys;
        rs_key_t*     key;
        THD* thd;
        su_err_t*     errh=NULL;

        DBUG_ENTER("ha_soliddb::info");
        ss_pprintf_1(("ha_solid::info, flag=%d\n", flag));

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

/*        ss_rc_assert(table != NULL && table->s->path != NULL, flag); */

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::info:table not found\n"));

#if MYSQL_VERSION_ID >= 50030
            DBUG_RETURN(0);
#else
            DBUG_VOID_RETURN;
#endif
        }

        cd = con->sc_cd;

        if (flag & HA_STATUS_TIME) {
            rs_relh_t*    s_relh;
            rs_entname_t  en;
            char*         relname;
            char*         schema;
            char*         name;
            int           len;

#if MYSQL_VERSION_ID >= 50100
            name = (char *)table->s->path.str;
#else
            name = (char *)table->s->path;
#endif

            len = strlen(name) + 1;
            relname = (char *)SsMemAlloc(len);
            schema = (char *)SsMemAlloc(len);
            solid_set_relname(schema,relname, (char*)name, lower_case_table_names);

            rs_entname_initbuf( &en, NULL, /* catalog*/ schema, relname);

            s_relh = tb_dd_getrelh(cd, con->sc_trans, &en, NULL, &errh);

            if (errh) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
                su_err_done(errh);
                errh = NULL;
            } else {
                dt_date_t     relh_createtime;
                dt_date_t     relh_updatetime;
                /* Get the create_time and update_time of the table */

                relh_createtime = tb_dd_get_createtime(cd, con->sc_trans, s_relh);
                relh_updatetime = tb_dd_get_updatetime(cd, con->sc_trans, s_relh);

#if MYSQL_VERSION_ID >= 50100
                stats.create_time = ha_solid_time_mysql(&relh_createtime);
                stats.update_time = ha_solid_time_mysql(&relh_updatetime);
#else
                create_time = ha_solid_time_mysql(&relh_createtime);
                update_time = ha_solid_time_mysql(&relh_updatetime);
#endif
                rs_relh_done(cd, s_relh);
            }
            SsMemFree(relname);
            SsMemFree(schema);
        }

    	if (flag & HA_STATUS_VARIABLE) {
            double avg_rowlen;

            ss_pprintf_1(("ha_solid::info, HA_STATUS_VARIABLE\n"));

            rsrelh = solid_rsrelh;

            i8 = rs_relh_ntuples(cd, rsrelh);
            succp = SsInt8ConvertToDouble(&ntuples, i8);
            ss_dassert(succp);

            if (ntuples <= 2.0) {
                ntuples = 2.0;
            }

#if MYSQL_VERSION_ID >= 50100
            stats.records = (ha_rows)ntuples;
            stats.deleted = 0;
#else
            records = (ha_rows)ntuples;
            deleted = 0;
#endif

            ss_pprintf_2(("ha_solid::info, HA_STATUS_VARIABLE, records=%.1lf\n", ntuples));


            i8 = rs_relh_nbytes(cd, rsrelh);
            succp = SsInt8ConvertToDouble(&nbytes, i8);
            ss_dassert(succp);

            if (nbytes <= 20.0) {
                nbytes = 20.0;
            }

            keys = rs_relh_keys(cd, rsrelh);

#if MYSQL_VERSION_ID >= 50100
            stats.data_file_length = (ulonglong)nbytes;
#else
            data_file_length = (ulonglong)nbytes;
#endif

            ss_pprintf_2(("ha_solid::info, HA_STATUS_VARIABLE, data_file_length=%.1lf\n", nbytes));

            nindexbytes = 0.0;
            avg_rowlen = nbytes / ntuples;

            su_pa_do(keys, i) {
                double maxstoragelen;
                key = (rs_key_t*)su_pa_getdata(keys, i);
                if (rs_key_isclustering(cd, key)) {
                    continue;
                }

                maxstoragelen = (double)rs_key_maxstoragelen(cd, key);

                if (maxstoragelen > avg_rowlen) {
                    maxstoragelen = avg_rowlen;
                }
                nindexbytes = nindexbytes + maxstoragelen * ntuples;
            }

#if MYSQL_VERSION_ID >= 50100
            stats.delete_length = 0;
            stats.check_time = 0;
            stats.mean_rec_length = (ulong)avg_rowlen;
            stats.index_file_length = (ulonglong)nindexbytes;
#else
            delete_length = 0;
            check_time = 0;
            mean_rec_length = (ulong)avg_rowlen;
            index_file_length = (ulonglong)nindexbytes;
#endif

            ss_pprintf_2(("ha_solid::info, HA_STATUS_VARIABLE, index_file_length=%.1lf\n", nindexbytes));
            ss_pprintf_2(("ha_solid::info, HA_STATUS_VARIABLE, mean_rec_length=%.1lf\n", avg_rowlen));
        }

    	if (flag & HA_STATUS_CONST) {
            int kpno;
            int ano;
            long ano_rowcount;

            ss_pprintf_1(("ha_solid::info, HA_STATUS_CONST\n"));

            rsrelh = solid_rsrelh;

            i8 = rs_relh_ntuples(cd, rsrelh);
            succp = SsInt8ConvertToDouble(&ntuples, i8);
            ss_dassert(succp);

            if (ntuples <= 2.0) {
                ntuples = 2.0;
            }

            if (ntuples > LONG_MAX) {
                ano_rowcount = LONG_MAX;
            } else {
                ano_rowcount = (long)ntuples;
            }

            if (rs_sysi_simpleoptimizerrules(cd, ntuples)) {

                ss_pprintf_2(("ha_solid::info, HA_STATUS_CONST, simple rules, ntuples=%.1ld, set all to 1\n", ntuples));

                for (i = 0; i < table->s->keys; i++) {
        			for (j = 0; j < table->key_info[i].key_parts; j++) {
                        table->key_info[i].rec_per_key[j] = (ha_rows)1;
                    }
                }

            } else {

                tb_est_ensureselectivityinfo(cd, rsrelh);

                for (i = 0; i < table->s->keys; i++) {
                    long diffrowcount;

                    key = solid_resolve_key(
                                cd,
                                solid_rsrelh,
                                &table->key_info[i],
                                i,
                                lower_case_table_names);

                    /* Skip all key parts that are not user defined. */
                    for (kpno = 0; rs_keyp_parttype(cd, key, kpno) != RSAT_USER_DEFINED; kpno++) {
                        continue;
                    }

        			for (j = 0; j < table->key_info[i].key_parts; j++, kpno++) {
                        ano = rs_keyp_ano(cd, key, kpno);
                        ss_dassert(ano != RS_ANO_NULL);

                        diffrowcount = rs_relh_getdiffrowcount(
                                            cd,
                                            rsrelh,
                                            ano,
                                            ano_rowcount);

                        if (diffrowcount > 0) {
                            table->key_info[i].rec_per_key[j] = (ha_rows)diffrowcount / 2;
                        } else {
                            table->key_info[i].rec_per_key[j] = 1;
                        }

                        if (table->key_info[i].rec_per_key[j] == 0) {
                            table->key_info[i].rec_per_key[j] = 1;
                        }

                        ss_pprintf_2(("ha_solid::info, HA_STATUS_CONST, key=%s, ano=%d, diffrowcount=%ld, set all to 1\n",
                            rs_key_name(cd, key), ano, diffrowcount));
                    }
                }
            }
        }

        if (flag & HA_STATUS_ERRKEY) {
            errkey = con->sc_errkey_num;
            con->sc_errkey_num = 0;
        }

#if MYSQL_VERSION_ID >= 50030
            DBUG_RETURN(0);
#else
        DBUG_VOID_RETURN;
#endif
}

/*#***********************************************************************\
 *
 *              ::extra
 *
 * extra() is called whenever the server wishes to send a hint to
 * the storage engine.
 *
 * Parameters :
 *
 *     enum ha_extra_function  operation, in
 *
 * Return value : 0 (or error code)
 *
 * Globals used :
 */
int ha_soliddb::extra(enum ha_extra_function operation)
{
        DBUG_ENTER("ha_soliddb::extra");
        ss_pprintf_1(("ha_solid::extra:operation=%d\n", (int)operation));

        switch (operation) {
#if MYSQL_VERSION_ID >= 50100
#else
            case HA_EXTRA_RESET:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_RESET\n"));
                extra_keyread = FALSE;
                extra_retrieve_primary_key = FALSE;
                extra_retrieve_all_cols = TRUE; /* TODO: Should be FALSE */
                break;
#endif
            case HA_EXTRA_RESET_STATE:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_RESET_STATE\n"));
                extra_keyread = FALSE;
                extra_retrieve_primary_key = FALSE;
                extra_retrieve_all_cols = TRUE; /* TODO: Should be FALSE */
                break;
            case HA_EXTRA_NO_KEYREAD:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_NO_KEYREAD\n"));
                extra_keyread = FALSE;
                extra_retrieve_all_cols = TRUE;
                break;
#if MYSQL_VERSION_ID >= 50100
#else
            case HA_EXTRA_RETRIEVE_ALL_COLS:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_RETRIEVE_ALL_COLS\n"));
                extra_retrieve_all_cols = TRUE;
                extra_keyread = FALSE;
                break;
            case HA_EXTRA_RETRIEVE_PRIMARY_KEY:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_RETRIEVE_PRIMARY_KEY\n"));
                extra_retrieve_primary_key = TRUE;
                break;
#endif
            case HA_EXTRA_KEYREAD:
                ss_pprintf_2(("ha_solid::extra:HA_EXTRA_KEYREAD\n"));
                extra_keyread = TRUE;
                extra_retrieve_all_cols = FALSE;
                break;
            default:
                break;
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::scan_time
 *
 * Called in test_quick_select to determine if indexes should be used.
 * At the moment returns number or tuples on a relation.
 *
 * Parameters : -
 *
 * Return value : double
 *
 * Globals used : THD* current_thd
 */
double ha_soliddb::scan_time()
{
        ss_int8_t     i8;
        SOLID_CONN*   con;
        rs_sysi_t*    cd;
        double        t;
        bool          succp=TRUE;
        THD* thd;

        ss_pprintf_1(("ha_solid::scan_time\n"));

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            return(rc);
        }

        cd = con->sc_cd;

        i8 = rs_relh_ntuples(
                cd,
                solid_rsrelh);

        succp = SsInt8ConvertToDouble(&t, i8);
        ss_dassert(succp);

        ss_pprintf_2(("ha_solid::scan_time, t=%.1lf\n", t));

        return t;
}

/*#***********************************************************************\
 *
 *              ::read_time
 *
 * The next method will never be called if you do not implement indexes.
 *
 * Parameters :
 *
 *     uint     index, in, index number
 *     uint     ranges, in
 *     ha_rows  rows,in
 *
 * Return value : double, read time
 *
 * Globals used : THD* current_thd
 */
double ha_soliddb::read_time(uint index, uint ranges, ha_rows rows)
{
        rs_key_t* key;
        SOLID_CONN* con;
        rs_sysi_t* cd;
        double t;
        THD* thd;

        ss_pprintf_1(("ha_solid::read_time\n"));

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            return(rc);
        }

        cd = con->sc_cd;

        key = solid_resolve_key(
                    cd,
                    solid_rsrelh,
                    &table->key_info[index],
                    index,
                    lower_case_table_names);

        if (rs_key_isclustering(cd, key)) {
            t = (double)ranges * (double)rows;
            ss_pprintf_2(("ha_solid::read_time, clustering key, t=%.1lf\n", t));
        } else {
            t = (double)ranges * (double)rows * 1.2;
            ss_pprintf_2(("ha_solid::read_time, secondary key, t=%.1lf\n", t));
        }

        return t;
}

/*#***********************************************************************\
 *
 *              solid_get_key_limits
 *
 * TODO: Provide description of this function
 *
 * Parameters : -
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
static int solid_get_key_limits(
        THD *thd,
        rs_sysi_t* cd,
        rs_relh_t* rsrelh,
        TABLE* table,
        uint keyindex,
        KEY* key_info,
        const byte* keyconstr_ptr,
        uint keyconstr_len,
        enum ha_rkey_function find_flag,
        rs_tval_t** tval,
        int* relopno,
        int** relops,
        int** anos,
        bool  lower_case_names)
{
        KEY_PART_INFO* key_part;
        int kpno;
        int start_kpno;
        int nkeyparts;
        rs_ttype_t* ttype;
        rs_key_t* key;
        const byte* ptr;
        uint key_len;
        enum ha_rkey_function current_find_flag;
        int relopsize;
        bool succp=TRUE;
        su_err_t* errh = NULL;

        DBUG_ENTER("ha_solid::solid_get_key_limits");
        ss_pprintf_1(("ha_solid::solid_get_key_limits:key %s, keyindex %d\n", key_info->name, keyindex));

        key = solid_resolve_key(cd, rsrelh, key_info, keyindex, lower_case_names);
        key_part = key_info->key_part;

        ttype = rs_relh_ttype(cd, rsrelh);
        *tval = rs_tval_create(cd, ttype);

        relopsize = rs_key_nparts(cd, key) * sizeof((*relops)[0]);
        *relops = (int *)SsMemAlloc(relopsize);
        *anos = (int *)SsMemAlloc(relopsize);

        /* Init with illegal relop values.
         */
        memset(*relops, '\xff', relopsize);

        nkeyparts = 0;

        /* Skip all key parts that are not user defined. */
        for (kpno = 0; rs_keyp_parttype(cd, key, kpno) != RSAT_USER_DEFINED; kpno++) {
            continue;
        }
        start_kpno = kpno;
        *relopno = 0;

        ptr = keyconstr_ptr;
        key_len = keyconstr_len;
        current_find_flag = find_flag;
        kpno = start_kpno;

        /* Get MySQL constraints.
         *
         * Some of the code here is copied from ha_federated.cpp.
         * TODO: Need to recheck this.
         */
        for (key_part = key_info->key_part; ; key_part++, kpno++) {
            int relop;
            rs_ano_t ano;
            rs_atype_t* atype;
            rs_aval_t* aval;
            Field* field;
            uint store_length;
            bool isnull;
            bool mysql_isnull;
            int offset;

            nkeyparts++;

            field = key_part->field;
            store_length = key_part->store_length;

            ss_dassert(rs_keyp_parttype(cd, key, kpno) == RSAT_USER_DEFINED);

            ano = rs_keyp_ano(cd, key, kpno);
            ss_dassert(ano != RS_ANO_NULL);
            (*anos)[*relopno] = ano;

            ss_pprintf_2(("ha_solid::solid_get_key_limits:kpno=%d, ano=%d\n", kpno, ano));

            atype = rs_ttype_atype(cd, ttype, ano);
            aval = rs_tval_aval(cd, ttype, *tval, ano);

            isnull = FALSE;
            offset = 0;

            if (key_part->null_bit) {
                if (*ptr) {
                    ss_pprintf_3(("ha_solid::solid_get_key_limits:NULL\n"));
                    rs_aval_setnull(
                        cd,
                        atype,
                        aval);
                    isnull = TRUE;
                    mysql_isnull = mysql_record_is_null(table, field, (char *)ptr);
                    switch (current_find_flag) {
                        case HA_READ_KEY_EXACT:
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:HA_READ_KEY_EXACT->RS_RELOP_EQUAL\n"));
                            relop = RS_RELOP_EQUAL;
                            break;
                        case HA_READ_KEY_OR_NEXT:
                        case HA_READ_PREFIX:
                        case HA_READ_PREFIX_LAST:
                        case HA_READ_PREFIX_LAST_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:HA_READ_KEY_OR_NEXT...->RS_RELOP_GE\n"));
                            relop = RS_RELOP_GE;
                            break;
                        case HA_READ_BEFORE_KEY:
                        case HA_READ_KEY_OR_PREV:
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:HA_READ_BEFORE_KEY...->RS_RELOP_GE\n"));
                            relop = RS_RELOP_LE;
                            break;
                        case HA_READ_AFTER_KEY:
                            if (mysql_isnull) {
                                ss_pprintf_3(("ha_solid::solid_get_key_limits:HA_READ_AFTER_KEY...->RS_RELOP_GE\n"));
                                relop = RS_RELOP_GE;
                            } else {
                                ss_pprintf_3(("ha_solid::solid_get_key_limits:HA_READ_AFTER_KEY...->RS_RELOP_GT\n"));
                                relop = RS_RELOP_GT;
                            }
                            break;
                        default:
                            ss_pprintf_1(("ha_solid::solid_get_key_limits:**UNKNOWN** find_flag=%d\n", current_find_flag));
                            DBUG_RETURN(1);
                    }
                }
                offset = 1;
            }
            if (!isnull) {
                switch (current_find_flag) {
                    case HA_READ_KEY_EXACT:
                        if (store_length >= key_len
                            || key_part->type == HA_KEYTYPE_BIT
                            || field->result_type() != STRING_RESULT)
                        {
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_EQUAL\n"));
                            relop = RS_RELOP_EQUAL;
                        } else {
                            /* TODO: LIKE */
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_GE\n"));
                            relop = RS_RELOP_GE;
                        }
                        break;
                    case HA_READ_AFTER_KEY:
                        if (store_length >= key_len) {
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_GT\n"));
                            relop = RS_RELOP_GT;
                            break;
                        }
                        /* FALLTHROUGH */
                    case HA_READ_KEY_OR_NEXT:
                        ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_GE\n"));
                        relop = RS_RELOP_GE;
                        break;
                    case HA_READ_BEFORE_KEY:
                        if (store_length >= key_len) {
                            ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_LT\n"));
                            relop = RS_RELOP_LT;
                            break;
                        }
                        /* FALLTHROUGH */
                    case HA_READ_KEY_OR_PREV:
                        ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_LE\n"));
                        relop = RS_RELOP_LE;
                        break;

                    case HA_READ_PREFIX:
                        ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_GE\n"));
                        relop = RS_RELOP_GE;
                        break;

                    case HA_READ_PREFIX_LAST:
                    case HA_READ_PREFIX_LAST_OR_PREV:
                        ss_pprintf_3(("ha_solid::solid_get_key_limits:RS_RELOP_LE\n"));
                        relop = RS_RELOP_LE;
                        break;

                    default:
                        ss_pprintf_1(("ha_solid::solid_get_key_limits:**UNKNOWN** find_flag=%d\n", current_find_flag));
                        DBUG_RETURN(1);
                }

                succp = set_mysql_field_to_aval_or_dynva(thd, table, cd, atype, aval, NULL, field, ptr + offset, TRUE, &errh);
                if (!succp) {
                    su_err_done(errh);
                    DBUG_RETURN(1);
                }
            }

            ss_poutput_2(ha_print_aval(cd, atype, aval, (char *)"ha_solid::solid_get_key_limits:"));

            (*relops)[*relopno] = relop;
            (*relopno)++;

            if (store_length >= key_len) {
                break;
            }
            key_len -= store_length;
            ptr += store_length;
        }
        ss_poutput_2(rs_tval_print(cd, ttype, *tval));
        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::records_in_range
 *
 * Given a starting key, and an ending key estimate the number of rows that
 * will exist between the two. end_key may be empty which in case determine
 * if start_key matches any rows.
 *
 * Parameters :
 *
 *     uint        in, index number
 *     key_range*  min_key, in, use
 *     key_range*  max_key, in, use
 *
 * Return value : ha_rows
 *
 * Globals used : THD* current_thd
 */
ha_rows ha_soliddb::records_in_range(uint inx, key_range *min_key,
                                     key_range *max_key)
{
        int rc;
        SOLID_CONN*   con;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        double        rowcount;
        THD* thd;
        rs_ttype_t* ttype;
        rs_tval_t* tval1 = NULL;
        int nvalues1 = 0;
        int* relops1 = NULL;
        int* anos1 = NULL;
        rs_tval_t* tval2 = NULL;
        int nvalues2 = 0;
        int* relops2 = NULL;
        int* anos2 = NULL;
        rs_relh_t* rsrelh;
        ss_int8_t i8;
        double new_rowcount;
        int nrange = 0;
        int i;
        bool succp=TRUE;
        ss_win_perf(__int64 startcount;)
        ss_win_perf(__int64 endcount;)

        ss_win_perf_start;

        DBUG_ENTER("ha_soliddb::records_in_range");
        ss_pprintf_1(("ha_solid::records_in_range\n"));
        SS_PMON_ADD(SS_PMON_MYSQL_RECORDS_IN_RANGE);

        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

        if (solid_rsrelh == NULL) {
            int rc;
            rc = ha_solid_mysql_error(thd, con->sc_errh);
            ss_pprintf_1(("ha_solid::table not found\n"));
            DBUG_RETURN(rc);
        }

        cd    = con->sc_cd;
        trans = con->sc_trans;
        rsrelh = solid_rsrelh;
        ttype = rs_relh_ttype(cd, rsrelh);

        i8 = rs_relh_ntuples(cd, rsrelh);

        succp = SsInt8ConvertToDouble(&rowcount, i8);
        ss_dassert(succp);
        ss_pprintf_2(("ha_solid::records_in_range, table ntuples=%.1lf\n", rowcount));

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, FALSE);
#else
        solid_trans_beginif(thd, FALSE);
#endif

        if (min_key == NULL) {
            min_key = max_key;
            max_key = NULL;
        }

        if ((min_key == NULL && max_key == NULL)
            || (rs_sysi_simpleoptimizerrules(cd, rowcount)
                && (min_key->flag == HA_READ_KEY_EXACT || max_key == NULL))) {
            double selectivity = 1.0;
            enum ha_rkey_function find_flag;

            ss_pprintf_2(("ha_solid::records_in_range, simple rules\n"));

            if (min_key) {
                find_flag = min_key->flag;
            } else {
                find_flag = HA_READ_KEY_EXACT;
            }

            switch (find_flag) {
                case HA_READ_KEY_EXACT:
                    selectivity = rs_sqli_equal_selectivity(rs_sysi_sqlinfo(cd));
                    break;
                case HA_READ_PREFIX:
                case HA_READ_PREFIX_LAST:
                case HA_READ_PREFIX_LAST_OR_PREV:
                case HA_READ_AFTER_KEY:
                case HA_READ_KEY_OR_NEXT:
                case HA_READ_BEFORE_KEY:
                case HA_READ_KEY_OR_PREV:
                    selectivity = rs_sqli_compare_selectivity(rs_sysi_sqlinfo(cd));
                    break;
                default:
                    ss_error;
            }

            ss_pprintf_2(("ha_solid::records_in_range, selectivity=%.1lf\n", selectivity));
            ss_pprintf_2(("ha_solid::records_in_range, rowcount=%.1lf\n", rowcount));

            rowcount = selectivity * rowcount;
        } else {

            tb_est_ensureselectivityinfo(cd, rsrelh);

            rc = solid_get_key_limits(
                    thd,
                    cd,
                    solid_rsrelh,
                    table,
                    inx,
                    &table->key_info[inx],
                    min_key->key,
                    min_key->length,
                    min_key->flag,
                    &tval1,
                    &nvalues1,
                    &relops1,
                    &anos1,
                    lower_case_table_names);

            if (rc) {

                if (tval1) {
                    rs_tval_free(cd, ttype, tval1);
                }

                if (anos1) {
                    SsMemFree(anos1);
                }

                if (relops1) {
                    SsMemFree(relops1);
                }

                DBUG_RETURN((ha_rows)0);
            }

            if (max_key != NULL) {

                rc = solid_get_key_limits(
                        thd,
                        cd,
                        solid_rsrelh,
                        table,
                        inx,
                        &table->key_info[inx],
                        max_key->key,
                        max_key->length,
                        max_key->flag,
                        &tval2,
                        &nvalues2,
                        &relops2,
                        &anos2,
                        lower_case_table_names);
                
                if (rc) {
                    
                    if (tval1) {
                        rs_tval_free(cd, ttype, tval1);
                    }

                    if (anos1) {
                        SsMemFree(anos1);
                    }

                    if (relops1) {
                        SsMemFree(relops1);
                    }

                    if (tval2) {
                        rs_tval_free(cd, ttype, tval2);
                    }

                    if (anos2) {
                        SsMemFree(anos2);
                    }

                    if (relops2) {
                        SsMemFree(relops2);
                    }

                    DBUG_RETURN((ha_rows)0);
                }

                nrange = SS_MIN(nvalues1, nvalues2);
            }

            for (i = 0; i < nrange; i++) {

                succp = rs_relh_getrangeselectivity(
                    	    cd,
                            rsrelh,
                            anos1[i],
                            rs_tval_va(cd, ttype, tval1, anos1[i]),
                            TRUE,
                            rs_tval_va(cd, ttype, tval2, anos1[i]),
                            TRUE,
                            FALSE,
                            &i8);

                if (succp) {
                    succp = SsInt8ConvertToDouble(&new_rowcount, i8);
                    ss_dassert(succp);
                    ss_pprintf_2(("ha_solid::records_in_range, rs_relh_getrangeselectivity=%.1lf\n", new_rowcount));
                    rowcount = SS_MIN(rowcount, new_rowcount);
                }
            }

            for (; i < nvalues1; i++) {

                int relop = relops1[i];

                /* At the moment you can't get selectivity for ISNULL and ISNOTNULL
                   relational operators */

                if (relop != RS_RELOP_ISNULL && relop != RS_RELOP_ISNOTNULL) {

                    succp = rs_relh_getrelopselectivity(
                    	    cd,
                            rsrelh,
                            anos1[i],
                            relop,
                            rs_tval_va(cd, ttype, tval1, anos1[i]),
                            '\\',
                            &i8);

                    if (succp) {
                        succp = SsInt8ConvertToDouble(&new_rowcount, i8);
                        ss_dassert(succp);
                        ss_pprintf_2(("ha_solid::records_in_range, rs_relh_getrelopselectivity=%.1lf\n",
                                      new_rowcount));
                        rowcount = SS_MIN(rowcount, new_rowcount);
                    }
                }
            }

            rs_tval_free(cd, ttype, tval1);

            SsMemFree(relops1);
            SsMemFree(anos1);

            if (relops2) {
                SsMemFree(relops2);
            }

            if (anos2) {
                SsMemFree(anos2);
            }

            if (tval2 != NULL) {
                rs_tval_free(cd, ttype, tval2);
            }
        }

        if (rowcount <= 1) {
            /* Return always at least one. */
            rowcount = 1;
        }

        ss_pprintf_2(("ha_solid::records_in_range:rowcount=%ld\n", (ha_rows)rowcount));

        ss_win_perf_stop(records_in_range_perfcount, records_in_range_callcount);

        DBUG_RETURN((ha_rows)rowcount);
}

/*#***********************************************************************\
 *
 *              ::delete_all_rows
 *
 * Used to delete all rows in a table. Both for cases of truncate and
 * for cases where the optimizer realizes that all rows will be
 * removed as a result of a SQL statement.
 *
 * Parameters : -
 *
 * Return value : 0 or error code
 *
 * Globals used : current_thd
 */
int ha_soliddb::delete_all_rows()
{
        int rc;
        rs_sysi_t* cd;
        tb_trans_t* trans;
        su_err_t* errh = NULL;
        solid_bool succp = FALSE;
        THD* thd;
        SOLID_CONN* con;
        int          len;
        char*        relname;
        char*        schema;
        char*        name;
        void*        cont=NULL;

        DBUG_ENTER("ha_soliddb::delete_all_rows");

#ifdef SOLID_TRUNCATE
        thd = current_thd;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
        name = (char *)table->s->path.str;
#else
        con = get_solid_connection(thd, table->s->path);
        name = (char *)table->s->path;
#endif
        ss_dassert(con != NULL);

        cd = con->sc_cd;
        trans = con->sc_trans;

        if (rs_relh_issysrel(cd, solid_rsrelh)) {
            DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        con->sc_table = table;

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        len = strlen(name) + 1;
        relname = (char *)SsMemAlloc(len);
        schema = (char *)SsMemAlloc(len);
        solid_set_relname(schema, relname, name, lower_case_table_names);

        succp = tb_truncaterelation(cd, trans, relname, schema,
                                    NULL, NULL, &cont, &errh);

        if (!succp) {
            sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
        }

        if (!succp) {
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
        } else {
            rc = 0;
        }
        
        DBUG_RETURN(0);
#else
        DBUG_RETURN(HA_ERR_WRONG_COMMAND);
#endif
}

/*#***********************************************************************\
 *
 *              ::external_lock
 *
 * First you should go read the section "locking functions for mysql" in
 * lock.cc to understand this.
 * This create a lock on the table. If you are implementing a storage engine
 * that can handle transacations look at ha_berkely.cc to see how you will
 * want to goo about doing this. Otherwise you should consider calling flock()
 * here.
 *
 * Parameters :
 *
 *     THD*  thd, in, use
 *     int   lock_type, in, use, MySQL lock type
 *
 * Return value : 0 or derror code
 *
 * Globals used :
 */
int ha_soliddb::external_lock(THD *thd, int lock_type)
{
        SOLID_CONN* con;
        su_err_t*   errh = NULL;
        int rc = 0;

        DBUG_ENTER("ha_soliddb::external_lock");
        ss_pprintf_1(("ha_solid::external_lock:lock_type=%d\n", lock_type));

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, table->s->path.str);
#else
        con = get_solid_connection(thd, table->s->path);
#endif

        ss_dassert(con != NULL);

        for_update = FALSE;
        extra_keyread = FALSE;
        extra_retrieve_primary_key = FALSE;
        extra_retrieve_all_cols = TRUE; /* TODO: Should be FALSE */

#if MYSQL_VERSION_ID >= 50100
        thd->set_current_stmt_binlog_row_based_if_mixed();
#endif

        if (con->sc_relcurdonelist != NULL) {
            solid_relcur_clearrelcurdonelist(con, 5);
        }

        if (lock_type != F_UNLCK) {
            ss_pprintf_2(("ha_solid::external_lock:LOCK\n"));
            con->con_n_tables++;

            /*
              Set up transaction isolation level at the start of the transaction
              i.e. when first table is accessed. Note that you can't change
              isolation level in the middle of the transaction.
            */

            if(con->con_n_tables == 1) {
                ss_pprintf_1(("ha_solid::external_lock:isolation_type=%d\n",
                             (int) thd->variables.tx_isolation));

                tb_trans_settransoption(con->sc_cd, con->sc_trans,
                    (tb_transopt_t)ha_solid_map_mysql_solid_isolation(
                            (enum_tx_isolation)thd->variables.tx_isolation));

                ss_pprintf_1(("ha_solid::external_lock:trans_isolation=%d\n",
                             (int)tb_trans_getisolation(con->sc_cd, con->sc_trans)));
            }

            if (thd->transaction.on) {
                if (thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN)) {

#if MYSQL_VERSION_ID >= 50100
                    trans_register_ha(thd, TRUE, ht);
#else
                    trans_register_ha(thd, TRUE, &solid_hton);
#endif
                }

#if MYSQL_VERSION_ID >= 50100
                trans_register_ha(thd, FALSE, ht);
#else
                trans_register_ha(thd, FALSE, &solid_hton);
#endif
            }

#if MYSQL_VERSION_ID >= 50100
            solid_trans_beginif(ht, thd, FALSE);
#else
            solid_trans_beginif(thd, FALSE);
#endif

            tb_trans_startnewsearch(con->sc_cd, con->sc_trans);

            if (lock_type == F_WRLCK) {
                ss_pprintf_2(("ha_solid::external_lock:WRITE LOCK\n"));

                if (thd->lex->sql_command == SQLCOM_SELECT ||
                    thd->lex->sql_command == SQLCOM_INSERT_SELECT) {
                    for_update = TRUE;
                }
            }
        } else {
            /*
             * When last table is unlocked in autocommit we execute commit.
             * This is needed because it seems that MySql does not call commit
             * for selects in autocommit mode.
             */
            ss_pprintf_2(("ha_solid::external_lock:UNLOCK\n"));
            con->con_n_tables--;
            if (con->con_n_tables == 0) {
                if (!(thd->options & (OPTION_NOT_AUTOCOMMIT | OPTION_BEGIN))) {
                    /* In case of autocommit we commit transaction on last unlock in this session */
                    bool succp=TRUE;
                    succp = internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, TRUE, &errh);

                    if (!succp) {
                        rc = ha_solid_mysql_error(thd, errh);
                        su_err_done(errh);
                    } else {
                        rc = 0;
                    }

                }
            }

        }
        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::store_lock
 *
 * The idea with handler::store_lock() is the following:
 *
 * The statement decided which locks we should need for the table
 * for updates/deletes/inserts we get WRITE locks, for SELECT... we get
 * read locks.
 *
 * Before adding the lock into the table lock handler (see thr_lock.c)
 * mysqld calls store lock with the requested locks.  Store lock can now
 * modify a write lock to a read lock (or some other lock), ignore the
 * lock (if we don't want to use MySQL table locks at all) or add locks
 * for many tables (like we do when we are using a MERGE handler).
 *
 * Berkeley DB for example  changes all WRITE locks to TL_WRITE_ALLOW_WRITE
 * (which signals that we are doing WRITES, but we are still allowing other
 * reader's and writer's.
 *
 * When releasing locks, store_lock() are also called. In this case one
 * usually doesn't have to do anything.
 *
 * In some exceptional cases MySQL may send a request for a TL_IGNORE;
 * This means that we are requesting the same lock as last time and this
 * should also be ignored. (This may happen when someone does a flush
 * table when we have opened a part of the tables, in which case mysqld
 * closes and reopens the tables and tries to get the same locks at last
 * time).  In the future we will probably try to remove this.
 *
 * Parameters :
 *
 *    THD*                thd, in, use
 *    THR_LOCK_DATA**     to, in, use
 *    enum thr_lock_type  lock_type, in
 *
 * Return value : THD_LOCK_DATA**
 *
 * Globals used :
 */
THR_LOCK_DATA **ha_soliddb::store_lock(THD *thd,
                                       THR_LOCK_DATA **to,
                                       enum thr_lock_type lock_type)
{
        ss_pprintf_1(("ha_solid::store_lock:lock_type=%d\n", (int)lock_type));

        /* The following code is copied from ha_archive.cc. The same
         * code seem to be also in ha_ndbcluster.cc.
         * TODO check this logic here, some other handlers have more complex logic.
         */
        if (lock_type != TL_IGNORE && lock.type == TL_UNLOCK)
        {
          /*
            Here is where we get into the guts of a row level lock.
            If TL_UNLOCK is set
            If we are not doing a LOCK TABLE or DISCARD/IMPORT
            TABLESPACE, then allow multiple writers
          */

          if ((lock_type >= TL_WRITE_CONCURRENT_INSERT &&
               lock_type <= TL_WRITE) && !thd->in_lock_tables
              && !thd->tablespace_op)
            lock_type = TL_WRITE_ALLOW_WRITE;

          /*
            In queries of type INSERT INTO t1 SELECT ... FROM t2 ...
            MySQL would use the lock TL_READ_NO_INSERT on t2, and that
            would conflict with TL_WRITE_ALLOW_WRITE, blocking all inserts
            to t2. Convert the lock to a normal read lock to allow
            concurrent inserts to t2.
          */

          if (lock_type == TL_READ_NO_INSERT && !thd->in_lock_tables)
            lock_type = TL_READ;

          lock.type=lock_type;
        }

        *to++= &lock;

        return to;
}

/*#***********************************************************************\
*              ::drop_table_autoincrements
* Parameters :
* rs_relh_t* relh, relation to drop autoincrements from. IN
* SOLID_CONN* con - solid connection structure. IN
*
* Return value : true or false on error.
*
*/
static bool drop_table_autoincrements(
        rs_relh_t* relh,
        SOLID_CONN* con)
{
    bool succp = TRUE;
    su_err_t* errh = NULL;
    char* seq_name = NULL;
    char* authid = NULL;
    char* catalog = NULL;
    rs_auth_t* auth = rs_sysi_auth(con->sc_cd);

    rs_auth_setsystempriv(con->sc_cd, auth, TRUE);

    if (tb_seq_findbyid(con->sc_cd, 
        rs_relh_readautoincrement_seqid(con->sc_cd, relh),
        &seq_name, &authid, &catalog)) {
            void* dummy = NULL;

            succp = tb_dropseq(con->sc_cd, con->sc_trans, seq_name, authid, catalog,
                               NULL, FALSE, &dummy, &errh);

            ss_pprintf_1(("SEQ NAME %s dropped succ = %d\n", seq_name, (int)succp));
    }

    if (seq_name) {
        SsMemFree(seq_name);
    }

    if (authid) {
        SsMemFree(authid);
    }

    if (catalog) {
        SsMemFree(catalog);
    }
    
    seq_name = authid = catalog = NULL;
    
    if (errh) {
        sql_print_error("%s", su_err_geterrstr(errh));
        su_err_done(errh);
        errh = NULL;
    }

    rs_auth_setsystempriv(con->sc_cd, auth, FALSE);

    return succp;
}

/*#***********************************************************************\
 *
 *              ::delete_table
 *
 * Used to delete a table. By the time delete_table() has been called all
 * opened references to this table will have been closed (and your globally
 * shared references released. The variable name will just be the name of
 * the table. You will need to remove any files you have created at this point.
 *
 * If you do not implement this, the default delete_table() is called from
 * handler.cc and it will delete all files with the file extentions returned
 * by bas_ext().
 *
 * Called from handler.cc by delete_table and  ha_create_table(). Only used
 * during create if the table_flag HA_DROP_BEFORE_CREATE was specified for
 * the storage engine.
 *
 * Parameters :
 *
 *     const char* name, in, MySQL table name
 *
 * Return value : 0 or error code
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::delete_table(const char *name)
{
        SDB_DBUG_ENTER("ha_solid::delete_table");
        bool succp = true;
        int rc = 0;
        su_err_t* errh = NULL;
        SOLID_CONN* con = NULL;
        rs_entname_t en;
        THD* thd = current_thd;
        solid_bool checkforkeys = TRUE;
        solid_bool cascade = FALSE;
        bool commit_succ_p = true;

        solid_relname(&en, name);

        ss_pprintf_1(("ha_solid::delete_table:'%s'->'%s.%s'\n", name, en.en_schema, en.en_name));

        if (is_soliddb_system_relation(en)) {
            ss_pprintf_1(("Trying to drop SolidDB system table %s\n", name));
            rs_entname_done_buf(&en);
            SDB_DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

        if (is_DDL_command( thd ) && !start_solid_DDL(rc)) {
            rs_entname_done_buf(&en);
            SDB_DBUG_RETURN(rc);
        }

        con = get_solid_connection(name);
        ss_dassert(con != NULL);

        if (this->solid_rsrelh == NULL) {
            ss_pprintf_1(("Trying to drop unknown SolidDB table %s\n", name));
            rs_entname_done_buf(&en);
            SDB_DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        if (operation_requires_xtable_lock(thd) &&
            #if defined(USE_DDL_TABLE_LOCKING)
                (errh = ::lock_table(con, &en, TRUE, DDL_LOCK_TIMEOUT))
            #else
                false
            #endif
            ) {
            succp = FALSE;
            goto epilogue;
        }

        if (rs_relh_isautoincinited(con->sc_cd, this->solid_rsrelh)) {
            succp = drop_table_autoincrements(this->solid_rsrelh, con);

            if (!succp) {
                sql_print_error("Can't drop table auto incremented fields.");
                rc = HA_ERR_NO_SUCH_TABLE;
                goto epilogue;
            }
        }

        /* In drop database and alter table allow dropping tables with foreign key definitions */
        if (thd->lex->sql_command == SQLCOM_DROP_DB ||
            thd->lex->sql_command == SQLCOM_ALTER_TABLE) {
            cascade      = TRUE;
            checkforkeys = FALSE;
        } else {
            /* In drop table allow dropping table with foreign key only
               if user has provided CASCADE options */
            cascade = (thd->lex->drop_mode == DROP_CASCADE);
            checkforkeys = TRUE;
        }

        succp = tb_admi_droprelation(con->sc_cd, con->sc_trans,
                                     en.en_name, en.en_schema, 
                                     NULL, NULL, cascade,
                                     checkforkeys, &errh);
        
        rs_relh_done(con->sc_cd, this->solid_rsrelh);
        this->solid_rsrelh = NULL;

epilogue:
        if (!succp) {
            rc = handle_error(errh);
        }

        /* Transaction is committed or rolled back depending on success predicate */
        commit_succ_p = internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, succp, &errh);

        if (!commit_succ_p) {
            rc = handle_error(errh);
        }

        rs_entname_done_buf(&en);
        finish_solid_DDL();
        SDB_DBUG_RETURN(rc);
}

/*
not implemented yet, possibly will be implemented.
int ha_soliddb::lock_table()
{
        SDB_DBUG_ENTER("ha_solid::lock_table");
        int rc = 0;
        su_err_t* errh = NULL;

        ss_dassert(this->solid_conn != NULL);
        ss_dassert(this->solid_rsrelh != NULL);
        //ss_dassert(this->solid_conn->sc_trans);
        // ss_assert( should be in an active transaction already! );

        errh = lock_table(this->solid_conn, this->solid_rsrelh);

        if (errh) {
            sql_print_error("%s", su_err_geterrstr(errh));
            rc = ha_solid_mysql_error(current_thd, errh);
            su_err_done(errh), errh = NULL;
        }

        SDB_DBUG_RETURN(rc);
}
*/

/*#***********************************************************************\
 *
 *              ::rename_table
 *
 * Renames a table from one name to another from alter table call.
 *
 * If you do not implement this, the default rename_table() is called from
 * handler.cc and it will delete all files with the file extentions returned
 * by bas_ext().
 *
 * Parameters :
 *
 *     const char*  from, in, old MySQL table name
 *     const char*  to, in, use, new MySQL table name
 *
 * Return value : 0 or return code
 *
 * Globals used : THD* current_the
 */
int ha_soliddb::rename_table(const char* from, const char* to)
{
        bool succp = true;
        int rc = 0;
        su_err_t* errh = NULL;
        SOLID_CONN* con = NULL;
        tb_relh_t* tbrelh = NULL;
        THD* thd = current_thd;
        rs_entname_t from_en;
        rs_entname_t to_en;
        bool commit_succ_p = true;
        
        SDB_DBUG_ENTER("ha_solid::rename_table");

        if( is_DDL_command( thd ) && !start_solid_DDL(rc) ) {
            SDB_DBUG_RETURN(rc);
        }

        solid_relname(&from_en, from);
        solid_relname(&to_en, to);

        ss_pprintf_1(("ha_solid::rename_table:'%s.%s'(%s)->'%s.%s'(%s)\n",
            from_en.en_schema, from_en.en_name, from, to_en.en_schema, to_en.en_name, to));

        if (is_soliddb_system_relation(from_en)) {
            ss_pprintf_1(("Trying to rename SolidDB system table %s\n", from));
            rs_entname_done_buf(&from_en);
            rs_entname_done_buf(&to_en);
            SDB_DBUG_RETURN(HA_ERR_NO_SUCH_TABLE);
        }

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_connection(ht, thd, from);
#else
        con = get_solid_connection(thd, from);
#endif

        ss_dassert(con != NULL);

        tb_trans_begintransandstmt(con->sc_cd, con->sc_trans);

        if (operation_requires_xtable_lock(thd) &&
            #if defined(USE_DDL_TABLE_LOCKING)
                (errh = ::lock_table(con, &from_en, TRUE, DDL_LOCK_TIMEOUT))
            #else
                false
            #endif
            ) {
            rs_entname_done_buf(&from_en);
            rs_entname_done_buf(&to_en);
            internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, FALSE, NULL);
            rc = handle_error(errh);
            SDB_DBUG_RETURN(rc);
        }

        tbrelh = tb_relh_create(con->sc_cd, con->sc_trans,
                                from_en.en_name, from_en.en_schema,
                                /* catalog=*/ NULL, &errh);

        succp = (tbrelh != NULL);

        if (!succp) {
            rc = handle_error(errh);
            errh = NULL;
        }

        /* Creating new schema if it does not exist. */
        if (succp && !tb_schema_find(con->sc_cd, con->sc_trans, to_en.en_schema, NULL)) {
            void* cont = NULL;

            ss_pprintf_2(("ha_solid::schema %s not found, creating it.\n", to_en.en_schema));

            succp = tb_schema_create( con->sc_cd, con->sc_trans,
                                      to_en.en_schema, NULL, NULL, &cont, &errh);
            ss_dassert(cont == NULL);
            
            if (!succp) {
                rc = handle_error(errh);
                errh = NULL;
            }
        }

        /* Renaming the table */
        if (succp) {
            succp = tb_dd_renametable_ex(con->sc_cd, con->sc_trans, tbrelh,
                                         to_en.en_name, to_en.en_schema, &errh);

            if (!succp) {
                rc = handle_error(errh);
                errh = NULL;
            }
        }

        if (tbrelh) {
            tb_relh_free(con->sc_cd, tbrelh);
            tbrelh = NULL;
        }

        commit_succ_p = internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, succp, &errh);

        if (succp && !commit_succ_p) {
            rc = handle_error(errh);
        }

        rs_entname_done_buf(&from_en);
        rs_entname_done_buf(&to_en);
        finish_solid_DDL();
        SDB_DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              solid_drop_database
 *
 * Drop database for solidDB
 *
 * Parameters :
 *
 *      char* path, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
void solid_drop_database(
#if MYSQL_VERSION_ID >= 50100
        handlerton* hton,
#endif
        char* path)
{
        SOLID_CONN* con;
        THD* thd;
        char* schema = NULL;
        char* tmp = NULL;
        void* cont = NULL;
        size_t len = 0;

        DBUG_ENTER("solid_drop_database");
        ss_pprintf_1(("solid_drop_database:path = %s\n", path));

        while (*path == '.' || *path == '/' || *path == '\\') {
            path++;
        }

        tmp = (char*) SsMemStrdup(path);
        
        /* SsStrupr(tmp); */
        len = strlen(tmp);

        /* Remove trailing directory character */
        if (len > 1 && (tmp[len-1] == '.' || tmp[len-1] == '/' || tmp[len-1] == '\\')) {
                tmp[len-1]='\0';
        }

        schema = (char *) SsMemStrdup(tmp);
        SsMemFree(tmp);

        ss_pprintf_2(("solid_drop_database:schema = %s\n", schema));

        thd = current_thd;
        ss_dassert(thd != NULL);

#if MYSQL_VERSION_ID >= 50100
        con = (SOLID_CONN*)thd->ha_data[hton->slot];
#else
        con = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        if (con) {
            bool succp = TRUE;
            rs_auth_t* auth = rs_sysi_auth(con->sc_cd);
            su_err_t*  errh = NULL;

            tb_trans_begintransandstmt(con->sc_cd, con->sc_trans);

            rs_auth_setsystempriv(con->sc_cd, auth, TRUE);

            succp = tb_schema_drop(
                        con->sc_cd,
                        con->sc_trans,
                        schema,
                        NULL,       // catalog
                        TRUE,       // cascade
                        &cont,
                        &errh);

            if (errh != NULL) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
                su_err_done(errh);
                errh = NULL;
            }

            rs_auth_setsystempriv(con->sc_cd, auth, FALSE);

            succp = internal_solid_end_trans_and_stmt(con->sc_cd, con->sc_trans, TRUE, NULL);
        }

        SsMemFree(schema);

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              check_solid_sysrel_create
 *
 * TODO: provide description to this function
 *
 * Parameters :
 *
 * Return value :
 *
 * Globals used :
 */
static bool check_solid_sysrel_create(
        rs_sysi_t* cd,
        tb_trans_t* trans,
        char* relname,
        rs_ttype_t* ttype)
{
        int i;
        bool foundp;
        rs_relh_t* sysrelh;
        rs_ttype_t* systtype;
        uint nattrs;
        rs_entname_t en;

        i = check_solid_sysrel(relname);
        if (i == -1) {
            return(FALSE);
        }

        rs_entname_initbuf(
            &en,
            NULL,       // catalog,
            (char *)RS_AVAL_SYSNAME,
            solid_sysrel_map[i].srm_solidname);

        sysrelh = tb_dd_getrelh(
                    cd,
                    trans,
                    &en,
                    NULL,
                    NULL);

        if (sysrelh == NULL) {
            ss_derror;
            return(FALSE);
        }

        /* Check that columns definitions are same.
         */
        foundp = TRUE;
        nattrs = rs_ttype_sql_nattrs(cd, ttype);
        systtype = rs_relh_ttype(cd, sysrelh);

        if (nattrs == rs_ttype_sql_nattrs(cd, systtype)) {
            uint j;

            for (j = 0; j < nattrs; j++) {
                rs_atype_t* atype;
                rs_atype_t* sysatype;

                atype = rs_ttype_sql_atype(cd, ttype, j);
                sysatype = rs_ttype_sql_atype(cd, systtype, j);

                if (!rs_atype_issame(cd, atype, sysatype)) {
                    foundp = FALSE;
                    break;
                }
            }
        } else {
            foundp = FALSE;
        }

        rs_relh_done(cd, sysrelh);

        return(foundp);
}


/*#***********************************************************************\
 *
 *              ha_map_mysql_solid_fkaction
 *
 * Maps a MySQL foreign key referential action code to solidDB foreign key
 * referential action.
 *
 * Parameters :
 *
 *     foreign_key* foreign_key, in, use, MySQL foreign key definition
 *
 * Return value : sqlrefact_t
 *
 * Globals used :
 */
static inline sqlrefact_t ha_map_mysql_solid_fkaction(
        uint fk_ref_action)
{
        sqlrefact_t solid_ref_action;

        switch (fk_ref_action) {
            case foreign_key::FK_OPTION_RESTRICT:
                solid_ref_action = SQL_REFACT_RESTRICT;
                ss_pprintf_1((" RESTRICT\n"));
                break;
            case foreign_key::FK_OPTION_CASCADE:
                solid_ref_action = SQL_REFACT_CASCADE;
                ss_pprintf_1((" CASCADE\n"));
                break;
            case foreign_key::FK_OPTION_SET_NULL:
                solid_ref_action = SQL_REFACT_SETNULL;
                ss_pprintf_1((" SET NULL\n"));
                break;
            case foreign_key::FK_OPTION_NO_ACTION:
                solid_ref_action = SQL_REFACT_NOACTION;
                ss_pprintf_1((" NO ACTION\n"));
                break;
            case foreign_key::FK_OPTION_DEFAULT:
                solid_ref_action = SQL_REFACT_SETDEFAULT;
                ss_pprintf_1((" SET DEFAULT\n"));
                break;
            default:
                /* If no foreign key action have been defined
                   use RESTRICT action */
                solid_ref_action = SQL_REFACT_RESTRICT;
                ss_pprintf_1((" RESTRICT\n"));
                break;
        }

        return (solid_ref_action);
}

/*#***********************************************************************\
 *
 *              solid_find_key
 *
 * Find a key_info for this foreign key definition.
 *
 * Parameters :
 *

 *
 * Return value :  MySQL key information
 *
 * Globals used :
 */
static st_key *solid_find_key(
        Key*    key,
        uint    n_keys,
        st_key* key_info)
{
        uint n_key;
        st_key* keyi = NULL;

        for(n_key = 0; n_key < n_keys; n_key++) {
            List_iterator<key_part_spec> col_iterator(key->columns);
            key_part_spec* key_part;

            keyi = &(key_info[n_key]);

            ss_pprintf_1(("Key name %s\n", keyi->name));
            ss_pprintf_1(("Key elements %ld == %ld\n", keyi->key_parts, key->columns.elements));

            if (keyi->key_parts == key->columns.elements) {
                uint n_col;

                n_col = 0;
                while((key_part = col_iterator++)) {
                    ss_pprintf_1(("Key col name '%s' == key_part->field_name '%s'\n", key_part->field_name,
                              keyi->key_part[n_col].field->field_name));

                    if ((strcmp(key_part->field_name, keyi->key_part[n_col].field->field_name))) {
                        break;
                    }

                    n_col++;
                }

                if (key_part == NULL) {
                    break;
                }
            }
        }

        if (n_key == n_keys) {
            keyi = NULL;
        }

        return (keyi);
}

/*#***********************************************************************\
 *
 *              create_foreign_key_definitions
 *
 * Create solid foreign key definitions based on MySQL lex structure.
 *
 * Parameters :
 *
 *    THD*            thd, in, use, MySQL thread data
 *    rs_sysi_t*      cd, in, use, system info client data
 *    rs_relh_t*      relh, in, use, relation
 *    rs_ttype_t*     ttype, in, use, table type information
 *    uint            n_keys, in, use, number of keys on MySQL key_info
 *    st_key*         key_info, in, use, MySQL key definitions
 *    const char*     name, in, use, catalog, schema name
 *    uint*           n_forkey, out, number of foreign keys
 *    tb_sqlforkey_t**   forkeys, out, foreign key definitions
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
static int create_foreign_key_definitions(
        THD*   thd,
        rs_sysi_t*    cd,
        rs_ttype_t*   ttype,
        uint          n_keys,
        st_key*       key_info,
        const char*   catalog,
        bool          lower_case_table_names,
        uint*         n_forkey,
        tb_sqlforkey_t** forkeys)
{
        Key          *key;
        uint         len, i;
        uint         n_fkeys;
        uint         keys;
        tb_sqlforkey_t* foreign_keys = NULL;
        List_iterator<Key> key_iterator(thd->lex->key_list);
        int          error = 0;

        SS_PUSHNAME("create_foreign_key_definitions");

        n_fkeys = 0;
        keys = thd->lex->key_list.elements;
        i = 0;

        while(i < keys) {
            key = key_iterator++;

            if (key && key->type == Key::FOREIGN_KEY) {

                n_fkeys++;
            }

            i++;
        }

        if (n_fkeys == 0) {
            *n_forkey = 0;
            SS_POPNAME;
            return (0);
        }

        foreign_keys = (tb_sqlforkey_t *) SsMemAlloc(sizeof(tb_sqlforkey_t) * n_fkeys);
        *n_forkey = n_fkeys;

        ss_pprintf_1(("Number of foreign key definitions %lu\n", n_fkeys));

        n_fkeys = 0;

        key_iterator.rewind();

        while(n_fkeys < *n_forkey) {
            key = key_iterator++;

            if (key && key->type == Key::FOREIGN_KEY) {
                char *fkname = NULL;
                char *relname = NULL;
                foreign_key *fk_key = (foreign_key*)key;
                key_part_spec *key_part;
                List_iterator<key_part_spec> col_iterator(fk_key->columns);
                List_iterator<key_part_spec> ref_iterator(fk_key->ref_columns);

                ss_pprintf_1(("FOREIGN KEY DEFINITION\n"));
                if (fk_key->name) {
                    ss_pprintf_1(("  Name %s\n", fk_key->name));
                } else {
                    ss_pprintf_1(("  Name NULL\n", fk_key->name));
                }

                ss_pprintf_1(("  Referenced db %s table %s\n",
                        fk_key->ref_table->db.str, fk_key->ref_table->table.str));

                if (fk_key->ref_table->db.str == NULL) {
                    foreign_keys[n_fkeys].refschema = (char*) SsMemStrdup((char *)catalog);
                } else {
                    foreign_keys[n_fkeys].refschema = (char*) SsMemStrdup(fk_key->ref_table->db.str);
                }

                /* Set up referenced table and schema */
                relname = (char *)SsMemStrdup((char *)fk_key->ref_table->table.str);

                if (lower_case_table_names) {
                    SsStrupr(relname);
                    SsStrupr(foreign_keys[n_fkeys].refschema);
                }

                foreign_keys[n_fkeys].reftable = relname;

                foreign_keys[n_fkeys].refcatalog = NULL;

                /* Set up foreign key name */

                if (fk_key->name) {
                    fkname = (char *)SsMemStrdup((char *)fk_key->name);

                    if (lower_case_table_names) {
                        SsStrupr(fkname);
                    }

                    /* keyname_from_mysql_to_solid(cd, relh, (char *)fk_key->name,
                       fkname, lower_case_table_names); */
                } else {
                    fkname = NULL;
                }

                if (fkname) {
                    ss_pprintf_1(("Foreign key constraint name %s\n", fkname));
                }

                foreign_keys[n_fkeys].name = fkname;

                /* Set up number of fields in the foreign key */
                len = fk_key->columns.elements;
                foreign_keys[n_fkeys].len = len;
                ss_pprintf_1(("  # referencing columns %lu\n", len));

                /* Set up array of indexes to fields in the foreign key */
                foreign_keys[n_fkeys].fields = (uint *)SsMemAlloc(sizeof(uint) * len);
                ss_pprintf_1(("  Referencing columns\n"));
                i = 0;

                while ((key_part = col_iterator++)) {
                    rs_ano_t col_ano;

                    if (lower_case_table_names) {
                        col_ano = rs_ttype_sql_anobyname(cd, ttype, (char *)SsStrupr((char *)key_part->field_name));
                    } else {
                        col_ano = rs_ttype_sql_anobyname(cd, ttype, (char *)key_part->field_name);
                    }

                    ss_dassert(col_ano != RS_ANO_NULL);

                    foreign_keys[n_fkeys].fields[i] = col_ano;

                    ss_pprintf_1(("    Field %d name %s length %d ano %ld\n",
                                  i, key_part->field_name, key_part->length, col_ano));
                    i++;
                }
                ss_rc_dassert(i == len, i);

                /* Set up fererential actions */
                ss_pprintf_1(("    ON DELETE"));
                foreign_keys[n_fkeys].delrefact = ha_map_mysql_solid_fkaction(fk_key->delete_opt);
                ss_pprintf_1(("    ON UPDATE"));
                foreign_keys[n_fkeys].updrefact = ha_map_mysql_solid_fkaction(fk_key->update_opt);

                /* This version of MySQL/solidDB does support only RESTRICT
                   foreign keys */
                if (foreign_keys[n_fkeys].delrefact != SQL_REFACT_RESTRICT ||
                    foreign_keys[n_fkeys].updrefact != SQL_REFACT_RESTRICT) {
                    error = HA_ERR_WRONG_COMMAND;
                    ss_pprintf(("Unsupported referential action in this version\n"));
                }

                len = fk_key->ref_columns.elements;
                ss_pprintf_1(("  References %lu columns\n", len));
                ss_pprintf_1(("  Referenced columns in table %s\n", foreign_keys[n_fkeys].reftable));

                if (len > 0) {

                    i = 0;

                    /* Set up referenced columns */
                    foreign_keys[n_fkeys].reffields = (char **)SsMemAlloc(sizeof(char *) * len);

                    while ((key_part = ref_iterator++)) {

                        foreign_keys[n_fkeys].reffields[i] = (char *)SsMemStrdup((char *)key_part->field_name);

                        if (lower_case_table_names) {
                            SsStrupr(foreign_keys[n_fkeys].reffields[i]);
                        }

                        ss_pprintf_1(("    Field %d name %s length %d\n",
                                    i, foreign_keys[n_fkeys].reffields[i], key_part->length));

                        i++;
                    }
                    ss_rc_dassert(i == len, i);

                } else {
                    foreign_keys[n_fkeys].reffields = NULL;
                }

                n_fkeys++;
            }
        }

        *n_forkey = n_fkeys;
        *forkeys  = foreign_keys;

        SS_POPNAME;

        return (error);
}

/*#***********************************************************************\
 *
 *              copy_foreign_key_definitions
 *
 * Copy solid foreign key definitions from another table.
 *
 * Parameters :
 *
 *    THD*            thd, in, use, MySQL thread data
 *    rs_sysi_t*      cd, in, use, system info client data
 *    schema          table schema
 *    table_name      table name
 *    uint*           n_forkey, out, number of foreign keys
 *    tb_sqlforkey_t**   forkeys, out, foreign key definitions
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
static int copy_foreign_key_definitions(
        THD* thd,
        rs_sysi_t* cd,
        tb_trans_t* trans,
        rs_ttype_t* new_ttype,
        const char* old_schema,
        const char* old_table_name,
        uint* n_forkey,
        tb_sqlforkey_t** forkeys)
{
        uint n_fkeys;
        tb_sqlforkey_t* foreign_keys = NULL;
        int error = 0;
        rs_relh_t* old_relh;
        rs_entname_t en;
        su_pa_t* refkeys;
        int i;
        int k;
        uint fki;

        ss_dprintf_3(("copy_foreign_key_definitions\n"));
        SS_PUSHNAME("copy_foreign_key_definitions");

        rs_entname_initbuf(
            &en,
            NULL,       // catalog,
            old_schema,
            old_table_name);

        old_relh = tb_dd_getrelh(
                    cd,
                    trans,
                    &en,
                    NULL,       // p_priv
                    NULL);      // rs_err_t** p_errh

        if (old_relh == NULL) {
            SS_POPNAME;
            /* This is ok when you alter e.g. myisam table to soliddb */
            return(0);
        }

        n_fkeys = 0;
        refkeys = rs_relh_refkeys(cd, old_relh);

        su_pa_do(refkeys, i) {
            rs_key_t* key;

            key = (rs_key_t*)su_pa_getdata(refkeys, i);

            if (rs_key_type(cd, key) == RS_KEY_FORKEYCHK) {
                rs_ano_t j;
                rs_ano_t k;
                rs_ano_t n_parts;
                rs_relh_t *refrelh;
                rs_ano_t k_part;
                rs_err_t* errh = NULL;
                bool first_part = TRUE;
                su_pa_t* forrefkeys = NULL;
                rs_key_t* forkey = NULL;
                tb_sqlforkey_t* newforkey;
                rs_ttype_t* old_ttype;
                rs_ttype_t* ref_ttype;

                n_fkeys++;

                old_ttype = rs_relh_ttype(cd, old_relh);

                if (foreign_keys == NULL) {
                    foreign_keys = (tb_sqlforkey_t*)SsMemAlloc(sizeof(tb_sqlforkey_t));
                } else {
                    foreign_keys = (tb_sqlforkey_t*)SsMemRealloc(foreign_keys, sizeof(tb_sqlforkey_t) * n_fkeys);
                }
                newforkey = &foreign_keys[n_fkeys - 1];

                /* Clear content so that we release only allocated fields if need to return
                 * in the middle. */
                memset(newforkey, '\0', sizeof(tb_sqlforkey_t));

                n_parts = rs_key_nparts(cd, key);

                newforkey->name = (char*)SsMemAlloc(strlen(old_table_name) + 1 + 20 + 1 + 6 + 1 + 20 + 1);

                SsSprintf(
                    newforkey->name, 
                    "%.255s_%d_FORKEY_%d", 
                    old_table_name, 
                    dbe_db_getnewrelid_log((dbe_db_t*)rs_sysi_db(cd)),
                    n_fkeys - 1);

                newforkey->fields = (uint*)SsMemAlloc(n_parts * sizeof(newforkey->fields));

                ss_dprintf_4(("copy_foreign_key_definitions:name=%s\n", newforkey->name));

                for (k_part = 0, fki = 0; k_part < n_parts; k_part++) {
                    if (rs_keyp_parttype(cd, key, k_part) == RSAT_USER_DEFINED) {
                        char* aname;
                        rs_ano_t kp_ano;
                        rs_ano_t new_ano;

                        kp_ano = rs_keyp_ano(cd, key, k_part);
                        aname = rs_ttype_aname(cd, old_ttype, kp_ano);
                        ss_dassert(aname != NULL);
                        new_ano = rs_ttype_anobyname(cd, new_ttype, aname);
                        ss_dassert((int)new_ano >= 0);

                        ss_dprintf_4(("copy_foreign_key_definitions:add field %d, index %d\n", new_ano, fki));

                        newforkey->fields[fki++] = new_ano;
                    }
                }
                newforkey->len = fki;

                refrelh = solid_getrelhbyid(
                            cd, 
                            trans,
                            rs_key_refrelid(cd, key), 
                            &errh);

                if (errh) {
                    /* Someone could have dropped the referenced table. */
                    ss_dassert(refrelh == NULL);
                    sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
                    su_err_done(errh);
                    errh = NULL;
                    error = -1;
                    break;
                }
                ss_dassert(refrelh != NULL);

                newforkey->reftable = (char*)SsMemStrdup(rs_relh_name(cd, refrelh));
                newforkey->refschema =(char*) SsMemStrdup(rs_relh_schema(cd, refrelh));
                newforkey->refcatalog = (char*)SsMemStrdup(rs_relh_catalog(cd, refrelh));
                newforkey->reffields = (char**)NULL;

                tb_dd_resolverefkeys(cd, refrelh);
                forrefkeys = rs_relh_refkeys(cd, refrelh);
                ref_ttype = rs_relh_ttype(cd, refrelh);
                k = 0;

                su_pa_do(forrefkeys, k) {
                    forkey = (rs_key_t*)su_pa_getdata(forrefkeys, k);
                    n_parts = rs_key_nparts(cd, forkey);

                    if (rs_key_type(cd, forkey) != RS_KEY_FORKEYCHK 
                        && rs_key_refrelid(cd, forkey) == rs_relh_relid(cd, old_relh))
                    {
                        ss_dassert(newforkey->reffields == NULL);

                        newforkey->reffields = (char**)SsMemAlloc(n_parts * sizeof(newforkey->reffields));

                        for (j = 0, fki = 0; j < n_parts; j++) {
                            rs_ano_t key_ano;

                            key_ano = rs_keyp_ano(cd, forkey, j);

                            if (rs_keyp_parttype(cd, forkey, j) == RSAT_USER_DEFINED) {
                                char* aname;

                                aname = rs_ttype_aname(cd, ref_ttype, key_ano);
                                ss_dprintf_4(("copy_foreign_key_definitions:add reffield %s, index %d\n", aname, fki));
                                newforkey->reffields[fki++] = (char*)SsMemStrdup(aname);
                            }
                        }
                        ss_dassert(newforkey->len == fki);
                    }
                }
                newforkey->delrefact = (sqlrefact_t)rs_key_delete_action(cd, key);
                newforkey->updrefact = (sqlrefact_t)rs_key_update_action(cd, key);

                rs_relh_done(cd, refrelh);
            }
        }

        rs_relh_done(cd, old_relh);

        *n_forkey = n_fkeys;
        *forkeys  = foreign_keys;

        SS_POPNAME;

        return (error);
}

/*#***********************************************************************\
 *
 *              solid_parse_relmode
 *
 * Parse comment field from MySQL create table and set relation mode
 * based on that.
 *
 * Parameters :
 *
 *     LEX_STRING* comment, in, use, comment string
 *
 * Return value : One of the below
 *
 * TB_RELMODE_SYSDEFAULT    if parse error or none given
 * TB_RELMODE_OPTIMISTIC    if MODE=OPTIMISTIC
 * TB_RELMODE_PESSIMISTIC   if MODE=PESSIMISTIC
 * TB_RELMODE_MAINMEMORY    if MODE=MAINMEMORY
 *
 * Limitations:
 *
 * Globals used :
 */
static tb_relmode_t solid_parse_relmode(
        LEX_STRING* comment_string)
{
        tb_relmode_t solid_relmode = TB_RELMODE_SYSDEFAULT;
        char* comment = NULL;
        const char* type_beg;
        size_t idst=0;

        if (comment_string == NULL ||
            comment_string->str == NULL ||
            comment_string->length == 0) {
            
            return TB_RELMODE_SYSDEFAULT;
        }

        ss_pprintf(("Create table comment = %s\n", comment_string));

        /* size of comment_string is unknown a priori - we have to allocate dyn. buffer */
        comment = (char*)SsMemAlloc(comment_string->length+1);
        ss_dassert(comment != NULL);

        /* copying the comment strind, converting it to upper case and
         * removing whitespaces.
         */
        
        for( size_t i = 0; i < comment_string->length; ++i ) {
            if ( !strchr( SQL_WHITESPACES, comment_string->str[i] ) ) {
              comment[idst++] = toupper( comment_string->str[i] );
            }
        }

        if (comment) {
            comment[idst]=0;

            if ((type_beg = strstr(comment, "MODE="))) {
                type_beg += sizeof("MODE=")-1; /* -1 - without '\0' */
                if(strncmp( type_beg, "PESSIMISTIC", sizeof("PESSIMISTIC")-1) == 0) {
                    solid_relmode = TB_RELMODE_PESSIMISTIC;
                } else if (strncmp( type_beg, "OPTIMISTIC", sizeof("OPTIMISTIC")-1 ) == 0) {
                    solid_relmode = TB_RELMODE_OPTIMISTIC;
                } else if (strncmp( type_beg, "MAINMEMORY", sizeof("MAINMEMORY")-1 ) == 0) {
                    solid_relmode = TB_RELMODE_MAINMEMORY;
                }
            }
        }
        
        if (comment) {
            SsMemFree(comment);
        }

        return solid_relmode;
}

/*#***********************************************************************\
 *
 *              solid_check_keys
 *
 * Check key information
 *
 * Parameters :
 *
 *     TABLE *table, in, use, information on table columns and indexes
 *     const char* table_name, in, use, table_name
 *
 * Return value : One of the below
 *
 *     TRUE   success
 *     FALSE  failure
 *
 * Limitations:
 *
 * Globals used :
 */
static bool solid_check_keys(
        TABLE *table,
        THD *thd,
        const char *table_name)
{
        uint		n_fields;
        Field*		field = NULL;
        uint        n_keys = table->s->keys;
        uint        n_key;
        uint        n_field;
        bool        blob = FALSE;

        n_key = 0;

        while(n_key < n_keys) {
            KEY*		key = &(table->key_info[n_key]);
            n_fields = key->key_parts;

            for (n_field = 0; n_field < n_fields; n_field++) {
                KEY_PART_INFO*	key_part;
                uint j;

                key_part = key->key_part + n_field;

                for (j = 0; j < table->s->fields; j++) {

                    field = table->field[j];

                    if (0 == my_strcasecmp(system_charset_info, field->field_name,
                                           key_part->field->field_name)) {
                        /* Found the corresponding column */

                        break;
                    }
                }

                if (field->has_charset()) {
                    CHARSET_INFO *charset = field->charset();

                    ss_dassert(charset != NULL);

                    /* Supported character sets and collations:
                       latin1, latin1_general_cs
                       ucs2, ucs2_bin */
                    if (charset->number != 49 && charset->number != 90) {
                        sql_print_error("This version of MySQL/solidDB does not support character "
                                        "set %s or collation %s used in table %s field %s",
                                        charset->csname,
                                        charset->name,
                                        table_name, field->field_name);
                        
                        return (FALSE);
                    }
                }

                ss_dassert(j < table->s->fields);

                switch(field->type()) {
                    case FIELD_TYPE_GEOMETRY:
                    case FIELD_TYPE_TINY_BLOB:
                    case FIELD_TYPE_MEDIUM_BLOB:
                    case FIELD_TYPE_BLOB:
                    case FIELD_TYPE_LONG_BLOB:
                        blob = TRUE;
                        break;
                    default:
                        blob = FALSE;
                        break;
                }

                if (blob
                    || (key_part->length < field->pack_length()
                        && field->type() != MYSQL_TYPE_VARCHAR)
                    || (field->type() == MYSQL_TYPE_VARCHAR
                        && key_part->length < field->pack_length()
                        -((Field_varstring*)field)->length_bytes)) {

                    switch(field->type()) {
                        case FIELD_TYPE_LONG:
                        case FIELD_TYPE_LONGLONG:
                        case FIELD_TYPE_TINY:
                        case FIELD_TYPE_SHORT:
                        case FIELD_TYPE_INT24:
                        case FIELD_TYPE_DATE:
                        case FIELD_TYPE_DATETIME:
                        case FIELD_TYPE_YEAR:
                        case FIELD_TYPE_NEWDATE:
                        case FIELD_TYPE_TIME:
                        case FIELD_TYPE_TIMESTAMP:
                        case FIELD_TYPE_FLOAT:
                        case FIELD_TYPE_DOUBLE:
                        case FIELD_TYPE_DECIMAL:
                            sql_print_error("Invalid column type for prefix index field %s table %s",
                                            table_name, field->field_name);
                            return (FALSE);
                        default:
                            if (blob && key_part->length == 255) {
                                /* For blob types, accept prefix index of 255 bytes which
                                 * if default prefix for solidDB.
                                 */
                                break;
                            } else {
                                sql_print_error("This version of MySQL/solidDB does not support prefix indexes: "
                                                "field %s table %s", table_name, field->field_name);
                                return (FALSE);
                            }
                    }
                }
            }
            n_key++;
        }

        return (TRUE);
}

/*#***********************************************************************\
*
*              solid_check_charset
*
* Check that used charset and collation for the table are supported
* in solidDB for MySQL
*
* Parameters :
*
* HA_CREATE_INFO*  create_info,in, use, table create information
* const char*      create_sql, in, use, create table SQL-query
* const char*      table_name, in, use, table name
*
* Return value : TRUE if charset and collation are supported
*                FALSE otherwise
*
* Globals used : -
*/
static inline bool solid_check_charset(
        HA_CREATE_INFO* create_info,
        const char*     create_sql,
        const char*     table_name)
{
        ss_dassert(create_info != NULL);
        ss_dassert(create_sql != NULL);

        /* If no table option COLLATE were given and default collation is not
           supported we change the collation to supported collation and
           give a warning unknown collation to user */

        if (create_info->used_fields | HA_CREATE_USED_DEFAULT_CHARSET) {
            if (create_info->default_table_charset &&
                create_info->default_table_charset->number != 49 &&
                create_info->default_table_charset->number != 90) {

                push_warning_printf(current_thd, MYSQL_ERROR::WARN_LEVEL_WARN,
                                    ER_UNKNOWN_COLLATION, ER(ER_UNKNOWN_COLLATION),
                                    create_info->default_table_charset->name);
            
                if (!(strcmp(create_info->default_table_charset->csname, "latin1"))) {
                    create_info->default_table_charset->number = 49; /* latin1_general_cs ! */
                } else if (!(strcmp(create_info->default_table_charset->csname, "ucs2"))) {
                    create_info->default_table_charset->number = 90; /* ucs2_bin */
                }
            }
        }
        
        /* Supported character sets and collations:
           latin1, latin1_general_cs and ucs2, ucs2_bin */
        
        if (create_info->table_charset &&
            (create_info->table_charset->number != 49 &&
             create_info->table_charset->number != 90)) {
            
            sql_print_error("This version of MySQL/solidDB does not support character "
                            "set %s or collation %s used in table %s",
                            create_info->table_charset->csname,
                            create_info->table_charset->name,
                            table_name);

            return (FALSE);
        } else {

            if (create_info->default_table_charset &&
                (create_info->default_table_charset->number != 49 &&
                 create_info->default_table_charset->number != 90)) {

                sql_print_error("This version of MySQL/solidDB does not support character "
                            "set %s or collation %s used in table %s",
                            create_info->default_table_charset->csname,
                            create_info->default_table_charset->name,
                            table_name);
                return (FALSE);
            }
        }

        return (TRUE);
}

/*#***********************************************************************\
*
*              ::create
*
* create() is called to create or alter a table. 
* ALTER TABLE is handled as following:
*   - a new table is created with the structure of final table, 
*     with a temporary name.
*   - data is copied from old table to new table, row by row.
*   - old table is renamed to a temporary name.
*   - new table is renamed to old table name.
*   - old table is dropped.
*
* All table modifications are handled really via ALTER TABLE
* even if the statement is not really ALTER (f.e. CREATE INDEX is
* handled by ALTER). Only trivial operations like renaming a table 
* are handled specially without copying data. In 5.1. an interface
* to handle creation or dropping of indices is also added so in 5.1
* indices are handled different also.
*
* FRM file has already been created so adjusting
* create_info will not do you any good. You can overwrite the .frm file 
* at this point if you need to change the table definition.
*
* Called from handle.cc by ha_create_table().
*
* Parameters :
*
*    const char*     name, in, full MySQL table name of new table
*    TABLE*          table_arg, in, use
*    HA_CREATE_INFO* create_info, in, use
*
* Return value : 0 or error code
*
* Globals used :
*/
int ha_soliddb::create(
        const char *name,
        TABLE *table_arg,               // in sql/table.h, struct st_table {
        HA_CREATE_INFO *create_info)    // in sql/handler.h, typedef struct st_ha_create_information
{
        SOLID_CONN*   con;
        tb_connect_t* tbcon;
        rs_sysi_t*    cd;
        tb_trans_t*   trans;
        rs_relh_t*    relh = NULL;

        int         rc;
        uint        i;
        uint        n_cols;
        tb_sqlunique_t* primkey = NULL;
        rs_ttype_t* ttype;
        su_err_t*   errh = NULL;
        bool        succp=TRUE;
        char*       aname;
        char*       relname;
        /* parameters to solid *create function(s) */
        char*       authid = NULL;
        char*       catalog = NULL;
        char*       extrainfo = NULL;
        int         namelen;
        void*       cont = NULL;
        long        seq_id = 0;
        rs_atype_t* atype = NULL;
        bool        auto_inc_defined = FALSE;
        bool        sysrelcreate;
        solid_bool  finished = FALSE;
        uint        n_forkeys = 0;
        tb_sqlforkey_t* forkeys = NULL;
        su_err_t*   commit_errh = NULL;
        bool commit_succ_p = true;

        THD* thd;
        uint soliddb_slot = 
            #if MYSQL_VERSION_ID >= 50100
                ht->
            #else
                solid_hton.
            #endif
                           slot;

        rc = 0;
        thd = current_thd;

        SDB_DBUG_ENTER("ha_solid::create");
        SS_PUSHNAME("ha_soliddb::create");
        ss_pprintf_1(("ha_solid::create SQL: type %d, text [%.*s].\n",
                       thd->lex->sql_command,
                       thd->query_length, thd->query));

        if( is_DDL_command( thd ) && !start_solid_DDL(rc) ) {
            SS_POPNAME;
            SDB_DBUG_RETURN(rc);
        }

        if (is_soliddb_system_relation(name)) {
            ss_pprintf_1(("Trying to create SolidDB system table %s\n", name));
            SDB_DBUG_RETURN(HA_ERR_TABLE_EXIST);
        }

        con = get_solid_connection(name);

        ss_dassert(con != NULL);
        tbcon = con->sc_tbcon;
        cd    = con->sc_cd;
        trans = con->sc_trans;

        succp = solid_check_keys(table_arg, thd, name);

        if (succp) {
            succp = solid_check_charset(create_info, thd->query, name);
        }
        
        if (!succp) {
            finish_solid_DDL();
            SS_POPNAME;
            SDB_DBUG_RETURN(HA_ERR_WRONG_COMMAND);
        }

#if MYSQL_VERSION_ID >= 50100
        solid_trans_beginif(ht, thd, TRUE);
#else
        solid_trans_beginif(thd, TRUE);
#endif

        /* this branch of code works only for ALTER TABLE statement. */
        if (operation_requires_xtable_lock(thd)) {
            /* getting name of the table being altered */
            rs_entname_t old_name;
            solid_relname(&old_name, ((TABLE_LIST*)thd->lex->select_lex.table_list.first)->db,
                                     ((TABLE_LIST*)thd->lex->select_lex.table_list.first)->table_name);
            if (is_soliddb_table(old_name.en_schema, old_name.en_name) &&
                #if defined(USE_DDL_TABLE_LOCKING)
                    (errh = ::lock_table(con, &old_name, TRUE, DDL_LOCK_TIMEOUT))
                #else
                    false
                #endif
                ) {
                rs_entname_done_buf(&old_name);
                rc = handle_error(errh);
                internal_solid_end_trans_and_stmt(cd, trans, FALSE, NULL);
                finish_solid_DDL();
                SS_POPNAME;
                SDB_DBUG_RETURN(rc);
            }
            rs_entname_done_buf(&old_name);
        }

        n_cols = table_arg->s->fields;

        /*
         * Create table type information (ttype),
         */
        ss_pprintf_1(("ha_solid::create:create table type, slot %d\n", soliddb_slot));

        ttype = rs_ttype_create(cd);

        for (i = 0; i < n_cols; i++) {
    	    Field*		field;

            field = table_arg->field[i];

            atype = ha_solid_mysql_atype(cd, field);

            rs_atype_setnullallowed(cd, atype, field->null_ptr != NULL);

            /* If this field is auto increment field, create a new sequence for it. */

            if (rs_atype_autoinc(cd, atype)) {
                    char* seq_name;
                    rs_auth_t*  auth = rs_sysi_auth(cd);
                    long seq_no = dbe_db_getnewrelid_log((dbe_db_t*)rs_sysi_db(cd));

                    seq_name = (char *)SsMemAlloc(13+32);

                    sprintf(seq_name,"AUTOINC_%lu_SEQ", seq_no);

                    ss_pprintf_1(("SEQ NAME = %s\n", seq_name));

                    rs_auth_setsystempriv(cd, auth, TRUE);

                    succp = tb_createseqandid(
                                cd,
                                trans,
                                seq_name,
                                (char *)RS_AVAL_SYSNAME,
                                catalog,
                                extrainfo,
                                FALSE,     /* sparse */
                                &cont,
                                &errh,
                                &seq_id);

                    if (errh != NULL) {
                        sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
                        su_err_done(errh);
                        errh = NULL;
                    }

                    rs_atype_setautoinc(cd, atype, TRUE, seq_id);

                    ss_pprintf_1(("Field %s (AUTO_INCREMENT_SEQ_ID = %lu)\n", rs_atype_name(cd, atype),
                                  (long) rs_atype_getautoincseqid(cd, atype)));

                    rs_auth_setsystempriv(cd, auth, FALSE);

                    SsMemFree(seq_name);

                    auto_inc_defined = TRUE;
            }

            rs_ttype_setatype(cd, ttype, i, atype);
            rs_atype_free(cd, atype);

            aname = (char *)SsMemAlloc(strlen(field->field_name)+1);
            strcpy(aname, (char*)field->field_name);

            if (lower_case_table_names) {
                SsStrupr(aname);
            }

            rs_ttype_setaname(cd, ttype, i, aname);

            ss_pprintf_1(("ha_solid::create:create table type:%d, %s, null_ptr %x\n",
                          i, aname, (unsigned long)field->null_ptr));

            SsMemFree(aname);

        }

        namelen = strlen(name) + 1;
        authid = (char *)SsMemAlloc(namelen);
        relname = (char *)SsMemAlloc(namelen);
        solid_set_relname(authid, relname, (char*)name, lower_case_table_names);

        ss_pprintf_1(("ha_solid::create:tb_createrelation:%s->%s.%s\n", name, authid, relname));

        succp = tb_schema_find(cd, trans, authid, catalog);

        if (!succp) {
            ss_pprintf_2(("ha_solid::schema %s not found, create it\n", authid));

            succp = tb_schema_create(
                        cd,
                        trans,
                        authid,
                        catalog,
                        NULL,
                        &cont,
                        &errh);

            if (errh != NULL) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
            }

            ss_dassert(cont == NULL);
        }

        sysrelcreate = check_solid_sysrel_create(cd, trans, relname, ttype);

        if (succp && !sysrelcreate) {
            tb_relmode_t    relmode = TB_RELMODE_SYSDEFAULT;

            i = 0;
            primkey = NULL;

            /* Create Primary key */
            while (i < table_arg->s->keys) {

                if (is_mysql_primary_key(&table_arg->key_info[i])) {

                    primkey = ha_mysql_solid_create_unique_init(
                                &table_arg->key_info[i],
                                table_arg->s->keys);

                    if (primkey != NULL) {
                        i++;
                        break;
                    }
                }

                i++;
            }

            if (table_arg->s->keys) {
                char* old_schema;
                char* old_table_name;

                rc = 0;

                if ( thd->lex->query_tables == NULL ) {
                    old_schema = NULL;
                    old_table_name = NULL;
                } else {
                    old_schema = (char*) SsMemStrdup(thd->lex->query_tables->db);
                    old_table_name = (char*) SsMemStrdup(thd->lex->query_tables->table_name);
                    if (lower_case_table_names) {
                        SsStrupr(old_schema);
                        SsStrupr(old_table_name);
                    }
                }

                if (thd->lex->query_tables == NULL
                    || (strcmp(authid, old_schema) == 0
                        && strcmp(relname, old_table_name) == 0))
                {
                    /* Createing a new table.
                     */
                    rc = create_foreign_key_definitions(
                            thd,
                            cd,
                            ttype,
                            (uint)table_arg->s->keys,
                            table_arg->key_info,
                            (const char *)authid,
                            (bool)lower_case_table_names,
                            (uint *)&n_forkeys,
                            &forkeys);
                } else {
                    /* Altering an existing table.
                     */
                    ss_dassert(old_schema != NULL);
                    ss_dassert(old_table_name != NULL);
                    ss_dprintf_2(("create %s.%s, original %s.%s\n", authid, relname, thd->lex->query_tables->db, thd->lex->query_tables->table_name));
                    rc = copy_foreign_key_definitions(
                                thd,
                                cd,
                                trans,
                                ttype,
                                old_schema,
                                old_table_name,
                                (uint *)&n_forkeys,
                                &forkeys);
                }

                if (old_schema != NULL) {
                    SsMemFree(old_schema);
                    SsMemFree(old_table_name);
                }

                if (rc != 0) {
                    goto error_handling;
                }
            }

            if (create_info->comment.str) {
                relmode = solid_parse_relmode(&create_info->comment);
            }

            /* Create relation */
            succp = tb_createrelation_ext(
                        cd,
                        trans,
                        relname,
                        authid,         // char*        authid
                        catalog,        // char*        catalog
                        extrainfo,      // char*        extrainfo,
                        ttype,          // rs_ttype_t*  ttype,
                        primkey,        // tb_sqlunique_t* primkey,
                        0,              // uint         unique_c,
                        NULL,           // tb_sqlunique_t* unique,
                        n_forkeys,      // uint         forkey_c,
                        forkeys,        // tb_sqlforkey_t* forkeys,
                        NULL,           // uint*        def,
                        NULL,           // rs_tval_t*   defvalue,
                        0,              // uint         check_c,
                        NULL,           // char**       checks,
                        NULL,           // char**       checknames,
                        TB_DD_CREATEREL_USER,
                        TB_DD_PERSISTENCY_PERMANENT,
                        TB_DD_STORE_DEFAULT,
                        TB_DD_DURABILITY_DEFAULT,
                        relmode,        // optimistic || pessimistic || mainmemory || sysdefaul
                        &relh,
                        &errh);         // rs_err_t**   err

            if (errh != NULL) {
                /* TODO is this correct way of error handling ? */
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
            }

            ha_mysql_solid_create_unique_done(primkey);
            primkey = NULL;

            if (n_forkeys) {
                ha_mysql_solid_foreign_done(forkeys, n_forkeys);
                forkeys = NULL;
                n_forkeys = 0;
            }
            ss_dassert(forkeys == NULL);

            /* Create secondary keys */
            if (succp) {
                i = 0;

                while (i < table_arg->s->keys) {

                    if (!is_mysql_primary_key(&table_arg->key_info[i])) {

                        succp = ha_solid_createindex(
                                relh,
                                &table_arg->key_info[i],
                                relname,
                                tbcon,
                                cd,
                                trans,
                                authid, catalog, extrainfo,
                                &errh,
                                lower_case_table_names);

                        if (errh != NULL) {
                            /* TODO is this correct way of error handling ? */
                            sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);

                            break;
                        }
                    }

                    i++;
                }
            }
        }

        /* If the auto increment field has been defined and
           query was ALTER TABLE...AUTO_INCREMENT = x; or
           CREATE TABLE ...AUTO_INCREMENT = x; Set a new value for the
           auto_increment field. */

        if (succp &&
            auto_inc_defined &&
            (create_info->used_fields & HA_CREATE_USED_AUTO) &&
	        (create_info->auto_increment_value != 0))
        {
            rs_atype_t* valatype;
            rs_aval_t*  aval;
            ss_int8_t   i8;
            rs_auth_t*  auth = rs_sysi_auth(cd);

            do {
                succp = tb_trans_commit(cd, trans, &finished, &errh);
            } while (rs_sysi_lockwait(cd) || !finished);

            if (errh != NULL) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query);
            }

            con = get_solid_connection(name);

            ss_dassert(con != NULL);
            tbcon = con->sc_tbcon;
            cd    = con->sc_cd;

#if MYSQL_VERSION_ID >= 50100
            solid_trans_beginif(ht, thd, TRUE);
#else
            solid_trans_beginif(thd, TRUE);
#endif
            trans = con->sc_trans;

            /* Set startup value for the auto_increment sequence */
            if (succp) {
                //dbe_ret_t rc;
                rs_auth_setsystempriv(cd, auth, TRUE);
                valatype = rs_atype_initbigint(cd);
                aval = rs_aval_create(cd, valatype);
                SsInt8SetNativeUint8(&i8, (create_info->auto_increment_value - 1));
                rs_aval_setint8_ext(cd, valatype, aval, i8, NULL);

                finished = FALSE;

                do {
                    succp = tb_seq_set(cd, trans, seq_id, FALSE, valatype, aval,
                                   &finished, &errh);
                } while (!finished);


                if (errh != NULL) {
                    sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query); 
                }

                ss_pprintf(("AUTO INC SEQ initial value = %lu\n",
                            (long) create_info->auto_increment_value));

                rs_auth_setsystempriv(cd, auth, FALSE);

                rs_aval_free(cd, valatype, aval);
                rs_atype_free(cd, valatype);

            }
            rs_relh_setautoincrement_seqid(cd, relh, seq_id);
        }

error_handling:

        ha_mysql_solid_create_unique_done(primkey);
        rs_ttype_free(cd, ttype);
        SsMemFree(relname);
        SsMemFree(authid);

        if (relh != NULL) {
            rs_relh_done(cd, relh);
        }

        commit_succ_p = internal_solid_end_trans_and_stmt(cd, trans, succp, &commit_errh);

        if (!commit_succ_p) {
            int commit_rc = 0;
            
            commit_rc = handle_error(commit_errh);

            if (!rc) {
                rc = commit_rc;
            }
        }

        if (rc != 0) {
            if (errh) {
                sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query); 
                su_err_done(errh);
                errh = NULL;
            }

            if (forkeys) {
                ha_mysql_solid_foreign_done(forkeys, n_forkeys);
            }
        } else {
            if (!succp) {
                rc = ha_solid_mysql_error(thd, errh);
                su_err_done(errh);
            } else {
                rc = 0;
            }
        }

        finish_solid_DDL();
        SS_POPNAME;
        SDB_DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              solid_next_insert_id
 *
 * Generate the next auto-increment number based on increment and offset
 *
 * In most cases increment= offset= 1, in which case we get:
 * 1,2,3,4,5,...
 * If increment=10 and offset=5 and previous number is 1, we get:
 * 1,5,15,25,35,...
 *
 * Parameters :
 *
 *     ulonglong nr, in, use
 *     ulonglong increment, in, use
 *     ulonglong offset, in, use
 *
 * Return value :
 *
 *     ulonglong
 *
 * Globals used :
 */
static inline ulonglong solid_next_insert_id(
        ulonglong nr,
        ulonglong increment,
        ulonglong offset)
{
        nr = (((nr + increment - offset)) / increment);

        return (nr * increment + offset);
}

#if MYSQL_VERSION_ID < 50100
/*#***********************************************************************\
 *
 *              ::update_auto_increment
 *
 * Update the auto_increment field if necessary
 *
 * IMPLEMENTATION
 *
 *   Updates columns with type NEXT_NUMBER if:
 *
 * - If column value is set to NULL (in which case
 *   auto_increment_field_not_null is 0)
 *
 *  - If column is set to 0 and (sql_mode & MODE_NO_AUTO_VALUE_ON_ZERO) is not
 *   set. In the future we will only set NEXT_NUMBER fields if one sets them
 *   to NULL (or they are not included in the insert list).
 *
 *
 * Parameters : -
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
bool ha_soliddb::update_auto_increment()
{
        ss_int8_t i8;
        THD* thd;
        SOLID_CONN* con;
        ulonglong mysql_value = 0;
        ulonglong new_value = 0;
        ulonglong solid_value = 0;
        struct system_variables* variables;
        bool auto_increment_not_null = FALSE;
        bool result = FALSE;
        rs_relh_t*   relh = NULL;
        long seq_id;
        rs_auth_t* auth;

        DBUG_ENTER("handler::update_auto_increment");

        thd = current_thd;
        variables = &thd->variables;

        con = get_solid_connection(thd, table->s->path);

        ss_dassert(con != NULL);
        relh = solid_rsrelh;
        ss_dassert(relh != NULL);
        auth = rs_sysi_auth(con->sc_cd);
        ss_dassert(auth != NULL);
        rs_auth_setsystempriv(con->sc_cd, auth, TRUE);

        seq_id = rs_relh_readautoincrement_seqid(con->sc_cd, relh);

        /* We must save the previous value to be able to restore it if the
	    row was not inserted */

        thd->prev_insert_id = thd->next_insert_id;
        auto_increment_not_null = table->auto_increment_field_not_null;
        table->auto_increment_field_not_null = FALSE;

        mysql_value = table->next_number_field->val_int();

        if (mysql_value != 0
            || auto_increment_not_null
            && variables->sql_mode & MODE_NO_AUTO_VALUE_ON_ZERO) {
            rs_atype_t* atype;
            rs_aval_t*  aval;
            solid_bool  finishedp;
            ss_int8_t   si8;
            bool succp=TRUE;

            /* Mark that we didn't generate a new value */
            auto_increment_column_changed = FALSE;
            atype = rs_atype_initbigint(con->sc_cd);
            aval = rs_aval_create(con->sc_cd, atype);

            succp = tb_seq_current(con->sc_cd, con->sc_trans, seq_id,
                           FALSE, atype, aval, &finishedp, NULL);
            ss_dassert(finishedp);
            ss_dassert(succp);

            si8 = rs_aval_getint8(con->sc_cd, atype, aval);

            solid_value = SsInt8GetNativeUint8(si8);

            if (mysql_value > solid_value) {
                dbe_ret_t rc;

                if (variables->auto_increment_increment != 1) {

                    new_value = solid_next_insert_id(mysql_value,
                                              variables->auto_increment_increment,
                                              variables->auto_increment_offset);

                    SsInt8SetNativeUint8(&i8, new_value-1);

                } else {
                    new_value = mysql_value;

                    SsInt8SetNativeUint8(&i8, new_value);
                }

                rs_aval_setint8_raw(con->sc_cd, atype, aval, i8, NULL);

                do {
                   rc = tb_seq_lock(con->sc_cd, con->sc_trans, seq_id, NULL);
                } while (rs_sysi_lockwait(con->sc_cd) || rc == DBE_RC_WAITLOCK);

                result = tb_seq_set(con->sc_cd, con->sc_trans, seq_id,
                           FALSE, atype, aval, &finishedp, NULL);

                ss_pprintf_1(("New auto_increment value = %lu\n", (long) new_value));

                /* Mark that we did generate a new value */

                auto_increment_column_changed = TRUE;
		    }

            rs_aval_free(con->sc_cd, atype, aval);
            rs_atype_free(con->sc_cd, atype);

            rs_auth_setsystempriv(con->sc_cd, auth, FALSE);

            DBUG_RETURN(0);
        }

        /* This is the case where user has not provided value for a auto_increment field. */

        if (mysql_value == 0) {
            rs_atype_t* atype;
            rs_aval_t*  aval;
            solid_bool  p_finishedp;
            ss_int8_t   si8;
            dbe_ret_t   rc;

            atype = rs_atype_initbigint(con->sc_cd);
            aval = rs_aval_create(con->sc_cd, atype);

            do {
                rc = tb_seq_lock(con->sc_cd, con->sc_trans, seq_id, NULL);
            } while (rs_sysi_lockwait(con->sc_cd) || rc == DBE_RC_WAITLOCK);

            result = tb_seq_next(con->sc_cd, con->sc_trans, seq_id,
                                 FALSE, atype, aval, &p_finishedp, NULL);

            if (result) {
                si8 = rs_aval_getint8(con->sc_cd, atype, aval);

                new_value = SsInt8GetNativeUint8(si8);

                if (variables->auto_increment_increment != 1) {
                    new_value = solid_next_insert_id(new_value-1,
                                          variables->auto_increment_increment,
                                          variables->auto_increment_offset);

                    SsInt8SetNativeUint8(&i8, new_value);
                    rs_aval_setint8_raw(con->sc_cd, atype, aval, i8, NULL);

                    result = tb_seq_set(con->sc_cd, con->sc_trans, seq_id,
                                    FALSE, atype, aval, &p_finishedp, NULL);
                }

            }

            rs_aval_free(con->sc_cd, atype, aval);
            rs_atype_free(con->sc_cd, atype);

            ss_pprintf_1(("New auto_increment value = %lu\n", (long) new_value));


            /* Note that we set next_insert_id to 0 to force handler
               call for every row on a multi-row statement */

            thd->next_insert_id = 0;
        }

        DBUG_PRINT("info",("auto_increment: %lu", (ulong) new_value));

        /* Mark that we should clear next_insert_id before next stmt */
        thd->clear_next_insert_id = 1;

        if (!table->next_number_field->store(new_value, TRUE)) {
            thd->insert_id(new_value);
        } else {

            thd->insert_id(table->next_number_field->val_int());
        }

        /* Mark that we generated a new value */
        auto_increment_column_changed = TRUE;

        rs_auth_setsystempriv(con->sc_cd, auth, FALSE);

        DBUG_RETURN(result);
}
#endif /* MYSQL_VERSION_ID < 50100 */

/*#***********************************************************************\
 *
 *              ::check
 *
 * Check table implementation for solidDB
 *
 * Parameters :
 *
 *      THD*           thd, in, use
 *      HA_CHECK_OPT*  check_opt, in, not used
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::check(
        THD* thd,
        HA_CHECK_OPT* check_opt)
{
        SOLID_CONN* con = NULL;
        dbe_db_t*   db  = NULL;
        rs_sysi_t*  cd  = NULL;
        bool        error = FALSE;

        DBUG_ENTER("ha_soliddb::check");

        ss_dassert(thd != NULL);
        ss_dassert(check_opt != NULL);

#if MYSQL_VERSION_ID >= 50100
        con = (SOLID_CONN*)thd->ha_data[ht->slot];
#else
        con = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        ss_dassert(con != NULL);
        cd = con->sc_cd;
        ss_dassert(cd != NULL);

        db = (dbe_db_t*)rs_sysi_db(cd);
        ss_dassert(db != NULL);

        /* TODO: These should be done:

           Check that data blocks for this table are sound
           Check that all indexes contain right number of rows
           Check that data blocks for all indexes are sound
        */

        if (error) {
            DBUG_RETURN(HA_ADMIN_CORRUPT);
        } else {
            DBUG_RETURN(HA_ADMIN_OK);
        }
}

/*#***********************************************************************\
 *
 *              soliddb_check_filespec
 *
 * Check syntax on a new database file specification
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      set_var*          var, in, use
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int soliddb_check_filespec(
        THD* thd,
        set_var *var)
{
        DBUG_ENTER("soliddb_check_filespec");

        char *filename = NULL;
        char* mismatch = NULL;
        char *p = NULL;
        ss_int8_t size;
        ss_int8_t coeff;
        bool b = TRUE;

        ss_dassert(thd != NULL);
        ss_dassert(var != NULL);

        filename = (char*) SsMemStrdup(var->value->name);
        p = filename;

        /* Find end of filename */
        while(*p != '\0' && !ss_isspace(*p) && *p != ',') {
            p++;
        }

        if (*p == '\0') {
            sql_print_error("solidDB: Error: Parse error on filespec %s", filename);
            SsMemFree(filename);
            DBUG_RETURN(-1);
        }

        *p = '\0';

        /* Find maxsize */
        p++;

        while(*p != '\0' && !ss_isdigit(*p)) {
            p++;
        }

        b = SsStrScanInt8(p, &size, &mismatch);

        if (!b) {
            sql_print_error("solidDB: Error: Illegal or missing argument filesize %s",
                    filename);
            SsMemFree(filename);
            DBUG_RETURN(-1);
        }

        p = mismatch;

        /* Set up database file size */
        switch (*p) {
            case 'k':
            case 'K':
                SsInt8SetUint4(&coeff, 1024U);
                break;
            case 'm':
            case 'M':
                SsInt8SetUint4(&coeff, (ss_uint4_t)1024U * 1024U);
                break;
            case 'g':
            case 'G':
                SsInt8SetUint4(&coeff,
                               (ss_uint4_t)1024U * 1024U * 1024U);
                break;
            default:
                SsInt8SetUint4(&coeff, 1U);
                break;
        }

        b = SsInt8MultiplyByInt8(&size, size, coeff);
        SsInt8SetUint4(&coeff, 102400L); /* FILESPEC_MIN_SIZE */

        if (!b || SsInt8Cmp(size,coeff) < 0) {
            sql_print_error("solidDB: Error: Illegal value for file size.");
            SsMemFree(filename);
            DBUG_RETURN(-1);
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              soliddb_add_filespec
 *
 * Add a new database file to solidDB filespec
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      set_var*          var, in, use
 *
 * Return value : bool
 *
 * Globals used :
 */
bool soliddb_add_filespec(
        THD* thd,
        set_var *var)
{
        DBUG_ENTER("soliddb_add_filespec");

        char *filename = NULL;
        char* mismatch = NULL;
        char *p = NULL;
        ss_int8_t size;
        ss_int8_t coeff;
        bool b = TRUE;
        su_ret_t rc;
        dbe_db_t* db;
        SOLID_CONN* con;

        ss_dassert(thd != NULL);
        ss_dassert(var != NULL);

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);
        filename = (char*) SsMemStrdup(var->value->name);
        p = filename;

        /* Find end of filename */
        while(*p != '\0' && !ss_isspace(*p) && *p != ',') {
            p++;
        }

        if (*p == '\0') {
            sql_print_error("solidDB: Error: Parse error on filespec %s", filename);
            SsMemFree(filename);
            ss_error;
        }

        *p = '\0';

        /* Find maxsize */
        p++;

        while(*p != '\0' && !ss_isdigit(*p)) {
            p++;
        }

        b = SsStrScanInt8(p, &size, &mismatch);

        if (!b) {
            sql_print_error("solidDB: Error: Illegal or missing argument from filespec %s",
                    filename);
            SsMemFree(filename);
            ss_error;
        }

        p = mismatch;

        /* Set up database file size */
        switch (*p) {
            case 'k':
            case 'K':
                SsInt8SetUint4(&coeff, 1024U);
                break;
            case 'm':
            case 'M':
                SsInt8SetUint4(&coeff, (ss_uint4_t)1024U * 1024U);
                break;
            case 'g':
            case 'G':
                SsInt8SetUint4(&coeff,
                               (ss_uint4_t)1024U * 1024U * 1024U);
                break;
            default:
                SsInt8SetUint4(&coeff, 1U);
                break;
        }

        /* Check size */
        b = SsInt8MultiplyByInt8(&size, size, coeff);
        SsInt8SetUint4(&coeff, 102400L); /* FILESPEC_MIN_SIZE */

        if (!b || SsInt8Cmp(size,coeff) < 0) {
            sql_print_error("solidDB: Error: Illegal value for file size.");
            SsMemFree(filename);
            ss_error;
        }

        /* Add new filespec to dbfile */
        rc = dbe_db_addnewfilespec(db, filename, size, 0);

        if (rc != SU_SUCCESS) {
            char *errmsg;

            sql_print_error("solidDB: Error: Adding a new database file failed.");
            errmsg = su_rc_givetext_noargs(rc);
            sql_print_error("%s", errmsg);
            SsMemFree(errmsg);
            SsMemFree(filename);
            DBUG_RETURN(FALSE);
        } else {

            /* If we have entered read only mode set readonly to false */
            if (dbe_db_isreadonly(db)) {
                dbe_db_setreadonly(db, FALSE);
            }

            SsMemFree(filename);
        }

        DBUG_RETURN(FALSE);
}

/*#***********************************************************************\
 *
 *              soliddb_update_durability_level
 *
 * Update durability level of transactions
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      enum_var_type     type, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
void soliddb_update_durability_level(
        THD* thd,
        enum_var_type type __attribute__ ((unused)))
{
        SOLID_CONN* con = NULL;
        dbe_db_t* db = NULL;
        su_inifile_t* inifile = NULL;

        DBUG_ENTER("soliddb_update_durability_level");

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        if (soliddb_durability_level < 1 || soliddb_durability_level > 3) {
            DBUG_VOID_RETURN; /* Ignore if error */
        }

        dbe_db_setdurabilitylevel(db, (dbe_durability_t) soliddb_durability_level);
        inifile = dbe_db_getinifile(db);

        if (inifile) {
            solid_set_long(inifile, soliddb_durability_level,
                           SU_DBE_LOGSECTION, SU_DBE_DURABILITYLEVEL);
        }

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              soliddb_update_checkpoint_time
 *
 * Update checkpoint time interval
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      enum_var_type     type, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
void soliddb_update_checkpoint_time(
        THD* thd,
        enum_var_type type __attribute__ ((unused)))
{
        SOLID_CONN* con = NULL;
        dbe_db_t* db = NULL;
        su_inifile_t* inifile = NULL;

        DBUG_ENTER("soliddb_update_checkpoint_time");

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        dbe_db_setcpmintime(db, soliddb_checkpoint_time);

        inifile = dbe_db_getinifile(db);

        if (inifile) {
            solid_set_long(inifile, soliddb_checkpoint_time,
                           SU_DBE_GENERALSECTION, SU_DBE_CPMINTIME);
        }

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              soliddb_update_lock_wait_timeout
 *
 * Update lock wait timeout interval
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      enum_var_type     type, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
void soliddb_update_lock_wait_timeout(
        THD* thd,
        enum_var_type type __attribute__ ((unused)))
{
        SOLID_CONN* con = NULL;
        dbe_db_t* db = NULL;
        su_inifile_t* inifile = NULL;

        DBUG_ENTER("soliddb_update_lock_wait_timeout");

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        dbe_db_setlocktimeout(db, &soliddb_lock_wait_timeout, &soliddb_lock_wait_timeout);

        inifile = dbe_db_getinifile(db);

        if (inifile) {
            solid_set_long(inifile, soliddb_lock_wait_timeout,
                           SU_DBE_GENERALSECTION, SU_DBE_PESSIMISTIC_LOCK_TO);

            solid_set_long(inifile, soliddb_lock_wait_timeout,
                           SU_DBE_GENERALSECTION, SU_DBE_OPTIMISTIC_LOCK_TO);
        }

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              soliddb_update_checkpoint_interval
 *
 * Update checkpoint interval i.e. how many log writes before a checkpoint
 *
 * Parameters :
 *
 *      THD*              thd, in, use
 *      enum_var_type     type, in, use
 *
 * Return value : -
 *
 * Globals used :
 */
void soliddb_update_checkpoint_interval(
        THD* thd,
        enum_var_type type __attribute__ ((unused)))
{
        SOLID_CONN* con = NULL;
        dbe_db_t* db = NULL;
        su_inifile_t* inifile = NULL;

        DBUG_ENTER("soliddb_update_checkpoint_interval");

#if MYSQL_VERSION_ID >= 50100
        /* TODO: Fix below when system variables get handlerton */
        con = get_solid_ha_data_connection(legacy_soliddb_hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        dbe_db_setcpinterval(db, soliddb_checkpoint_interval);

        inifile = dbe_db_getinifile(db);

        if (inifile) {
            solid_set_longlong(inifile, soliddb_checkpoint_interval,
                               SU_DBE_GENERALSECTION, SU_DBE_CPINTERVAL);
        }

        DBUG_VOID_RETURN;
}

/*#***********************************************************************\
 *
 *              soliddb_show_mutex_status
 *
 * Implementation for show soliddb mutex status and show engine soliddb mutex
 * commands
 *
 * Parameters :
 *
 *      handlerton*       hton,in, use, handlerton (used on 5.1.x)
 *      THD*              thd, in, use, MySQL query thread
 *      stat_print_fn*    in, use, print function  (used on 5.1.x)
 *
 * Return value : bool
 *
 * Globals used :
 */
bool soliddb_show_mutex_status(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton,
        THD* thd,
        stat_print_fn *stat_print
#else
        THD* thd
#endif
    )
{
#if MYSQL_VERSION_ID < 50100
        //char buff[256];
#else
        uint size1;
        uint size2;
        char buff[IO_SIZE];
        char buff2[IO_SIZE];
#endif
        uint hton_name_len = strlen(solid_hton_name);
#ifdef SS_PROFILE
        SsSemDbgT** sd_list;
        SsSemDbgT* sd;
#endif
#if MYSQL_VERSION_ID < 50100
        Protocol* protocol = thd->protocol;
        List<Item> field_list;
#endif

        DBUG_ENTER("solid_show_mutex_status");

#ifndef SS_PROFILE

#if MYSQL_VERSION_ID < 50100
        field_list.push_back(new Item_empty_string("Mutex status", FN_REFLEN));
        protocol->send_fields(&field_list,
                              Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF);
        protocol->prepare_for_resend();
        protocol->store("Mutex status not available for solidDB storage engine", system_charset_info);
        protocol->write();
        send_eof(thd);
#else
        size1 = my_snprintf(buff, sizeof(buff), "%s",
                            "Mutex status not available for solidDB storage engine");
        stat_print(thd, solid_hton_name, hton_name_len, buff, size1, NULL, 0);
#endif

        DBUG_RETURN(FALSE);

#else

        sd_list = SsGetSemList();

        if (sd_list != NULL) {
            int i;

#if MYSQL_VERSION_ID < 50100
            field_list.push_back(new Item_empty_string("File", FN_REFLEN));
            field_list.push_back(new Item_uint("Line", 21));
            field_list.push_back(new Item_uint("CallCnt", 21));
            field_list.push_back(new Item_uint("WaitCnt", 21));
            field_list.push_back(new Item_uint("WaitPrc", 21));
            field_list.push_back(new Item_empty_string("Gate", FN_REFLEN));
            field_list.push_back(new Item_empty_string("Note", FN_REFLEN));

            if (protocol->send_fields(&field_list,
                                      Protocol::SEND_NUM_ROWS | Protocol::SEND_EOF)) {
                send_eof(thd);
                DBUG_RETURN(FALSE);
            }

#endif

            for (i = 0; i < SS_SEMNUM_LAST; i++) {
                sd = sd_list[i];

                if (sd != NULL) {
                    ulong waitprc;

                    waitprc = sd->sd_callcnt == 0
                                ? 0L
                                : (ulonglong)(sd->sd_waitcnt * 100.0 / sd->sd_callcnt);

#if MYSQL_VERSION_ID < 50100
                    protocol->prepare_for_resend();
                    protocol->store(sd->sd_file, system_charset_info);
                    protocol->store((ulonglong)sd->sd_line);
                    protocol->store((ulonglong)sd->sd_callcnt);
                    protocol->store((ulonglong)sd->sd_waitcnt);
                    protocol->store((ulonglong)waitprc);
                    protocol->store(sd->sd_gatep ? "gate" : "", system_charset_info);
                    protocol->store(waitprc >= 10 ? "***" : "", system_charset_info);

                    if (protocol->write()) {
                        send_eof(thd);
                        DBUG_RETURN(FALSE);
                    }
#else /* MYSQL_VERSION_ID >= 50100 i.e. MySQL 5.1.x */

                    size1 = my_snprintf(buff, sizeof(buff), "%s",
                                        sd->sd_file);

                    size2 = my_snprintf(buff2, sizeof(buff2),
                                        "line =%d, callcnt=%lu, waitcnt=%lu"
                                        "waitprc=%lu, gate=%s, note=%s",
                                        sd->sd_line, sd->sd_callcnt, sd->sd_waitcnt,
                                        waitprc, sd->sd_gatep ? "gate" : "",
                                        waitprc >= 10 ? "***" : "");

                    if (stat_print(thd, solid_hton_name, hton_name_len, buff, size1,
                                   buff2, size2)) {
                        DBUG_RETURN(FALSE);
                    }
#endif

                }
            }
        }

#if MYSQL_VERSION_ID < 50100
        send_eof(thd);
#endif

        DBUG_RETURN(FALSE);
#endif /* SS_PROFILE */
}

/*#***********************************************************************\
 *
 *              soliddb_show_status
 *
 * Implementation for show soliddb status and show engine soliddb commands
 *
 * Parameters :
 *
 *      handlerton*       hton, in, use, handlerton (used on 5.1.x)
 *      THD*              thd, in, use, MySQL query thread
 *      stat_print_fn*    in, use, print function   (used on 5.1.x)
 *      enum ha_stat_type stat_type, in, use        (used on 5.1.x)
 *
 * Return value : bool
 *
 * Globals used :
 */
bool soliddb_show_status(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton,
        THD* thd,
        stat_print_fn *stat_print,
        enum ha_stat_type stat_type
#else
        THD* thd
#endif
        )
{
        const long MAX_STATUS_SIZE = 64000;
        char* status_str = NULL;
        char* buf = NULL;
        char buf2[256];
        long real_length = 0;
        dbe_dbstat_t dbs;
        SsQmemStatT qms;
        double hitrate = 0.0;
        dbe_db_t* db;
        SOLID_CONN* con;
#if MYSQL_VERSION_ID < 50100
        Protocol* protocol = thd->protocol;
#endif

        DBUG_ENTER("solid_show_status");

        if (have_soliddb != SHOW_OPTION_YES) {
            DBUG_RETURN(FALSE);
        }

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(hton, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        status_str = (char *)SsMemAlloc(MAX_STATUS_SIZE);
        buf = (char *) SsMemAlloc(MAX_STATUS_SIZE);

        dbe_db_getstat(db, &dbs);
        dbe_db_getmeminfo(db, &qms);

        SsSprintf(status_str, "Memory statistics:\n");

#ifdef SS_NT64
        SsSprintf(buf, "    %I64lu kilobytes\n",
                  (qms.qms_sysbytecount + qms.qms_slotbytecount) / 1024);
#else
        SsSprintf(buf,   "    %lu kilobytes\n",
                  (qms.qms_sysbytecount + qms.qms_slotbytecount) / 1024);
#endif

        strcat(status_str, buf);
        SsSprintf(buf, "Transaction count statistics:\n");
        strcat(status_str, buf);
        SsSprintf(buf, "    Commit Abort Rollback    Total Read-only  Trxbuf  Active Validate\n");
        strcat(status_str, buf);

#ifdef SS_NT64
        SsSprintf(buf,   "    %6I64lu %5I64lu %8I64lu %8I64lu %9I64lu %7I64lu %7I64lu %8I64lu\n",
                  dbs.dbst_trx_commitcnt, dbs.dbst_trx_abortcnt, dbs.dbst_trx_rollbackcnt,
                  dbs.dbst_trx_commitcnt + dbs.dbst_trx_abortcnt + dbs.dbst_trx_rollbackcnt,
                  dbs.dbst_trx_readonlycnt,
                  dbs.dbst_trx_bufcnt, dbs.dbst_trx_activecnt, dbs.dbst_trx_validatecnt);
#else
        SsSprintf(buf,   "    %6lu %5lu %8lu %8lu %9lu %7lu %7lu %8lu\n",
                  dbs.dbst_trx_commitcnt, dbs.dbst_trx_abortcnt, dbs.dbst_trx_rollbackcnt,
                  dbs.dbst_trx_commitcnt + dbs.dbst_trx_abortcnt + dbs.dbst_trx_rollbackcnt,
                  dbs.dbst_trx_readonlycnt,
                  dbs.dbst_trx_bufcnt, dbs.dbst_trx_activecnt, dbs.dbst_trx_validatecnt);
#endif

        strcat(status_str, buf);

        if (dbs.dbst_cac_findcnt == 0) {
            hitrate = 0.0;
        } else {
            hitrate = (double)100.0 * (dbs.dbst_cac_findcnt - dbs.dbst_cac_readcnt) /
                dbs.dbst_cac_findcnt;
        }

#ifdef SS_WIN
        SsDoubleToAscii(hitrate, buf2, 5);
#else
        SsSprintf(buf2, "%.1lf", hitrate);
#endif

        SsSprintf(buf, "Cache count statistics:\n");
        strcat(status_str, buf);
        SsSprintf(buf, "    Hit rate      Find      Read     Write\n");
        strcat(status_str, buf);
        SsSprintf(buf,   "    %-13s %-9lu %-9lu %-9lu\n",
                  buf2, dbs.dbst_cac_findcnt, dbs.dbst_cac_readcnt, dbs.dbst_cac_writecnt);
        strcat(status_str, buf);
        SsSprintf(buf, "Database statistics:\n");
        strcat(status_str, buf);

        if (dbs.dbst_ind_mergeact) {
            SsSprintf(buf, "    Index merge is active.\n");
            strcat(status_str, buf);
        }

        SsSprintf(buf,   "    Index writes    %9lu  After last merge %6lu\n",
                  dbs.dbst_ind_writecnt, dbs.dbst_ind_writecntsincemerge);
        strcat(status_str, buf);

        SsSprintf(buf,   "    Log writes      %9lu  After last cp    %6lu\n",
                  dbs.dbst_log_writecnt, dbs.dbst_log_writecntsincecp);
        strcat(status_str, buf);

        SsSprintf(buf,   "    Active searches %9lu  Average          %6lu\n",
                  dbs.dbst_sea_activecnt, dbs.dbst_sea_activeavg);
        strcat(status_str, buf);

        SsSprintf(buf,   "    Database size %10lu kilobytes\n", dbs.dbst_ind_filesize);
        strcat(status_str, buf);

        SsSprintf(buf,   "    Log size      %10lu kilobytes\n", dbs.dbst_log_filesize);
        strcat(status_str, buf);
        SsMemFree(buf);

#if MYSQL_VERSION_ID > 50100
        switch (stat_type) {
            case HA_ENGINE_STATUS: {
                   real_length = strlen(status_str) + 1;

                   stat_print(thd, solid_hton_name, strlen(solid_hton_name),
                       STRING_WITH_LEN(""), status_str, real_length);

                   break;
            }
            case HA_ENGINE_MUTEX: break;

            default: ss_error; break;
        }

        SsMemFree(status_str);
#else
    {
	    List<Item> field_list;

        real_length = strlen(status_str) + 1;
        field_list.push_back(new Item_empty_string("Status", real_length));

        if (protocol->send_fields(&field_list, Protocol::SEND_NUM_ROWS |
                                               Protocol::SEND_EOF)) {
            SsMemFree(status_str);
            DBUG_RETURN(TRUE);
        }

        protocol->prepare_for_resend();
        protocol->store(status_str, real_length, system_charset_info);
        SsMemFree(status_str);

        if (protocol->write()) {
        	DBUG_RETURN(TRUE);
        }

        send_eof(thd);
    }
#endif

    DBUG_RETURN(FALSE);
}

/*#***********************************************************************\
 *
 *              soliddb_export_status
 *
 * Export solidDB status variables to MySQL
 *
 * Parameters : -
 *
 * Return value : -
 *
 * Globals used : current_thd
 */
void soliddb_export_status(void)
{
        SOLID_CONN* con = NULL;
        dbe_dbstat_t dbs;
        SsQmemStatT qms;
        double hitrate = 0.0;
        dbe_db_t* db;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(legacy_soliddb_hton, current_thd);
#else
        con = get_solid_ha_data_connection(current_thd);
#endif
        
        ss_dassert(con != NULL);
        db = (dbe_db_t*)rs_sysi_db(con->sc_cd);
        ss_dassert(db != NULL);

        dbe_db_getstat(db, &dbs);
        dbe_db_getmeminfo(db, &qms);

        export_vars.soliddb_memory_statistics = ((qms.qms_sysbytecount + qms.qms_slotbytecount) / 1024);
        export_vars.soliddb_transaction_commits = dbs.dbst_trx_commitcnt;
        export_vars.soliddb_transaction_rollbacks = dbs.dbst_trx_abortcnt;
        export_vars.soliddb_transaction_aborts = dbs.dbst_trx_rollbackcnt;
        export_vars.soliddb_transaction_total = dbs.dbst_trx_commitcnt + dbs.dbst_trx_abortcnt + dbs.dbst_trx_rollbackcnt;
        export_vars.soliddb_transaction_readonly = dbs.dbst_trx_readonlycnt;
        export_vars.soliddb_transaction_trxbuf = dbs.dbst_trx_bufcnt;
        export_vars.soliddb_transaction_active = dbs.dbst_trx_activecnt;
        export_vars.soliddb_transaction_validate = dbs.dbst_trx_validatecnt;

        if (dbs.dbst_cac_findcnt == 0) {
            hitrate = 0.0;
        } else {
            hitrate = (double)100.0 * (dbs.dbst_cac_findcnt - dbs.dbst_cac_readcnt) /
                dbs.dbst_cac_findcnt;
        }

        export_vars.soliddb_cache_hitrate = (longlong)hitrate;        
        export_vars.soliddb_cache_find = dbs.dbst_cac_findcnt;
        export_vars.soliddb_cache_read = dbs.dbst_cac_readcnt;
        export_vars.soliddb_cache_write = dbs.dbst_cac_writecnt;
        export_vars.soliddb_indexmerge_active = dbs.dbst_ind_mergeact;
        export_vars.soliddb_index_writes = dbs.dbst_ind_writecnt;
        export_vars.soliddb_index_writesincemerge = dbs.dbst_ind_writecntsincemerge;
        export_vars.soliddb_log_writes = dbs.dbst_log_writecnt;
        export_vars.soliddb_log_writessincecp = dbs.dbst_log_writecntsincecp;
        export_vars.soliddb_active = dbs.dbst_sea_activecnt;
        export_vars.soliddb_activeavg = dbs.dbst_sea_activeavg;
        export_vars.soliddb_filesize = dbs.dbst_ind_filesize;
        export_vars.soliddb_logsize = dbs.dbst_log_filesize;
        export_vars.soliddb_cp_active= dbe_db_isbackupactive(db);
        export_vars.soliddb_backup_active= dbe_db_iscpactive(db);
}

/*#***********************************************************************\
 *
 *              ::referenced_by_foreign_key()
 *
 * Checks if a table is referenced by a foreign key. The MySQL manual states that
 * a REPLACE is either equivalent to an INSERT, or DELETE(s) + INSERT. Only a
 * delete is then allowed internally to resolve a duplicate key conflict in
 * REPLACE, not an update.
 *
 * Parameters : void
 *
 * Return value : 0 if not referenced, 1 if is referenced
 *
 * Globals used : current_thd
 */
uint ha_soliddb::referenced_by_foreign_key(void)
{
        SOLID_CONN* con;
        THD*        thd = (THD *)current_thd;
        rs_relh_t*  relh = NULL;
        rs_err_t*   errh = NULL;
        solid_bool  referenced = FALSE;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(ht, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        relh = solid_rsrelh;
        ss_dassert(relh != NULL);

        referenced = tb_dd_isforkeyref(con->sc_cd, con->sc_trans, relh, &errh);

        if (errh) {
            sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query); 
            su_err_done(errh);
            errh = NULL;
        }

        if (referenced) {
            return (1);
        } else {
            return (0);
        }
}

/*#***********************************************************************\
 *
 *              ::get_foreign_key_create_info()
 *
 * Return foreign key create information for solidDB tables
 *
 * Parameters : void
 *
 * Return value : char*
 *
 * Globals used : current_thd
 */
char* ha_soliddb::get_foreign_key_create_info(void)
{
        SOLID_CONN* con;
        THD *   thd = (THD *)current_thd;
        char*   foreign_str	= NULL;
        long    foreign_len = 0;
        rs_relh_t*   relh = NULL;
        su_pa_t* refkeys = NULL;
        rs_key_t* key = NULL;
        uint i = 0;

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(ht, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        relh = solid_rsrelh;
        ss_dassert(relh != NULL);

        foreign_str = my_malloc(MAX_FOREIGN_LEN + 1, MYF(0));
        foreign_str[0]='\0';

        refkeys = rs_relh_refkeys(con->sc_cd, relh);

        su_pa_do(refkeys, i) {
            key = (rs_key_t*)su_pa_getdata(refkeys, i);

            if (rs_key_type(con->sc_cd, key) == RS_KEY_FORKEYCHK) {
                rs_ano_t j;
                rs_ano_t k;
                rs_ano_t n_parts;
                rs_relh_t *refrelh;
                rs_ano_t k_part;
                rs_err_t* errh = NULL;
                bool first_part = TRUE;
                su_pa_t* forrefkeys = NULL;
                rs_key_t* forkey = NULL;

                strcat(foreign_str, ",\n  CONSTRAINT `");
                strcat(foreign_str, rs_key_name(con->sc_cd, key));
                strcat(foreign_str, "` FOREIGN KEY (");
                n_parts = rs_key_nparts(con->sc_cd, key);

                for(k_part = 0; k_part < n_parts; k_part++) {
                    rs_keypart_t *key_part;
                    rs_ano_t     kp_ano;

                    key_part = &key->k_parts[k_part];
                    kp_ano = key_part->kp_ano;

                    if (rs_keyp_parttype(con->sc_cd, key, k_part) == RSAT_USER_DEFINED) {

                        if (first_part) {
                            first_part = FALSE;
                        } else {
                            strcat(foreign_str, ", ");
                        }

                        strcat(foreign_str, "`");
                        strcat(foreign_str, rs_ttype_aname(con->sc_cd, rs_relh_ttype(con->sc_cd, relh), kp_ano));
                        strcat(foreign_str, "`");
                    }
                }

                strcat(foreign_str, ") REFERENCES ");

                refrelh = tb_dd_getrelhbyid(
                                con->sc_cd, 
                                con->sc_trans,
                                rs_key_refrelid(con->sc_cd, key),
                                NULL,
                                &errh);

                if (refrelh) {
                    if (errh) {
                        sql_print_error("%s (%s)", su_err_geterrstr(errh), thd->query); 
                        su_err_done(errh);
                        errh = NULL;
                    }

                    strcat(foreign_str, "`");
                    strcat(foreign_str, rs_relh_name(con->sc_cd, refrelh));
                    strcat(foreign_str, "`");
                    strcat(foreign_str, " (");
                    first_part = TRUE;

                    tb_dd_resolverefkeys(con->sc_cd, refrelh);
                    forrefkeys = rs_relh_refkeys(con->sc_cd, refrelh);
                    k = 0;

                    su_pa_do(forrefkeys, k) {
                        forkey = (rs_key_t*)su_pa_getdata(forrefkeys, k);
                        n_parts = rs_key_nparts(con->sc_cd, forkey);

                        if (rs_key_type(con->sc_cd, forkey) != RS_KEY_FORKEYCHK &&
                            rs_key_refrelid(con->sc_cd, forkey) == rs_relh_relid(con->sc_cd, relh)) {

                            for(j = 0; j < n_parts; j++) {
                                rs_ano_t key_ano;

                                key_ano = rs_keyp_ano(con->sc_cd, forkey, j);

                                if(rs_keyp_parttype(con->sc_cd, forkey, j) == RSAT_USER_DEFINED) {
                                    if (first_part) {
                                        first_part = FALSE;
                                    } else {
                                        strcat(foreign_str, ", ");
                                    }

                                    strcat(foreign_str, "`");
                                    strcat(foreign_str, rs_ttype_aname(con->sc_cd,
                                                               rs_relh_ttype(con->sc_cd, refrelh), key_ano));
                                    strcat(foreign_str, "`");
                                }
                            }
                        }
                    }

                    strcat(foreign_str, ") ON DELETE ");
                    ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_delete_action(con->sc_cd, key), foreign_str);
                    strcat(foreign_str, " ON UPDATE ");
                    ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_update_action(con->sc_cd, key), foreign_str);

                    rs_relh_done(con->sc_cd, refrelh);
                } else {
                    foreign_str[0]='\0';
                }
            }
        }

        return(foreign_str);
}

/*#***********************************************************************\
 *
 *              ::free_foreign_key_create_info()
 *
 * Free foreign key create information if given
 *
 * Parameters :
 *
 *     char* str, in, use, foreign key definition
 *
 * Return value : -
 *
 * Globals used : -
 */
void ha_soliddb::free_foreign_key_create_info(
        char* str)
{
        if (str) {
            my_free(str, MYF(0));
        }
}

/*#***********************************************************************\
 *
 *              ::get_foreign_key_list()
 *
 * Return foreign key list for solidDB tables
 *
 * Parameters :
 *
 *      THD*                    thd, in, use
 *      List<FOREIGN_KEY_INFO>* f_key_list
 *
 * Return value : 0 or error code
 *
 * Globals used : -
 */
int ha_soliddb::get_foreign_key_list(
        THD *thd,
        List<FOREIGN_KEY_INFO> *f_key_list)
{
        SOLID_CONN* con;
        rs_relh_t*  relh = NULL;
        su_pa_t*    refkeys = NULL;
        rs_key_t*   key = NULL;
        uint        i = 0;

        DBUG_ENTER("ha_soliddb::get_foreign_key_list()");

#if MYSQL_VERSION_ID >= 50100
        con = get_solid_ha_data_connection(ht, thd);
#else
        con = get_solid_ha_data_connection(thd);
#endif

        ss_dassert(con != NULL);
        relh = solid_rsrelh;
        ss_dassert(relh != NULL);

        refkeys = rs_relh_refkeys(con->sc_cd, relh);

        su_pa_do(refkeys, i) {
            key = (rs_key_t*)su_pa_getdata(refkeys, i);

            if (rs_key_type(con->sc_cd, key) == RS_KEY_FORKEYCHK) {
                rs_ano_t j;
                rs_ano_t n_parts;
                rs_relh_t *refrelh;
                rs_ano_t k_part;
                rs_ano_t k;
                FOREIGN_KEY_INFO f_key_info;
                FOREIGN_KEY_INFO *pf_key_info;
                char buf[256];
                LEX_STRING *name= 0;
                char *cname = NULL;
                su_pa_t* forrefkeys = NULL;

                refrelh = tb_dd_getrelhbyid(
                                con->sc_cd,
                                con->sc_trans,
                                rs_key_refrelid(con->sc_cd, key),
                                NULL,
                                NULL);

                if (refrelh == NULL) {
                    /* Concurrent ddl, skip this table, */
                    continue;
                }

                cname = rs_key_name(con->sc_cd, key);
                f_key_info.forein_id= make_lex_string(thd, 0, cname,
                                                      (uint) strlen(cname), 1);

                cname = rs_relh_schema(con->sc_cd, relh);
                f_key_info.referenced_db= make_lex_string(thd, 0, cname,
                                                          (uint)strlen(cname), 1);

                cname = rs_relh_name(con->sc_cd, refrelh);
                f_key_info.referenced_table= make_lex_string(thd, 0, cname,
                                                             (uint) strlen(cname), 1);

                n_parts = rs_key_nparts(con->sc_cd, key);

                for(k_part = 0; k_part < n_parts; k_part++) {
                    rs_keypart_t *key_part;
                    rs_ano_t     kp_ano;

                    key_part = &key->k_parts[k_part];
                    kp_ano = key_part->kp_ano;

                    if (rs_keyp_parttype(con->sc_cd, key, k_part) == RSAT_USER_DEFINED) {

                        cname = rs_ttype_aname(con->sc_cd, rs_relh_ttype(con->sc_cd, relh), kp_ano);

                        name = make_lex_string(thd, name, cname, (uint) strlen(cname), 1);
                        f_key_info.foreign_fields.push_back(name);
                    }
                }

                tb_dd_resolverefkeys(con->sc_cd, refrelh);
                forrefkeys = rs_relh_refkeys(con->sc_cd, refrelh);
                k = 0;

                su_pa_do(forrefkeys, k) {
                    rs_key_t* forkey = NULL;

                    forkey = (rs_key_t*)su_pa_getdata(forrefkeys, k);
                    n_parts = rs_key_nparts(con->sc_cd, forkey);

                    if (rs_key_type(con->sc_cd, forkey) != RS_KEY_FORKEYCHK &&
                        rs_key_refrelid(con->sc_cd, forkey) == rs_relh_relid(con->sc_cd, relh)) {

                        for(j = 0; j < n_parts; j++) {

                            rs_ano_t key_ano;

                            key_ano = rs_keyp_ano(con->sc_cd, forkey, j);

                            if(rs_keyp_parttype(con->sc_cd, forkey, j) == RSAT_USER_DEFINED) {

                                cname = rs_ttype_aname(con->sc_cd, rs_relh_ttype(con->sc_cd, refrelh), key_ano);
                                name = make_lex_string(thd, name, cname, (uint) strlen(cname), 1);
                                f_key_info.referenced_fields.push_back(name);
                            }
                        }
                    }
                }

                rs_relh_done(con->sc_cd, refrelh);

                buf[0] = '\0';
                strcat(buf, "ON DELETE ");
                ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_delete_action(con->sc_cd, key), buf);

#if MYSQL_VERSION_ID >= 50100
                f_key_info.update_method = make_lex_string(thd,
                                                           f_key_info.update_method,
                                                           buf, (uint) strlen(buf), 1);
                buf[0] = '\0';
#endif
                strcat(buf, " ON UPDATE ");
                ha_map_solid_mysql_fkaction((sqlrefact_t)rs_key_update_action(con->sc_cd, key), buf);

#if MYSQL_VERSION_ID >= 50100
                f_key_info.delete_method = make_lex_string(thd,
                                                           f_key_info.delete_method,
                                                           buf, (uint) strlen(buf), 1);
#else
                f_key_info.constraint_method = make_lex_string(thd,
                                                               f_key_info.constraint_method,
                                                               buf, (uint) strlen(buf), 1);
#endif

                pf_key_info= ((FOREIGN_KEY_INFO *)thd->memdup((gptr) &f_key_info,
                                                              sizeof(FOREIGN_KEY_INFO)));

                f_key_list->push_back(pf_key_info);
            }
        }

        DBUG_RETURN(0);
}

/*#***********************************************************************\
*
*              solid_set_relname
*
* Convert MySQL relation name to solidDB
*
* Parameters :
*
*     char* schema, in, use
*     char* relname, in out, use
*     char* mysqlname, in, use
*
* Return value : -
*
* Globals used :
*/

/* THIS FUNC IS OUTDATED, IT MUST BE DROPPED !! Use solid_relname() instead. */
void solid_set_relname(
                       char* schema,
                       char* relname,
                       char* mysqlname,
                       bool  lower_case_names)
{
    char ch;
    char* p;

    ss_dprintf_2(("solid_set_relname:mysqlname='%s'\n", mysqlname));

    for (;;) {
        ch = *mysqlname;
        if (ch != '.' && ch != '/' && ch != '\\') {
            break;
        }
        mysqlname++;
    }

    p = schema;
    for (;;) {
        ch = *mysqlname++;
        if (ch == '\0' || ch == '/' || ch == '\\') {
            break;
        }

        if (lower_case_names) {
            *p++ = toupper(ch);
        } else {
            *p++ = ch;
        }
    }

    *p = '\0';

    p = relname;
    for (;;) {
        ch = *mysqlname++;
        if (ch == '\0' || ch == '.') {
            break;
        }

        if (lower_case_names) {
            *p++ = toupper(ch);
        } else {
            *p++ = ch;
        }
    }

    *p = '\0';

    ss_pprintf_2(("solid_set_relname:schema='%s', relname='%s'\n", schema, relname));
}

/*#***********************************************************************\
*
*              solid_split_myrelname
*
* Splits MySQL full table name in form ./<db-name>/<table-name> in two parts:
* database name and table name.
*
* Parameters :
*
*     char*  mysqlname, in, use, MySQL table name
*     char** schema, inout, database name
*     char** relname, inout, table name
*
* Return value : -
*
* Globals used : -
*/

void solid_split_myrelname(
        const char* mysqlname,
        char** schema,
        char** relname )
{
    const char* p;
    const char* q;

    ss_dassert(mysqlname != NULL);

    *schema = *relname = NULL;
    if (mysqlname == NULL) {
        return;
    }

    for (p = mysqlname; *p; ++p) {
        if (*p != '.' && *p != '/' && *p != '\\') {
            break;
        }
    }

    for (q = p; *q; ++q) {
        if (*q == '/' || *q == '\\') {
            break;
        }
    }

    *schema = (char*)SsMemAlloc(q - p + 1);
    SsTcsncpy(*schema, p, q - p);
    (*schema)[q - p] = '\0';
    
    for (; *q; ++q) {
        if (*q != '/' && *q != '\\') {
            break;
        }
    }
    
    *relname = (char*)SsMemAlloc(SsTcslen(q) + 1);
    
    SsTcscpy(*relname, q);
}

void solid_relname( rs_entname_t *solrel, const char* mysqlname)
{
    rs_entname_initbuf(solrel, NULL, NULL, NULL);
    solid_split_myrelname(mysqlname, &solrel->en_schema, &solrel->en_name);
    if (lower_case_table_names) {
        my_caseup_str( system_charset_info, solrel->en_schema );
        my_caseup_str( system_charset_info, solrel->en_name );
    }
}

void solid_relname( rs_entname_t *solrel, const char* mysqldb, const char* mysqlrel)
{
    rs_entname_initbuf(solrel, NULL, NULL, NULL);
    solrel->en_schema = (char*)SsMemStrdup((char*)mysqldb);
    solrel->en_name = (char*)SsMemStrdup((char*)mysqlrel);
    if (lower_case_table_names) {
        my_caseup_str( system_charset_info, solrel->en_schema );
        my_caseup_str( system_charset_info, solrel->en_name );
    }
}

/*#***********************************************************************\
*
*              is_soliddb_table
*
* Tells if the .FRM file is for solidDB table.
*
* Parameters :
*
*   THD* thd, in, - current my-sql thread.
*   File &frm_file, in - descriptor of .FRM file
*
* Return value :
* true - solidDB table,
* false - something else, can be other table or not a table at all.
*
* See also table.cpp::openfrm()
*/
bool is_soliddb_table( File &frm_file )
{

    SDB_DBUG_ENTER("solid:global:is_soliddb_table( THD* thd, File &frm_file )");
    uchar head[288];

    my_seek( frm_file, 0, MY_SEEK_SET, MYF(0) );

    if ( my_read( frm_file, (byte*) head, 64, MYF(MY_NABP) ) ) {
        SDB_DBUG_RETURN(false);
    }

    if ( !memcmp( head, STRING_WITH_LEN("TYPE=") ) ) {
        SDB_DBUG_RETURN(false);
    }

    if ( head[0] != (uchar)254 || head[1] != 1 ) {
        SDB_DBUG_RETURN(false);
    }

    if ( head[2] != FRM_VER && head[2] != FRM_VER + 1 &&
        !(head[2] >= FRM_VER + 3 && head[2] <= FRM_VER + 4) ) {
            SDB_DBUG_RETURN(false);
    }

    bool solid_table =
#if MYSQL_VERSION_ID >= 50100
        (enum legacy_db_type)
#else
        (enum db_type)
#endif
        (uint) head[3] == DB_TYPE_SOLID_DB;
    SDB_DBUG_RETURN(solid_table);
}

/*#***********************************************************************\
*
*              is_soliddb_table
*
* Tells if the table is solidDB table. Opens .FRM file for it.
*
* Parameters :
*
*   THD* thd, in, - current my-sql thread.
*   const char* dbname, in    - my-sql database name
*   const char* tablename  in - my-sql table name
*
* Return value :
* true - solidDB table,
* false - something else, can be other table or not a table at all.
*
* See also table.cpp::openfrm()
*/
bool is_soliddb_table( const char* dbname, const char* tablename )
{
    File  file;
    char  path[FN_REFLEN];
    //char  buff[FN_REFLEN];
    bool  result = false;

    SDB_DBUG_ENTER("solid:global:is_soliddb_table(THD* thd, const char* dbname, const char* tablename)");

    strxmov(path, mysql_data_home, "/", dbname, "/", tablename, NullS);

    file = my_open( path, // fn_format( buff, path, "", reg_ext, MY_UNPACK_FILENAME ),
        O_RDONLY | O_SHARE, MYF(0) );

    if (file  < 0) {
        SDB_DBUG_RETURN(false);
    }

    result = is_soliddb_table( file );

    my_close( file, MYF(MY_WME) );
    SDB_DBUG_RETURN(result);
}

/*#***********************************************************************\
*              is_DDL_command
* Tells if the currently executed command in the MySQL thread is a DDL
* command.
* Parameters :
*
*    THD* thd, in, MySQL current thread/process
*
* Return value : true or false
* Globals used : none
*/
bool is_DDL_command( THD* thd )
{
    static enum enum_sql_command DDL_commands[] = {
        SQLCOM_CREATE_TABLE,
        SQLCOM_CREATE_INDEX,
        SQLCOM_ALTER_TABLE,
        SQLCOM_RENAME_TABLE,
        SQLCOM_DROP_TABLE,
        SQLCOM_DROP_INDEX,
        /* These commands may be removed from the list as they do not influence
        * SolidDB database file actually,
        * although formally they are DDL commands.
        */
        SQLCOM_CREATE_DB,
        SQLCOM_DROP_DB,
        SQLCOM_ALTER_DB,
        SQLCOM_CREATE_FUNCTION,
        SQLCOM_DROP_FUNCTION,
        SQLCOM_CREATE_PROCEDURE,
        SQLCOM_CREATE_SPFUNCTION,
        SQLCOM_DROP_PROCEDURE,
        SQLCOM_ALTER_PROCEDURE,
        SQLCOM_ALTER_FUNCTION,
        SQLCOM_CREATE_VIEW,
        SQLCOM_DROP_VIEW,
        SQLCOM_CREATE_TRIGGER,
        SQLCOM_DROP_TRIGGER
    };

    bool is_DDL = false;
    size_t i;

    SDB_DBUG_ENTER("solid:is_DDL_command");

    if( !thd || !thd->lex ) {
        SDB_DBUG_RETURN(false);
    }

    is_DDL = false;
    for( i = 0; i < sizeof(DDL_commands)/sizeof(DDL_commands[0]); ++i ) {
        if( thd->lex->sql_command == DDL_commands[i] ) {
            is_DDL = true;
            break;
        }
    }

    SDB_DBUG_RETURN(is_DDL);
}

/*#***********************************************************************\
*              operation_requires_xtable_lock
* Tells if the currently executed command in the MySQL thread is a DDL
* command which requires an exclusive table lock on the table.
* Parameters :
*
*    THD* thd, in, MySQL current thread/process
*
* Return value : true or false
* Globals used : none
*/
bool operation_requires_xtable_lock( THD* thd )
{
    static enum enum_sql_command XDDL_commands[] = {
        SQLCOM_CREATE_INDEX,
        SQLCOM_ALTER_TABLE,
        SQLCOM_RENAME_TABLE,
        SQLCOM_DROP_INDEX,
        SQLCOM_DROP_TABLE
    };

    SDB_DBUG_ENTER("operation_requires_xtable_lock");

    if( !thd || !thd->lex ) {
        SDB_DBUG_RETURN(false);
    }

    bool xlock_required = false;
    for(size_t i = 0; i < sizeof(XDDL_commands)/sizeof(XDDL_commands[0]); ++i ) {
        if( thd->lex->sql_command == XDDL_commands[i] ) {
            xlock_required = true;
            break;
        }
    }

    SDB_DBUG_RETURN(xlock_required);
}

/*#***********************************************************************\
*
*              copy_file
*
* Copies source file to specified directory. The destination file is
* given a new name and extension. If the destination directory does not
* exist it is created.
*
* Parameters :
*
*   THD* thd, in, - current my-sql thread.
*   File &src, in   - my-sql descriptor of opened source file
*   const char* dstdir, in - destination directory
*   const char* fname,  in - destination file name
*   const char* ext     in - destination file extension
*
* Return value : true - ok, false - failure.
*/
bool copy_file( File &src, const char* dstdir, const char* fname, const char* ext )
{
    char dstname[FN_REFLEN];
    byte block[2048];
    unsigned int read;

    DBUG_ENTER("solid:global:copy_file");

    fn_format( dstname, fname, dstdir, ext, MY_UNPACK_FILENAME );

    my_mkdir( dstdir, 0777, MYF(0) );

    File dst = my_open( dstname, O_WRONLY|O_CREAT|O_TRUNC|O_BINARY, MYF(0) );

    if ( dst < 0) {
        DBUG_RETURN(false);
    }

    my_seek( src, 0, MY_SEEK_SET, MYF(0) );

    while ( (read = my_read( src, block, sizeof(block), MYF(MY_FAE) ) ) > 0 &&
        read != MY_FILE_ERROR ) {
            my_write( dst, block, read, MYF(MY_NABP) );
    }

    DBUG_RETURN( !my_close( dst, MYF(MY_WME|MY_WME) ) && read != MY_FILE_ERROR );
}

/*#***********************************************************************\
 *
 *              solid_flush_logs
 *
 * Flush logs to the disk
 *
 * Parameters : -
 *
 * Return value : -
 *
 * Globals used : current_thd
 */
bool solid_flush_logs(
#if MYSQL_VERSION_ID >= 50100
        handlerton *hton
#endif
    )
{
        SOLID_CONN* con = NULL;
        THD* thd = current_thd;
        DBUG_ENTER("solid_flush_logs");

        ss_dassert(thd != NULL);

#if MYSQL_VERSION_ID >= 50100
        con = (SOLID_CONN*)thd->ha_data[hton->slot];
#else
        con = (SOLID_CONN*)thd->ha_data[solid_hton.slot];
#endif

        if (con) {
            dbe_db_t* db = NULL;
            rs_sysi_t* cd = NULL;

            cd = con->sc_cd;
            ss_dassert(cd != NULL);

            db = (dbe_db_t*)rs_sysi_db(cd);
            ss_dassert(db != NULL);

            dbe_db_logflushtodisk(db);
        }

        DBUG_RETURN(0);
}

#if MYSQL_VERSION_ID >= 50100

/*#***********************************************************************\
 *
 *              ::check_if_incompatible_data
 *
 * Check wheather alter table changes are compatible or not
 *
 * Parameters :
 *
 *      HA_CREATE_INFO* info, in, use
 *      uint            table_changes, in, use
 *
 * Return value :
 *
 *      COMPATIBLE_DATA_YES if data is compatible or
 *      COMPATIBLE_DATA_NO
 *
 * Globals used :
 */

bool ha_soliddb::check_if_incompatible_data(
	HA_CREATE_INFO*	info,
	uint		table_changes)
{
	if (table_changes != IS_EQUAL_YES) {

		return COMPATIBLE_DATA_NO;
	}

	/* Check that auto_increment value was not changed */
	if ((info->used_fields & HA_CREATE_USED_AUTO) &&
		info->auto_increment_value != 0) {

		return COMPATIBLE_DATA_NO;
	}

	/* Check that row format didn't change */
	if ((info->used_fields & HA_CREATE_USED_AUTO) &&
		get_row_type() != info->row_type) {

		return COMPATIBLE_DATA_NO;
	}

	return COMPATIBLE_DATA_YES;
}

/*#***********************************************************************\
 *
 *              mysql_declare_plugin
 *
 * Declare solidDB plugin
 *
 * Parameters :
 *
 * Return value :
 *
 * Globals used :
 */

static int show_soliddb_vars(THD *thd, SHOW_VAR *var, char *buff)
{
        soliddb_export_status();
        var->type = SHOW_ARRAY;
        var->value= (char *)soliddb_status_variables;
        return 0;
}

SHOW_VAR soliddb_status_variables_export[]= {
  {"solidDB",                   (char*) &show_soliddb_vars, SHOW_FUNC},
  {NullS, NullS, SHOW_LONG}
};

struct st_mysql_storage_engine soliddb_storage_engine=
{ MYSQL_HANDLERTON_INTERFACE_VERSION};

mysql_declare_plugin(soliddb)
{
      MYSQL_STORAGE_ENGINE_PLUGIN,
      &soliddb_storage_engine,
      solid_hton_name,
      "Solid Information Technology Ltd",
      solid_hton_comment,
      PLUGIN_LICENSE_GPL,
      solid_init, /* Plugin Init */
      NULL, /* Plugin Deinit */
      0x0100 /* 1.0 */,
      soliddb_status_variables_export, /* TODO: status variables */
      NULL, /* TODO: system variables */
      NULL  /* TODO: config options   */
}
mysql_declare_plugin_end;

/*#***********************************************************************\
 *
 *              solid_alter_table_flags
 *
 * Alter table flags currently implemented on solidDB
 *
 * Parameters :
 *
 *      uint  flags, in, not ussed
 *
 * Return value : uint
 *
 * Globals used :
 */
static inline uint solid_alter_table_flags(uint flags)
{
        /* No primary key support at the moment */

        return(HA_ONLINE_ADD_INDEX_NO_WRITES
		| HA_ONLINE_DROP_INDEX_NO_WRITES
		| HA_ONLINE_ADD_UNIQUE_INDEX_NO_WRITES
		| HA_ONLINE_DROP_UNIQUE_INDEX_NO_WRITES);
}

/*#***********************************************************************\
 *
 *              ::add_index
 *
 * Add indexes to a table. Note that at the moment we use FULL table scan
 * for every index added to a table. TODO: use only one FULL table scan
 * and build indexes using this information.
 *
 * Parameters :
 *
 *      TABLE* table_arg, in, use
 *      KEY*   key_info, in, use
 *      uint   num_of_keys, in, use
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::add_index(TABLE* table_arg, KEY *key_info, uint num_of_keys)
{
        rs_relh_t* relh = NULL;
        st_key* key = NULL;
        char* mysqltabname = NULL;
        SOLID_CONN* con = NULL;
        rs_sysi_t* cd = NULL;
        tb_trans_t* trans = NULL;
        char*  authid = NULL;
        char*  catalog = NULL;
        char*  extrainfo = NULL;
        char* relname = NULL;
        tb_connect_t* tbcon = NULL;
        su_err_t*  errh = NULL;
        uint nkey = 0;
        bool succp = TRUE;
        solid_bool finished = FALSE;
        THD* thd = NULL;
        int rc = 0;
        int len;

        DBUG_ENTER("ha_soliddb::add_index");
        ss_dassert(table_arg != NULL);
        ss_dassert(key_info != NULL);
        ss_dassert(num_of_keys != 0);

        thd = current_thd;

        len = strlen(table_arg->s->path.str) + 1;
        mysqltabname = table_arg->s->path.str;

        relname = (char *)SsMemAlloc(len);
        authid = (char *)SsMemAlloc(len);

        solid_set_relname(authid, relname, (char *)mysqltabname, lower_case_table_names);

        con = get_solid_connection(ht, thd, relname);
        ss_dassert(con != NULL);

        cd = con->sc_cd;
        tbcon = con->sc_tbcon;
        trans = con->sc_trans;
        relh = solid_rsrelh;

        solid_trans_beginif(ht, thd, TRUE);

        for(nkey = 0; succp && nkey < num_of_keys; nkey++) {

            key = &key_info[nkey];
            ss_dassert(key != NULL);

            succp = ha_solid_createindex(relh, key, relname, tbcon, cd, trans,
                                         authid, catalog, extrainfo, &errh, lower_case_table_names);
        }

        if (succp) {
            do {
                succp = tb_trans_commit(cd, trans, &finished, &errh);

                /* Store MySQL key number to connection if we had a duplicate key error */
                if (errh != NULL) {
                    int solrc = su_err_geterrcode(errh);

                    if (solrc == DBE_ERR_UNIQUE_S || solrc == DBE_ERR_PRIMUNIQUE_S) {
                        char *nerrmsg;
                        char *tmp;
                        char *saveptr = NULL;

                        uint len = strlen(su_err_geterrstr(errh))+1;

                        nerrmsg = (char*) SsMemAlloc(len);
                        strcpy(nerrmsg, su_err_geterrstr(errh));
                        tmp = strtok_r(nerrmsg, "(", &saveptr);
                        tmp = strtok_r(NULL, ")", &saveptr);

                        for(nkey = 0; nkey < num_of_keys; nkey++) {
                            char *indexname;
                            indexname = (char *)SsMemAlloc(strlen(key->name)+64);

                            keyname_from_mysql_to_solid(cd, relh, key_info[nkey ].name,
                                                        indexname, lower_case_table_names);

                            if (!(strcmp(tmp, indexname))) {
                                con->sc_errkey_num = nkey;
                                con->sc_table = table_arg;
                                break;
                            }

                            SsMemFree(indexname);
                        }

                        SsMemFree(nerrmsg);
                    }
                }

            } while (rs_sysi_lockwait(cd) || !finished);

        }

        SsMemFree(relname);
        SsMemFree(authid);

        if (!succp) {
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
        } else {
            rc = 0;
        }

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::prepare_drop_index
 *
 * Mark provided indexes to be dropped
 *
 * Parameters :
 *
 *      TABLE* table_arg
 *      uint*  key_num
 *      uint   num_of_keys
 *
 * Return value : 0 or error code
 *
 * Globals used :
 */
int ha_soliddb::prepare_drop_index(TABLE* table_arg, uint *key_num, uint num_of_keys)
{
        SOLID_CONN* con = NULL;
        rs_sysi_t* cd = NULL;
        tb_trans_t* trans = NULL;
        su_err_t*  errh = NULL;
        rs_key_t* solid_key = NULL;
        rs_relh_t* relh = NULL;
        char *extrainfo = NULL;
        void*  cont = NULL;
        char* relname = NULL;
        char* mysqltabname = NULL;
        char* authid = NULL;
        bool succp = TRUE;
        THD* thd = NULL;
        int rc = 0;
        uint len = 0;
        uint nkey;

        DBUG_ENTER("ha_soliddb::prepare_drop_index");
        ss_dassert(table_arg != NULL);
        ss_dassert(key_num != NULL);
        ss_dassert(num_of_keys != 0);

        thd = current_thd;

        len = strlen(table_arg->s->path.str) + 1;
        mysqltabname = table_arg->s->path.str;

        relname = (char *)SsMemAlloc(len);
        authid = (char *)SsMemAlloc(len);

        solid_set_relname(authid, relname, (char *)mysqltabname, lower_case_table_names);

        con = (SOLID_CONN*)thd->ha_data[ht->slot];

        relh = solid_rsrelh;

        SsMemFree(relname);
        SsMemFree(authid);

        ss_dassert(con != NULL);
        cd = con->sc_cd;
        ss_dassert(cd != NULL);
        trans = con->sc_trans;

        solid_trans_beginif(ht, thd, TRUE);

        for (nkey = 0; succp && nkey < num_of_keys; nkey++) {
            rs_entname_t* en;
            uint knum = key_num[nkey];

            solid_key = solid_resolve_key(cd, solid_rsrelh,
                                          &table_arg->key_info[knum], knum, lower_case_table_names);

            ss_dassert(solid_key != NULL);

            en = rs_relh_entname(cd, solid_rsrelh);

            succp = tb_dropindex_relh(
                        cd,
                        trans,
                        relh,
                        rs_key_name(cd, solid_key),
                        rs_entname_getschema(en),
                        rs_entname_getcatalog(en),
                        extrainfo,
                        &cont,
                        &errh);
        }

        if (!succp) {
            rc = ha_solid_mysql_error(thd, errh);
            su_err_done(errh);
        } else {
            rc = 0;
        }

        DBUG_RETURN(rc);
}

/*#***********************************************************************\
 *
 *              ::final_drop_index
 *
 * Finalize index drop, nothing really to do because index is really dropped
 * at commit.
 *
 * Parameters :
 *
 *      TABLE* table_arg
 *
 * Return value : 0
 *
 * Globals used :
 */
int ha_soliddb::final_drop_index(TABLE* table_arg)
{
        DBUG_ENTER("ha_soliddb::final_drop_index");
        ss_dassert(table_arg != NULL);

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::reset
 *
 * Reset to statement defaults
 *
 * Parameters : -
 *
 * Return value : 0
 *
 *
 * Globals used : THD* current_thd
 */
int ha_soliddb::reset()
{
        DBUG_ENTER("ha_soliddb:reset");

        THD* thd = current_thd;

        extra_keyread = FALSE;
        extra_retrieve_primary_key = FALSE;
        extra_retrieve_all_cols = TRUE; /* TODO: Should be FALSE */

        /* Set binlog to row based */
        thd->set_current_stmt_binlog_row_based_if_mixed();
        thd->current_stmt_binlog_row_based = TRUE;

        DBUG_RETURN(0);
}

/*#***********************************************************************\
 *
 *              ::get_auto_increment
 *
 * Reserves an interval of auto_increment values from the handler.
 * offset and increment means that we want values to be of the form
 * offset + N * increment, where N>=0 is integer.
 * If the function sets *first_value to ~(ulonglong)0 it means an error.
 * If the function sets *nb_reserved_values to ULONGLONG_MAX it means it has
 * reserved to "positive infinite".
 *
 * Parameters :
 *
 *     ulonglong  offset, in, use
 *     ulonglong  increment, in, use
 *     ulonglong  nb_desired_values, in, use
 *     ulonglong* first_value, out
 *     ulonglong* nb_reserved_values, out
 *
 * Return value : -
 *
 * Globals used : THD* current_thd
 */
void ha_soliddb::get_auto_increment(
        ulonglong offset,
        ulonglong increment,
        ulonglong nb_desired_values,
        ulonglong *first_value,
        ulonglong *nb_reserved_values)
{
        DBUG_ENTER("ha_soliddb:get_auto_increment");

        ss_int8_t i8;
        SOLID_CONN* con;
        ulonglong solid_value = 0;
        struct system_variables* variables;
        bool result = FALSE;
        rs_relh_t*  relh = NULL;
        long        seq_id=0;
        rs_auth_t*  auth;
        rs_atype_t* atype;
        rs_aval_t*  aval;
        solid_bool  p_finishedp;
        rs_err_t*   errh = NULL;
        ulonglong   n_value;
        THD*        thd = current_thd;

        ss_dassert(first_value != NULL);
        ss_dassert(nb_reserved_values != NULL);

        con = (SOLID_CONN*)thd->ha_data[ht->slot];

        ss_dassert(con != NULL);
        relh = solid_rsrelh;
        ss_dassert(relh != NULL);
        auth = rs_sysi_auth(con->sc_cd);
        ss_dassert(auth != NULL);
        rs_auth_setsystempriv(con->sc_cd, auth, TRUE);

        seq_id = rs_relh_readautoincrement_seqid(con->sc_cd, relh);
        *first_value = ~(ulonglong)0;

        atype = rs_atype_initbigint(con->sc_cd);
        aval = rs_aval_create(con->sc_cd, atype);

        result = tb_seq_next(con->sc_cd, con->sc_trans, seq_id,
                             FALSE, atype, aval, &p_finishedp, &errh);

        if (result) {
            i8 = rs_aval_getint8(con->sc_cd, atype, aval);

            solid_value = SsInt8GetNativeUint8(i8);

            for (n_value = 0; result && n_value < nb_desired_values; n_value++, solid_value++) {

                solid_value = solid_next_insert_id(solid_value-1, increment, offset);

                if (n_value == 0) {
                    *first_value = solid_value;
                }
            }

            *nb_reserved_values = n_value;

            SsInt8SetNativeUint8(&i8, solid_value);
            rs_aval_setint8_raw(con->sc_cd, atype, aval, i8, NULL);

            result = tb_seq_set(con->sc_cd, con->sc_trans, seq_id,
                                FALSE, atype, aval, &p_finishedp, &errh);
        }

        rs_aval_free(con->sc_cd, atype, aval);
        rs_atype_free(con->sc_cd, atype);
        rs_auth_setsystempriv(con->sc_cd, auth, FALSE);

        if (errh) {
            ss_pprintf_1(("solidDB: [Error]::get_auto_increment: %s\n",  su_err_geterrstr(errh)));
            su_err_done(errh);
        }

        DBUG_VOID_RETURN;
}

#endif /* MYSQL_VERSION_ID >= 50100 */

#endif /* if defined WITH_SOLIDDB_STORAGE ENGINE || defined HAVE_SOLID_DB */
